; File:  basescan.pro
; Author:  Ted Fisher 2013-11-21 <ted@specinc.com> <fishnuke@gmail.com>
; Pupose:  reads and parsed 3V-CPI base files and records the number of H and V particles, overloads, etc.
;          approaching a replacement of the obsolete and heavivly spaghettified VB6/VC++ Playback


; (need to set up process second by second function to process each second's worth of data)


; wishlist:

; - plot images with H 50 micron to scale to the V 10 micron (as a flag)
; - plot gray lines between images, between H and V
; - fix up overloads
; - open multiple consecutive files (ie midnight files)
; - finish multi-part images
; - mask bit filling
; - input file table (view raw input data in a table format)
; - make basescan 2D-S/HVPS/2D-128 compatible
; - find out if CPI min size (pixels) field is in HK, add to plots and csv
; - disable generating hoverload file
; - maybe disable generating basescan_out file if just_HK is set



; 2.0.26 2016-03-15 modified to use recompressed images from processparticle.pro for the SORTED files
; 2.0.25 2016-02-23 merged the T1 to T2 functionality from basescan_mo.pro, abandonded other half-finished changes from that version and cleaned up
; 2.0.24 2015-06-29 modified to read to end of file instead of spinning
; 2.0.23 2014-10-02 integrated into 3VCPIView_1.1.33_TAF, note, no changes were made to basescan, its main proceedure
;                   is just called from a button in 3V-CPI View
;                   this isn't a streamlined integration for batch processing, but just a way to call basescan directly from the View,
;                   and makes base->SORTED processing more consistent with SORTED->bin and bin->archive processing
;                   the new basescan button is between the CPI button and the SORTED->bin button
;                   added !show_images system variable to override/complement command line show_images_input, for use with 3V-CPI integration
; 2.0.22 2014-10-02 set up to use universal processparticles.pro (same file as basescan, 3V-CPI View, 2D-S View, and HVPS View now all use)
;                   improved image display, added show H 50 micron to scale, separator lines, and image scaling
; 2.0.21 2014-10-01 fixed show_images display bug, added !zoom_factor for scaling up images
; 2.0.20 2014-09-23 added pnger.pro to program directory
; 2.0.19 2014-09-18 added output_TAS_file system variable to generate a TAS file from the Housekeeping Data,
;                   which could be used as input for an (Hawkeye) FCDP, FSSP
; 2.0.18 2014-09-02 added infilepath + to all output filenames, so that the output files are all put into the base file directory instead of into
;                   the program directory, also changed to use pnger instead of jpger -- smaller, lossless output files with better colors for line plots
;                   than jpgs
; 2.0.17 2014-08-27 added extraction and plots for CPI modes, HK words 55 and 62 show Triggering and Min Pixels before Fire,
;                   as well as CPI frame rate, added new system variable !output_CPI_frames_csv to generate csv of CPI frame rate, min pixels before fire
; 2.0.16 2014-07-31 updated typo in fix below, wrong parens for n_elements, added catch in write_overflow_to_SORTED for undefined particles
;        would be better to correct the logic for overflows to ensure both are defined
;        added bracketing test of new probecount to be between max and min HK counts
;        added test for valid particle of NULL packet AFTER the current particle to accept, using test_read_data
; 2.0.15 2014-07-01 fixed freeing lun's when only one SORTED file was generated (had been trying to free the second lun always, and hung)
; 2014-06-30 version 2.0.14 updated processparticle.pro for image decompression error (skipping the first slice)
; 2014-05-19 version 2.0.13 added just_HK flag to just show the HK and exit
; 2014-05-09 version 2.0.12 improved HK bracketing, H and V particle buffers
; 2014-05-09 version 2.0.11 fixed TAS packets to SORTED file start/stop -- which enabled full run through 3V-CPI View
; 2014-05-09 version 2.0.10 put time labels on HK plots, adding HK_TXT and HK_BIN outputs, TT labels in SORTED files
; 2014-05-09 version 2.0.9 added overload printout, working on multipart images, some parse/timing issues, added SORTED file splitting
; 2014-05-08 version 2.0.8 adding SORTED file type outputs, changed to 2.#.# because this is Version 2.0 of the Playback
; 2014-05-05 version 1.7 putting wrapper function around getting raw blocks to simplify data access, fixed circular buffer error
; 2014-05-02 version 1.6 added reading HK file via read_hk, calculating times of particles
; 2014-04-18 version 1.5 modified input state machine to more robustly scan through file and not get hung up on false "2S" flags in timestamps, etc
; 2013-12-24 version 1.4 modified to display images
; 2013-12-20 version 1.3 modified to deal with NH or NV eq 0, also to deal with MS timestamps interleaved with data
; 2013-12-05 version 1.2
; 2013-12-05 version 1.1 added support for Cancel in dialog_pickfile
; 2013-12-05 version 1.0 initial version





pro setup_globals
    ; initialized system variables and heap variables
    common basescan_common, show_images, particle, timestamp, prev_sec, timestamp_printed, basefile_lun, file_size, $
        raw_data_buffer, raw_data_buffer_size, buffer_size, buffer_ptr, raw_data, hk_array, mask_array, mask_timestamp_array, $
        timestamp_index, timestamp_array_size, timestamp_array, hk_timestamp_array, hk_probecount_array, $
        probecount_index, probecount_array_size, probecount_array, iteration_counter, $
        HK1_timestamp, HK2_timestamp, HK_timediff_nanos, block_timestamp

    defsysv, '!basescan_version', '2.0.26' ; version number, used in output file headers

    defsysv, '!use_test_file',    0 ; use the test_data/base130909131733.2DSCPI file instead of opening a dialog file picker

    defsysv, '!testfile_infilename', 'C:\fisher\SPEC\basescan\test_data\base130909131733.2DSCPI' ; fully qualified filename opened if !use_test_file is set to 1
    defsysv, '!testfile_infilepath', 'C:\fisher\SPEC\basescan\test_data\'                        ; full path for the testfile above

    ;defsysv, '!testfile_infilename', 'infilename = 'C:\fisher\SPEC\NavyLear2014\FLIGHT_DATA\2014-08-21_NavyLear2014_ScienceFlight_S_WY\Lear__Hawkeye_2D-S\base140821231403.2DSCPI' ; fully qualified filename opened if !use_test_file is set to 1
    ;defsysv, '!testfile_infilepath', 'infilepath = 'C:\fisher\SPEC\NavyLear2014\FLIGHT_DATA\2014-08-21_NavyLear2014_ScienceFlight_S_WY\Lear__Hawkeye_2D-S\'                        ; full path for the testfile above

    defsysv, '!show_images',       1 ; if set to 1, override the command line show_images flag, useful if running from the View program
    defsysv, '!show_masks',       0 ; display the current mask over the images

    defsysv, '!just_HK',          0 ; just process the HK file and display the HK plots, then exit.  Useful for quick look at the HK file.
    defsysv, '!show_HK_plots',    1 ; set to 1 to show plots of the HK data, set to 0 to not show plots (overrides !save_HK_plots below)
    defsysv, '!save_HK_plots',    1 ; set to 1 to save images of the HK plots in the output directory, set to 0 to not save plots

    defsysv, '!output_TAS_file',  0 ; set to 1 to output TAS file from HK data

    defsysv, '!output_CPI_frames_csv', 0 ; output CPI frame rate and min pixels before fire csv file

    defsysv, '!reprint_raw_data', 0 ; flag to print out raw data into an ascii file for debugging parsing and analyzing raw data

    defsysv, '!log_lun',          0 ; used to store the log file logical unit number so any subroutine can write to the log file


    defsysv, '!generate_SORTED',         1   ; flag to generate SORTED file outputs
    defsysv, '!file_splitting_at', 100000L   ; number of particles to split SORTED files after, +2 seconds of overlap after this

    defsysv, '!emoflag', 0

    defsysv, '!manual_override_H50vs10micron',  0  ; 0 or 1 -- manually override the automatic 50 vs 10 micron detection
    ;(for files prior to Playback 1.1.17)
    defsysv, '!H_resolution_microns',          10  ; 10 or 50 microns for the H channel resulotion.
    ; Use with !manual_override_H50vs10micron = 1 to manually set the H resolution

    defsysv, '!H_res_set',                      0  ; flag to determine if the H resolution has been set this run, to avoid stops in manual override
    defsysv, '!H_res_set_manually',             0  ; flag for internal logic to detect if the manual selection window has been used or not

    defsysv, '!zoom_factor',                    1  ; when basescan is run with show_images (basescan, 1), this is the zoom scale used
    ; 1 is 1 display pixel for each image pixle, 5 makes a 5x5 square in the display for each image pixel
    defsysv, '!draw_separator_line',            1  ; draw a separator line after each image is drawn, if show_images is set
    defsysv, '!draw_50um_to_scale',             1  ; magnify the images of the 50 micron channel by 5X, so that they are to scale with the 10 micron channel
    defsysv, '!line_color',                    42

    defsysv, '!flag_3VCPI',                     1  ; define this system variable so that processparticle.pro handles the 3V-CPI type data format instead of 2D-S/HVPS format (slice value 0x7FFF handled differently)

    defsysv, '!PIXELS', 128

    ; system variables for Roundness
    defsysv, '!Roundness1_min',  .8
    defsysv, '!Roundness1_max', 2.
    defsysv, '!Roundness2_min', 0.
    defsysv, '!Roundness2_max',  .2
    defsysv, '!Roundness3_min', 0.
    defsysv, '!Roundness3_max',  .25

    defsysv, '!disp_window',      12
    defsysv, '!disp_window_xs',  257 ; default 2 * !PIXELS + 1, window may be expanded if !zoom_factor or !draw_50um_to_scale are set (handled below)
    defsysv, '!disp_window_ys',  768 ; not adjusted for !zoom_factor or !draw_50um_to_scale, user can adjust to fit the screen


    ; define a maximum forward time jump between particles
    ; 1 slice * 10um/slice * 1s/170m * 1m/1e6um = 5.88e-8 s per slice or 58.8ns
    ; 0x00FFFFFFFF slices * 5.88e-8s/slice =   252.5s
    ; 0x0FFFFFFFFF slices * 5.88e-8s/slice =  4042.3s
    ; 0xFFFFFFFFFF slices * 5.88e-8s/slice = 64677.2s which seems reasonably long enough never to happen other than by error

    ;defsysv, '!max_probecount_diff', ulong64('000000FFFFFFFFFF'X)
    ;defsysv, '!max_probecount_diff', ulong64('00000000FFFFFFFF'X)
    defsysv, '!max_probecount_diff', ulong64('0000000000FFFFFF'X)

    defsysv, '!DATA_BLOCK_LENGTH',   4096L ; in bytes, note that raw_data_buffer and raw_data are indexed in words
    ; NOTE  ; + 2 for checksum = 4098 byte blocks

    defsysv, '!array_growth_factor', 100000L ; size to alocate initial arrays, and to grow arrays when size exceeded


    defsysv, '!npad',                  512L ; number of particles to keep before the current one
    defsysv, '!particle_array_size',  4096L ; total number of H or V particles to buffer


    ; SORTED file header types (comments from VB code)
    ;    MASK_HEADER = &H4D4D
    ;    HK_HEADER = &H4B4B
    ;    TAS_HEADER_H = &H5454
    ;    TAS_HEADER_V = &H4141
    ;    H_PARTICLE_HEADER = &H4848
    ;    V_PARTICLE_HEADER = &H5656
    ;    H_DEADTIME_HEADER = &H6C6C    ' (-100/-101)
    ;    V_DEADTIME_HEADER = &H7C7C    ' (-100/-101)
    ;    H_MASK_HEADER = &H1C1C    '(-102/-103)
    ;    V_MASK_HEADER = &H2C2C    '(-102/-103)

    ;    V_DUMMY_DEADTIME_HEADER = &H1414    ' 5140 decimal (-105/-106 in IDL)
    ;    H_DUMMY_DEADTIME_HEADER = &H1424    ' 5156 decimal (-105/-106 in IDL)
    ;    V_BACKUP_DEADTIME_HEADER = &H2414    ' 9236 decimal (marked as -105/-106)
    ;    H_BACKUP_DEADTIME_HEADER = &H2424    ' 9252 decimal
    ;    V_STREAKER_HEADER = &H1818    '  6168 decimal  (-107/108)
    ;    H_STREAKER_HEADER = &H1919    ' 6425 decimal  (-107/108)

    defsysv, '!header_length',             256 ; number of bytes of the SORTED, HK file headers

    defsysv, '!MASK_HEADER',               '4D4D'X
    defsysv, '!HK_HEADER',                 '4B4B'X
    defsysv, '!TAS_HEADER_H',              '5454'X
    defsysv, '!TAS_HEADER_V',              '4141'X
    defsysv, '!H_PARTICLE_HEADER',         '4848'X
    defsysv, '!V_PARTICLE_HEADER',         '5656'X
    defsysv, '!H_DEADTIME_HEADER',         '6C6C'X    ; (-100/-101) ; used for FIFO Overflow
    defsysv, '!V_DEADTIME_HEADER',         '7C7C'X    ; (-100/-101)
    defsysv, '!H_MASK_HEADER',             '1C1C'X    ; (-102/-103)
    defsysv, '!V_MASK_HEADER',             '2C2C'X    ; (-102/-103)

    defsysv, '!V_DUMMY_DEADTIME_HEADER',   '1414'X    ; 5140 decimal (-105/-106 in IDL)
    defsysv, '!H_DUMMY_DEADTIME_HEADER',   '1424'X    ; 5156 decimal (-105/-106 in IDL)
    defsysv, '!V_BACKUP_DEADTIME_HEADER',  '2414'X    ; 9236 decimal (marked as -105/-106)
    defsysv, '!H_BACKUP_DEADTIME_HEADER',  '2424'X    ; 9252 decimal
    defsysv, '!V_STREAKER_HEADER',         '1818'X    ; 6168 decimal  (-107/108)
    defsysv, '!H_STREAKER_HEADER',         '1919'X    ; 6425 decimal  (-107/108)



    checksum = uint(0)


    timestamp  = { _timestamp_struct, $
        year  : uint(0), $
        month : uint(0), $
        dow   : uint(0), $
        day   : uint(0), $
        hour  : uint(0), $
        minute: uint(0), $
        second: uint(0), $
        ms    : uint(0)  }

    hires_timestamp = {_hires_timestamp_struct, $
        year   : uint(0), $
        month  : uint(0), $
        day    : uint(0), $
        hour   : uint(0), $
        minute : uint(0), $
        second : uint(0), $
        ms     : uint(0), $
        us     : uint(0), $
        ns     : uint(0), $
        counter: ulong64(0)}

    HK1_timestamp = hires_timestamp
    HK2_timestamp = hires_timestamp

    block_timestamp = timestamp ; timestamp from last data block, filled by read_raw_block


    ;    ; particle
    ;    flag   = uint(0) ; '3253'X = "2S"
    ;    NHraw  = uint(0) ; bits 0-12, high bits are flagged defined below
    ;    NVraw  = uint(0) ; bits 0-12, high bits are flagged defined below
    ;    partID = uint(0)
    ;    nslice = uint(0)
    ;    Himage = uint(0) ; variable length
    ;    Vimage = uint(0) ; variable length
    ;    time1  = uint(0) ; timing word [15: 0], LSW
    ;    time2  = uint(0) ; timing word [31:16], ISW
    ;    time3  = uint(0) ; timing word [47:32], MSW
    ;    checksum = uint(0)

    ;    particle = {_particle_struct, $
    ;                        flag  : uint(0), $
    ;                        NHraw : uint(0), $
    ;                        NVraw : uint(0), $
    ;                        partID: uint(0), $
    ;                        nslice: uint(0), $
    ;                        image_data:ptr_new(), $
    ;                        counter: ulong64(0) }


    ; particle.type is a single character used as a flag
    ; U = Uninitialized
    ; R = Reject
    ; H = Valid H particle
    ; V = Valid V particle


    particle = {_particle_struct, $
        type:   'U',      $
        flag  : uint(0),  $
        NHraw : uint(0),  $
        NVraw : uint(0),  $
        partID: uint(0) , $
        nwords: long(0),  $
        nslice: uint(0),  $
        CPI_trigger: uint(0), $
        image_data:ptr_new(), $
        timestamp: hires_timestamp }

    raw_data_buffer_size = (!DATA_BLOCK_LENGTH/2L + 2L) * 512L ; TAF 2014-05-06 can expand once debugging is finished...

    raw_data_buffer = uintarr(raw_data_buffer_size) ; implement circular buffer

    ; data blocks are 4096 bytes, or 4096/2 words
    raw_data = uintarr(!DATA_BLOCK_LENGTH/2)

    buffer_size = 0L
    buffer_ptr  = 0L
end ; setup_globals



pro define_pixel_resolutions;  This subroutine is called when data are loaded.  if it it a 150um, or 50/10um or 10/10um probe.
    ;defsysv, '!MICRONS_PER_PIXELH',      10.; default. calibrated size of pixels for calculating real lengths in microns, temporary for 2D-S
    defsysv, '!MICRONS_PER_PIXELV',      10.; default. calibrated size of pixels for calculating real lengths in microns, temporary for 2D-S
    DEFSYSV, '!DWW_main', 50.8D ; mm Distance Window to Window for defining the Depth of Field (DOF), 50.8mm for 3V-CPI
    ;DEFSYSV, '!DWW_2DS',  63.0D ; mm Distance Window to Window for defining the Depth of Field (DOF) 63.0mm for 2D-S

    if(!manual_override_H50vs10micron eq 1 or !H_res_set_manually eq 1) then begin ; user has selected to manually override automatic detection of H 50 vs 10 microns
        print, ""                                      ; use the value set in !H_resolution_microns
        print, "!!!WARNING!!!
        print, "Using manual override for H 50 vs 10 microns"
        print, "!manual_override_H50vs10micron is set for manual override"
        print, "If you want to change this, set !manual_override_H50vs10micron = 0 and restart program!!"
        print, ""

        if(!H_res_set_manually ne 1) then begin
            ; if we're going for a user setup manual override,
            ; the popup will not have been displayed yet,
            ; if the automatic detetction in SORTED or binary fails, the popup will already have been shown
            H50micron_pop_up ; pop up window for User to manually select H 10 or 50 microns
        endif

        if(!H_resolution_microns eq 10) then begin
            defsysv, '!MICRONS_PER_PIXELH', 10.

            print, "using manually set H resolution: ", !MICRONS_PER_PIXELH
            print, ""
            ;print, ".cont to continue, or .res to reset IDL, then restart program with new value"
            ;print, ""
        endif else if(!H_resolution_microns eq 50) then begin
            defsysv, '!MICRONS_PER_PIXELH', 50.

            print, "using manually set H resolution: ", !MICRONS_PER_PIXELH
            print, ""
            ;print, ".cont to continue, or .res to reset IDL, then restart program with new value"
            ;print, ""
        endif else begin
            print, "Error, !H_resolution_microns must be either 10 or 50 microns"
            print, "correct value then restart program"
            print, ""

            stop
        endelse

        ;stop ; use .cont to continue from here or .res to reset IDL
        ; if you want to use automatic 50 vs 10 micron detection for SORTED files from Playback 1.1.17 or later,
        ; set !manual_override_H50vs10micron to 0
        ; otherwise, set !H_resolution_microns to 10 or 50 microns
    endif else begin ; using automatic detection was successful
        if(!H_resolution_microns eq 10) then begin
            defsysv, '!MICRONS_PER_PIXELH', 10.
        endif else if(!H_resolution_microns eq 50) then begin
            defsysv, '!MICRONS_PER_PIXELH', 50.
        endif

        print, ""
        print, "Using automatic H resolution detection"
        print, "H resolution: ", !MICRONS_PER_PIXELH, " microns"
        print, "V resolution: ", !MICRONS_PER_PIXELV, " microns"
        print, ""
    endelse

    !H_res_set = 1 ; set the flag to prevent repeating setting the H resolution
end ; define_pixel_resolutions



pro print_raw_hex, probecount_in
    ; debug utility to print the probecounter in hex as well as in the raw file format hex, to aid finding the timestamp in the hex dump
    ; example:

    ; ulong64 value = 0x00000F013445

    ; prints         45 34 01 0F 00 00

    ; byte order is swapped, but byte values are the same


    byte0 = ishft(probecount_in and ulong64('0000000000FF'X),   0)
    byte1 = ishft(probecount_in and ulong64('00000000FF00'X),  -8)
    byte2 = ishft(probecount_in and ulong64('000000FF0000'X), -16)
    byte3 = ishft(probecount_in and ulong64('0000FF000000'X), -24)
    byte4 = ishft(probecount_in and ulong64('00FF00000000'X), -32)
    byte5 = ishft(probecount_in and ulong64('FF0000000000'X), -40)

    print, format='(%"0x%016X    raw:    %02X %02X %02X %02X %02X %02X")', probecount_in, byte0, byte1, byte2, byte3, byte4, byte5
end ; print_raw_hex



pro write_particle_to_SORTED, in_olun, in_header, in_particle
    ; prints the particle data to the SORTED file

    ; the file format from the VB is:

    ; Put #SORTED_FNUM, , H_PARTICLE_HEADER
    ; Put #SORTED_FNUM, , gPart_H.part_id
    ; Put #SORTED_FNUM, , gPart_H.h_slices + prev_NH1 + prev_NH2 + prev_NH3 + prev_NH4
    ; Put #SORTED_FNUM, , gPart_H.true_slice_count
    ; Put #SORTED_FNUM, , H_HR_Timestamp
    ; Put #SORTED_FNUM, , gPart_H.h_timing_word_hi
    ; Put #SORTED_FNUM, , gPart_H.h_timing_word
    ; Put #SORTED_FNUM, , CPI_triggerH

    ; Put #SORTED_FNUM, , image_data_h

    ; and the IDL basescan particle for reference:

    ; particle = {_particle_struct, $
    ;            flag  : uint(0), $
    ;            NHraw : uint(0), $
    ;            NVraw : uint(0), $
    ;            partID: uint(0), $
    ;            nwords: uint(0), $
    ;            nslice: uint(0), $
    ;            CPI_trigger: uint(0), $
    ;            image_data:ptr_new(), $
    ;            timestamp: hires_timestamp }


    ;2014-05-15 particle differences in SORTED files
    ;basescan
    ;Playback

    ;48 48 66 D1 01 00 01 00 DD 07 09 00 09 00 0D 00 33 00 39 00 02 00 78 03 DD 00 06 00 00 00 76 40 A4 CA 00 00 8A 40
    ;48 48 66 D1 01 00 01 00 DD 07 09 00 09 00 0D 00 33 00 39 00 00 00 02 00 07 03 06 00 00 00 76 40 A4 CA 00 00 8A 40

    ;48 48 67 D1 01 00 01 00 DD 07 09 00 09 00 0D 00 33 00 39 00 02 00 CE 03 68 01 06 00 00 00 08 45 A4 CA 00 00 C3 40
    ;48 48 67 D1 01 00 01 00 DD 07 09 00 09 00 0D 00 33 00 39 00 00 00 58 00 B9 00 06 00 00 00 08 45 A4 CA 00 00 C3 40


    ;if(in_particle.timestamp.us ge 1000) then stop
    ;if(in_particle.timestamp.ns ge 1000) then stop

    writeu, in_olun, in_header,          $
        in_particle.partID, $
        uint(in_particle.nwords), $
        in_particle.nslice, $
        in_particle.timestamp.year,   $ ; NOTE, the particle counter is in the IDL basescan timestamp,
        in_particle.timestamp.month,  $ ; but split up in the VB, so to match SORTED we have to do this muck
        in_particle.timestamp.day,    $ ; can certainly be cleaned up when View is being updated to Version 2.
        in_particle.timestamp.hour,   $
        in_particle.timestamp.minute, $
        in_particle.timestamp.second, $
        in_particle.timestamp.ms,     $
        in_particle.timestamp.us,     $
        in_particle.timestamp.ns,     $
        ulong(ishft(in_particle.timestamp.counter and ulong64('0000FFFF00000000'X), -32)), $ ; .h_timing_word_hi
        ulong(      in_particle.timestamp.counter and ulong64('00000000FFFFFFFF'X)      ), $ ; .h_timing_word
        in_particle.CPI_trigger, $
        *in_particle.image_data

end ; write_particle_to_SORTED



pro write_overflow_to_SORTED, in_olun, in_header, in_particle1, in_particle2
    ; writes the FIFO Overflow packet type to the SORTED file

    ;Put #SORTED_FNUM, , V_DEADTIME_HEADER
    ;Put #SORTED_FNUM, , First_Overflow_Timestamp    'timestamp from first overflow
    ;Put #SORTED_FNUM, , Second_Overflow_Timestamp
    ;Put #SORTED_FNUM, , First_Overflow_gPart_V.v_timing_word_hi    ' start overload probe counter
    ;Put #SORTED_FNUM, , First_Overflow_gPart_V.v_timing_word
    ;Put #SORTED_FNUM, , gPart_V.v_timing_word_hi                ' end overload probe counter
    ;Put #SORTED_FNUM, , gPart_V.v_timing_word

    if(n_elements(in_particle1) eq 0 or n_elements(in_particle2) eq 0) then begin
        print,            'Overload write error in write_overflow_to_SORTED, particle1: ', n_elements(in_particle1), ' particle2: ', n_elements(in_particle2)
        printf, !log_lun, 'Overload write error in write_overflow_to_SORTED, particle1: ', n_elements(in_particle1), ' particle2: ', n_elements(in_particle2)

        help, in_particle1
        help, in_particle2

        return
    endif

    writeu, in_olun, in_header,                     $
        in_particle1.timestamp.year,   $ ; NOTE, the particle counter is in the IDL basescan timestamp,
        in_particle1.timestamp.month,  $ ; but split up in the VB, so to match SORTED we have to do this muck
        in_particle1.timestamp.day,    $ ; can certainly be cleaned up when View is being updated to Version 2.
        in_particle1.timestamp.hour,   $
        in_particle1.timestamp.minute, $
        in_particle1.timestamp.second, $
        in_particle1.timestamp.ms,     $
        in_particle1.timestamp.us,     $
        in_particle1.timestamp.ns,     $
        in_particle2.timestamp.year,   $ ; NOTE, the particle counter is in the IDL basescan timestamp,
        in_particle2.timestamp.month,  $ ; but split up in the VB, so to match SORTED we have to do this muck
        in_particle2.timestamp.day,    $ ; can certainly be cleaned up when View is being updated to Version 2.
        in_particle2.timestamp.hour,   $
        in_particle2.timestamp.minute, $
        in_particle2.timestamp.second, $
        in_particle2.timestamp.ms,     $
        in_particle2.timestamp.us,     $
        in_particle2.timestamp.ns,     $
        ulong(ishft(in_particle1.timestamp.counter and ulong64('0000FFFF00000000'X), -32)), $ ; .h_timing_word_hi
        ulong(      in_particle1.timestamp.counter and ulong64('00000000FFFFFFFF'X)      ), $
        ulong(ishft(in_particle2.timestamp.counter and ulong64('0000FFFF00000000'X), -32)), $ ; .h_timing_word_hi
        ulong(      in_particle2.timestamp.counter and ulong64('00000000FFFFFFFF'X)      )

end ; write_overflow_to_SORTED



pro write_multiparticle_to_SORTED, in_olun, in_header, in_nmulti, in_multi_array
    ; writes the multi-packet particle data to the SORTED file

    ; the file format from the VB is:
    ; the main thing with the multi is the Number of Words is added, and the image fragments are written in the correct order

    ;    Put #SORTED_FNUM, , H_PARTICLE_HEADER
    ;    Put #SORTED_FNUM, , gPart_H.part_id
    ;    Put #SORTED_FNUM, , gPart_H.NH + prev_NH1 + prev_NH2 + prev_NH3 + prev_NH4
    ;    Put #SORTED_FNUM, , gPart_H.true_slice_count
    ;    Put #SORTED_FNUM, , H_HR_Timestamp
    ;    Put #SORTED_FNUM, , gPart_H.h_timing_word_hi
    ;    Put #SORTED_FNUM, , gPart_H.h_timing_word
    ;    Put #SORTED_FNUM, , CPI_triggerH


    ;        Put #SORTED_FNUM, , prev_image_data_v1
    ;    End If
    ;    If (gVPendingMultiPartCount >= 2) Then
    ;        Put #SORTED_FNUM, , prev_image_data_v2
    ;    End If
    ;    If (gVPendingMultiPartCount >= 3) Then
    ;        Put #SORTED_FNUM, , prev_image_data_v3
    ;    End If
    ;    If (gVPendingMultiPartCount = 4) Then
    ;        Put #SORTED_FNUM, , prev_image_data_v4
    ;    End If
    ;    Put #SORTED_FNUM, , image_data_v

    ; and the IDL basescan particle for reference:

    ; particle = {_particle_struct, $
    ;            flag  : uint(0), $
    ;            NHraw : uint(0), $
    ;            NVraw : uint(0), $
    ;            partID: uint(0), $
    ;            nwords: uint(0), $
    ;            nslice: uint(0), $
    ;            CPI_trigger: uint(0), $
    ;            image_data:ptr_new(), $
    ;            timestamp: hires_timestamp }


    print, ''
    print, 'write_multiparticle_to_SORTED start ', in_header, ' ', in_nmulti

    total_nwords = total(in_multi_array[0:in_nmulti-1].nwords)

    full_image = bytarr(total_nwords*2L)

    image_start = 0L
    image_stop  = 0L ;long(in_multi_array[0].nwords-1L)

    for i=0, in_nmulti-1 do begin

        image_stop = image_start + in_multi_array[i].nwords * 2L - 1L

        print, 'i = ', i, ' image_start = ', image_start, ' image_stop = ', image_stop

        if(image_stop gt 0L) then begin ; found a weird case where the first image was Nwords = 0, second was Nword = 2
            full_image[image_start:image_stop] = *in_multi_array[i].image_data ; could use pointers/temporary?
        endif else begin
            print,            'annomalous multi part image part ID: ', in_multi_array[i].partID, ' probe count: ', in_multi_array[i].timestamp.counter
            printf, !log_lun, 'annomalous multi part image part ID: ', in_multi_array[i].partID, ' probe count: ', in_multi_array[i].timestamp.counter
        endelse

        image_start += image_stop + 1L
    endfor

    ;stop


    writeu, in_olun, in_header,          $
        in_multi_array[in_nmulti-1].partID, $
        total_nwords, $
        in_multi_array[in_nmulti-1].nslice, $
        in_multi_array[in_nmulti-1].timestamp.year,   $ ; NOTE, the particle counter is in the IDL basescan timestamp,
        in_multi_array[in_nmulti-1].timestamp.month,  $ ; but split up in the VB, so to match SORTED we have to do this muck
        in_multi_array[in_nmulti-1].timestamp.day,    $ ; can certainly be cleaned up when View is being updated to Version 2.
        in_multi_array[in_nmulti-1].timestamp.hour,   $
        in_multi_array[in_nmulti-1].timestamp.minute, $
        in_multi_array[in_nmulti-1].timestamp.second, $
        in_multi_array[in_nmulti-1].timestamp.ms,     $
        in_multi_array[in_nmulti-1].timestamp.us,     $
        in_multi_array[in_nmulti-1].timestamp.ns,     $
        ulong(ishft(in_multi_array[in_nmulti-1].timestamp.counter and ulong64('0000FFFF00000000'X), -32)), $ ; .h_timing_word_hi
        ulong(      in_multi_array[in_nmulti-1].timestamp.counter and ulong64('00000000FFFFFFFF'X)      ), $ ; .h_timing_word
        in_multi_array[in_nmulti-1].CPI_trigger, $
        full_image



    print, 'write_multiparticle_to_SORTED exitting'
    print, ''

    ;stop
end ; write_multiparticle_to_SORTED



function calc_timediff_nanos, in_time1, in_time2
    ; calculates the time difference in nanoseconds between two hires_timestamp structures
    ; in_time1 and in_time2 are hires_timestamp structs
    ; returns a ulong64 which is the difference, in_time2 - in_time1, in nanoseconds

    ; first converts both dates to Julian for easy comparison that is year/month/day robust, but will not work correctly if the
    ; input years are more than 584 years appart, so don't try comparing data from 2014 with data from 1430 or from 2598, unless you use a different function


    ; Note ulong64 is sufficient to hold these values, since 584 years of nanoseconds can be held comfortably in a 64 bit number

    ; print, ulong64(365)*ulong64(24)*ulong64(3600)*ulong64(1e9)
    ; 31536000000000000


    ; printh,ulong64(365)*ulong64(24)*ulong64(3600)*ulong64(1e9)
    ; 0x007009D32DA30000

    ; print, ulong64('FFFFFFFFFFFFFFFF'X) / (ulong64(365)*ulong64(24)*ulong64(3600)*ulong64(1e9))
    ; 584


    juldate1 = calc_jul_date(in_time1.year, month=in_time1.month, in_time1.day)
    juldate2 = calc_jul_date(in_time2.year, month=in_time2.month, in_time2.day)

    if(juldate1 gt juldate2) then begin
        message, 'calc_timediff_nanos error, in_time2 is not greater than in_time1', in_time1, in_time2
        return, ulong64('FFFFFFFFFFFFFFFF'X)
    endif

    day_diff = juldate2 - juldate1

    julnanos2 = ulong64(day_diff) * 24ULL * 3600ULL * 1000000000ULL + $
        ulong64(in_time2.hour)    * 3600ULL * 1000000000ULL + $
        ulong64(in_time2.minute)  *   60ULL * 1000000000ULL + $
        ulong64(in_time2.second)            * 1000000000ULL + $
        ulong64(in_time2.ms)                *    1000000ULL + $
        ulong64(in_time2.us)                *       1000ULL + $
        ulong64(in_time2.ns)

    julnanos1 = ulong64(day_diff) * 24ULL * 3600ULL * 1000000000ULL + $
        ulong64(in_time1.hour)    * 3600ULL * 1000000000ULL + $
        ulong64(in_time1.minute)  *   60ULL * 1000000000ULL + $
        ulong64(in_time1.second)            * 1000000000ULL + $
        ulong64(in_time1.ms)                *    1000000ULL + $
        ulong64(in_time1.us)                *       1000ULL + $
        ulong64(in_time1.ns)

    timediff_nanos = julnanos2 - julnanos1

    return, timediff_nanos
end ; calc_timediff_nanos



function timestamp2string, in_timestamp
    ; generates a human readable string from a _timestamp_struct or a _hires_timestamp_struct

    ;    timestamp  = { _timestamp_struct, $
    ;                    year  : uint(0), $
    ;                    month : uint(0), $
    ;                    dow   : uint(0), $
    ;                    day   : uint(0), $
    ;                    hour  : uint(0), $
    ;                    minute: uint(0), $
    ;                    second: uint(0), $
    ;                    ms    : uint(0)  }

    ;    hires_timestamp = {_hires_timestamp_struct, $
    ;                        year   : uint(0), $
    ;                        month  : uint(0), $
    ;                        day    : uint(0), $
    ;                        hour   : uint(0), $
    ;                        minute : uint(0), $
    ;                        second : uint(0), $
    ;                        ms     : uint(0), $
    ;                        us     : uint(0), $
    ;                        ns     : uint(0), $
    ;                        counter: ulong64(0)}


    structure_type = tag_names(in_timestamp, /structure_name)

    if(strcmp(structure_type, '_timestamp_struct', /fold_case)) then begin
        time_string = string(format='(%"%04d-%02d-%02d %02d:%02d:%02d.%03d")', $
            in_timestamp.year, in_timestamp.month, in_timestamp.day, in_timestamp.hour, in_timestamp.minute, in_timestamp.second, in_timestamp.ms)
    endif else if(strcmp(structure_type, '_hires_timestamp_struct', /fold_case)) then begin
        time_string = string(format='(%"%04d-%02d-%02d %02d:%02d:%02d.%03d.%03d.%03d 0x%012X")', $
            in_timestamp.year, in_timestamp.month, in_timestamp.day, in_timestamp.hour, in_timestamp.minute, in_timestamp.second, in_timestamp.ms, in_timestamp.us, in_timestamp.ns, in_timestamp.counter)
    endif else begin
        time_string = 'print_timestamp, incorrect structure type: ' + structure_type
    endelse

    return, time_string
end ; timestamp2string



function timestamp2hkstring, in_timestamp
    ; generates a timestamp string as used in the HK_TXT files from a _timestamp_struct or a _hires_timestamp_struct
    ; HH:MM:SS: ms: us: ns
    ; 13:55:11:812:000:000

    ;    timestamp  = { _timestamp_struct, $
    ;                    year  : uint(0), $
    ;                    month : uint(0), $
    ;                    dow   : uint(0), $
    ;                    day   : uint(0), $
    ;                    hour  : uint(0), $
    ;                    minute: uint(0), $
    ;                    second: uint(0), $
    ;                    ms    : uint(0)  }

    ;    hires_timestamp = {_hires_timestamp_struct, $
    ;                        year   : uint(0), $
    ;                        month  : uint(0), $
    ;                        day    : uint(0), $
    ;                        hour   : uint(0), $
    ;                        minute : uint(0), $
    ;                        second : uint(0), $
    ;                        ms     : uint(0), $
    ;                        us     : uint(0), $
    ;                        ns     : uint(0), $
    ;                        counter: ulong64(0)}


    structure_type = tag_names(in_timestamp, /structure_name)

    if(strcmp(structure_type, '_timestamp_struct', /fold_case)) then begin
        time_string = string(format='(%"%02d:%02d:%02d:%03d:%03d:%03d")', $
            in_timestamp.hour, in_timestamp.minute, in_timestamp.second, in_timestamp.ms, 0, 0)
    endif else if(strcmp(structure_type, '_hires_timestamp_struct', /fold_case)) then begin
        time_string = string(format='(%"%02d:%02d:%02d:%03d:%03d:%03d")', $
            in_timestamp.hour, in_timestamp.minute, in_timestamp.second, in_timestamp.ms, in_timestamp.us, in_timestamp.ns)
    endif else begin
        time_string = 'print_timestamp, incorrect structure type: ' + structure_type
    endelse

    return, time_string
end ; timestamp2hkstring



pro write_HK_TXT, in_olun, in_timestamp, in_tas
    ; writes to the HK_TXT file, timestamp and TAS
    ; 13:55:12:812:000:000, 139.00000

    timestamp_string = timestamp2hkstring(in_timestamp)

    output_string = timestamp_string + string(format='(%", %9.5f")', in_tas)

    printf, in_olun, output_string
end ; write_HK_TXT



pro write_HK_BIN, in_olun, in_timestamp, in_hk_packet
    ; writes to the HK_BIN file, raw binary HK data
    ; !HK_HEADER
    ;    The HK_BIN format:
    ;    Put #SECOND_HK_FNUM, , HK_HEADER
    ;    Put #SECOND_HK_FNUM, , HK_HR_Timestamp
    ;    Put #SECOND_HK_FNUM, , hk_cpi

    ;    HK_HR_Timestamp:
    ;        year As Integer
    ;        month As Integer
    ;        day As Integer
    ;        hours As Integer
    ;        minutes As Integer
    ;        seconds As Integer
    ;        milliseconds As Integer
    ;        microseconds As Integer
    ;        nanoseconds As Integer

    ;    hk_cpi
    ;     'Public Type CPI_HOUSEKEEPING_TYPE
    ;        '    header As Integer
    ;        '    packet_length As Integer
    ;        '    data1(69) As Integer
    ;        '    timestamp_hi As Integer
    ;        '    timestamp_mid As Integer
    ;        '    timestamp_lo As Integer
    ;        '    TAS_lo As Integer
    ;        '    TAS_hi As Integer
    ;        '    data2(4) As Integer
    ;        '    checksum As Integer
    ;        'End Type

    out_timestamp = intarr(9)
    out_timestamp[0] = in_timestamp.year
    out_timestamp[1] = in_timestamp.month
    out_timestamp[2] = in_timestamp.day
    out_timestamp[3] = in_timestamp.hour
    out_timestamp[4] = in_timestamp.minute
    out_timestamp[5] = in_timestamp.second
    out_timestamp[6] = in_timestamp.ms
    out_timestamp[7] = in_timestamp.us
    out_timestamp[8] = in_timestamp.ns

    writeu, in_olun, out_timestamp, uint('484B'X), in_hk_packet
end ; write_HK_BIN



pro write_TT_marker, in_olun, in_timestamp, in_tas, in_counter
    ; writes to the double TAS header marker to file, these are used as flags at the start of each SORTED file

    ; ts_TAS is a _timestamp struct, not a hires timestamp
    ; CSng(TAS) is VB6 type "Single" -- apparently single precision Floating point
    ; kind of a hack from the Playback to use the timing_word-1, but will propagate it here, with hopes to eliminate this step in the future

    ;    Put #SORTED_FNUM, , TAS_HEADER_H    ' TAF 2013-10-30 Playback seems to be hitting this quite a bit...
    ;    Put #SORTED_FNUM, , CSng(TAS)
    ;    Put #SORTED_FNUM, , ts_TAS
    ;    Put #SORTED_FNUM, , gPart_H.h_timing_word_hi    ' TS48_hi(hk_timingword) ' gPart_H.h_timing_word_hi
    ;    Put #SORTED_FNUM, , gPart_H.h_timing_word - 1   ' TS48_lo(hk_timingword) ' gPart_H.h_timing_word

    ;    Put #SORTED_FNUM, , TAS_HEADER_V
    ;    Put #SORTED_FNUM, , CSng(TAS)
    ;    Put #SORTED_FNUM, , ts_TAS
    ;    Put #SORTED_FNUM, , gPart_H.h_timing_word_hi    ' TS48_hi(hk_timingword) ' gPart_H.h_timing_word_hi
    ;    Put #SORTED_FNUM, , gPart_H.h_timing_word - 1   ' TS48_lo(hk_timingword) ' gPart_H.h_timing_word

    ;    writeu, in_olun, !TAS_HEADER_H, $
    ;                     float(in_tas), $
    ;                     in_timestamp,  $
    ;                     uint(0),       $ ; some missalignment in the output, not sure why
    ;                     ulong(ishft(in_counter and ulong64('0000FFFF00000000'X), -32)), $ ; h_timing_word_hi
    ;                     ulong(     (in_counter and ulong64('00000000FFFFFFFF'X)) - 1 )    ; h_timing_word - 1

    ;    writeu, in_olun, !TAS_HEADER_V, $
    ;                     float(in_tas), $
    ;                     in_timestamp,  $
    ;                     uint(0),       $ ; some missalignment in the output, not sure why
    ;                     ulong(ishft(in_counter and ulong64('0000FFFF00000000'X), -32)), $ ; h_timing_word_hi
    ;                     ulong(     (in_counter and ulong64('00000000FFFFFFFF'X)) - 1 )    ; h_timing_word - 1


    writeu, in_olun, !TAS_HEADER_H, $
        float(in_tas), $
        in_timestamp.year,   $
        in_timestamp.month,  $
        in_timestamp.day,    $
        in_timestamp.hour,   $
        in_timestamp.minute, $
        in_timestamp.second, $
        in_timestamp.ms,     $
        in_timestamp.us,     $
        in_timestamp.ns,     $
        ulong(ishft(in_counter and ulong64('0000FFFF00000000'X), -32)), $ ; h_timing_word_hi
        ulong(     (in_counter and ulong64('00000000FFFFFFFF'X)) - 1 )    ; h_timing_word - 1

    writeu, in_olun, !TAS_HEADER_V, $
        float(in_tas), $
        in_timestamp.year,   $
        in_timestamp.month,  $
        in_timestamp.day,    $
        in_timestamp.hour,   $
        in_timestamp.minute, $
        in_timestamp.second, $
        in_timestamp.ms,     $
        in_timestamp.us,     $
        in_timestamp.ns,     $
        ulong(ishft(in_counter and ulong64('0000FFFF00000000'X), -32)), $ ; h_timing_word_hi
        ulong(     (in_counter and ulong64('00000000FFFFFFFF'X)) - 1 )    ; h_timing_word - 1

end ; write_TT_marker



function add_2_seconds, in_timestamp
    ; adds 2 seconds to a timestamp structure, does not adjust the counter in a hires_timestamp input

    ;    timestamp  = { _timestamp_struct, $
    ;                    year  : uint(0), $
    ;                    month : uint(0), $
    ;                    dow   : uint(0), $
    ;                    day   : uint(0), $
    ;                    hour  : uint(0), $
    ;                    minute: uint(0), $
    ;                    second: uint(0), $
    ;                    ms    : uint(0)  }

    ;    hires_timestamp = {_hires_timestamp_struct, $
    ;                        year   : uint(0), $
    ;                        month  : uint(0), $
    ;                        day    : uint(0), $
    ;                        hour   : uint(0), $
    ;                        minute : uint(0), $
    ;                        second : uint(0), $
    ;                        ms     : uint(0), $
    ;                        us     : uint(0), $
    ;                        ns     : uint(0), $
    ;                        counter: ulong64(0)}


    ; number of days in each month, used for time calculations in GenerateTimestamp, indexed based on 1, 0 index not used
    month_days = [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31] ; Feb = 28, except for leap year, handled below
    if((in_timestamp.year mod 4) eq 0) then begin ; handle most leap years
        month_days[2]++
    endif
    this_month_days = month_days[in_timestamp.month]

    new_timestamp = in_timestamp ; initialize the generated timestamp

    new_timestamp.second += 2

    if(new_timestamp.second ge 60) then begin
        new_timestamp.second -= 60
        new_timestamp.minute++
    endif

    if(new_timestamp.minute ge 60) then begin
        new_timestamp.minute -= 60
        new_timestamp.hour++
    endif

    if(new_timestamp.hour ge 24) then begin
        new_timestamp.hour -= 24
        new_timestamp.day++
    endif

    if(new_timestamp.day ge this_month_days) then begin
        new_timestamp.day = 1
        new_timestamp.month++
    endif

    if(new_timestamp.month ge 12) then begin
        new_timestamp.month = 1
        new_timestamp.year++
    endif

    return, new_timestamp
end ; add_2_seconds



function timestamp1_ge_timestamp2, in_time1, in_time2
    ; returns 1 if in_time1 occurs later than in_time2, 0 otherwise

    ; implemented for short timestamps only

    ;    timestamp  = { _timestamp_struct, $
    ;                    year  : uint(0), $
    ;                    month : uint(0), $
    ;                    dow   : uint(0), $
    ;                    day   : uint(0), $
    ;                    hour  : uint(0), $
    ;                    minute: uint(0), $
    ;                    second: uint(0), $
    ;                    ms    : uint(0)  }


    ; check the year/month/day first by converting to Julian

    jul_date1 = calc_jul_date(in_time1.year, month=in_time1.month, in_time1.day)
    jul_date2 = calc_jul_date(in_time2.year, month=in_time2.month, in_time2.day)

    if(jul_date1 gt jul_date2) then begin ; easy case, date of time1 is after time2
        return, 1
    endif else if(jul_date1 lt jul_date2) then begin ; easy case, date of time1 is before time2
        return, 0
    endif else begin ; harder case, dates must be equal, compare times
        ms_time1 = ulong64(in_time1.hour) * 3600ULL * 1000ULL + ulong64(in_time1.minute) * 60ULL * 1000ULL + ulong64(in_time1.second) * 1000ULL + ulong64(in_time1.ms)
        ms_time2 = ulong64(in_time2.hour) * 3600ULL * 1000ULL + ulong64(in_time2.minute) * 60ULL * 1000ULL + ulong64(in_time2.second) * 1000ULL + ulong64(in_time2.ms)

        if(ms_time1 ge ms_time2) then begin
            return, 1
        endif else begin
            return, 0
        endelse
    endelse
end ; timestamp1_ge_timestamp2



function generate_timestamp, in_counter, in_HK1_timestamp, in_HK2_timestamp, in_HK_timediff_nanos
    ; generates a high resolution timestamp using the pixel clock counter input, as well as two Housekeeping MS timestamps and counters
    ; uses the ratio of nanoseconds between the HKs to the pixel clock counter difference, then scales the output timestamp based on the
    ; input counter value

    ;print, 'HK1 ', HK1_timestamp
    ;print, 'HK2 ', HK2_timestamp
    ;print, 'in_counter', in_counter
    ;print, 'timediff_nanos ', calc_timediff_nanos(HK1_timestamp, HK2_timestamp)

    ; number of days in each month, used for time calculations in GenerateTimestamp, indexed based on 1, 0 index not used
    month_days = [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31] ; Feb = 28, except for leap year, handled below
    if((in_HK1_timestamp.year mod 4) eq 0) then begin ; handle most leap years
        month_days[2]++
    endif
    this_month_days = month_days[in_HK1_timestamp.month]

    new_timestamp = in_HK1_timestamp ; initialize the generated timestamp to the timebase from HK1

    ; check to make sure in_counter is after HK1 and before HK2

    if(in_counter le in_HK1_timestamp.counter or in_counter ge in_HK2_timestamp.counter or $ ; check for bracketting HK1 <= counter <= HK2
        in_HK1_timestamp.counter eq in_HK2_timestamp.counter) then begin                      ; chekc for inequality HK1 != HK2

        ;print,                'generate_timestamp error, in_counter is not bracketed by HK1 and HK2 ' + string(in_HK1_timestamp.counter) + string(in_counter) + string(in_HK2_timestamp.counter)
        ;printf, !log_lun, 'generate_timestamp error, in_counter is not bracketed by HK1 and HK2 ' + string(in_HK1_timestamp.counter) + string(in_counter) + string(in_HK2_timestamp.counter)
        new_timestamp.year   = 0
        new_timestamp.month  = 0
        new_timestamp.day    = 0
        new_timestamp.hour   = 0
        new_timestamp.minute = 0
        new_timestamp.second = 0
        new_timestamp.ms     = 0
        new_timestamp.us     = 0
        new_timestamp.ns     = 0

        return, new_timestamp
    endif

    HK_counter_diff = in_HK2_timestamp.counter - in_HK1_timestamp.counter ; the pixel clock difference between the HKs, note zero diff is checked above

    counts_after_HK1 = in_counter - in_HK1_timestamp.counter

    nanos_after_HK1 = in_HK_timediff_nanos * counts_after_HK1 / HK_counter_diff

    time_temp = nanos_after_HK1

    ; add the nanos_after_HK1 to the timebase

    new_timestamp.ns     += uint(time_temp mod 1000ULL)
    time_temp /= 1000ULL

    if(new_timestamp.ns ge 1000) then begin
        new_timestamp.ns -= 1000
        new_timestamp.us++
    endif

    new_timestamp.us     += uint(time_temp mod 1000ULL)
    time_temp /= 1000ULL

    if(new_timestamp.us ge 1000) then begin
        new_timestamp.us -= 1000
        new_timestamp.ms++
    endif

    new_timestamp.ms     += uint(time_temp mod 1000ULL)
    time_temp /= 1000ULL

    if(new_timestamp.ms ge 1000) then begin
        new_timestamp.ms -= 1000
        new_timestamp.second++
    endif

    new_timestamp.second += uint(time_temp mod   60ULL)
    time_temp /= 60ULL

    if(new_timestamp.second ge 60) then begin
        new_timestamp.second -= 60
        new_timestamp.minute++
    endif

    new_timestamp.minute += uint(time_temp mod   60ULL)
    time_temp /= 60ULL

    if(new_timestamp.minute ge 60) then begin
        new_timestamp.minute -= 60
        new_timestamp.hour++
    endif

    new_timestamp.hour   += uint(time_temp mod   24ULL)
    time_temp /= 24ULL

    if(new_timestamp.hour ge 24) then begin
        new_timestamp.hour -= 24
        new_timestamp.day++
    endif

    new_timestamp.day   += uint(time_temp mod ulong64(this_month_days))
    time_temp /= ulong64(this_month_days)

    if(new_timestamp.day ge this_month_days) then begin
        new_timestamp.day = 1
        new_timestamp.month++
    endif

    new_timestamp.month += uint(time_temp mod 12ULL) ; really should never get here
    time_temp /= 12ULL

    if(new_timestamp.month ge 12) then begin
        new_timestamp.month = 1
        new_timestamp.year++
    endif

    new_timestamp.year += uint(time_temp mod total(month_days))

    new_timestamp.counter = in_counter

    return, new_timestamp
end ; generate_timestamp



function read_raw_block
    ; reads one 16 byte timestamp and one 4096 byte data block, and a 2 byte checksum, returns 0 on success, and -1 on file size exceeded error
    common basescan_common, show_images, particle, timestamp, prev_sec, timestamp_printed, basefile_lun, file_size, $
        raw_data_buffer, raw_data_buffer_size, buffer_size, buffer_ptr, raw_data, hk_array, mask_array, mask_timestamp_array, $
        timestamp_index, timestamp_array_size, timestamp_array, hk_timestamp_array, hk_probecount_array, $
        probecount_index, probecount_array_size, probecount_array, iteration_counter, $
        HK1_timestamp, HK2_timestamp, HK_timediff_nanos, block_timestamp

    prev_sec = block_timestamp.second

    ; read the 16 byte MS timestamp
    ; test explicitly for end of file before each read
    point_lun, -basefile_lun, current_pos

    if(current_pos ge file_size) then begin
        print, 'at end of file, exiting'
        return, 1
    endif

    next_read_length = n_tags(block_timestamp, /data_length)
    if(current_pos + next_read_length gt file_size) then begin
        print, 'file size exceeded in read_raw_block'

        ;stop

        return, -1
    endif

    readu, basefile_lun, block_timestamp

    timestamp_array[timestamp_index] = block_timestamp
    timestamp_index++

    if(timestamp_index ge timestamp_array_size) then begin
        timestamp_array_size += !array_growth_factor

        empty_timestamp = timestamp
        empty_timestamp.year   = 0
        empty_timestamp.month  = 0
        empty_timestamp.day    = 0
        empty_timestamp.hour   = 0
        empty_timestamp.minute = 0
        empty_timestamp.second = 0
        empty_timestamp.ms     = 0

        new_timestamp_array = replicate(empty_timestamp, timestamp_array_size)
        new_timestamp_array[0:timestamp_index-1] = temporary(timestamp_array[0:timestamp_index-1])
        timestamp_array = temporary(new_timestamp_array)
    endif

    if(prev_sec ne block_timestamp.second) then begin
        timestamp_printed = 0
    endif

    ; read the 4098 byte data block
    ; test explicitly for end of file before each read
    point_lun, -basefile_lun, current_pos
    next_read_length = 2L*(size(raw_data))[1] + 2L ; note, + 2 for checksum
    if(current_pos + next_read_length gt file_size) then begin
        print, 'file size exceeded in read_raw_block'

        return, -1
    endif

    checksum = uint(0)

    readu, basefile_lun, raw_data, checksum

    point_lun, -basefile_lun, current_pos

    printf, !log_lun, format='(%"%04d-%02d-%02d, %02d:%02d:%02d.%03d file pointer: 0x%08X   %d")', $
        block_timestamp.year,                  $
        block_timestamp.month,                 $
        block_timestamp.day,                   $
        block_timestamp.hour,                  $
        block_timestamp.minute,                $
        block_timestamp.second,                $
        block_timestamp.ms,      $
        current_pos, current_pos

    if(buffer_size le raw_data_buffer_size - !DATA_BLOCK_LENGTH/2) then begin
        if(!emoflag eq 1) then begin
            print, '    read_raw_block appending, buffer_size = ', buffer_size, ' after will be: ', buffer_size + !DATA_BLOCK_LENGTH/2
        endif

        ; simple case, add the next data chunk to the circular buffer
        raw_data_buffer[buffer_size:buffer_size + !DATA_BLOCK_LENGTH/2 - 1] = raw_data

        buffer_size += !DATA_BLOCK_LENGTH/2
    endif else begin ; complex case, prepend the last particle data to the new raw_data, and start at the beginning of the buffer
        if(!emoflag eq 1) then begin
            print, '    read_raw_block restarting circular, buffer_size = ', buffer_size, ' after will be: ',((raw_data_buffer_size - buffer_ptr - 1) + !DATA_BLOCK_LENGTH/2) mod raw_data_buffer_size
        endif

        if(buffer_ptr eq raw_data_buffer_size - 1) then begin ; trivial case -- we're already at the end so just restart at the beginning
            raw_data_buffer[0:!DATA_BLOCK_LENGTH/2 - 1] = raw_data

            buffer_size = !DATA_BLOCK_LENGTH/2
        endif else begin ; complex case, copy the data from the end to the beginning, and append the new data
            read_length = buffer_size - buffer_ptr

            if(read_length eq 0) then begin ; we're at the end but the new block of raw data won't fit
                raw_data_buffer[0:!DATA_BLOCK_LENGTH/2 - 1] = raw_data

                buffer_size = !DATA_BLOCK_LENGTH/2
            endif else begin
                raw_data_buffer[0:read_length-1] = raw_data_buffer[buffer_ptr:buffer_ptr+read_length-1]
                raw_data_buffer[read_length:read_length + !DATA_BLOCK_LENGTH/2 - 1] = raw_data

                buffer_size = read_length + !DATA_BLOCK_LENGTH/2
            endelse
        endelse

        buffer_ptr = 0L
    endelse

    point_lun, -basefile_lun, current_pos

    if(current_pos ge file_size) then begin
        return, 1
    endif

    return, 0
end ; read_raw_block



function read_data, retval, size_to_read
    common basescan_common, show_images, particle, timestamp, prev_sec, timestamp_printed, basefile_lun, file_size, $
        raw_data_buffer, raw_data_buffer_size, buffer_size, buffer_ptr, raw_data, hk_array, mask_array, mask_timestamp_array, $
        timestamp_index, timestamp_array_size, timestamp_array, hk_timestamp_array, hk_probecount_array, $
        probecount_index, probecount_array_size, probecount_array, iteration_counter, $
        HK1_timestamp, HK2_timestamp, HK_timediff_nanos, block_timestamp
    ; reads (size(in_data))[1] of data from the raw data stream,
    ; using read_raw_block to fill the circular buffer and avoid the ms timestamps
    ; returns 1 if read_raw_block hits end of file, 0 for success
    ; on success, in_data will have the next available data from the base file, using the single global file pointer, buffer_ptr

    ; due to IDL pass by reference limitations, this function is a bit funky for now, maybe pointers?

    ;size_to_read = (size(in_data))[2] ; size is acting funny on struct element inputs, must be a way to detect them

    if(buffer_ptr + size_to_read ge buffer_size) then begin
        retval = read_raw_block()

        if(retval eq -1) then begin
            print, 'read_raw_block attempted to read past end of file, exiting' ; should likely deal with the particles in the last block???

            return, retval
        endif else if(retval eq 1) then begin
            print, 'read_raw_block found end of file, exiting' ; should likely deal with the particles in the last block???

            return, retval
        endif else if(retval lt 0) then begin
            print, 'read_raw_block returned: ', retval, ' in read_data buffer update'

            return, retval
        endif
    endif

    retval = 0 ; success

    orig_buffer_ptr = buffer_ptr ; need to use this for the return values

    buffer_ptr += size_to_read ;(buffer_ptr + size_to_read) mod buffer_size

    if(size_to_read eq 1) then begin
        return, raw_data_buffer[orig_buffer_ptr]
    endif else begin
        return, raw_data_buffer[orig_buffer_ptr:orig_buffer_ptr+size_to_read-1]
    endelse
end ; read_data



function test_read_data
    common basescan_common, show_images, particle, timestamp, prev_sec, timestamp_printed, basefile_lun, file_size, $
        raw_data_buffer, raw_data_buffer_size, buffer_size, buffer_ptr, raw_data, hk_array, mask_array, mask_timestamp_array, $
        timestamp_index, timestamp_array_size, timestamp_array, hk_timestamp_array, hk_probecount_array, $
        probecount_index, probecount_array_size, probecount_array, iteration_counter, $
        HK1_timestamp, HK2_timestamp, HK_timediff_nanos, block_timestamp
    ; reads the next value in the data buffer, but does not increment the buffer_ptr

    return, raw_data_buffer[buffer_ptr]
end ; test_read_data



;pro read_one_second_data, in_HK_index1, in_HK_index2
;    common basescan_common, show_images, particle, timestamp, prev_sec, timestamp_printed, basefile_lun, file_size, $
;                            raw_data_buffer, raw_data_buffer_size, buffer_size, buffer_ptr, raw_data, hk_array, mask_array, mask_timestamp_array, $
;                            timestamp_index, timestamp_array_size, timestamp_array, hk_timestamp_array, hk_probecount_array, $
;                            probecount_index, probecount_array_size, probecount_array, iteration_counter, $
;                            HK1_timestamp, HK2_timestamp, HK_timediff_nanos







;end ; read_one_second_data



pro basescan, basefilename, StartTime, EndTime, Startfpos, Endfpos  ; inputs  H50microns is 1 or zero from the read_hk data file. HK should be read first.
    Common HKdatablock, HKseconds,HKtas, num_hrz_dect, num_vrt_dect, pHKtimefpos, H50micron
    common basescan_common, show_images, particle, timestamp, prev_sec, timestamp_printed, basefile_lun, file_size, $
        raw_data_buffer, raw_data_buffer_size, buffer_size, buffer_ptr, raw_data, hk_array, mask_array, mask_timestamp_array, $
        timestamp_index, timestamp_array_size, timestamp_array, hk_timestamp_array, hk_probecount_array, $
        probecount_index, probecount_array_size, probecount_array, iteration_counter, $
        HK1_timestamp, HK2_timestamp, HK_timediff_nanos, block_timestamp

    setup_globals

    ;if(keyword_set(show_images_input) or !show_images or !showimg) then begin ; NOTE !showimg defined in a3vcpiview_browse_images.pro -- these should be cleaned up and unified
    if(keyword_set(show_images_input) or !show_images) then begin ; NOTE !showimg defined in a3vcpiview_browse_images.pro -- these should be cleaned up and unified
        show_images = 1
    endif else begin
        show_images = 0
    endelse

    ; binary reverse the 128 bit manual mask
    ; user entry and Realtime 2.X and IDL storage are backwards -- this allows user readable order in both programs
    if(!manual_mask) then begin
        ; reverse the 8x 16 bit uint array order
        !manual_H_mask = reverse(!manual_H_mask)
        !manual_V_mask = reverse(!manual_V_mask)

        ; bit reverse each binary 16 bit uint
        !manual_H_mask[*] = bit_reverse_16(!manual_H_mask[*])
        !manual_V_mask[*] = bit_reverse_16(!manual_V_mask[*])
    endif

    empty_timestamp = timestamp ; used for generating structures and resizing arrays
    empty_particle  = particle

    temp_uint = 0U

    timestamp_index = 0L
    timestamp_array_size = !array_growth_factor
    timestamp_array = replicate(empty_timestamp, timestamp_array_size)

    probecount_index = 0L
    probecount_array_size = !array_growth_factor
    probecount_array = ulon64arr(probecount_array_size)

    harray_size = !array_growth_factor
    varray_size = !array_growth_factor
    hprobecount_array = ulon64arr(harray_size)
    vprobecount_array = ulon64arr(varray_size)
    hpartcount = 0L
    vpartcount = 0L

    h_particle_index = 0L
    h_particle_array_size = !particle_array_size
    h_particle_array = replicate(empty_particle, h_particle_array_size)

    v_particle_index = 0L
    v_particle_array_size = !particle_array_size
    v_particle_array = replicate(empty_particle, v_particle_array_size)

    current_H_mask = uintarr(8)
    current_V_mask = uintarr(8)
    current_H_mask[*] = 'FFFF'X    ; initialize mask bits to all 1's, ie not masked
    current_V_mask[*] = 'FFFF'X

    if(!manual_mask) then begin ; AND the manual masks with the current_*_masks to generate composite masks for bit filling
        current_H_mask = !manual_H_mask and current_H_mask
        current_V_mask = !manual_V_mask and current_V_mask
    endif

    current_mask = -1

    hpos = !disp_window_ys - 1
    vpos = !disp_window_ys - 1

    inword = uint(0) ; read 2 bytes
    Hwords = uint(0)
    Vwords = uint(0)

    total_Hpart     = 0L
    total_Vpart     = 0L
    total_Hmulti    = 0L
    total_Vmulti    = 0L
    total_Hoverflow = 0L
    total_Voverflow = 0L
    total_Hcpi_trigger = 0L
    total_Vcpi_trigger = 0L
    total_Hflag_bit13  = 0L
    total_Vflag_bit13  = 0L
    total_Hflag_bit12  = 0L
    total_vflag_bit12  = 0L
    total_NH_and_NV_ne_zero = 0L
    total_NULL_count   = 0L
    total_reject_particle_count = 0L

    H_particles_written = 0L
    V_particles_written = 0L
    H_overflow_written = 0L
    V_overflow_written = 0L
    nothing_to_write = 0L

    Hpart_this_second = 0L
    Vpart_this_second = 0L
    Hmulti_this_second = 0L
    Vmulti_this_second = 0L
    Hoverflow_this_second = 0L
    Voverflow_this_second = 0L
    NULL_this_second = 0L

    probecountflag_set = 0 ; debug flag for stopping at specified probecount values

    probecount = ulong64(0)

    nHmulti = 0L
    Hmulti_flag = 0
    stored_Hflags = 0U
    Hmulti_ID = uint(0)
    Hmulti_array_size = 10
    Hmulti_array = replicate(particle, Hmulti_array_size)

    nVmulti = 0L
    Vmulti_flag = 0
    stored_Vflags = 0U
    Vmulti_ID = 0U
    Vmulti_array_size = 10
    Vmulti_array = replicate(particle, Vmulti_array_size)



    ; MAIN FILE OPEN

    infilename=basefilename
    infilepath=file_dirname(infilename)+'\'

    stripped_infilename = file_basename(infilename); the file name with the path removed, for generating the output file names

    ; get the starttime end time string to make a direct for the processd time period.
    StartTimestring=SPEC_seconds2HHMMSS(StartTime)
    EndTimeString  =SPEC_seconds2HHMMSS(EndTime)

    outfilepath = infilepath+strmid(stripped_infilename, 0, strlen(stripped_infilename)-7)+'_'+StartTimestring+'_'+EndTimestring+'\'
    FILE_MKDIR, outfilepath
    print, ''
    print, 'basescan opening file: ', stripped_infilename
    print, ''
    print, 'Scan starting at:'
    print, systime()
    print, ''

    logfile_name = outfilepath + 'basescan_out_log.txt'
    openw, lun, /get, logfile_name
    !log_lun = lun

    printf, !log_lun, ''
    printf, !log_lun, 'basescan opening file: ', stripped_infilename
    printf, !log_lun, ''
    printf, !log_lun, 'Scan starting at:'
    printf, !log_lun, systime()
    printf, !log_lun, ''

    startseconds = systime(/seconds)

    openr, lun, /get, infilename

    infilefstat = fstat(lun)

    file_size = infilefstat.size

    basefile_lun = lun

    outfilename = outfilepath + 'basescan_out_' + stripped_infilename + '.csv'

    openw,  olun, /get, outfilename
    printf, olun, 'YYYY-MM-DD, HH:MM:SS.mmm, Hex Time, H Part, V Part, H Overflow, V Overflow, NULL Packets'

    outfilename2 = outfilepath + 'basescan_out_' + stripped_infilename + '_hoverload.csv'

    openw , olun2, /get, outfilename2
    printf, olun2, 'HOVERLOAD Count, partID, (from IDL basescan)'

    if(!reprint_raw_data) then begin
        outfilename3 = outfilepath + 'basescan_out_reprint_' + stripped_infilename + '.csv'

        openw,  olun3, /get, outfilename3
        printf, olun3, 'Header, NHraw,  NVraw,  partID, nslices, time stamp,             NH, H flags,     NV, V flags,   NULL size'
    endif

    n_hk = (size(hk_probecount_array))[1]

    if(H50micron eq 1) then begin
        !H_resolution_microns = 50
    endif else begin
        !H_resolution_microns = 10
    endelse

    define_pixel_resolutions

    ; set up the image display window, and initialize the separator bar and mask display variables
    if(show_images) then begin
        ; scale the image display window for the selected zoom, and 50um to scale settings
        if(!draw_50um_to_scale and !H_resolution_microns eq 50) then begin
            !disp_window_xs =  1 + 6 * !PIXELS * !zoom_factor
            H_horz_bar = bytarr(3, 5 * !PIXELS * !zoom_factor, 1, /nozero)
            H_horz_bar = reform(H_horz_bar, 3, 5 * !PIXELS * !zoom_factor, 1)
        endif else begin
            !disp_window_xs =  1 + 2 * !PIXELS * !zoom_factor
            H_horz_bar = bytarr(3,     !PIXELS * !zoom_factor, 1, /nozero)
            H_horz_bar = reform(H_horz_bar, 3,     !PIXELS * !zoom_factor, 1)
        endelse

        V_horz_bar = bytarr(3, !PIXELS * !zoom_factor, 1, /nozero)
        H_horz_bar[*, *, *] = !line_color
        V_horz_bar[*, *, *] = !line_color

        ; for some reason bytarr is skipping the single dimension, so reform must be used
        V_horz_bar = reform(V_horz_bar, 3, !PIXELS * !zoom_factor, 1)


        window, !disp_window, xs=!disp_window_xs, ys=!disp_window_ys

        ; draw the vertical separator line between H and V images
        vert_bar = bytarr(3, 1, !disp_window_ys, /nozero)
        vert_bar[*, *, *] = !line_color

        tv, vert_bar, !disp_window_xs - !PIXELS*!zoom_factor - 1, 0, /true
    endif

    first_hk_timestamp = hk_probecount_array[0]
    last_hk_timestamp = hk_probecount_array[n_hk-1]

    print, ''
    print, 'HK start/end times for file: ', infilename
    print, timestamp2string(hk_timestamp_array[0])
    print, timestamp2string(hk_timestamp_array[n_hk-1])
    printh, first_hk_timestamp
    printh,  last_hk_timestamp
    print, ''

    printf, !log_lun, ''
    printf, !log_lun, 'HK start/end times for file: ', infilename
    printf, !log_lun, timestamp2string(hk_timestamp_array[0])
    printf, !log_lun, timestamp2string(hk_timestamp_array[n_hk-1])
    printf, !log_lun, ''

    point_lun,  basefile_lun, 0
    point_lun, -basefile_lun, current_pos

    disable_H = 0

    if(show_images) then begin
        wset, !disp_window
    endif

    prev_Hprobecount = 0ULL
    prev_Vprobecount = 0ULL

    counter = 0L
    iteration_counter = counter
    done = 0

    ; find the begining fpos and ending fpos for the specific time period.


    ; find a workable start of file loop
    ; could likely use more error handling and make this robust enough to deal with any broken or truncated file...

    current_pos =Startfpos
    fpos=current_pos

    ;need to assign the HK1_timesamp and HK2_time-stamp for the first sorted file's filename.

    index=where(HKseconds ge startTime,contG);  ; startTime correspond to the startfpos.

    if contG gt 0 then hk_index=index(0) else hk_index=0

    HK1_timestamp.year    = hk_timestamp_array[hk_index].year
    HK1_timestamp.month   = hk_timestamp_array[hk_index].month
    HK1_timestamp.day     = hk_timestamp_array[hk_index].day
    HK1_timestamp.hour    = hk_timestamp_array[hk_index].hour
    HK1_timestamp.minute  = hk_timestamp_array[hk_index].minute
    HK1_timestamp.second  = hk_timestamp_array[hk_index].second
    HK1_timestamp.ms      = hk_timestamp_array[hk_index].ms
    HK1_timestamp.us      = 0
    HK1_timestamp.ns      = 0
    HK1_timestamp.counter = hk_probecount_array[hk_index]

    HK2_timestamp.year    = hk_timestamp_array[hk_index+1].year
    HK2_timestamp.month   = hk_timestamp_array[hk_index+1].month
    HK2_timestamp.day     = hk_timestamp_array[hk_index+1].day
    HK2_timestamp.hour    = hk_timestamp_array[hk_index+1].hour
    HK2_timestamp.minute  = hk_timestamp_array[hk_index+1].minute
    HK2_timestamp.second  = hk_timestamp_array[hk_index+1].second
    HK2_timestamp.ms      = hk_timestamp_array[hk_index+1].ms
    HK2_timestamp.us      = 0
    HK2_timestamp.ns      = 0
    HK2_timestamp.counter = hk_probecount_array[hk_index+1]

    ; since read_hk_3v was for some reason butchered and functionality copied insted of using it as a procedure as it was written, these was no longer being defined
    hk_count = !hk_count
    mask_count = !mask_count

    start_second = hk_timestamp_array[0].hour*3600L          + hk_timestamp_array[0].minute*60L          + hk_timestamp_array[0].second
    end_second = hk_timestamp_array[hk_count-1].hour*3600L + hk_timestamp_array[hk_count-1].minute*60L + hk_timestamp_array[hk_count-1].second
    nseconds = calc_timediff_seconds(hk_timestamp_array[0], hk_timestamp_array[hk_count-1])+1
    current_second = start_second


    ;while  (fpos ge startfpos and fpos le endfpos ) do begin
    while(~EOF(basefile_lun) and ~done) do begin

        readu, lun, inword ; line up with the first MS timestamp start, should streamline this and put it into read_raw_block

        if(inword ge '07C0'XU and inword le '0800'XU) then begin ; years 1984 through 2048
            ; found first MS 16 byte timestamp

            ; backup file pointer to start of timestamp, then call read_raw_block to get the timestamp, and the data
            point_lun, -basefile_lun, current_pos
            point_lun,  basefile_lun, current_pos-2

            retval = read_raw_block()

            if(retval eq 1) then begin
                print, 'read_raw_block found end of file, exiting'

                done = 1
            endif else if(retval lt 0) then begin
                print, 'read_raw_block returned: ', retval, ' in initial buffer read'
                print, 'exiting'
                return
            endif

            done = 1
        endif
    endwhile

    file_split_particle_counter1 = 0L
    file_split_particle_counter2 = 0L
    extending_file_2_seconds = 0
    start_time = HK1_timestamp
    end_time = HK1_timestamp

    if(!generate_SORTED) then begin
        ; Generate the SORTED file names, open the first file, and print the header.
        ; The SORTED file names here are generated from the timestamp information, NOT from the filename


        ;tms_h = "Date YYYY-MM-DD, Time hh:mm:ss.ms, hex timestamp, HK H counts, HK V counts, HK H Number Overloads,
        ;, HK V Number Overloads, H count, V count, Mask count, TAS count, H Multi count, V Multi count, H Overload count, V Overload count, Realtime Version " & App.Major & "." & App.Minor & "." & App.Revision
        ; Print #SORTED_TMS_FNUM, tms_h

        print, 'infilepath: ', infilepath
        print, 'stripped_infilename: ', stripped_infilename

        ;   base130909131733.2DSCPI
        ; SORTED130909131733_131743_135159_0000.2DSCPI

        print, 'first timestamp: ', timestamp2string(HK1_timestamp)

        SORTED_file_record_number = 0L ; the last four digits of the filename _0000, incremented with split files

        main_time_base = HK1_timestamp
        fSfn_timestamp = HK1_timestamp ; store the timestamp for renaming the SORTED file when it is split

        namestr =string(format='(%"%02d%02d%02d%02d%02d%02d_%02d%02d%02d_%04d.2DSCPI")', $
            main_time_base.year-2000, main_time_base.month, main_time_base.day, $
            main_time_base.hour, main_time_base.minute, main_time_base.second,  $
            main_time_base.hour, main_time_base.minute, main_time_base.second,  $
            SORTED_file_record_number)

        full_first_SORTED_filename = outfilepath + 'SORTED'+namestr
        full_first_HK_TXT_filename = outfilepath + 'HK_TXT'+namestr
        full_first_HK_BIN_filename = outfilepath + 'HK_BIN'+namestr

        print, ''
        print, 'first_SORTED_filename: ', full_first_SORTED_filename
        print, ''

        openw, olun_S1, /get, full_first_SORTED_filename
        openw, olun_HK_TXT1, /get, full_first_HK_TXT_filename
        openw, olun_HK_BIN1, /get, full_first_HK_BIN_filename

        ; Headers are all 256 bytes, padded with spaces after the information
        ; header HK_BIN...  Ver 3.00 HK BINARY DATA (no 50 micron flag)
        ; header HK_TXT...  Ver 3.00 HK TEXT DATA   (no 50 micron flag)
        ; header SORTED...  Ver 3.00 SORTED PARTICLE DATA   Playback Version 1.1.25 H50Micron = True [False]

        HK_BIN_header_string = 'Ver 3.00 HK BINARY DATA         basescan Version ' + !basescan_version
        HK_TXT_header_string = 'Ver 3.00 HK TEXT DATA           basescan Version ' + !basescan_version
        SORTED_header_string = 'Ver 3.00 SORTED PARTICLE DATA   basescan Version ' + !basescan_version

        SORTED_header_string = SORTED_header_string + ' H50Micron = ' ; View does a bad job of detecting H50Micron, must be at exactly the right spacing... need to fix that!

        if(!H_resolution_microns eq 50) then begin
            SORTED_header_string = SORTED_header_string + 'True'
        endif else begin
            SORTED_header_string = SORTED_header_string + 'False'
        endelse

        ; pad the headers with spaces to 256 bytes
        pad = bytarr((!header_length-strlen(SORTED_header_string)))
        pad[*] = '20'X
        SORTED_header_string = SORTED_header_string + string(pad)

        pad = bytarr((!header_length-strlen(HK_TXT_header_string)))
        pad[*] = '20'X
        HK_TXT_header_string = HK_TXT_header_string + string(pad)

        pad = bytarr((!header_length-strlen(HK_BIN_header_string)))
        pad[*] = '20'X
        HK_BIN_header_string = HK_BIN_header_string + string(pad)

        writeu, olun_S1, SORTED_header_string

        ; fragments of the HK_TXT header, which is too long to fit on a single line in IDL, should finish this, but let's see if the View can deal with fragments for a start
        frag1 = 'TIME, TAS'
        frag2 = 'TIME, TAS'
        frag3 = 'TIME, TAS'
        frag4 = 'TIME, TAS'

        writeu, olun_HK_TXT1, HK_TXT_header_string
        printf, olun_HK_TXT1, ''
        printf, olun_HK_TXT1, frag1
        printf, olun_HK_TXT1, frag2
        printf, olun_HK_TXT1, frag3
        printf, olun_HK_TXT1, frag4

        writeu, olun_HK_BIN1, HK_BIN_header_string

        ; initialize the SORTED file with double TAS packets
        hk_tas = two_uints2float(hk_array[0, 74], hk_array[0, 75])

        write_TT_marker, olun_S1, HK1_timestamp, hk_tas, HK1_timestamp.counter
    endif ; !generate_SORTED

    first_Hoverflow_flag = 0
    second_Hoverflow_flag = 0
    first_Voverflow_flag = 0
    second_Voverflow_flag = 0

    done = 0

    ; main processing loop -- scans through the base file to the end

    ; two main modes:  fill and process

    ; fill fills the H and V arrays with particles, up to just after HK2 time

    ; process loops through all H and V between HK1 and HK2, generating timestamps and writing data

    while (current_pos le endfpos) do begin
        counter++
        iteration_counter = counter

        if(counter mod 10000L eq 0L) then begin
            ; print out current file completion status, also wait so that breaks can be input

            point_lun, -basefile_lun, current_pos

            ;print,                format='(%"iteration: %8d,  0x%08X %s 0x%08X,  %7.2f%s read")', counter, current_pos, '/', file_size, double(current_pos)/double(file_size) * 100., '%'
            ;printf, !log_lun, format='(%"iteration: %8d,  0x%08X %s 0x%08X,  %7.2f%s read")', counter, current_pos, '/', file_size, double(current_pos)/double(file_size) * 100., '%'

            last_time_string = timestamp2string(HK1_timestamp)

            print,            format='(%"iteration: %8d,  0x%08X %s 0x%08X,  %7.2f%s read,    %s")', $
                counter, current_pos, '/', file_size, double(current_pos)/double(file_size) * 100., '%', last_time_string
            printf, !log_lun, format='(%"iteration: %8d,  0x%08X %s 0x%08X,  %7.2f%s read,    %s")', $
                counter, current_pos, '/', file_size, double(current_pos)/double(file_size) * 100., '%', last_time_string
            ; print, format='(%"%0X")', current_pos
            wait, .001
        endif

        ; this was printing the 1Hz from the read_raw_block timestamp (now block_timestamp)
        ; need to do this printing 1Hz from the HK1/HK2 update section instead, then delete these comments
        ;        ; print 1Hz data to file, and zero 1Hz counters
        ;        if(~timestamp_printed) then begin ; TAF primitive but quick change detected
        ;            timestamp_printed = 1

        ;            ; print, format='(%"%04d-%02d-%02d, %02d:%02d:%02d.%03d, %5d, %5d, %5d, %5d")', year, month, day, hour, minute, second, ms, Hpart_this_second, Vpart_this_second, Hoverflow_this_second, Voverflow_this_second
        ;            ; print, format='(%"%04d-%02d-%02d, %02d:%02d:%02d.%03d")', year, month, day, hour, minute, second, ms

        ;            printf, olun, format='(%"%04d-%02d-%02d, %02d:%02d:%02d.%03d, 0x%016X, %5d, %5d, %5d, %5d, %5d")', $
        ;                          timestamp.year,                  $
        ;                          timestamp.month,                 $
        ;                          timestamp.day,                   $
        ;                          timestamp.hour,                  $
        ;                          timestamp.minute,                $
        ;                          timestamp.second,                $
        ;                          timestamp.ms,                    $
        ;                          probecount,            $
        ;                          Hpart_this_second,     $
        ;                          Vpart_this_second,     $
        ;                          Hoverflow_this_second, $
        ;                          Voverflow_this_second, $
        ;                          NULL_this_second

        ;            ;print, "H =", Hpart_this_second, "     V =", Vpart_this_second,   "     H OF =", Hoverflow_this_second, "     V OF=", Voverflow_this_second

        ;                Hpart_this_second = 0L
        ;                Vpart_this_second = 0L
        ;               Hmulti_this_second = 0L
        ;               Vmulti_this_second = 0L
        ;            Hoverflow_this_second = 0L
        ;            Voverflow_this_second = 0L
        ;                 NULL_this_second = 0L

        ;            ;stop
        ;            ;wait, .001
        ;        endif

        prev_particle = particle

        particle.flag        = 0
        particle.NHraw       = 0
        particle.NVraw       = 0
        particle.partID      = 0
        particle.nwords      = 0
        particle.nslice      = 0
        particle.CPI_trigger = 0
        particle.type = 'U' ; U for Unitialized
        ;particle.image_data = ptr_new()
        particle.timestamp.year    = 0
        particle.timestamp.month   = 0
        particle.timestamp.day     = 0
        particle.timestamp.hour    = 0
        particle.timestamp.minute  = 0
        particle.timestamp.second  = 0
        particle.timestamp.ms      = 0
        particle.timestamp.us      = 0
        particle.timestamp.ns      = 0
        particle.timestamp.counter = 0

        NH = 0
        NV = 0

        Hflags = 0
        Vflags = 0

        reject_particle = 0 ; initialize to accept, conditions below may set this to 1 to prevent writing a particle to output, or adversely affecting program state
        particle.flag   = read_data(retval, 1)

        if(particle.flag ne uint('3253'X) and particle.flag ne uint('4E4C'X)) then begin ; 2S or NL
            print, 'particle.flag incorrect'
            print,  particle.flag
            printh, particle.flag

            printf, !log_lun, 'particle.flag incorrect ', particle.flag

            ;stop

            ;while(particle.flag ne uint('3253'X) and particle.flag ne uint('4E4C'X)) do begin
            ;    if(buffer_ptr ge 2) then begin
            ;        print, 'backing up'

            ;        buffer_ptr -= 2
            ;    endif else begin
            ;        print, 'fetching new block'

            ;        retval = read_raw_block()
            ;    endelse

            ;    particle.flag   = read_data(retval, 1)
            ;endwhile

            print, buffer_ptr
            while(particle.flag ne uint('3253'X) and particle.flag ne uint('4E4C'X)) do begin ; 2S or NL
                buffer_ptr++

                particle.flag   = read_data(retval, 1)
            endwhile
            print, buffer_ptr

            ;stop
        endif

        ; check the timestamps and increment the mask, if needed
        block_second = block_timestamp.hour*3600L + block_timestamp.minute*60 + block_timestamp.second
        if(block_second gt current_second) then begin
            ; check the mask timestamp and update mask variables if a new mask is needed
            if(current_mask lt mask_count-1) then begin ; make sure there are more masks available
                if(calc_timediff_seconds(mask_timestamp_array[current_mask+1], block_timestamp) ge 0) then begin
                    current_mask++

                    current_H_mask = mask_array[current_mask,  3:10]
                    current_V_mask = mask_array[current_mask, 11:18]

                    if(!manual_mask) then begin ; AND the manual masks with the current_*_masks to generate composite masks for bit filling
                        current_H_mask = !manual_H_mask and current_H_mask
                        current_V_mask = !manual_V_mask and current_V_mask
                    endif
                endif
            endif

            current_second = block_second
        endif

        particle.NHraw  = read_data(retval, 1)
        particle.NVraw  = read_data(retval, 1)
        particle.partID = read_data(retval, 1)
        particle.nslice = read_data(retval, 1)

        if(particle.flag eq uint('3253'X)) then begin ; flag 2S, 0x3253, 12883 decimal, particle flag
            ; '2S' flag, particle detected

            ; parse out the NH and NV to determine next read lengths
            ; TAF NOTE: 2013-12-05 need to do some error checking on the dynamic read lengths
            ; also identify all flags in NH, NV

            NH = particle.NHraw and uint('0FFF'X)
            NV = particle.NVraw and uint('0FFF'X)

            Hflags = particle.NHraw and uint('F000'X)
            Vflags = particle.NVraw and uint('F000'X)

            ; print, format='(%"0x %04X %04X %04X %04X %04X")', flag, NHraw, NVraw, partID, nslice
            ; print, format='(%"%d %d 0x%04X 0x%04X")', NH, NV, NVraw, Hflags, Vflags

            if((NH gt 0) and (NV gt 0)) then begin
                ;print, 'major error, NH and NV both greater than 0, skip garbage particle'
                total_NH_and_NV_ne_zero++

                ;point_lun, -lun, current_pos
                ;print, format='(%"current_pos: 0x%0X")', current_pos

                ;print, format='(%"0x %04X %04X %04X %04X %04X")', particle.flag, particle.NHraw, particle.NVraw, particle.partID, particle.nslice
                ;print, format='(%"%d %d 0x%04X 0x%04X")', NH, NV, Hflags, Vflags

                ;stop
            endif else begin
                ; Frame types: 2S normal particle,
                ;              exceeded frame size image -- seems to be same as particle, but maybe lacking timing words???
                ;              overflow start, overflow finish


                ; parse H or V particle or overflow [NOTE: currently no H xor V]

                if((Hflags and uint('8000'X)) ne '0'X) then begin
                    ;print, 'Hflag: 0x8000 bit 15 H FIFO overflow'
                    Hoverflow_this_second++
                    total_Hoverflow++

                    ;print, format='(%"0x%04X   %d     %d")', partID, partID, 2L^16 - partID
                    ;stop

                    ; this business with "tempID" is to match the poorly formatted VB output
                    if(particle.partID gt 2L^16/2-1) then begin
                        tempID = 2L^16 - particle.partID
                        tempID = -tempID
                    endif else begin
                        tempID = particle.partID
                    endelse

                    printf, olun2, format='(%"%7d, %7d, %04X")', total_Hoverflow, tempID, particle.partID

                    ;print, 'H overflow, partID = ', particle.partID

                    if(first_Hoverflow_flag eq 0) then begin
                        first_Hoverflow_flag = 1
                    endif else begin
                        first_Hoverflow_flag = 0
                        second_Hoverflow_flag = 1  ; NOTE, second overflow may have an image!!!!
                    endelse
                endif

                if(NH gt 0) then begin ; this is a horizontal particle
                    ; parse the flags
                    if((Hflags and uint('4000'X)) ne uint('0000'X)) then begin
                        ;print, 'Hflag: 0x4000 bit 14 unused, should be 0'
                        particle.CPI_trigger = 1

                        total_Hcpi_trigger++
                    endif
                    if((Hflags and uint('2000'X)) ne uint('0000'X)) then begin
                        ;print, 'Hflag: 0x2000 bit 13 unused, should be 0'
                        total_Hflag_bit13++
                    endif
                    if((Hflags and uint('1000'X)) ne uint('0000'X)) then begin
                        ; multi-packet particle
                        ;print, 'Hflag: 0x1000 bit 12 particle words exceeded max frame size'
                        total_Hflag_bit12++

                        ;                        if(Hmulti_flag eq 1) then begin
                        ;                            if(Hmulti_ID ne particle.partID) then begin ; this is a second multi-packet image, with a different particle ID, write out the first and clear
                        ;                                ;stop

                        ;                                if(nHmulti lt 2) then begin
                        ;                                    print, 'nHmulti too low 1 ', nHmulti
                        ;                                endif else begin
                        ;                                    write_multiparticle_to_SORTED, olun_S1, !H_PARTICLE_HEADER, nHmulti, Hmulti_array
                        ;                                endelse

                        ;                                stored_Hflags = uint(0)

                        ;                                ; reset variables and clear the multi_array -- can/should this be done in the function call?
                        ;                                for i=0, Hmulti_array_size-1 do begin
                        ;                                    ptr_free, Hmulti_array[i].image_data
                        ;                                endfor

                        ;                                nHmulti = 0
                        ;                            endif
                        ;                        endif

                        ;                        ; start, continue, or second multi-packet particle
                        ;                        ; the finished particle gets added to the multi_array below
                        ;                        Hmulti_flag = 1 ; signal to accumulate Hmulti packets
                        ;                        Hmulti_ID = particle.partID
                    endif

                    ;                    if(Hmulti_flag eq 1) then begin
                    ;                        stored_Hflags = Hflags
                    ;                    endif

                    total_Hpart++
                    Hpart_this_second++

                    Hwords = NH - 3 ; NH is number of words plus 3 timing words  ; TAF 2013-12-24 NH -3 or -2???
                    Vwords = uint(0)

                    particle.nwords = Hwords

                    if(NH gt 3 and Hwords gt 0) then begin ; and Hwords+buffer_ptr lt raw_data_buffer_size) then begin
                        ;Himage_data2 = uintarr(Hwords)

                        Himage_data2 = read_data(retval, Hwords)

                        if(retval eq 1) then begin
                            done = 1
                            break
                        endif

                        ;readu, lun, Himage_data

                        ; should likely do a check on the buffer size before reading from it here
                        ;Himage_data2[0:Hwords-1] = raw_data_buffer[buffer_ptr:buffer_ptr+Hwords-1]

                        Himage_data = bytarr(Hwords*2L)

                        ; TAF convert the image words to bytes
                        for i=0L, Hwords-1L do begin
                            Himage_data[i*2L+1] = ishft((Himage_data2[i] and 'FF00'X), -8)
                            Himage_data[i*2L]   =        Himage_data2[i] and '00FF'X
                        endfor

                        ptr_free, particle.image_data
                        particle.image_data = ptr_new(Himage_data, /no_copy)
                    endif else begin
                        ; NH = 0, usually a deadtime begin/end
                    endelse

                    ;disable_H = 1

                    if((show_images eq 1 or !generate_SORTED) and particle.nslice gt 0 and particle.nslice lt 32000 and ~disable_H) then begin
                        Area = 0UL
                        Midpoint = 0
                        Edges = 0
                        L3 = 0.0
                        L4 = 0
                        L5 = 0
                        PC3x = 0.0
                        PC3y = 0.0
                        PC4x = 0.0
                        PC4y = 0.0
                        A3 = 0.0
                        Roundness1 = 0.0
                        Roundness2 = 0.0
                        Roundness3 = 0.0
                        L1r = 0.0
                        maxLength = 0.0
                        maxWidth  = 0.0
                        perimeter = 0UL
                        perimeterum = 0.
                        csp = 0.
                        num_satellites = 0.
                        Err=0

                        Himage = bytarr(!PIXELS, particle.nslice)

                        HorV = 'H'

                        ;Himage_data = *particle.image_data

                        processparticle, *particle.image_data, Himage, Hwords, particle.nslice, Sums, L2, area, Midpoint, Edges, Err,$
                            L3, L4, L5, PC3x, PC3y, PC4x, PC4y, A3, Roundness1, Roundness2, Roundness3, L1r, maxLength, maxWidth, num_perimeter_points, crystal_type, perimeterum, csp, $
                            ml_p1x=ml_p1x, ml_p1y=ml_p1y, ml_p2x=ml_p2x, ml_p2y=ml_p2y, mw_p1x=mw_p1x, mw_p1y=mw_p1y, mw_p2x=mw_p2x, mw_p2y=mw_p2y, $
                            perimeter_points_x=perimeter_points_x, perimeter_points_y=perimeter_points_y, num_satellites=num_satellites, HorV=HorV, $
                            mask=current_H_mask, recompressed_image=recompressed_image, new_wordcount=new_wordcount

                        if(!mask_bit_filling and n_elements(recompressed_image) ne 0) then begin ; mask bits were potentially filled, changing the image and the number of words
                            Hwords = new_wordcount

                            particle.nwords = Hwords

                            ;temp = *particle.image_data
                            ptr_free, particle.image_data
                            ;temp = recompressed_image
                            ;particle.image_data = ptr_new(temp, /no_copy)
                            particle.image_data = ptr_new(recompressed_image, /no_copy)
                        endif

                        Hsize = (size(Himage))[2]

                        if(show_images eq 1 and Hsize*!zoom_factor lt !disp_window_ys) then begin
                            if(!draw_50um_to_scale and !H_resolution_microns eq 50) then begin
                                Hscale = 5
                            endif else begin
                                Hscale = 1
                            endelse

                            if(!zoom_factor gt 1 or Hscale gt 1) then begin
                                if(Hsize eq 1) then begin
                                    Himage = rebin(Himage, !PIXELS, Hscale*Hsize*!zoom_factor)
                                endif

                                Himage = congrid(Himage, Hscale*!PIXELS*!zoom_factor, Hscale*Hsize*!zoom_factor)
                            endif

                            hpos = hpos - Hscale*Hsize*!zoom_factor

                            if(hpos lt 0) then begin
                                hpos = !disp_window_ys - Hscale*Hsize*!zoom_factor - 1
                            endif

                            tv, Himage, 0, hpos

                            if(!draw_separator_line) then begin
                                hpos--

                                if(hpos lt 0) then begin
                                    hpos = !disp_window_ys - 1 ;+ !zoom_factor
                                endif

                                tv, H_horz_bar, 0, hpos, /true
                            endif

                        endif
                    endif
                endif

                if((Vflags and uint('8000'X)) ne uint('0000'X)) then begin
                    ;print, 'Vflag: 0x8000 bit 15 H FIFO overflow'
                    Voverflow_this_second++
                    total_Voverflow++

                    if(first_Voverflow_flag eq 0) then begin
                        first_Voverflow_flag = 1
                    endif else begin
                        first_Voverflow_flag = 0
                        second_Voverflow_flag = 1
                    endelse
                endif

                if(NV gt 0) then begin ; this is a vertical particle
                    ; parse the flags
                    if((Vflags and uint('4000'X)) ne uint('0000'X)) then begin
                        ;if((Vflags and uint('0004'X)) ne uint('0000'X)) then begin
                        ;print, 'Vflag: 0x4000 bit 14 unused, should be 0'
                        particle.CPI_trigger = 1

                        total_Vcpi_trigger++
                    endif
                    if((Vflags and uint('2000'X)) ne uint('0000'X)) then begin
                        ;print, 'Vflag: 0x2000 bit 13 unused, should be 0'
                        total_Vflag_bit13++
                    endif
                    if((Vflags and uint('1000'X)) ne uint('0000'X)) then begin
                        ; multi-packet particle
                        ;print, 'Vflag: 0x1000 bit 12 particle words exceeded max frame size'
                        total_Vflag_bit12++

                        ;                        if(Vmulti_flag eq 1) then begin
                        ;                            if(Vmulti_ID ne particle.partID) then begin ; this is a second multi-packet image, with a different particle ID, write out the first and clear
                        ;                                ;stop


                        ;                                if(nHmulti lt 2) then begin
                        ;                                    print, 'nVmulti too low 1 ', nHmulti
                        ;                                endif else begin
                        ;                                    write_multiparticle_to_SORTED, olun_S1, !V_PARTICLE_HEADER, nVmulti, Vmulti_array
                        ;                                endelse

                        ;                                stored_Vflags = uint(0)

                        ;                                ;write_multiparticle_to_SORTED, olun_S1, !V_PARTICLE_HEADER, nVmulti, Vmulti_array

                        ;                                ; reset variables and clear the multi_array -- can/should this be done in the function call?
                        ;                                for i=0, Vmulti_array_size-1 do begin
                        ;                                    ptr_free, Vmulti_array[i].image_data
                        ;                                endfor

                        ;                                nVmulti = 0
                        ;                            endif
                        ;                        endif

                        ;                        ; start, continue, or second multi-packet particle
                        ;                        ; the finished particle gets added to the multi_array below
                        ;                        Vmulti_flag = 1 ; signal to accumulate Hmulti packets
                        ;                        Vmulti_ID = particle.partID
                    endif

                    ;                    if(Vmulti_flag eq 1) then begin
                    ;                        stored_Vflags = Vflags
                    ;                    endif

                    total_Vpart++
                    Vpart_this_second++

                    Vwords = NV - 3 ; NV is number of words plus 3 timing words
                    Hwords = uint(0)

                    particle.nwords = Vwords

                    if(NV gt 3 and Vwords gt 0) then begin ; and Vwords+buffer_ptr lt raw_data_buffer_size) then begin
                        ;Vimage_data2 = uintarr(Vwords)

                        Vimage_data2 = read_data(retval, Vwords)

                        if(retval eq 1) then begin
                            done = 1
                            break
                        endif

                        ;readu, lun, Vimage_data

                        ; should likely do a check on the buffer size before reading from it here
                        ;Vimage_data2[0:Vwords-1] = raw_data_buffer[buffer_ptr:buffer_ptr+Vwords-1]

                        Vimage_data = bytarr(Vwords*2L)

                        ; TAF convert the image words to bytes
                        for i=0L, Vwords-1L do begin
                            Vimage_data[i*2L+1] = ishft((Vimage_data2[i] and 'FF00'X), -8)
                            Vimage_data[i*2L]   =        Vimage_data2[i] and '00FF'X
                        endfor

                        ptr_free, particle.image_data
                        particle.image_data = ptr_new(Vimage_data, /no_copy)

                        ;buffer_ptr += Vwords
                    endif else begin
                        ; NV = 0, usually a deadtime begin/end
                    endelse

                    if(show_images eq 1 and particle.nslice gt 0 and particle.nslice lt 32000) then begin
                        Area = 0UL
                        Midpoint = 0
                        Edges = 0
                        L3 = 0.0
                        L4 = 0
                        L5 = 0
                        PC3x = 0.0
                        PC3y = 0.0
                        PC4x = 0.0
                        PC4y = 0.0
                        A3 = 0.0
                        Roundness1 = 0.0
                        Roundness2 = 0.0
                        Roundness3 = 0.0
                        L1r = 0.0
                        maxLength = 0.0
                        maxWidth  = 0.0
                        perimeter = 0UL
                        perimeterum = 0.
                        csp = 0.
                        num_satellites = 0.
                        Err=0

                        Vimage = bytarr(!PIXELS, particle.nslice)

                        HorV = 'V'

                        processparticle, *particle.image_data, Vimage, Vwords, particle.nslice, Sums, L2, area, Midpoint, Edges, Err,$
                            L3, L4, L5, PC3x, PC3y, PC4x, PC4y, A3, Roundness1, Roundness2, Roundness3, L1r, maxLength, maxWidth, num_perimeter_points, crystal_type, perimeterum, csp, $
                            ml_p1x=ml_p1x, ml_p1y=ml_p1y, ml_p2x=ml_p2x, ml_p2y=ml_p2y, mw_p1x=mw_p1x, mw_p1y=mw_p1y, mw_p2x=mw_p2x, mw_p2y=mw_p2y, $
                            perimeter_points_x=perimeter_points_x, perimeter_points_y=perimeter_points_y, num_satellites=num_satellites, HorV=HorV, $
                            mask=current_V_mask, recompressed_image=recompressed_image, new_wordcount=new_wordcount

                        if(!mask_bit_filling and n_elements(recompressed_image) ne 0) then begin ; mask bits were potentially filled, changing the image and the number of words
                            Vwords = new_wordcount

                            particle.nwords = Vwords

                            ;temp = *particle.image_data
                            ptr_free, particle.image_data
                            ;temp = recompressed_image
                            ;particle.image_data = ptr_new(temp, /no_copy)
                            particle.image_data = ptr_new(recompressed_image, /no_copy)
                        endif

                        Vsize = (size(Vimage))[2]

                        if(Vsize*!zoom_factor lt !disp_window_ys) then begin
                            if(!zoom_factor gt 1) then begin
                                if(Vsize eq 1) then begin
                                    Vimage = rebin(Vimage, !PIXELS, Vsize*!zoom_factor)
                                endif

                                Vimage = congrid(Vimage, !PIXELS*!zoom_factor, Vsize*!zoom_factor)
                            endif

                            vpos = vpos - Vsize*!zoom_factor

                            if(vpos lt 0) then begin
                                vpos = !disp_window_ys - Vsize*!zoom_factor - 1
                            endif

                            tv, Vimage, !disp_window_xs - !PIXELS*!zoom_factor, vpos

                            if(!draw_separator_line) then begin
                                vpos--

                                if(vpos lt 0) then begin
                                    vpos = !disp_window_ys - 1
                                endif

                                tv, V_horz_bar, !disp_window_xs - !PIXELS*!zoom_factor, vpos, /true
                            endif
                        endif
                    endif else begin
                        ; NV = 0, usually a deadtime begin/end
                    endelse
                endif

                time1 = read_data(retval, 1)
                time2 = read_data(retval, 1)
                time3 = read_data(retval, 1)

                ;junk = read_data(retval, 1)
                next_flag = test_read_data()


                ;printh, particle.flag
                ;printh, time1
                ;printh, time2
                ;printh, time3
                ;printh, next_flag

                ;stop

                if(next_flag eq uint('3253'X) or next_flag eq uint('4E4C'X)) then begin ; valid particle
                    ; store the previous probecount
                    if(NH gt 0) then begin
                        prev_probecount  = prev_Hprobecount
                        prev_Hprobecount = probecount
                    endif else if(NV gt 0) then begin
                        prev_probecount  = prev_Vprobecount
                        prev_Vprobecount = probecount
                    endif

                    ; assemble the times 1-3 into the 48 bit hex timestamp
                    probecount = ulong64(ishft(ulong64(time3),  32) or ishft(ulong64(time2),  16) or ulong64(time1))

                    ; put the probecount into the particle data, and generate a high resolution timestamp based on HK data and the counter
                    particle.timestamp.counter = probecount


                    ;                    if(probecount ge '00062B4B036D'X) then begin
                    ;                        print, 'activating the megaflaga'
                    ;                        printh, probecount
                    ;
                    ;                        ;!megaflaga = 1
                    ;                    endif
                    ;
                    ;                    if(probecount ge '00062B4B0DE7'X) then begin
                    ;
                    ;
                    ;                        print, 'probecount hit:'
                    ;                        printh, probecount
                    ;
                    ;                        ;stop
                    ;                    endif



                    ; rough HK incrementer, needs refinement!!!
                    if(probecount eq prev_probecount) then begin ; check particle ID as well
                        printf, !log_lun, 'probecounts equal, likely multipart?'

                        reject_particle = 1
                    endif else if(probecount lt prev_probecount) then begin ; if it goes down -- NOTE this is wrong, since the H and Vs can be interleaved from different times
                        printf, !log_lun, 'probecount went down, check H vs V?'

                        if(!reprint_raw_data) then begin
                            ; printf, olun3, 'Header, NHraw, NVraw, part ID, n slices, time stamp, NH, H flags, NV, V flags'
                            printf, olun3, format='(%"0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%016X, %7d, 0x%04X, %7d, 0x%04X")', $
                                particle.flag, particle.NHraw, particle.NVraw, particle.partID, particle.nslice, probecount, NH, Hflags, NV, Vflags
                        endif

                        ;print_raw_hex, probecount

                        reject_particle = 1
                    endif else if(probecount gt prev_probecount) then begin
                        if((probecount - prev_probecount) gt !max_probecount_diff or $
                            probecount lt first_hk_timestamp                      or $
                            probecount gt  last_hk_timestamp) then begin
                            printf, !log_lun, 'time jump too big, or out of HK range, dicarding particle, should flag as bad'

                            reject_particle = 1
                        endif else if(probecount gt HK2_timestamp.counter) then begin
                            ; get new bracketing HKs
                            ; set HK1 = HK2, and get a new HK2

                            ;                        index = where(hk_probecount_array lt probecount, count)

                            ;                        if(count eq 0) then begin
                            ;                            ;print, ''
                            ;                            printf, !log_lun, 'particle cannot be bracketted by HK, cannot calculate times, exiting'
                            ;                            ;print, ''

                            ;                            done = 1
                            ;                            break
                            ;                        endif

                            ;                        if(count eq 1 or count eq (size(hk_probecount_array))[1]) then begin
                            ;                            ;print, ''
                            ;                            printf, !log_lun, 'particle after last HK, cannot calculate times, exiting'
                            ;                            ;print, ''

                            ;                            done = 1
                            ;                            break
                            ;                        endif

                            ;                        if(count gt 1) then begin
                            ;                            ;HK1_timestamp = hk_timestamp_array

                            ;                            index1 = index[count-1]
                            ;                            index2 = index1 + 1


                            hk_index++

                            index1 = hk_index
                            index2 = hk_index + 1L

                            HK1_timestamp.year    = hk_timestamp_array[index1].year
                            HK1_timestamp.month   = hk_timestamp_array[index1].month
                            HK1_timestamp.day     = hk_timestamp_array[index1].day
                            HK1_timestamp.hour    = hk_timestamp_array[index1].hour
                            HK1_timestamp.minute  = hk_timestamp_array[index1].minute
                            HK1_timestamp.second  = hk_timestamp_array[index1].second
                            HK1_timestamp.ms      = hk_timestamp_array[index1].ms
                            HK1_timestamp.us      = 0
                            HK1_timestamp.ns      = 0
                            HK1_timestamp.counter = hk_probecount_array[index1]

                            HK2_timestamp.year    = hk_timestamp_array[index2].year
                            HK2_timestamp.month   = hk_timestamp_array[index2].month
                            HK2_timestamp.day     = hk_timestamp_array[index2].day
                            HK2_timestamp.hour    = hk_timestamp_array[index2].hour
                            HK2_timestamp.minute  = hk_timestamp_array[index2].minute
                            HK2_timestamp.second  = hk_timestamp_array[index2].second
                            HK2_timestamp.ms      = hk_timestamp_array[index2].ms
                            HK2_timestamp.us      = 0
                            HK2_timestamp.ns      = 0
                            HK2_timestamp.counter = hk_probecount_array[index2]


                            HK_timediff_nanos = calc_timediff_nanos(HK1_timestamp, HK2_timestamp)

                            if(!generate_SORTED) then begin
                                ; write out the HK_BIN and HK_TXT files
                                hk_tas = two_uints2float(hk_array[index1, 74], hk_array[index1, 75])

                                write_HK_TXT, olun_HK_TXT1, HK1_timestamp, hk_tas[0]
                                write_HK_BIN, olun_HK_BIN1, HK1_timestamp, hk_array[index1, *]

                                if(extending_file_2_seconds) then begin
                                    write_HK_TXT, olun_HK_TXT2, HK1_timestamp, hk_tas
                                    write_HK_BIN, olun_HK_BIN2, HK1_timestamp, hk_array[index1, *]
                                endif

                                ; also should write the HK timestamp to the SORTED file, not sure if this is actually used by View though

                                ;Put #SORTED_FNUM, , HK_HR_Timestamp
                                ;'Put #SORTED_FNUM, , HKRolloverCount
                                ;Put #SORTED_FNUM, , HK_timingword_hi
                                ;Put #SORTED_FNUM, , hk_timingword
                                ;If (gSecondary_File_Opened = True) Then
                                ;    Put #SECOND_SORTED_FNUM, , HK_HR_Timestamp
                                ;    'Put #SECOND_SORTED_FNUM, , HKRolloverCount
                                ;    Put #SECOND_SORTED_FNUM, , HK_timingword_hi
                                ;    Put #SECOND_SORTED_FNUM, , hk_timingword
                            endif ; !generate_SORTED for writing HK_TXT and HK_BIN
                            ;endif
                        endif
                    endif
                    ; rough HK incrementer, needs refinement!!!



                    new_timestamp = generate_timestamp(particle.timestamp.counter, HK1_timestamp, HK2_timestamp, HK_timediff_nanos)

                    if(new_timestamp.year eq 0) then begin
                        reject_particle = 1
                        ;stop
                    endif
                endif else begin
                    print,            format='(%"rejecting 0x%04X, 0x%016X, next_flag incorrect 0x%04X")', particle.flag, particle.timestamp.counter, next_flag
                    printf, !log_lun, format='(%"rejecting 0x%04X, 0x%016X, next_flag incorrect 0x%04X")', particle.flag, particle.timestamp.counter, next_flag
                    reject_particle = 1 ; the next
                endelse

                if(reject_particle eq 1) then begin
                    total_reject_particle_count++

                    printf, !log_lun, format='(%"rejecting 0x%04X, 0x%016X")', particle.flag, particle.timestamp.counter
                endif else begin ; particle is accepted, add it to the H or V particle buffer array ([hv]_particle_array)
                    particle.timestamp.year   = new_timestamp.year
                    particle.timestamp.month  = new_timestamp.month
                    particle.timestamp.day    = new_timestamp.day
                    particle.timestamp.hour   = new_timestamp.hour
                    particle.timestamp.minute = new_timestamp.minute
                    particle.timestamp.second = new_timestamp.second
                    particle.timestamp.ms     = new_timestamp.ms
                    particle.timestamp.us     = new_timestamp.us
                    particle.timestamp.ns     = new_timestamp.ns

                    probecount_array[probecount_index] = particle.timestamp.counter
                    if(probecount_index ge probecount_array_size-1) then begin
                        probecount_array_size += !array_growth_factor

                        new_probecount_array = temporary(ulon64arr(probecount_array_size))
                        new_probecount_array[0:probecount_index] = temporary(probecount_array[0:probecount_index])
                        probecount_array = temporary(new_probecount_array)
                    endif
                    probecount_index++

                    if(!reprint_raw_data) then begin ; could be moved outside of reject test
                        ; printf, olun3, 'Header, NHraw, NVraw, part ID, n slices, time stamp, NH, H flags, NV, V flags,'
                        printf, olun3,format='(%"0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%016X, %7d, 0x%04X, %7d, 0x%04X")', $
                            particle.flag, particle.NHraw, particle.NVraw, particle.partID, particle.nslice, particle.timestamp.counter, NH, Hflags, NV, Vflags
                    endif

                    if(NH gt 0) then begin
                        particle.type = 'H'
                        h_particle_array[h_particle_index] = particle
                        temp = *particle.image_data
                        h_particle_array[h_particle_index].image_data = ptr_new(temp, /no_copy)

                        h_particle_index++

                        if(h_particle_index ge h_particle_array_size) then begin
                            ; move all the particles back, but keep the first hundred at the beginning of the array

                            new_h_particle_array = replicate(empty_particle, h_particle_array_size)
                            new_h_particle_array[0:!npad-1] = temporary(h_particle_array[h_particle_index-!npad:h_particle_index-1])

                            ; free memory before h_particle_index-!npad
                            ptr_free, h_particle_array[0:h_particle_index-!npad-1].image_data

                            h_particle_index = !npad
                            h_particle_array = temporary(new_h_particle_array)
                        endif

                        ; put the 48 bit particle timestamps into H and V counter arrays
                        hprobecount_array[hpartcount] = probecount
                        if(hpartcount ge harray_size-1) then begin
                            harray_size += !array_growth_factor

                            new_hprobecount_array = ulon64arr(harray_size)
                            new_hprobecount_array[0:hpartcount] = temporary(hprobecount_array[0:hpartcount])
                            hprobecount_array = temporary(new_hprobecount_array)
                        endif
                        hpartcount++
                    endif else if(NV gt 0) then begin
                        particle.type = 'V'
                        v_particle_array[v_particle_index] = particle
                        temp = *particle.image_data
                        v_particle_array[v_particle_index].image_data = ptr_new(temp, /no_copy)

                        v_particle_index++

                        if(v_particle_index ge v_particle_array_size) then begin
                            ; move all the particles back, but keep the first hundred at the beginning of the array

                            new_v_particle_array = replicate(empty_particle, v_particle_array_size)
                            new_v_particle_array[0:!npad-1] = temporary(v_particle_array[v_particle_index-!npad:v_particle_index-1])

                            ; free memory before v_particle_index-!npad
                            ptr_free, v_particle_array[0:v_particle_index-!npad-1].image_data

                            v_particle_index = !npad
                            v_particle_array = temporary(new_v_particle_array)
                        endif

                        ; put the 48 bit particle timestamps into H and V counter arrays
                        vprobecount_array[vpartcount] = probecount
                        if(vpartcount ge varray_size-1) then begin
                            varray_size += !array_growth_factor

                            new_vprobecount_array = ulon64arr(varray_size)
                            new_vprobecount_array[0:vpartcount] = temporary(vprobecount_array[0:vpartcount])
                            vprobecount_array = temporary(new_vprobecount_array)
                        endif
                        vpartcount++
                    endif



                    ;                    if((h_particle_index gt 0 and h_particle_index mod 1000 eq 0) or $
                    ;                       (v_particle_index gt 0 and v_particle_index mod 1000 eq 0)) then begin
                    ;;                        date_labels = label_date(date_format='%H:%I:%S')

                    ;;                        jultimes = timegen(start=julday(hk_timestamp_array[0].month,          hk_timestamp_array[0].day,             hk_timestamp_array[0].year,             $
                    ;;                                                        hk_timestamp_array[0].hour,           hk_timestamp_array[0].minute,          hk_timestamp_array[0].second),          $
                    ;;                                           final=julday(hk_timestamp_array[hk_count-1].month, hk_timestamp_array[hk_count-1].day,    hk_timestamp_array[hk_count-1].year,    $
                    ;;                                                        hk_timestamp_array[hk_count-1].hour,  hk_timestamp_array[hk_count-1].minute, hk_timestamp_array[hk_count-1].second), $
                    ;;                                                        units='Seconds')

                    ;;                        nxticks = 10

                    ;                        window, 7, xs = 600, ys = 600
                    ;                        device, decomposed = 0
                    ;                        loadct, 39, /silent

                    ;                        plot,  h_particle_array[0:h_particle_index-1].timestamp.counter
                    ;                        oplot, h_particle_array[0:h_particle_index-1].timestamp.counter, color = 120
                    ;                        oplot, v_particle_array[0:v_particle_index-1].timestamp.counter, color = 170

                    ;                        print, ''
                    ;                        print, 'HK1: ', timestamp2string(HK1_timestamp)
                    ;                        print, 'H:   ', timestamp2string(h_particle_array[h_particle_index-1].timestamp)
                    ;                        print, 'V:   ', timestamp2string(v_particle_array[v_particle_index-1].timestamp)
                    ;                        print, 'HK2: ', timestamp2string(HK2_timestamp)


                    ;                        hindex = where(h_particle_array[0:h_particle_index-1].timestamp.counter ge HK1_timestamp.counter and $
                    ;                                       h_particle_array[0:h_particle_index-1].timestamp.counter lt HK2_timestamp.counter,    $
                    ;                                       hcount)


                    ;                        vindex = where(v_particle_array[0:v_particle_index-1].timestamp.counter ge HK1_timestamp.counter and $
                    ;                                       v_particle_array[0:v_particle_index-1].timestamp.counter lt HK2_timestamp.counter,    $
                    ;                                       vcount)

                    ;                        print, ''
                    ;                        print, 'hcount: ', hcount, ' vcount: ', vcount
                    ;                        print, ''



                    ;stop
                    ;                    endif



                    ; MAIN SORTED OUTPUT SECTION

                    if(!generate_SORTED) then begin
                        ; print the particle data into the SORTED file
                        ; also handles logic of OVERFLOW pairs
                        ; and assembling Multi-Packet images

                        ; test number of particle to open new SORTED files
                        ; two conditions
                        ; 1) number of particles to split exceeded, open second file
                        ; 2) 2 seconds after split, close first file and rename

                        if(file_split_particle_counter1 gt !file_splitting_at and ~extending_file_2_seconds) then begin
                            extending_file_2_seconds = 1

                            ; calculate the current time as the start time of the second file
                            ; add 2 seconds to this for the end time of the first file

                            start_time = HK1_timestamp
                            end_time = add_2_seconds(start_time)

                            SORTED_file_record_number++ ; the last four digits of the filename _0000, incremented with split files

                            ; open the second file
                            second_SORTED_filename = string(format='(%"SORTED%02d%02d%02d%02d%02d%02d_%02d%02d%02d_%04d.2DSCPI")', $
                                main_time_base.year-2000, main_time_base.month, main_time_base.day, $
                                main_time_base.hour, main_time_base.minute, main_time_base.second,  $
                                start_time.hour, start_time.minute, start_time.second,  $
                                SORTED_file_record_number)

                            full_second_SORTED_filename = outfilepath + second_SORTED_filename
                            openw,  olun_S2, /get, full_second_SORTED_filename
                            writeu, olun_S2, SORTED_header_string

                            ; initialize the SORTED file with double TAS packets
                            ;hk_tas = two_uints2float(hk_array[index1, 74], hk_array[index1, 75]) ; should have hk_tas from updating HKs
                            ;write_TT_marker, olun_S2, start_time, hk_tas, HK1_timestamp.counter
                            write_TT_marker, olun_S2, HK1_timestamp, hk_tas, HK1_timestamp.counter

                            second_HK_TXT_filename = string(format='(%"HK_TXT%02d%02d%02d%02d%02d%02d_%02d%02d%02d_%04d.2DSCPI")', $
                                main_time_base.year-2000, main_time_base.month, main_time_base.day, $
                                main_time_base.hour, main_time_base.minute, main_time_base.second,  $
                                start_time.hour, start_time.minute, start_time.second,  $
                                SORTED_file_record_number)

                            full_second_HK_TXT_filename = outfilepath + second_HK_TXT_filename
                            openw,  olun_HK_TXT2, /get, full_second_HK_TXT_filename
                            writeu, olun_HK_TXT2, HK_TXT_header_string
                            printf, olun_HK_TXT2, ''
                            printf, olun_HK_TXT2, frag1
                            printf, olun_HK_TXT2, frag2
                            printf, olun_HK_TXT2, frag3
                            printf, olun_HK_TXT2, frag4

                            second_HK_BIN_filename = string(format='(%"HK_BIN%02d%02d%02d%02d%02d%02d_%02d%02d%02d_%04d.2DSCPI")', $
                                main_time_base.year-2000, main_time_base.month, main_time_base.day, $
                                main_time_base.hour, main_time_base.minute, main_time_base.second,  $
                                start_time.hour, start_time.minute, start_time.second,  $
                                SORTED_file_record_number)

                            full_second_HK_BIN_filename = outfilepath + second_HK_BIN_filename
                            openw,  olun_HK_BIN2, /get, full_second_HK_BIN_filename
                            writeu, olun_HK_BIN2, HK_BIN_header_string
                        endif else if(extending_file_2_seconds) then begin
                            timecomp = timestamp1_ge_timestamp2(particle.timestamp, end_time) ;;;  TAF this particle.timestamp needs to be the next one in the HVarray

                            if(timecomp) then begin ; timestamp is greater than or equal to end_time
                                ; close the first file and rename
                                ; move the second file to the first lun

                                ; close the SORTED file with double TAS packets
                                ;write_TT_marker, olun_S1, end_time, hk_tas, HK2_timestamp.counter ;prev_particle.timestamp.counter
                                write_TT_marker, olun_S1, HK2_timestamp, hk_tas, HK2_timestamp.counter ;prev_particle.timestamp.counter

                                free_lun, olun_S1
                                new_SORTED_filename = outfilepath + string(format='(%"SORTED%02d%02d%02d%02d%02d%02d_%02d%02d%02d_%02d%02d%02d_%04d.2DSCPI")', $
                                    main_time_base.year-2000, main_time_base.month, main_time_base.day, $
                                    main_time_base.hour, main_time_base.minute, main_time_base.second,  $
                                    fSfn_timestamp.hour, fSfn_timestamp.minute, fSfn_timestamp.second,  $
                                    end_time.hour,       end_time.minute,       end_time.second,  $
                                    SORTED_file_record_number-1)

                                file_move, full_first_SORTED_filename, new_SORTED_filename, /overwrite
                                full_first_SORTED_filename = full_second_SORTED_filename
                                olun_S1 = olun_S2

                                free_lun, olun_HK_TXT1
                                new_HK_TXT_filename = outfilepath + string(format='(%"HK_TXT%02d%02d%02d%02d%02d%02d_%02d%02d%02d_%02d%02d%02d_%04d.2DSCPI")', $
                                    main_time_base.year-2000, main_time_base.month, main_time_base.day, $
                                    main_time_base.hour, main_time_base.minute, main_time_base.second,  $
                                    fSfn_timestamp.hour, fSfn_timestamp.minute, fSfn_timestamp.second,  $
                                    end_time.hour,       end_time.minute,       end_time.second,  $
                                    SORTED_file_record_number-1)

                                file_move, full_first_HK_TXT_filename, new_HK_TXT_filename, /overwrite
                                full_first_HK_TXT_filename = full_second_HK_TXT_filename
                                olun_HK_TXT1 = olun_HK_TXT2

                                free_lun, olun_HK_BIN1
                                new_HK_BIN_filename = outfilepath + string(format='(%"HK_BIN%02d%02d%02d%02d%02d%02d_%02d%02d%02d_%02d%02d%02d_%04d.2DSCPI")', $
                                    main_time_base.year-2000, main_time_base.month, main_time_base.day, $
                                    main_time_base.hour, main_time_base.minute, main_time_base.second,  $
                                    fSfn_timestamp.hour, fSfn_timestamp.minute, fSfn_timestamp.second,  $
                                    end_time.hour,       end_time.minute,       end_time.second,  $
                                    SORTED_file_record_number-1)

                                file_move, full_first_HK_BIN_filename, new_HK_BIN_filename, /overwrite
                                full_first_HK_BIN_filename = full_second_HK_BIN_filename
                                olun_HK_BIN1 = olun_HK_BIN2

                                fSfn_timestamp = start_time
                                extending_file_2_seconds = 0
                                file_split_particle_counter1 = file_split_particle_counter2
                                file_split_particle_counter2 = 0L
                            endif
                        endif


                        ;appended_multi = 0

                        ;                        ; assemble the finished particle packet into the H or V multi_array
                        ;                        if((NH gt 0) and (Hmulti_flag eq 1)) then begin
                        ;                            Hmulti_array[nHmulti] = particle

                        ;                            ; make a unique pointer to the preserved image data, direct copies get deallocated,
                        ;                            ; might be able to avoid that with more logic around the deallocation section...

                        ;                            temp = *particle.image_data
                        ;                            Hmulti_array[nHmulti].image_data = ptr_new(temp, /no_copy)

                        ;                            if(nHmulti ge Hmulti_array_size-1) then begin
                        ;                                print,                'expanding Hmulti_array, nHmulti = ', nHmulti
                        ;                                printf, !log_lun, 'expanding Hmulti_array, nHmulti = ', nHmulti

                        ;                                Hmulti_array_size += 10 ; we really don't expect to deal with more than about 5, but might as well implement for extremes...

                        ;                                new_Hmulti_array_array = replicate({_particle}, Hmulti_array_size)
                        ;                                new_Hmulti_array_array[0:nHmulti] = temporary(Hmulti_array_array[0:nHmulti])
                        ;                                Hmulti_array_array = temporary(new_Hmulti_array_array)
                        ;                            endif
                        ;                            nHmulti++

                        ;                            appended_multi = 1

                        ;                            if((stored_Hflags and uint('1000'X)) eq '0'X and nHmulti ge 2) then begin ; end of multi, write it out
                        ;                                ;stop


                        ;                                if(nHmulti lt 2) then begin
                        ;                                    print, 'nHmulti too low 2 ', nHmulti
                        ;                                endif else begin
                        ;                                    write_multiparticle_to_SORTED, olun_S1, !H_PARTICLE_HEADER, nHmulti, Hmulti_array
                        ;                                endelse

                        ;                                stored_Hflags = uint(0)

                        ;                                ;write_multiparticle_to_SORTED, olun_S1, !H_PARTICLE_HEADER, nHmulti, Hmulti_array

                        ;                                ; reset variables and clear the multi_array
                        ;                                for i=0, Hmulti_array_size-1 do begin
                        ;                                    ptr_free, Hmulti_array[i].image_data
                        ;                                endfor

                        ;                                nHmulti = 0
                        ;                                Hmulti_flag = 0
                        ;                            endif

                        ;                            ;stop
                        ;                        endif

                        ;                        if((NV gt 0) and (Vmulti_flag eq 1)) then begin
                        ;                            Vmulti_array[nVmulti] = particle

                        ;                            ; make a unique pointer to the preserved image data, direct copies get deallocated,
                        ;                            ; might be able to avoid that with more logic around the deallocation section...

                        ;                            temp = *particle.image_data
                        ;                            Vmulti_array[nVmulti].image_data = ptr_new(temp, /no_copy)

                        ;                            if(nVmulti ge Vmulti_array_size-1) then begin
                        ;                                print,                'expanding Vmulti_array, nHmulti = ', nVmulti
                        ;                                printf, !log_lun, 'expanding Vmulti_array, nHmulti = ', nVmulti

                        ;                                Vmulti_array_size += 10 ; we really don't expect to deal with more than about 5, but might as well implement for extremes...

                        ;                                new_Vmulti_array_array = replicate({_particle}, Vmulti_array_size)
                        ;                                new_Vmulti_array_array[0:nVmulti] = temporary(Vmulti_array_array[0:nVmulti])
                        ;                                Vmulti_array_array = temporary(new_Vmulti_array_array)
                        ;                            endif
                        ;                            nVmulti++

                        ;                            appended_multi = 1

                        ;                            ; is not detecting prior vflags -- wiped out by h read...
                        ;                            if((stored_Vflags and uint('1000'X)) eq '0'X and nVmulti ge 2) then begin ; end of multi, write it out
                        ;                                ;stop

                        ;                                if(nVmulti lt 2) then begin
                        ;                                    print, 'nVmulti too low 2 ', nHmulti
                        ;                                endif else begin
                        ;                                    write_multiparticle_to_SORTED, olun_S1, !V_PARTICLE_HEADER, nVmulti, Vmulti_array
                        ;                                endelse

                        ;                                stored_Vflags = uint(0)

                        ;                                ;write_multiparticle_to_SORTED, olun_S1, !V_PARTICLE_HEADER, nVmulti, Vmulti_array

                        ;                                ; reset variables and clear the multi_array
                        ;                                for i=0, Vmulti_array_size-1 do begin
                        ;                                    ptr_free, Vmulti_array[i].image_data
                        ;                                endfor

                        ;                                nVmulti = 0
                        ;                                Vmulti_flag = 0
                        ;                            endif

                        ;                            ;stop
                        ;                        endif

                        ;if(appended_multi eq 0) then begin ; don't do a normal particle write if it was a multi part image being appended


                        if(NH gt 0) then begin ; on H particle
                            if(first_Hoverflow_flag eq 1) then begin ; first H overflow, store particle

                                first_Hoverflow_particle = particle ; should be OK not copying image, since image should be empty for first one anyway
                                ;temp = *particle.image_data
                                ;first_Hoverflow_particle.image_data = ptr_new(temp, /no_copy)
                            endif else if(second_Hoverflow_flag eq 1) then begin ; second H overflow, write out and reset flags
                                write_overflow_to_SORTED, olun_S1, !H_DEADTIME_HEADER, first_Hoverflow_particle, particle

                                if(extending_file_2_seconds) then begin
                                    write_overflow_to_SORTED, olun_S2, !H_DEADTIME_HEADER, first_Hoverflow_particle, particle
                                endif

                                ; write a normal particle here in addition to the last overflow?

                                H_overflow_written++

                                second_Hoverflow_flag = 0
                            endif else begin ; normal particle
                                write_particle_to_SORTED, olun_S1, !H_PARTICLE_HEADER, particle
                                file_split_particle_counter1++

                                if(extending_file_2_seconds) then begin
                                    write_particle_to_SORTED, olun_S2, !H_PARTICLE_HEADER, particle
                                    file_split_particle_counter2++
                                endif

                                H_particles_written++
                            endelse
                        endif else if(NV gt 0) then begin ; on V particle
                            if( first_Voverflow_flag eq 1) then begin ; first V overflow, store particle
                                first_Voverflow_particle = particle
                            endif else if(second_Voverflow_flag eq 1) then begin ; second V overflow, write out and reset flags
                                write_overflow_to_SORTED, olun_S1, !V_DEADTIME_HEADER, first_Voverflow_particle, particle

                                if(extending_file_2_seconds) then begin
                                    write_overflow_to_SORTED, olun_S2, !V_DEADTIME_HEADER, first_Voverflow_particle, particle
                                endif

                                ; write a normal particle here in addition to the last overflow?

                                V_overflow_written++

                                first_Voverflow_flag = 0
                                second_Voverflow_flag = 0
                            endif else begin ; normal particle
                                write_particle_to_SORTED, olun_S1, !V_PARTICLE_HEADER, particle
                                file_split_particle_counter1++

                                if(extending_file_2_seconds) then begin
                                    write_particle_to_SORTED, olun_S2, !V_PARTICLE_HEADER, particle
                                    file_split_particle_counter2++
                                endif

                                V_particles_written++
                            endelse
                        endif else begin
                            print, 'neither NH and NV gt 0, nothing to write!' ; should not get here

                            nothing_to_write++
                        endelse
                    endif ; !generate_SORTED -- main particle output section



                    if(probecountflag_set eq 1) then begin ; debug section to halt on specific probecount values
                        ; debug printouts:
                        print, ''

                        point_lun, -lun, current_pos
                        print, format='(%" 0x%016X    File Position")',  current_pos

                        print, ''
                        print, format='(%"   0x %04X        flag")',  particle.flag
                        print, format='(%"   0x %04X        NHraw")', particle.NHraw
                        print, format='(%"   0x %04X        NH      = %d")', NH, NH
                        print, format='(%"   0x %04X        Hflags")',Hflags
                        print, format='(%"   0x %04X        NVraw")', particle.NVraw
                        print, format='(%"   0x %04X        NV      = %d")', NV, NV
                        print, format='(%"   0x %04X        Vflags")',Vflags
                        print, format='(%"   0x %04X        partID  = %d")', particle.partID, particle.partID
                        print, format='(%"   0x %04X        nslice  = %d")', particle.nslice, particle.nslice
                        print, format='(%"   0x %04X        time3")',  time3
                        print, format='(%"   0x %04X        time2")',  time2
                        print, format='(%"   0x %04X        time1")',  time1
                        print, ''
                        print, format='(%"   0x%016X    probecount")', probecount
                        print, ''

                        if(NH gt 0) then begin
                            print, 'Himage: ', (size(Himage))[1], ' Words'
                        endif else if(NV gt 0) then begin
                            print, 'Vimage: ', (size(Vimage))[1], ' Words'
                        endif

                        print, ''

                        stop
                    endif


                endelse ; particle rejected or not for writing
            endelse ; NH or NV zero (2S normal particle, overflow, or multi-part image)
        endif else if(particle.flag eq uint('4E4C'X)) then begin ; flag NL, NULL packet header detected, deal with NULL packet
            NULL_this_second++
            total_NULL_count++

            if((particle.NHraw ne 3) or (particle.NVraw ne 3)) then begin
                point_lun, -lun, current_pos

                print, format='(%"NULL packet error at 0x%016X, NH = 0x%04X, NV = 0x%04X")', current_pos, particle.NHraw, particle.NVraw
            endif

            time1 = read_data(retval, 1)
            time2 = read_data(retval, 1)
            time3 = read_data(retval, 1)

            ; assemble the times 1-3 into the 48 bit hex timestamp
            null_probecount = ulong64(ishft(ulong64(time3),  32) or ishft(ulong64(time2),  16) or ulong64(time1))

            particle.timestamp.counter = null_probecount

            ;probecount_array[probecount_index] = probecount
            ;probecount_index++


            ;if(probecount_index ge probecount_array_size) then begin
            ;    probecount_array_size += !array_growth_factor

            ;    new_probecount_array = temporary(ulon64arr(probecount_array_size))
            ;    new_probecount_array[0:probecount_index-2] = temporary(probecount_array[0:probecount_index-2])
            ;    probecount_array = temporary(new_probecount_array)
            ;endif

            ; not all NULL packets are 2048 long, so cycle through the NULL packet

            ; 2014-05-05 NULL packets send zeros until the end of the block, checksum and timestamp, so will be variable length
            ; should be able to calculate exact length to speed up reads, but searching zeros shouldn't hurt for now

            nullsize_counter = 0L
            checksum = uint(0)

            while(checksum eq uint(0)) do begin
                checksum = read_data(retval, 1)
                ;buffer_ptr++
                nullsize_counter++
            endwhile

            buffer_ptr-- ;; back up for the checksum or next packet header???

            if(!reprint_raw_data) then begin
                ; printf, olun3, 'Header, NHraw, NVraw, part ID, n slices, time stamp, NH, H flags, NV, V flags, NULL length'
                printf, olun3, format='(%"0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%016X, %7d, 0x%04X, %7d, 0x%04X, %7d")', $
                    particle.flag, particle.NHraw, particle.NVraw, particle.partID, particle.nslice, particle.timestamp.counter, NH, Hflags, NV, Vflags, nullsize_counter
            endif
        endif else begin
            print, ''
            print, 'incorrect particle flag, '

            point_lun, -lun, current_pos
            ;print, format='(%"current_pos: 0x%0X")', current_pos

            print, format='(%"%04X, %016X, %016X")', particle.flag, probecount, current_pos

            if(!reprint_raw_data) then begin
                ; printf, olun3, 'Header, NHraw, NVraw, part ID, n slices, time stamp, NH, H flags, NV, V flags,'
                printf, olun3,format='(%"0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%016X, %7d, 0x%04X, %7d, 0x%04X")', $
                    particle.flag, particle.NHraw, particle.NVraw, particle.partID, particle.nslice, particle.timestamp.counter, NH, Hflags, NV, Vflags
            endif

            print, 'iteration counter = ', counter

            ; printf, olun3, 'Header, NHraw, NVraw, part ID, n slices, time stamp, NH, H flags, NV, V flags, NULL length'
            printf, !log_lun ,format='(%"0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%04X, 0x%016X, %d, 0x%04X, %d, 0x%04X")', particle.flag, particle.NHraw, particle.NVraw, particle.partID, particle.nslice, probecount, NH, Hflags, NV, Vflags

            probecount = 0ULL
        endelse

        if(retval ne 0) then begin
            print, 'likely need a more robust exit and more testing of read_data retval'

            done = 1
            break
        endif

        ;if(probecount eq ulong64('00000006C7303340'X) or $
        ;   probecount eq ulong64('00000006C7751D45'X) or $
        ;   probecount eq ulong64('000000076AF2C9CD'X) or $
        ;   probecount eq ulong64('000000076AF2CDCF'X) or $
        ;   probecount eq ulong64('000040F300011BAF'X) or $
        ;   probecount eq ulong64('000000076C78EA0A'X)) then begin

        ;    ; first error occurs here... try to catch in probecount
        ;    ; 2013-09-09, 13:55:07.014, 0x000000076AF2CDCF,  3332, 10998,     0,     0,     0
        ;    ; 2013-09-09, 13:55:08.015, 0x000040F300011BAF, 11416, 16321,     0,   252,     0
        ;    ; 2013-09-09, 13:55:09.015, 0x000000076C78EA0A, 11225, 13105,     2,   259,     0

        ;    point_lun, -lun, current_pos
        ;    ;print, format='(%"current_pos: 0x%0X")', current_pos

        ;    print, format='(%"%04X, %016X, %016X")', particle.flag, probecount, current_pos

        ;    print_raw_hex, probecount

        ;    ;!emoflag = 1

        ;    ;stop
        ;endif

        ; break out of the while loop afer data file has been readed pasting the EndTime of the desired period.
        Time_read=   (particle.timestamp.hour)*3600LL + (particle.timestamp.minute)*60LL + (particle.timestamp.second)*1LL
        if Time_read gt EndTime+1L then break; -----added by Mo on Januray 17, 2015.

    endwhile

    free_lun,  lun
    free_lun, olun
    free_lun, olun2

    if(!reprint_raw_data) then begin
        free_lun, olun3
    endif

    if(!generate_SORTED) then begin
        free_lun, olun_S1
        free_lun, olun_HK_TXT1
        free_lun, olun_HK_BIN1

        if(n_elements(olun_S2) ne 0) then begin
            free_lun, olun_S2
            free_lun, olun_HK_TXT2
            free_lun, olun_HK_BIN2
        endif
    endif

    print, ''
    print, 'basescan finished processing file: ', infilename
    print, ''
    print, 'Total H Particles: ', total_Hpart,     '   written: ', H_particles_written
    print, 'Total V Particles: ', total_Vpart,     '   written: ', V_particles_written
    print, 'Total H Overflows: ', total_Hoverflow, '   written: ', H_overflow_written
    print, 'Total V Overflows: ', total_Voverflow, '   written: ', V_overflow_written
    print, ''
    print, 'nothing_to_write: ', nothing_to_write
    print, ''
    print, 'Total H cpi trigger: ', total_Hcpi_trigger
    print, 'Total V cpi trigger: ', total_Vcpi_trigger
    print, 'Total H flag bit13:  ', total_Hflag_bit13
    print, 'Total V flag bit13:  ', total_Vflag_bit13
    print, 'Total H flag bit12:  ', total_Hflag_bit12
    print, 'Total V flag bit12:  ', total_Vflag_bit12
    print, 'Total NH and NV gt 0: ', total_NH_and_NV_ne_zero
    print, ''
    print, 'Total NULL count: ', total_NULL_count
    print, 'Total Rejected Particles: ', total_reject_particle_count
    print, ''
    print, 'Finished at:'
    print, systime()

    printf, !log_lun, ''
    printf, !log_lun, 'basescan finished processing file: ', infilename
    printf, !log_lun, ''
    printf, !log_lun, 'Total H Particles: ', total_Hpart, ' written: ', H_particles_written
    printf, !log_lun, 'Total V Particles: ', total_Vpart, ' written: ', V_particles_written
    printf, !log_lun, 'Total H Overflows: ', total_Hoverflow, ' written: ', H_overflow_written
    printf, !log_lun, 'Total V Overflows: ', total_Voverflow, ' written: ', V_overflow_written
    printf, !log_lun, ''
    printf, !log_lun, 'nothing_to_write: ', nothing_to_write
    printf, !log_lun, ''
    printf, !log_lun, 'Total H cpi trigger: ', total_Hcpi_trigger
    printf, !log_lun, 'Total V cpi trigger: ', total_Vcpi_trigger
    printf, !log_lun, 'Total H flag bit13:  ', total_Hflag_bit13
    printf, !log_lun, 'Total V flag bit13:  ', total_Vflag_bit13
    printf, !log_lun, 'Total H flag bit12:  ', total_Hflag_bit12
    printf, !log_lun, 'Total V flag bit12:  ', total_Vflag_bit12
    printf, !log_lun, 'Total NH and NV gt 0: ', total_NH_and_NV_ne_zero
    printf, !log_lun, ''
    printf, !log_lun, 'Total NULL count: ', total_NULL_count
    printf, !log_lun, 'Total Rejected Particles: ', total_reject_particle_count
    printf, !log_lun, ''
    printf, !log_lun, 'Finished at:'
    printf, !log_lun, systime()

    endseconds = systime(/seconds)

    elapsed_seconds = endseconds - startseconds

    elapsed_hours   = floor(elapsed_seconds / 3600)
    elapsed_minutes = floor((elapsed_seconds - elapsed_hours * 3600) / 60)
    elapsed_seconds = elapsed_seconds mod 60

    print,                format='(%"processing time: %02d:%02d:%02d")', elapsed_hours, elapsed_minutes, elapsed_seconds
    printf, !log_lun, format='(%"processing time: %02d:%02d:%02d")', elapsed_hours, elapsed_minutes, elapsed_seconds

    print, ''

    free_lun, !log_lun

    ;window, 5

    ;plot,    hprobecount_array, /xst, /yst, title='Probe Counts, H (red), V (green), HK (blue)'
    ;oplot,   hprobecount_array, color = 254
    ;oplot,   vprobecount_array, color = 160
    ;oplot, hk_probecount_array, color = 60


    ; TAF need to make a second by second array to plot against time, index for the probecount_arrays vs hk arrays are different times...
    ;    date_labels = label_date(date_format='%H:%I:%S')

    ;    jultimes = timegen(start=julday(hk_timestamp_array[0].month,          hk_timestamp_array[0].day,             hk_timestamp_array[0].year,             $
    ;                                    hk_timestamp_array[0].hour,           hk_timestamp_array[0].minute,          hk_timestamp_array[0].second),          $
    ;                       final=julday(hk_timestamp_array[hk_count-1].month, hk_timestamp_array[hk_count-1].day,    hk_timestamp_array[hk_count-1].year,    $
    ;                                    hk_timestamp_array[hk_count-1].hour,  hk_timestamp_array[hk_count-1].minute, hk_timestamp_array[hk_count-1].second), $
    ;                                    units='Seconds')


    ;    nxticks = 10

    ;    plot,  jultimes,   hprobecount_array, /yst, title='Probe Counts, H (red), V (green), HK (blue)', $
    ;           xtickunits=['Time'], xtickformat='LABEL_DATE', xstyle=1, xticks=nxticks
    ;    oplot, jultimes,   hprobecount_array, color = 254
    ;    oplot, jultimes,   vprobecount_array, color = 160
    ;    oplot, jultimes, hk_probecount_array, color = 60




    ;    window, 6

    ;    plot, probecount_array, title='All Probecounts'




    ;    window, 7

    ;    sods = dblarr(timestamp_index)

    ;    sods[*] = timestamp_array[0:timestamp_index-1].hour * 3600. + timestamp_array[0:timestamp_index-1].minute * 60. + timestamp_array[0:timestamp_index-1].second + timestamp_array[0:timestamp_index-1].ms / 1000.

    ;    plot, sods, /xst, /yst, title='basefile Timestamp Seconds of Day', ytitle='seconds', xtitle='index'




    ; now plot hk_array counts and overloads vs actual counts and overloads


    ;num_hrz_dect    =        hk_array[*, 56]  ; Horizontal particles detected
    ;num_vrt_dect    =        hk_array[*, 57]  ; Vertical particles detected

    ;dead_time       = double(hk_array[*, 58]) ; dead time in ms seconds

    ;num_hrz_ovrld   =        hk_array[*, 65]  ; # of overload TWs found
    ;num_vrt_ovrld   =        hk_array[*, 66]  ; # of overload TWs found




    ;stop



end ; basescan











