; File:  common_lib.pro




function CalcStrobeTime, T1, T2, version_num
    ; calculates the frame time using the ArrivalT1, and ArrivalT2 values
    ; returns seconds time as a float


    ;stop

    ; returns time in msec by combining the upper and lower words into a time with correct scaling
    if version_num ne 40 then RETURN, float(T1)*62.5 + float(T2)*.0008
    if version_num eq 40 then RETURN, float(T1)*62.5 + float(T2)*.0013333; changed to four 3s on dec7/2004



end ; CalcStrobeTime




pro times__define
    start={times,month:0U,day:0U,hour:0U,minute:0U,sec:0U,msec:0U}
end



;===============================================================================
; Utility functions for Options bit-fields
; Processing options

; Background options
;FUNCTION Subtract_background, Options
; returns true if user chooses to subtract a background from the particle
;   return, (Options AND 1)
;end
;FUNCTION SkipBackground, Options
; returns true if user chooses to NOT display background frames
;   return, (Options AND 2)/2
;end
function convertToseconds,start,year
    ; purpose: to convert a structure of time values into seconds from start of year
    ; parameters
    ;   start : contains the times in a structure with form : {times,month:0,day:0,hour:0,minute:0,sec:0}
    ;   year : year of file

    ; these arrays hold the days count at the start of each month
    days = [0,31,59,90,120,151,181,212,243,273,304,334,365]
    leap = [0,31,60,91,121,152,182,213,244,274,305,335,366]
    ; handle leap years
    if((year mod 4) eq 0 and year ne 2000)then days = leap[start.month-1] + start.day - 1 else days = days[start.month-1] + start.day - 1

    return, (days*86400L + (start.hour)*3600L + (start.minute)*60 + start.sec); + start.msec/1000.)
end

function ConvertFromSeconds,secs,year

    ; extract number of seconds to current month
    if((year mod 4) eq 0 and year ne 2000)then begin
        monthSec =  [2678400,5184000,7862400,10454400,13132800,15724800,18403200,21081600,23673600,26352000,28944000,31622400]
    endif else begin
        monthSec =  [2678400,5097600,7776000,10368000,13046400,15638400,18316800,20995200,23587200,26265600,28857600,31536000]
    endelse

    output = {times}
    indx = where(secs le monthsec,cnt)
    if(cnt ne 0)then begin
        output.month = fix(indx[0])
        if(output.month gt 0)then time = secs - monthSec[output.month-1] else time = secs

        output.month = output.month + 1LL
        output.day = fix(time/86400L) + 1LL
        output.hour = fix((time mod 86400LL)/3600LL)
        output.minute = fix(((time mod 86400LL) mod 3600LL) / 60LL)
        output.sec =fix(((time mod 86400LL) mod 3600LL) mod 60LL)
    endif else begin
        msg = 'Error: input time >(365*24*3600=3156000):  Sec value = '+ strtrim(secs,2)
        ;R = DIALOG_MESSAGE(msg,/info)
        secs=(secs mod 3156000LL)
        indx = where(secs le monthsec,cnt)
        output.month = fix(indx[0])
        if(output.month gt 0)then time = secs - monthSec[output.month-1] else time = secs

        output.month = output.month + 1LL
        output.day = fix(time/86400L) + 1LL
        output.hour = fix((time mod 86400LL)/3600LL)
        output.minute = fix(((time mod 86400LL) mod 3600LL) / 60LL)
        output.sec =fix(((time mod 86400LL) mod 3600LL) mod 60LL)
    endelse

    return,output
end



;===============================================================================
FUNCTION GetFrameByTime, pStateInfo, Time, StartPos=Pos, absolute=absolute, type = type


;keywords
;   type determines if this is a start or stop time

if( not keyword_set(Pos)) THEN Pos = 0

if(not keyword_set(absolute))then begin
    ; scan image frames only
    imgList = (*((*pStateInfo).pFileScan))[*((*pStateInfo).pImgPosIndx)]
endif else begin
    ; scan all frame blocks including backgrounds and empty frames
    imgListIndx = where((*(*pStateInfo).pFileScan).imgtype ge 0 and not (*(*pStateInfo).pFileScan).badframe)
    imgList = (*(*pStateInfo).pFileScan)[imgListIndx]
endelse

guiTime = convertToSeconds(time,(*pStateInfo).filehead.year)
allTimes = imglist.HTime
lastFrame = n_elements(allTimes)-1

if(imgList[LastFrame].imgtype eq 2)then begin
    ; last frame is a background, skip it
    LastFrame = LastFrame - 1
    allTimes = allTimes[0:LastFrame-1]
    LastFrame = LastFrame - 1
endif

;test for first and last frames
if(guitime le allTimes[0])then return,long(0)  ; ULong(0)
if(guitime gt allTimes[lastFrame])then return,long(lastFrame)  ;ULong(lastFrame)

upperTIndex = where(allTimes ge guiTime and allTimes ne 0,UpCnt)
lowerTIndex = where(allTimes lt guiTime and allTimes ne 0,DnCnt)  ;jeff's old code. do work for uAV
stopTindex   =where(allTimes ge guiTime and allTimes ne 0,snCnt)  ;mo's new code.
; handle start and stop times differently
if(keyword_set(type))then begin
    if(type eq 'start')then begin
        return, Long(upperTIndex[0]);ULong(upperTIndex[0])
    endif else begin
        ; stop time case, choose previous frame below stop time
        ;print,'mo stopfrm=', long(stopTindex(0)),'jeff_stopfrm=',Long(lowerTIndex[DnCnt-1])
        ;return,ULong(lowerTIndex[DnCnt-1])  ;jeff's old code.
        return,long(stopTindex(0))
    endelse
endif else return,-1

return,-1
end

;===============================================================================
FUNCTION GetFrameByTime_3VCPI, pStateInfo, inputTime, StartPos=Pos, absolute=absolute, type = type


;keywords
;   type determines if this is a start or stop time

if( not keyword_set(Pos)) THEN Pos = 0

if(not keyword_set(absolute))then begin
    ; scan image frames only
    imgList = (*((*pStateInfo).pFileScan))[*((*pStateInfo).pImgPosIndx)]
endif else begin
    ; scan all frame blocks including backgrounds and empty frames
    imgListIndx = where((*(*pStateInfo).pFileScan).imgtype ge 0)
    imgList = (*(*pStateInfo).pFileScan)[imgListIndx]
endelse

nnnn=n_elements(imglist)

for frame=0LL,nnnn-1LL do begin
    frameData = read_frame((*pStateInfo).InFileLun,(*pStateInfo).FileHead.Ver,imgList[frame].fpos,date=(*pStateInfo).date, TSBlanked = imglist[frame].TSBlanked)
    frameTime = frameData.FrameHeader.time  ;  This time should be  corrected to = Htime*16^8+ ArivelaT1*16^4 + ArriveT2  August 6, 2010.
    time = convertFromSeconds(frameTime , (*pStateInfo).FileHead.year) ;  Time={month, day, hh, mm, ss, msec} ; change to 4/19/19:20:00.00 format.
    time.month=((*pStateInfo).date.month);  added this on August, 2014.
    Time.day   =framedata.FrameHeader.day
    Time.hour  =framedata.FrameHeader.hour
    Time.minute=framedata.FrameHeader.minute
    Time.sec   =framedata.FrameHeader.sec
    Time.msec  =framedata.FrameHeader.msec; ;   HOUR =   14;  MINUTE=0; SEC= 11;MSEC=  791  from frameHeader.
    This_time =convertToSeconds(Time , (*pStateInfo).FileHead.year) 
    if this_time ge inputTime then begin
        frame_num =frame
        break
    endif
endfor

if frame eq nnnn-1LL then frame_num=nnnn-1LL

return,frame_num
end


