;===============================================================================
FUNCTION Mode, Array
    IF N_ELEMENTS(Array) LT 1 THEN RETURN, !VALUES.F_NAN
    IF N_ELEMENTS(Array) EQ 1 THEN RETURN, Array[0]

    ; Compute mode
    Modes = Array[UNIQ(Array, SORT(Array))]
    Mode = Modes[0]
    ModesCount = N_ELEMENTS(Modes)

    IF ModesCount GT 1 THEN BEGIN
        ModesTotals = ULONARR(ModesCount)

        FOR i=0, ModesCount-1 DO BEGIN
            Indx = WHERE(Array EQ Modes[i], MCount)
            ModesTotals[i] = MCount
        ENDFOR
        ; Which one is the most common
        MaxModeTot = MAX(ModesTotals, MaxModeIndx)
        ; Multiple modes?
        Indx = WHERE(ModesTotals EQ MaxModeTot, MCount)
        IF MCount GT 1 THEN RETURN, !VALUES.F_NAN

        Mode = Modes[MaxModeIndx]
    ENDIF

    RETURN, Mode
END; can this be replaced with mode_mo ?????---to be explored later--Feb 23,2008.

;===============================================================================
FUNCTION ShapeLen, ObjPts, xlen, lp1, lp2, YXASPECT=YXAspect

    IF N_ELEMENTS(YXASPECT) EQ 0 THEN YXASPECT = 1
    ; convert to 2D coords from 1D
    ObjPtsX = ObjPts MOD xlen
    ObjPtsY = ObjPts / xlen
    ; JS this could be passed in
    ObjPtsCount = N_ELEMENTS(ObjPts)

    MaxDist = 0.0
    FOR i=0L, ObjPtsCount-2 DO BEGIN
        ; these are the delta x and delta y from the ith point to all the others
        dx = long(ObjPtsX[i] - ObjPtsX[i+1:ObjPtsCount-1])
        dy = long(ObjPtsY[i] - ObjPtsY[i+1:ObjPtsCount-1])

        ; length from ith point to all others
        len = sqrt(FLOAT(dx*YXASPECT)^2 + FLOAT(dy)*dy)
        ;update MaxDist to be the longest so far
        MaxDist = max(len) > MaxDist

        indx = where(len EQ MaxDist, Matches)
        IF (Matches GT 0) THEN BEGIN
            ; here if there are more than 1 longest vectors
            IF ObjPtsX[i] LT ObjPtsX[i+1+indX[0]] THEN BEGIN
                lp1 = ObjPts[i]
                lp2 = ObjPts[i+1+indX[0]]
            ENDIF ELSE BEGIN
                lp2 = ObjPts[i]
                lp1 = ObjPts[i+1+indX[0]]
            ENDELSE
        ENDIF
    ENDFOR
    return, MaxDist
END

;===============================================================================
FUNCTION ShapeWidth, ObjPts, xlen, angle, wp1, wp2, YXASPECT=YXAspect
    IF N_ELEMENTS(YXASPECT) EQ 0 THEN YXASPECT = 1
    ;Slope = -tan(angle)
    IF angle LT 0 THEN angle = angle + !PI
    IF angle LT !PI/2 THEN angle = angle + !PI/2 $
        ELSE angle = angle - !PI/2

    ObjPtsX = ObjPts MOD xlen
    ObjPtsY = ObjPts / xlen
    ObjPtsCount = N_ELEMENTS(ObjPts)

    ; search along this direction for the largest distance ( ie. the width) (see shapeLength)
    MaxDist = 0.0
    FOR i=0L, ObjPtsCount-2 DO BEGIN
        dx = long(ObjPtsX[i+1:ObjPtsCount-1] - ObjPtsX[i])
        dy = long(ObjPtsY[i+1:ObjPtsCount-1] - ObjPtsY[i])
        AngleNow = atan(dy,dx)
        WIndx = where(AngleNow LT 0, Matches)
        IF (Matches GT 0) THEN AngleNow[WIndx] = AngleNow[WIndx] + !PI

        WIndx = where(abs(AngleNow - angle) LT 0.1, Matches)
        IF (Matches GT 0) THEN BEGIN
            len = sqrt(FLOAT(YXASPECT*dx[WIndx])^2 + FLOAT(dy[WIndx])^2)
            MaxDist = max(len) > MaxDist
            indx = where(len EQ MaxDist, Matches)
            IF (Matches GT 0) THEN BEGIN
                IF ObjPtsX[i] LT ObjPtsX[i+1+WIndx[indx[0]]] THEN BEGIN
                    wp1 = ObjPts[i]
                    wp2 = ObjPts[i+1+WIndx[indx[0]]]
                ENDIF ELSE BEGIN
                    wp2 = ObjPts[i]
                    wp1 = ObjPts[i+1+WIndx[indx[0]]]
                ENDELSE
            ENDIF
        ENDIF
    ENDFOR
    return, MaxDist
END

;===============================================================================
FUNCTION ParticleShapeType, dx, dy, Len, Width, Area, Perimeter,hole_count,total_hole_area,mean_brightness, $
    HARMONICS=Harmonics, Roundness1=Roundness1, Roundness2=Roundness2, Roundness3=Roundness3, BoundaryPts=BoundaryPts, rShape=rShape
;print, "in ParticleShapeType"
;stop
if(keyword_set(harmonics) eq 0) then begin
    harmonics = -1
endif

if harmonics(0) ne -1 then begin
    h2gt001=Harmonics(2) gt 0.01
    h4gt001=Harmonics(4) gt 0.01
    h6gt001=Harmonics(6) gt 0.01

    h3lt002=Harmonics(3) lt 0.02
    h5lt002=Harmonics(5) lt 0.02
    h4lt002=harmonics(4) lt 0.02

    L_lt_18W= len lt 1.8*width

    PA_ratio=perimeter/Area lt 2.6*(Len+width)/(Len*width)
    column =len gt 1.9*width and PA_ratio

    plate1=(h2gt001 and h3lt002 and h4gt001 and h5lt002 and h6gt001); and holearea gt 20)
    plate2=(h2gt001 and h3lt002 and h5lt002 and h6gt001 and L_lt_18W and len gt 50 and h4lt002)
    plate3=(len gt 75 and h2gt001 and L_lt_18W and (h4gt001 or h6gt001) and h3lt002); and holearea gt 10)
    plateF=(plate1 or plate2 or plate3 ) and (len gt 20.0) and not(column)
    ;if the particle satisfy plate1+plate2+plate3 > 1 and it is big and it is not column, then it must be a plate.
    ;print,'plateF=',plateF

    ; if PlateF eq 1 then begin
    ; Return,15
    ;  endif
endif else begin
    PlateF=0
endelse
;************************************************************************************

IF N_ELEMENTS(rShape) LT 4 THEN RETURN, 15


; Classify only spherical water drops for now
IF ABS(Roundness1 - 1) LT 0.25 THEN BEGIN
    Sigma = STDEV(rShape, MeanRadius)
    MinR = MIN(rShape, MAX=MaxR)
    ; Ignore deviations less than 2 pixels, as that's just noise
    IF Sigma LE 2 THEN Sigma = 0
    IF ABS(MinR - MeanRadius) LT 2 THEN MinR = MeanRadius
    IF ABS(MaxR - MeanRadius) LT 2 THEN MaxR = MeanRadius

    ; Looking for shapes that deviate less than 10% from mean
    MaxDeviate = 0.1*MeanRadius
    ;PRINT, MaxDeviate, Sigma, ABS(MaxR - MeanRadius), ABS(MinR - MeanRadius)
    IF Sigma LT MaxDeviate AND ABS(MaxR - MeanRadius) LT MaxDeviate $
        AND ABS(MinR - MeanRadius) LT MaxDeviate THEN RETURN, 1   ; Water = 1

ENDIF




; Return unknown type (TAF NOTE: 15 is now ice, not unknown)
RETURN, 15

END

;===============================================================================
PRO DeleteROINode, pNode
    ; Is it the last node in the list?
    IF (ptr_valid((*pNode).Next)) THEN BEGIN
        DeleteROINode, (*pNode).Next
        (*pNode).Next = PTR_NEW()
        DeleteROINode, pNode
    ENDIF ELSE BEGIN
        IF (ptr_valid((*pNode).pImg)) THEN PTR_FREE, (*pNode).pImg
        PTR_FREE, pNode
    ENDELSE

    RETURN
END

;===============================================================================
PRO DeleteROIList, pImgs
    IF NOT PTR_VALID(pImgs) THEN RETURN
    IF N_TAGS(*pImgs) LT 2 THEN BEGIN
        PTR_FREE, pImgs
        RETURN
    ENDIF

    ; Is it a multiple ROI list
    IF (size((*pImgs).Next))(1) NE 10 THEN BEGIN
        IF (ptr_valid((*pImgs).pImg)) THEN PTR_FREE, (*pImgs).pImg
        PTR_FREE, pImgs
        RETURN
    ENDIF
    DeleteROINode, pImgs

END

;===============================================================================
PRO LinkLists, pListHead, pNewList
    ; Sanity checks
    IF NOT PTR_VALID(pListHead) THEN RETURN
    IF N_TAGS(*pListHead) LT 2 THEN RETURN
    IF NOT PTR_VALID(pNewList) THEN RETURN
    IF N_TAGS(*pNewList) LT 2 THEN RETURN

    ; How many items?
    List1Len = (*pListHead).CountNext +1
    List2Len = (*pNewList).CountNext +1
    ; Error checks
    IF List1Len LT 1 THEN RETURN
    IF List2Len LT 1 THEN RETURN

    TotLen = List1Len + List2Len
    ; Move to the tail of list 1
    pListTail = pListHead
    FOR i=1, List1Len DO BEGIN
        ; Correct node counters
        (*pListTail).CountNext = TotLen -i
        IF i NE List1Len THEN $
            pListTail = (*pListTail).Next
    ENDFOR
    ; Link list 2 to list 1
    (*pListTail).Next = pNewList

END

;===============================================================================
FUNCTION CopyLinkList, pList
    ; Sanity checks
    IF NOT PTR_VALID(pList) THEN RETURN, PTR_NEW()
    IF N_TAGS(*pList) LT 2 THEN RETURN, PTR_NEW()
    ; How many items?
    ListLen = (*pList).CountNext +1
    ; Error check
    IF ListLen LT 1 THEN RETURN, PTR_NEW()

    ; Copy each node
    pSrcNode = pList
    FOR i=1, ListLen DO BEGIN
        pNode = PTR_NEW(*pSrcNode)
        ; Copy pImg item
        IF (ptr_valid((*pSrcNode).pImg)) THEN $
            (*pNode).pImg = PTR_NEW(*(*pSrcNode).pImg)

        ; Link the nodes
        IF i EQ 1 THEN pCopyList = pNode $
            ELSE (*pPrevNode).Next = pNode

        pPrevNode = pNode
        pNode = (*pNode).Next
    ENDFOR

    RETURN, pCopyList
END

;===============================================================================
FUNCTION DelRepObjs, pObjList
    IF NOT PTR_VALID(pObjList) THEN RETURN, PTR_NEW()
    IF N_TAGS(*pObjList) LT 2 THEN RETURN, PTR_NEW()

    ; How many ROIs?
    ListLen = (*pObjList).CountNext +1
    IF ListLen LT 1 THEN RETURN, PTR_NEW()
    IF ListLen LT 2 THEN RETURN, pObjList

    pPrevObj = pObjList
    pPrevPrevObj = PTR_NEW()
    WHILE (ptr_valid(pPrevObj)) DO BEGIN
        dx = (*pPrevObj).dx
        sx = (*pPrevObj).SPBox MOD dx
        ex = ((*pPrevObj).EPBox MOD dx)
        sy = (*pPrevObj).SPBox / dx
        ey = ((*pPrevObj).EPBox / dx)

        DelPrev = 0
        pThisObj = (*pPrevObj).Next
        pPrevThisObj = pPrevObj
        WHILE (ptr_valid(pThisObj)) DO BEGIN
            tsx = (*pThisObj).SPBox MOD dx
            tex = ((*pThisObj).EPBox MOD dx)
            tsy = (*pThisObj).SPBox / dx
            tey = ((*pThisObj).EPBox / dx)

            DelThis = 0
            IF (tsx GE sx) AND (tsx LE ex) AND (tsy GE sy) AND (tsy LE ey) THEN $
                IF (tex GE sx) AND (tex LE ex) AND (tey GE sy) AND (tey LE ey) THEN $
                DelThis = 1

            IF NOT DelThis THEN $
                IF (sx GE tsx) AND (sx LE tex) AND (sy GE tsy) AND (sy LE tey) THEN $
                IF (ex GE tsx) AND (ex LE tex) AND (ey GE tsy) AND (ey LE tey) THEN $
                DelPrev = 1

            pOldObj = pThisObj
            IF DelPrev THEN pThisObj = PTR_NEW() $
                ELSE pThisObj = (*pThisObj).Next

            IF DelThis THEN BEGIN
                ListLen = ListLen -1
                ; Remove this node
                (*pOldObj).Next = PTR_NEW()
                DeleteROIList, pOldObj
                (*pPrevThisObj).Next = pThisObj
            ENDIF ELSE pPrevThisObj = pOldObj
        ENDWHILE

        pOldObj = pPrevObj
        pPrevObj = (*pPrevObj).Next
        IF DelPrev THEN BEGIN
            ListLen = ListLen -1
            ; Remove this node
            (*pOldObj).Next = PTR_NEW()
            DeleteROIList, pOldObj
            ; Are we deleting the head of the list?
            IF pPrevPrevObj EQ PTR_NEW() THEN pObjList = pPrevObj $
                ELSE (*pPrevPrevObj).Next = pPrevObj
        ENDIF ELSE pPrevPrevObj = pOldObj
    ENDWHILE
    ; Fix counter for each element in our new list
    pNextNode = pObjList
    FOR i=1, ListLen DO BEGIN
        (*pNextNode).CountNext = ListLen -i
        pNextNode = (*pNextNode).Next
    ENDFOR

    RETURN, pObjList
END

;===============================================================================
; OrderBoundaryPts - sorts boundary pts in order tracing the periphery of the object

FUNCTION OrderBoundaryPts, FeatureID, BoundaryImg, Centroid, PERIMETER=Perimeter, BoundaryThinning=BoundaryThinnning
    ; MODIFY - We ignore pixel aspect ratio here
    ImgSize = SIZE(BoundaryImg)
    dx = ImgSize[1]
    dy = ImgSize[2]

    BoundaryIndxs = WHERE(BoundaryImg EQ FeatureID, BFCount)
    IF BFCount LT 4 THEN RETURN, -1

    ; Make a copy of the boundary image, as we need to modify it
    TmpBoundary = BoundaryImg

    FinalCount = 1
    Done = 0
    Pos = BoundaryIndxs[0]
    TmpBoundary[Pos] = 0
    PosX = Pos MOD dx
    PosY = Pos / dx

    ; Orthogonal neighbors
    OrthoNx = [-1,0,1,0]
    OrthoNy = [0,1,0,-1]

    ; Diagonal neighbors
    DiagNx = [-1,1,1,-1]
    DiagNy = [1,1,-1,-1]

    ; 7x7 Neighbothood
    Nx7b7 = INTARR(7,7)
    Ny7b7 = INTARR(7,7)
    FOR i=-3, 3 DO BEGIN
        Nx7b7[i+3, *] = i
        Ny7b7[*, i+3] = i
    ENDFOR

    Perimeter = 0.0
    ; For now assume the boundary is closed
    ; Trace boundary pts
    REPEAT BEGIN
        FoundNeighbor = 1
        ; For each point, find it's closest neighbor
        ; If more than one, find the one farthest away from the centroid
        ; Orthogonal neighbors
        PosXN = (PosX + OrthoNx) > 0
        PosXN = PosXN < (dx-1)
        PosYN = (PosY + OrthoNy) > 0
        PosYN = PosYN < (dy-1)
        NIndx = WHERE(TmpBoundary[PosXN, PosYN] EQ FeatureID, NCount)
        ; If only one near neighbor, then we are done
        IF NCount EQ 1 THEN BEGIN
            BoundaryIndxs[FinalCount] = PosYN[NIndx]*dx + PosXN[NIndx]
            ;Perimeter = Perimeter + 1
            Perimeter = Perimeter + 0.948
        ENDIF ELSE IF NCount GT 1 THEN BEGIN
            ; Pick the one farthest away from the centroid
            PosXN = PosXN[NIndx]
            PosYN = PosYN[NIndx]
            DeltaX = PosXN - Centroid[0]
            DeltaY = PosYN - Centroid[1]
            d = SQRT(DeltaX*DeltaX + DeltaY*DeltaY)
            Maxd = MAX(d, NIndx)

            BoundaryIndxs[FinalCount] = PosYN[NIndx]*dx + PosXN[NIndx]
            ;Perimeter = Perimeter + 1
            Perimeter = Perimeter + 0.948
        ENDIF ELSE BEGIN
            ; No orthogonal neighbors; look for diagonal
            PosXN = (PosX + DiagNx) > 0
            PosXN = PosXN < (dx-1)
            PosYN = (PosY + DiagNy) > 0
            PosYN = PosYN < (dy-1)
            NIndx = WHERE(TmpBoundary[PosXN, PosYN] EQ FeatureID, NCount)
            ; If only one neighbor, then we are done
            IF NCount EQ 1 THEN BEGIN
                BoundaryIndxs[FinalCount] = PosYN[NIndx]*dx + PosXN[NIndx]
                ;Perimeter = Perimeter + 1
                Perimeter = Perimeter + 1.34
            ENDIF ELSE IF NCount GT 1 THEN BEGIN
                ; Pick the one farthest away from the centroid
                PosXN = PosXN[NIndx]
                PosYN = PosYN[NIndx]
                DeltaX = PosXN - Centroid[0]
                DeltaY = PosYN - Centroid[1]
                d = SQRT(DeltaX*DeltaX + DeltaY*DeltaY)
                Maxd = MAX(d, NIndx)

                BoundaryIndxs[FinalCount] = PosYN[NIndx]*dx + PosXN[NIndx]
                ;Perimeter = Perimeter + 1
                Perimeter = Perimeter + 1.34
            ENDIF ELSE BEGIN
                ; No near neighbors
                ; Search for closest neighbor in a 7x7 neighborhood
                PosXN = (PosX + Nx7b7) > 0
                PosXN = PosXN < (dx-1)
                PosYN = (PosY + Ny7b7) > 0
                PosYN = PosYN < (dy-1)
                NIndx = WHERE(TmpBoundary[PosXN, PosYN] EQ FeatureID, NCount)
                IF NCount GT 0 THEN BEGIN
                    ; Pick the closest neighbor
                    PosXN = PosXN[NIndx]
                    PosYN = PosYN[NIndx]
                    DeltaX = PosXN - PosX
                    DeltaY = PosYN - PosY
                    d = SQRT(DeltaX^2 + DeltaY^2)
                    Mind = MIN(d, NIndx)

                    BoundaryIndxs[FinalCount] = PosYN[NIndx]*dx + PosXN[NIndx]
                    Perimeter = Perimeter + Mind
                ENDIF ELSE FoundNeighbor = 0
            ENDELSE
        ENDELSE

        IF FoundNeighbor THEN BEGIN
            ; Remove neighbor from boundary
            TmpBoundary[BoundaryIndxs[FinalCount]] = 0
            PosX = BoundaryIndxs[FinalCount] MOD dx
            PosY = BoundaryIndxs[FinalCount] / dx
            FinalCount = FinalCount + 1
        ENDIF ELSE Done = 1

        IF FinalCount GE BFCount THEN Done = 1
        if FinalCount ge 32767 then begin
            print, "warning, FinalCount exceeds integer size 32768, likely bad boundary"
            done = 1
        endif
    ENDREP UNTIL (Done)

    BoundaryThinnning = 1 - FinalCount/FLOAT(BFCount)

    RETURN, BoundaryIndxs[0:FinalCount-1]
END

;===============================================================================
FUNCTION ComputeFWHM, Y, StartPt, PEAKVAL=PeakVal

    ; Default is to find the tallest peak
    IF N_ELEMENTS(StartPt) NE 1 THEN StartPt = -1

    Pts = N_ELEMENTS(Y)

    ; First, need to find the closest peak to StartPt
    ; If StartPt is negative, then find the tallest peak
    IF StartPt LT 0 THEN PeakVal = MAX(Y, PeakPos) $
        ELSE BEGIN
        ; Create a smoothed version of Y, just to find peaks
        YSmooth = SMOOTH(Y, 3)
        YSmooth = SMOOTH(YSmooth, 3)

        ; We define a peak to be a max pt surrounded by +/- 3 pts of lower values
        ; Search for peak on the left
        LeftPeak = 0
        Pos = StartPt
        PosArr = INDGEN(7) -3

        WHILE Pos GE 3 DO BEGIN
            Neighbors = PosArr + Pos
            Neighbors = Neighbors > 0
            Neighbors = Neighbors < (Pts -1)

            MaxPt = MAX(YSmooth[Neighbors], PeakPos)
            IF PeakPos EQ 3 THEN BEGIN
                LeftPeak = Neighbors[3]
                Pos = -1
            ENDIF
            Pos = Pos -1
        ENDWHILE

        ; Search for peak on the right
        RightPeak = Pts-1
        Pos = StartPt +1
        WHILE Pos LT Pts-3 DO BEGIN
            Neighbors = PosArr + Pos
            Neighbors = Neighbors > 0
            Neighbors = Neighbors < (Pts -1)

            MaxPt = MAX(YSmooth[Neighbors], PeakPos)
            IF PeakPos EQ 3 THEN BEGIN
                RightPeak = Neighbors[3]
                Pos = Pts
            ENDIF
            Pos = Pos +1
        ENDWHILE
        ; Pick the closest peak
        IF (StartPt - LeftPeak) LE (RightPeak - StartPt) THEN PeakPos = LeftPeak $
            ELSE PeakPos = RightPeak

        ; We now have the smoothed peak, but we want the real peak
        Neighbors = PosArr + PeakPos
        Neighbors = Neighbors > 0
        Neighbors = Neighbors < (Pts -1)
        MaxPt = MAX(Y[Neighbors], PeakPos)
        PeakPos = Neighbors[PeakPos]

        PeakVal = Y[PeakPos]
    ENDELSE

    BasePts = Pts/3
    ; Compute a baseline from the lowest 1/3 of the points
    SortedIndx = SORT(Y)
    BaseLine = TOTAL(Y[SortedIndx[0:BasePts-1]])/FLOAT(BasePts)

    HalfMax = (PeakVal + BaseLine)/2
    ; Find HalfMax to the left
    LeftPos = 0
    Pos = PeakPos-1
    WHILE Pos GE 0 DO BEGIN
        IF Y[Pos] LE HalfMax THEN BEGIN
            LeftPos = Pos
            Pos = -1
        ENDIF
        Pos = Pos -1
    ENDWHILE
    ; Find HalfMax to the right
    RightPos = Pts-1
    Pos = PeakPos+1
    WHILE Pos LT Pts DO BEGIN
        IF Y[Pos] LE HalfMax THEN BEGIN
            RightPos = Pos
            Pos = Pts
        ENDIF
        Pos = Pos +1
    ENDWHILE

    RETURN, RightPos - LeftPos
END

;===============================================================================
FUNCTION ComputeFocus, BoundaryPtsX, BoundaryPtsY, GradientImg

    FORWARD_FUNCTION ComputeFWHM

    BCount = N_ELEMENTS(BoundaryPtsX)
    IF BCount LT 7 THEN RETURN, [0,0, -1,-1,-1,-1,-1,-1,-1,-1]

    ImgSize = SIZE(GradientImg)
    Cols = ImgSize[1]
    Rows = ImgSize[2]

    ; Look at the gradient profile for 14 cross-sections along the particle boundary
    Samples = 16 < BCount
    SampleStep = BCount / FLOAT(Samples)

    SamplesFound = 0
    FocusVals = FLTARR(Samples)

    ; TESTING window
    ;LastWind = -1
    ;IF !D.WINDOW GT -1 THEN LastWind = !D.WINDOW
    ;WINDOW, 3, xsize = 300, ysize=300,TITLE='Gradient Cross Section'

    ;stop
    ; TAF focus experiment
    ;if((size(gradientimg))[1] gt 785) then begin
    ;    a=gradientimg
    ;    gradientimg[*]=0
    ;    gradientimg[540:785,395:640]=a[540:785,395:640]
    ;endif


    FOR i=0, Samples-1 DO BEGIN
        ; Current point
        SampleIndx = ROUND(i*SampleStep)
        SampleIndx = (BCount-1) < SampleIndx
        ; Get 4 neighbor points also
        IF SampleIndx EQ 0 THEN BEGIN
            x = [BoundaryPtsX[BCount-2], BoundaryPtsX[BCount-1], BoundaryPtsX[SampleIndx:SampleIndx+2]]
            y = [BoundaryPtsY[BCount-2], BoundaryPtsY[BCount-1], BoundaryPtsY[SampleIndx:SampleIndx+2]]
        ENDIF ELSE IF SampleIndx EQ 1 THEN BEGIN
            x = [BoundaryPtsX[BCount-1], BoundaryPtsX[SampleIndx:SampleIndx+3]]
            y = [BoundaryPtsY[BCount-1], BoundaryPtsY[SampleIndx:SampleIndx+3]]
        ENDIF ELSE IF SampleIndx EQ BCount-1 THEN BEGIN
            x = [BoundaryPtsX[SampleIndx-2:SampleIndx], BoundaryPtsX[0], BoundaryPtsX[1]]
            y = [BoundaryPtsY[SampleIndx-2:SampleIndx], BoundaryPtsY[0], BoundaryPtsY[1]]
        ENDIF ELSE IF SampleIndx EQ BCount-2 THEN BEGIN
            x = [BoundaryPtsX[SampleIndx-3:SampleIndx], BoundaryPtsX[0]]
            y = [BoundaryPtsY[SampleIndx-3:SampleIndx], BoundaryPtsY[0]]
        ENDIF ELSE BEGIN
            x = BoundaryPtsX[SampleIndx-2:SampleIndx+2]
            y = BoundaryPtsY[SampleIndx-2:SampleIndx+2]
        ENDELSE
        x0 = x[2]
        y0 = y[2]

        ; Find perpendicular cross section to current points
        ; Find slope, offset for current points
        ; Check for vertical lines
        IF N_ELEMENTS(UNIQ(x,SORT(x))) LT 3 THEN BEGIN
            ; The cross-sectional line is horizontal
            sy = y0
            ey = y0
            sx = (x0 - 15) > 0
            ex = (x0 + 15) < (Cols-1)
            ; Check for horizontal lines
        ENDIF ELSE IF N_ELEMENTS(UNIQ(y,SORT(y))) LT 3 THEN BEGIN
            ; The cross-sectional line is vertical
            sx = x0
            ex = x0
            sy = (y0 - 15) > 0
            ey = (y0 + 15) < (Rows-1)
        ENDIF ELSE BEGIN
            Fit = POLY_FIT(x,y,1)
            ; We need the angle of the perpendicular
            Slope = -1/Fit[1]
            Theta = ATAN(Slope)
            dx = ABS(15*COS(Theta))
            dy = Slope*dx
            sx = x0 - dx
            ex = x0 + dx
            sy = y0 - dy
            ey = y0 + dy
            ; Check for bounds on the left
            IF sx LT 0 THEN BEGIN
                sx = 0
                sy = y0 - Slope*x0
            ENDIF
            IF sy LT 0 THEN BEGIN
                sy = 0
                sx = x0 - y0/Slope
            ENDIF
            IF sy GT Rows-1 THEN BEGIN
                sy = Rows-1
                sx = x0 - (y0 - sy)/Slope
            ENDIF
            ; Check for bounds on the right
            IF ex GT Cols-1 THEN BEGIN
                ex = Cols-1
                ey = y0 + Slope*(ex - x0)
            ENDIF
            IF ey GT Rows-1 THEN BEGIN
                ey = Rows-1
                ex = x0 + (y0)/Slope
            ENDIF
            IF ey LT 0 THEN BEGIN
                ey = 0
                ex = x0 + (ey - y0)/Slope
            ENDIF
        ENDELSE
        ; How many pts do we have in the cross section?
        Pts = FIX(SQRT((ex - sx)^2 + (ey - sy)^2))
        dx = (ex - sx)/FLOAT(Pts)
        dy = (ey - sy)/FLOAT(Pts)

        ; Do we have enough?
        IF Pts GT 15 THEN BEGIN
            PtsX = FIX(ROUND(sx + INDGEN(Pts)*dx))
            PtsY = FIX(ROUND(sy + INDGEN(Pts)*dy))
            PtsX[Pts-1] = ROUND(ex)
            PtsY[Pts-1] = ROUND(ey)

            ; Find the location of the point which intersects the boundary
            BoundPt = 15
            IF Pts GT 26 THEN BoundPt = Pts/2 $
                ELSE BEGIN
                ; Find the closest point to x0, y0
                dist = SQRT((x0 - PtsX)^2 + (y0 - PtsY)^2)
                Mindist = MIN(dist, BoundPt)
            ENDELSE

            ; Compute the FWHM for the closest peak to x0,y0
            FWHM = ComputeFWHM(GradientImg[PtsX, PtsY], BoundPt, PEAKVAL=PeakVal)
            IF FWHM LT 3 THEN FWHM = 30
            FocusVals[SamplesFound] = 100.0*PeakVal/FLOAT(FWHM)/110.0

            SamplesFound = SamplesFound +1
        ENDIF
    ENDFOR

    IF SamplesFound LT 1 THEN RETURN, [0,0, -1,-1,-1,-1,-1,-1,-1,-1]
    IF SamplesFound EQ 1 THEN RETURN, [100/FLOAT(Samples),1,FocusVals[0],FocusVals[0],FocusVals[0],FocusVals[0],FocusVals[0],0,!VALUES.F_NAN,!VALUES.F_NAN]

    IF SamplesFound LT Samples THEN $
        FocusVals = FocusVals[0:SamplesFound-1]

    MinVal = MIN(FocusVals, MAX=MaxVal)
    Moments = MOMENT(FocusVals, SDEV=Sigma); moments(0) is the mean values of focusVals.

    ; TESTING window
    ;IF LastWind GT -1 THEN WSET, LastWind

    RETURN, [100*SamplesFound/FLOAT(Samples),SamplesFound,MinVal,MaxVal,Moments[0],MEDIAN(FocusVals,/EVEN),MODE(FocusVals),Sigma,Moments[2],Moments[3]]
END

;===============================================================================
FUNCTION ShapeHarmonics, BoundaryPtsX, BoundaryPtsY, Centroid, rShape=rShape
    BFCount = N_ELEMENTS(BoundaryPtsX)
    IF BFCount LT 4 THEN RETURN, -1

    ; Compute r, pd w.r.t. centroid (pd: perimeter distance)
    ; First compute r, pd for each boundary pt.
    DeltaX = FLOAT(BoundaryPtsX - Centroid[0])
    DeltaY = FLOAT(BoundaryPtsY - Centroid[1])
    BoundaryR = SQRT(DeltaX^2 + DeltaY^2)

    DeltaX = ABS(BoundaryPtsX - SHIFT(BoundaryPtsX, 1))
    DeltaY = ABS(BoundaryPtsY - SHIFT(BoundaryPtsY, 1))
    BoundaryPD = FLTARR(BFCount)
    BoundaryPD[0] = 0

    FOR i=1, BFCount-1 DO BEGIN
        IF DeltaX[i] LT 2 AND DeltaY[i] LT 2 THEN BEGIN
            IF DeltaX[i] + DeltaY[i] EQ 2 THEN BoundaryPD[i] = BoundaryPD[i-1] + 1.34 $
                ELSE BoundaryPD[i] = BoundaryPD[i-1] + 0.948
        ENDIF ELSE BEGIN
            BoundaryPD[i] = BoundaryPD[i-1] + SQRT(DeltaX[i]^2 + DeltaY[i]^2)
        ENDELSE
    ENDFOR

    ; Using a fixed sampling period of ~1 pixel
    Samples = ROUND(BoundaryPD[BFCount-1] +1)
    r = FLTARR(Samples)
    r[0] = BoundaryR[0]
    dpd = BoundaryPD[BFCount-1]/(Samples-1)
    pd = dpd
    pt2 = 1

    ; Now do Samples steps along perimeter
    i = 1
    REPEAT BEGIN
        WHILE pd GT (BoundaryPD[pt2] + 0.001) AND pt2 LT BFCount-1 DO pt2 = pt2 + 1
        IF pt2 GE BFCount-1 THEN BEGIN
            r[i] = BoundaryR[BFCount-1]
        ENDIF ELSE BEGIN
            ; Interpolate for r
            Slope = (BoundaryR[pt2] - BoundaryR[pt2-1])/(BoundaryPD[pt2] - BoundaryPD[pt2-1])
            r[i] = BoundaryR[pt2-1] + (pd - BoundaryPD[pt2-1])*Slope
        ENDELSE

        i = i + 1
        pd = pd + dpd
    ENDREP UNTIL (pd GT (BoundaryPD[BFCount-1] + 0.001))

    IF i LT Samples THEN r[i:Samples-1] = r[i-1]

    LastWind = -1
    ;IF !D.WINDOW GT -1 THEN LastWind = !D.WINDOW
    ;WINDOW, 1, XPOS=0, YPOS=0, xsize = 300, ysize=300,TITLE='Object Shape'
    ;WINDOW, 2, XPOS=0, YPOS=330, xsize = 300, ysize=300,TITLE='Object Harmonics'

    ; Plot it
    ;WSET, 1
    ;PLOT, r

    rShape = r

    r = (ABS(FFT(r)))[0:Samples/2]
    r = r/MAX(r)
    PlotPts = N_ELEMENTS(r)
    ;WSET, 2
    ;PLOT, r[0:(PlotPts-1) < 40]
    ;IF LastWind GT -1 THEN WSET, LastWind
    if n_elements(r) lt 7 then r=[r,0,0,0,0,0];  added this on May, 5, 2009.
    RETURN, r
END

;===============================================================================
PRO ComputeROIHoleStats, pROIStruct, FeatureID, Image, GradientImg, SegmentedImg, BoundaryImg, HolesImg
    ; Mark each hole
    HolesImg = LABEL_REGION(HolesImg)
    ; How many total holes
    TotHoles = Max(HolesImg)
    HolesInside = 0

    ; Find which holes are inside our particle
    IF TotHoles GT 0 THEN BEGIN
        HoleAreas = FLTARR(TotHoles)
        HoleIds = BYTARR(TotHoles)
        Hole1Area = 0
        Hole2Area = 0
        Hole3Area = 0

        FOR i=1, TotHoles DO BEGIN
            HIndx = WHERE(i EQ HolesImg, HPts)
            ; WARNING - we don't check to see if all hole points are within
            ; the particle. That would be a really weird and rare error, though.
            InHIndx = WHERE(SegmentedImg[HIndx] EQ FeatureID, HPts2)

            IF HPts2 GT HPts/2 THEN BEGIN
                HoleAreas[HolesInside] = HPts
                HoleIds[HolesInside] = i
                HolesInside = HolesInside + 1

                IF N_ELEMENTS(HoleTotIndx) LT 1 THEN HoleTotIndx = HIndx $
                    ELSE HoleTotIndx = [HoleTotIndx, HIndx]

                ; Save three largest holes
                IF HolesInside EQ 1 THEN BEGIN
                    Hole1Indx = TEMPORARY(HIndx)
                    Hole1Area = HPts
                ENDIF ELSE BEGIN
                    IF HPts GT Hole1Area THEN BEGIN
                        IF N_ELEMENTS(Hole2Indx) GT 0 THEN BEGIN
                            Hole3Indx = TEMPORARY(Hole2Indx)
                            Hole3Area = Hole2Area
                        ENDIF
                        Hole2Indx = TEMPORARY(Hole1Indx)
                        Hole2Area = Hole1Area
                        Hole1Indx = TEMPORARY(HIndx)
                        Hole1Area = HPts
                    ENDIF ELSE IF HPts GT Hole2Area THEN BEGIN
                        IF N_ELEMENTS(Hole2Indx) GT 0 THEN BEGIN
                            Hole3Indx = TEMPORARY(Hole2Indx)
                            Hole3Area = Hole2Area
                        ENDIF
                        Hole2Indx = TEMPORARY(HIndx)
                        Hole2Area = HPts
                    ENDIF ELSE IF HPts GT Hole3Area THEN BEGIN
                        Hole3Indx = TEMPORARY(HIndx)
                        Hole3Area = HPts
                    ENDIF
                ENDELSE
            ENDIF
        ENDFOR
    ENDIF

    IF HolesInside GT 0 THEN BEGIN
        ; Save largest 3 holes
        (*pROIStruct).Hole1_Area = Hole1Area
        HoleVals = Image(Hole1Indx)
        IF Hole1Area GT 1 THEN BEGIN
            MinVal = MIN(HoleVals, MAX=MaxVal)
            Moments = MOMENT(HoleVals, SDEV=Sigma)
            (*pROIStruct).Hole1_Intensity_Stats = [Hole1Area,MinVal,MaxVal,Moments[0],MEDIAN(HoleVals,/EVEN),Mode(HoleVals),Sigma,Moments[2],Moments[3]]
        ENDIF ELSE (*pROIStruct).Hole1_Intensity_Stats = [Hole1Area,HoleVals[0],HoleVals[0],HoleVals[0],HoleVals[0],HoleVals[0],0.0,!VALUES.F_NAN,!VALUES.F_NAN]

        IF N_ELEMENTS(Hole2Indx) GT 0 THEN BEGIN
            (*pROIStruct).Hole2_Area = Hole2Area
            HoleVals = Image(Hole2Indx)
            IF Hole2Area GT 1 THEN BEGIN
                MinVal = MIN(HoleVals, MAX=MaxVal)
                Moments = MOMENT(HoleVals, SDEV=Sigma)
                (*pROIStruct).Hole2_Intensity_Stats = [Hole2Area,MinVal,MaxVal,Moments[0],MEDIAN(HoleVals,/EVEN),Mode(HoleVals),Sigma,Moments[2],Moments[3]]
            ENDIF ELSE (*pROIStruct).Hole2_Intensity_Stats = [Hole2Area,HoleVals[0],HoleVals[0],HoleVals[0],HoleVals[0],HoleVals[0],0.0,!VALUES.F_NAN,!VALUES.F_NAN]

            IF N_ELEMENTS(Hole3Indx) GT 0 THEN BEGIN
                (*pROIStruct).Hole3_Area = Hole3Area
                HoleVals = Image(Hole3Indx)
                IF Hole3Area GT 1 THEN BEGIN
                    MinVal = MIN(HoleVals, MAX=MaxVal)
                    Moments = MOMENT(HoleVals, SDEV=Sigma)
                    (*pROIStruct).Hole3_Intensity_Stats = [Hole3Area,MinVal,MaxVal,Moments[0],MEDIAN(HoleVals,/EVEN),Mode(HoleVals),Sigma,Moments[2],Moments[3]]
                ENDIF ELSE (*pROIStruct).Hole3_Intensity_Stats = [Hole3Area,HoleVals[0],HoleVals[0],HoleVals[0],HoleVals[0],HoleVals[0],0.0,!VALUES.F_NAN,!VALUES.F_NAN]
            ENDIF
        ENDIF

        ; ========= Hole Area Stats =================
        HoleAreas = HoleAreas[0:HolesInside-1]
        (*pROIStruct).Hole_Count = HolesInside
        if n_elements(HolesInsdie) gt 1 then stop;   mooo-feb 23, 2008.  why shold have two elements.

        IF HolesInside GT 1 THEN BEGIN
            MinVal = MIN(HoleAreas, MAX=MaxVal)
            Moments = MOMENT(HoleAreas, SDEV=Sigma)
            (*pROIStruct).Hole_Area_Stats = [HolesInside,MinVal,MaxVal,Moments[0],MEDIAN(HoleAreas,/EVEN),Mode(HoleAreas),Sigma,Moments[2],Moments[3]]
        ENDIF ELSE (*pROIStruct).Hole_Area_Stats = [HolesInside,HoleAreas[0],HoleAreas[0],HoleAreas[0],HoleAreas[0],HoleAreas[0],0.0,!VALUES.F_NAN,!VALUES.F_NAN]

        ; ========= Hole Intensity Stats =================
        HoleVals = Image(HoleTotIndx)
        IF N_ELEMENTS(HoleVals) GT 1 THEN BEGIN
            MinVal = MIN(HoleVals, MAX=MaxVal)
            Moments = MOMENT(HoleVals, SDEV=Sigma)
            (*pROIStruct).Hole_Intensity_Stats = [N_ELEMENTS(HoleVals),MinVal,MaxVal,Moments[0],MEDIAN(HoleVals,/EVEN),Mode(HoleVals),Sigma,Moments[2],Moments[3]]
        ENDIF ELSE (*pROIStruct).Hole_Intensity_Stats = [N_ELEMENTS(HoleVals),HoleVals[0],HoleVals[0],HoleVals[0],HoleVals[0],HoleVals[0],0.0,!VALUES.F_NAN,!VALUES.F_NAN]

        ; Take out holes
        SegmentedImg[HoleTotIndx] = 0
        ; Get points for our particle excluding holes
        FIndx = WHERE(SegmentedImg EQ FeatureID, FCount)
        ; ========= Intensity Stats =================
        ImgVals = Image(FIndx)
        IF N_ELEMENTS(ImgVals) GT 1 THEN BEGIN
            MinVal = MIN(ImgVals, MAX=MaxVal)
            Moments = MOMENT(ImgVals, SDEV=Sigma)
            (*pROIStruct).Intensity_HoleLess_Stats = [N_ELEMENTS(ImgVals),MinVal,MaxVal,Moments[0],MEDIAN(ImgVals,/EVEN),Mode(ImgVals),Sigma,Moments[2],Moments[3]]
        ENDIF ELSE (*pROIStruct).Intensity_HoleLess_Stats = [N_ELEMENTS(ImgVals),ImgVals[0],ImgVals[0],ImgVals[0],ImgVals[0],ImgVals[0],0.0,!VALUES.F_NAN,!VALUES.F_NAN]
        ; Fill holes back up
        SegmentedImg[HoleTotIndx] = FeatureID
    ENDIF
END



function cpi_getangle, inAx, inAy, inBx, inBy, inCx, inCy
    ; calculates the angle ABC in degrees, always positive 0 to 90

    ; TAF NOTE 2016-03-24 identical name to a function in 3V-2DS -- this function is causing floating point errors,
    ; likely dividing by zero or sqrt of a negative when the inputs are the same points, renaming       

    Ax = float(inAx)
    Ay = float(inAy)
    Bx = float(inBx)
    By = float(inBy)
    Cx = float(inCx)
    Cy = float(inCy)

    ; test for identical points, which cause floating point errors below (which slows down processing)
    if((Ax eq Bx and Ay eq By) or $
        (Ax eq Cx and Ay eq Cy) or $
        (Cx eq Bx and Cy eq By)) then begin
        return, 0.
    endif   

    one   = (Ax - Bx) * (Bx - Cx) + (Ay - By) * (By - Cy)
    two   = (Ax - Bx) * (Ax - Bx) + (Ay - By) * (Ay - By)
    three = (Bx - Cx) * (Bx - Cx) + (By - Cy) * (By - Cy)

    pre_result = abs(one / (sqrt(two) * sqrt(three)))

    return, acos(pre_result) / !DtoR
end ; cpi_getangle



;===============================================================================
PRO CalcObjParams, FeatureID, Image, SegmentedImg, BoundaryImg, GradientImg, GivenLength=GivenLen, GivenWidth=GivenWidth, YXAspect=YXAspect, $
    AREA=Area, PERIMETER=Perimeter, CENTROID=Centroid, Roundness1=Roundness1, Roundness2=Roundness2, Roundness3=Roundness3, Roun, BoundaryPts=BoundaryPts, HARMONICS=Harmonics, Focus=Focus, $
    rShape=rShape, Radii_Stats=Radii_Stats, Gradient_Stats=GradStats, Intensity_Stats=Intensity_Stats

FORWARD_FUNCTION OrderBoundaryPts, ShapeHarmonics, ComputeFocus

IF N_ELEMENTS(YXAspect) EQ 0 THEN YXAspect = 1



; ========= AREA =================
FIndx = WHERE(SegmentedImg EQ FeatureID, FCount)
; MODIFY - How do we handle YXAspect?
Area = FCount*YXAspect

ImgSize = SIZE(SegmentedImg)
dx = ImgSize[1]
dy = ImgSize[2]

; ========= Intensity Stats =================
ImgVals = Image(FIndx)
MinVal = MIN(ImgVals, MAX=MaxVal)
Moments = MOMENT(ImgVals, SDEV=Sigma)
Intensity_Stats = [N_ELEMENTS(ImgVals),MinVal,MaxVal,Moments[0],MEDIAN(ImgVals,/EVEN),Mode(ImgVals),Sigma,Moments[2],Moments[3]]

; ========= Gradient Stats =================
GradVals = GradientImg(FIndx)
MinVal = MIN(GradVals, MAX=MaxVal)
Moments = MOMENT(GradVals, SDEV=Sigma)
GradStats = [N_ELEMENTS(GradVals),MinVal,MaxVal,Moments[0],MEDIAN(GradVals,/EVEN),Mode(GradVals),Sigma,Moments[2],Moments[3]]

; ========= Centroid =================
ObjPtsX = FIndx MOD dx;   objptsX has same number of n_elements(Findx).
ObjPtsY = FIndx / dx
CGx = TOTAL(ObjPtsX)/FLOAT(FCount); Fcount=n_elements(ObjPtsX); Mrach12,2009.
CGy = TOTAL(ObjPtsY)/FLOAT(FCount)
Centroid = [CGx, CGy]

; ========= Roundness1 ================
IF N_ELEMENTS(GivenLen) NE 0 THEN BEGIN
    Roundness1 = (4*Area)/(!PI*GivenLen^2)
ENDIF

; ========= PERIMETER =================
ObjPts = WHERE(BoundaryImg EQ FeatureID, BFCount)
Perimeter1 = BFCount*YXAspect

; ========= Boundary-Dependent Measurements ======
BoundaryPts = OrderBoundaryPts(FeatureID, BoundaryImg, Centroid, PERIMETER=Perimeter, BoundaryThinning=BoundaryThinnning)

BoundaryPtsX = BoundaryPts MOD dx
BoundaryPtsY = BoundaryPts / dx
; If OrderBoundaryPts() failed, then we can't really do any harmonics or
; boundary dependent computations
Focus = [0,0, -1,-1,-1,-1,-1,-1,-1,-1]
;JS MOD
Harmonics = -1L
rShape = -1L

IF N_ELEMENTS(BoundaryPts) GT 3 AND BoundaryThinnning LT 0.3 THEN BEGIN
    Harmonics = ShapeHarmonics(BoundaryPtsX, BoundaryPtsY, Centroid, rShape=rShape)


    ;*******************************************************
    ; window,1,ysize=72*4.5,xsize=72*8
    ; !P.thick=1
    ; !P.CHARSIZE=1
    ; !P.multi=[0,1,1]
    ; loadct,0
    ; white=!D.N_COLORS-1
    ;!P.BACKGROUND=white

    ; plot,rshape, color=0,title='  old Rshape',$
    ;xstyle=1,ystyle=1
    ;write_png,'C:/New_rshape_x_old.png',tvrd(true=1),/order;;   is the the same as teh new Rshape, but new Rshpae has 3 times more points.
    ;**************************************************************************


    IF N_ELEMENTS(rShape) GT 1 THEN BEGIN
        MinVal = MIN(rShape, MAX=MaxVal)
        Moments = MOMENT(rShape, SDEV=Sigma)
        Radii_Stats = [N_ELEMENTS(rShape),MinVal,MaxVal,Moments[0],MEDIAN(rShape,/EVEN),Mode(rShape),Sigma,Moments[2],Moments[3]]

    ENDIF

    ; Out of focus computation
    Focus = ComputeFocus(BoundaryPtsX, BoundaryPtsY, GradientImg)

    ;print,'Focus: ', Focus
ENDIF

; ========= Roundness2 ================
; Algorithm based on Chris Wieland's June 2012 Roundness 2 implemented for the 2D-S Process Particles dll

; calculates the standard deviation of distances from the centroid to the perimeter pixels

; find the length of the longest line between perimeter points, the effective max diameter of the particle

num_perimeter_points = (size(BoundaryPtsX, /dim))[0]

max_dist = -1.

; TAF using the wrong distance here, need distance from centroid, not edge distances!!!
; also can use distance_measure to speed up distance calcs instead of nested for loops...

if(num_perimeter_points gt 2) then begin

    pdist = fltarr(num_perimeter_points-1)

    for from=0, num_perimeter_points-2 do begin
        for to=from+1, num_perimeter_points-1 do begin
            pdist[from] = sqrt((float(BoundaryPtsX[from]) - float(BoundaryPtsX[to]))^2. + (float(BoundaryPtsY[from]) - float(BoundaryPtsY[to]))^2.)

            max_dist = max_dist > pdist[from]
        endfor
    endfor
endif

if(num_perimeter_points gt 2) then begin

    radii = fltarr(num_perimeter_points)

    for i=0, num_perimeter_points-1 do begin
        radii[i] = sqrt((float(Centroid[0]) - float(BoundaryPtsX[i]))^2. + (float(Centroid[1]) - float(BoundaryPtsY[i]))^2.)
    endfor
endif

Roundness2 = -1.
Roundness3 = -1.

if(num_perimeter_points gt 4) then begin
    estimated_radius = 0.5 * max_dist

    dist_variance = 0.

    for i=0, num_perimeter_points-1 do begin
        diff = radii[i] - estimated_radius
        dist_variance = dist_variance + diff^2.
    endfor

    dist_variance = dist_variance / (num_perimeter_points-1)

    radius_SD = sqrt(dist_variance)
    Roundness2 = radius_SD / estimated_radius ; scale to the presumed radius


    ; ========= Roundness3 ================
    ; Algorithm based on Chris Wieland's June 2012 Roundness 3 implemented for the 2D-S Process Particles dll

    ; calculates the ratio of the average angles of the perimeter pixels with respect to the centroid, and compares
    ; to the expected angles for a circle

    one = 0
    two = 1

    ; catch the NaN on larger particles, div by zero???

    angles = fltarr(num_perimeter_points)

    estimated_angle  = 360. / num_perimeter_points

    angle_variance = 0.

    mean_angle = 0.

    for i=0, num_perimeter_points-1 do begin
        angles[i] = cpi_getangle(BoundaryPtsX[one], BoundaryPtsY[one], Centroid[0], Centroid[1], BoundaryPtsX[two], BoundaryPtsY[two])

        one = (one+1) mod num_perimeter_points
        two = (two+1) mod num_perimeter_points

        mean_angle = mean_angle + angles[i]
    endfor

    mean_angle = mean_angle / num_perimeter_points

    Roundness3 = mean_angle / estimated_angle
endif ; (num_perimeter_points gt 4)

END

;===============================================================================
; Compute various particle characterization parameters

FUNCTION ComputeParticleParams, pObjList, Image, EdgeImg, SegmentedImg, BoundaryImg, HolesImg, YXAspect=YXAspect, IMGPOS=ImgPos, MARKIMAGE=MarkImage, Roundness2=Roundness2, Roundness3=Roundness3
    ;edgeIMg is the gradientIMg, BoundaryIMg is teh BinImg, HolesIMg is the HoleImGs all ready got from earlier calls.
    IF NOT PTR_VALID(pObjList) THEN RETURN, PTR_NEW()
    IF N_TAGS(*pObjList) LT 2 THEN RETURN, PTR_NEW()
    IF N_ELEMENTS(ImgPos) NE 2 THEN ImgPos = [0,0]

    ; How many Objects in list?
    ListLen = (*pObjList).CountNext +1
    IF ListLen LT 1 THEN RETURN, PTR_NEW()

    pPriorNode = PTR_NEW()
    pListHead = pObjList
    pThisNode = pObjList

    FOR i=0, ListLen-1 DO BEGIN
        ; .obj are the boundary points of the object (1 pixel thick)
        ; .dx is the width of the image ROI
        ; BinImg is the Binary image for the boundary
        ; YXAspect is the aspect ratio defined in GUI


        ;Calculate area and perimeter
        Area = 20
        Perimeter = 10
        Crystal = 0
        ROIFeatNum = SegmentedImg[(*pThisNode).Obj[0]]; VALUES CHNAGES FROM 1 TO 3 ETC.

        CalcObjParams, ROIFeatNum, Image, SegmentedImg, BoundaryImg, EdgeImg, GivenLen=(*pThisNode).Len, $
            GivenWidth=(*pThisNode).Width, YXAspect=YXAspect, $
            AREA=Area, PERIMETER=Perimeter, CENTROID=Centroid, Roundness1=Roundness1, Roundness2=Roundness2, Roundness3=Roundness3, $
            BoundaryPts=BoundaryPts, HARMONICS=Harmonics, Focus=Focus, rShape=rShape, $
            Gradient_Stats=Gradient_Stats, Intensity_Stats=Intensity_Stats

        ;stop

        ;Crystal = ParticleShapeType((*pThisNode).dx, (*pThisNode).dy, (*pThisNode).Len, (*pThisNode).Width, Area, Perimeter, $
        ;          HARMONICS=Harmonics, ROUNDNESS=Roundness, BoundaryPts=BoundaryPts, rShape=rShape)

        pNextNode =    (*pThisNode).Next

        ; Replace object structure with new one
        IF N_ELEMENTS(BoundaryPts) GT 4 THEN Pts = [LONG(ImgPos[0]), ImgPos[1], TEMPORARY(BoundaryPts)] $
            ELSE Pts = [LONG(ImgPos[0]), ImgPos[1], (*pThisNode).Obj]

        pNewNode = PTR_NEW({Obj:Pts, SPBox:(*pThisNode).SPBox, EPBox:(*pThisNode).EPBox, $
            Len:(*pThisNode).Len, Len1:(*pThisNode).Len1, Len2:(*pThisNode).Len2, Width:(*pThisNode).Width, $
            Wid1:(*pThisNode).wid1, Wid2:(*pThisNode).wid2, dx:(*pThisNode).dx, dy:(*pThisNode).dy, Dark:FIX(0), $
            Flags:FIX(0), Area:FLOAT(Area), Perimeter:FLOAT(Perimeter), Roundness1:Roundness1, Centroid:Centroid, $
            CutOff:(*pThisNode).CutOff, Crystal:BYTE(15), Harmonics:TEMPORARY(Harmonics), Focus:TEMPORARY(Focus), $
            Radii:TEMPORARY(rShape), Radii_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  $
            Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],   $
            Intensity_HoleLess_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  $
            Gradient_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],    $
            Curvature_len:0U, Curvature_Window:0U, Curvature:-1, $
            Curvature_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],   $
            Hole_Count:0U, Hole_Area_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], $
            Hole_Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], $
            Hole_Other_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], $
            Hole1_Area:0.0,  Hole1_Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], $
            Hole1_Perim:0.0, Hole1_Other_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], $
            Hole2_Area:0.0,  Hole2_Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], $
            Hole2_Perim:0.0, Hole2_Other_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], $
            Hole3_Area:0.0,  Hole3_Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], $
            Hole3_Perim:0.0, Hole3_Other_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], $
            Roundness2:Roundness2, Roundness3:Roundness3, $
            pImg:PTR_NEW(), Next:(*pThisNode).Next, CountNext:(*pThisNode).CountNext})

        ;print, "Roundness2", string(Roundness2), "  Roundness3", string(Roundness3)

        ;stop



        ComputeROIHoleStats, pNewNode, ROIFeatNum, Image, EdgeImg, SegmentedImg, BoundaryImg, HolesImg

        ;*************************  move to here on Mya 8, 2009 to reclassify crystal habit with hole area count.
        hole_count= (*pThisNode).Hole_count
        total_hole_area=(*pThisNode).hole1_area+(*pThisNode).hole2_area+(*pThisNode).hole3_area
        mean_brightness= 150.0

        Crystal = ParticleShapeType((*pThisNode).dx, (*pThisNode).dy, (*pThisNode).Len, (*pThisNode).Width, Area, Perimeter,$
            hole_count,total_hole_area, mean_brightness,HARMONICS=Harmonics, ROUNDNESS1=Roundness1, BoundaryPts=BoundaryPts,$
            rShape=rShape, Roundness2=Roundness2, Roundness3=Roundness3)

        (*pNewNode).crystal=byte(crystal)
        ;***************************************************************************************************

        IF N_ELEMENTS(Gradient_Stats) GT 2 THEN (*pNewNode).Gradient_Stats = Gradient_Stats
        IF N_ELEMENTS(Intensity_Stats) GT 2 THEN (*pNewNode).Intensity_Stats = Intensity_Stats

        ; Mark image?
        IF KEYWORD_SET(MarkImage) THEN BEGIN
            Image[(*pNewNode).Obj[2:*]] = 255
        ENDIF

        OutFocus = (100 - (*pNewNode).Focus[0]) > 0
        Crystal = Crystal AND '000F'X
        OutFocus = BYTE(OutFocus) + ISHFT(Crystal,8)
        (*pNewNode).Flags = FIX(OutFocus)


        ; Now replace pThisNode with pNewNode
        (*pThisNode).Next = PTR_NEW()
        DeleteROIList, pThisNode

        ; Are we deleting the head of the list?
        ;if listLen gt 1 then begin
        ;  aaa=ppriorNode eq PTR_new()
        ;  print,'i=',I, '  AAA=',AAA
        ; stop
        ; endif

        IF pPriorNode EQ PTR_NEW() THEN pListHead = pNewNode $
            ELSE (*pPriorNode).Next = pNewNode
        pPriorNode = pNewNode

        ; Move on to the next node
        pThisNode =    pNextNode

        ;if listLen gt 1 then stop

    ENDFOR

    ;if listLen gt 1 then stop

    RETURN, pListHead

END

;===============================================================================
; Object rejection criteria
FUNCTION ScreenObjects, pObjList

    IF NOT PTR_VALID(pObjList) THEN RETURN, PTR_NEW()
    IF N_TAGS(*pObjList) LT 2 THEN RETURN, PTR_NEW()

    ; How many Objects in list?
    ListLen = (*pObjList).CountNext +1
    IF ListLen LT 1 THEN RETURN, PTR_NEW()

    IF N_ELEMENTS(DarkThresh) NE 1 THEN DarkThresh = 3
    DarkThresh = FLOAT(DarkThresh)
    FinalCount = 0

    pPriorNode = PTR_NEW()
    pListHead = pObjList
    pNextNode = pObjList
    FOR i=0, ListLen-1 DO BEGIN
        ; convert to 2D from 1D coord
        ObjPtsX = (*pNextNode).Obj MOD (*pNextNode).dx
        ObjPtsY = (*pNextNode).Obj / (*pNextNode).dx
        ; find bounding box points
        sx = min(ObjPtsX)
        ex = max(ObjPtsX)
        sy = min(ObjPtsY)
        ey = max(ObjPtsY)
        Reject = 0
        ; size of bounding box
        Objdx = ex-sx +1
        Objdy = ey-sy +1

        BoundaryCutOff = 0

        IF Objdx LT 2 OR Objdy LT 2 OR (*pNextNode).Len LT 1 OR (*pNextNode).Width LT 1 THEN BEGIN
            ; TOO SMALL
            Reject = 1
        ENDIF ELSE BEGIN
            ; If the object is at a boundary, then it
            ; may be cut-off
            IF sx GT 1 AND sy GT 1 AND ex LT (*pNextNode).dx -2 AND $
                ey LT (*pNextNode).dy -2 THEN BEGIN
                ; Check aspect ratio
                Aspect = float((*pNextNode).Width)/(*pNextNode).Len
                IF Aspect LT 1 THEN Aspect = 1/Aspect
                IF Aspect GT 20 THEN BEGIN
                    Reject = 1
                ENDIF
            ENDIF ELSE BEGIN
                ; Don't allow any more than 40% of boundary to be cut-off
                CutoffCount = 0
                Tmp = WHERE(ObjPtsX LE 1, CutoffCount1)
                CutoffCount = CutoffCount + CutoffCount1
                Tmp = WHERE(ObjPtsX GE (*pNextNode).dx -2, CutoffCount1)
                CutoffCount = CutoffCount + CutoffCount1
                Tmp = WHERE(ObjPtsY LE 1, CutoffCount1)
                CutoffCount = CutoffCount + CutoffCount1
                Tmp = WHERE(ObjPtsY GE (*pNextNode).dy -2, CutoffCount1)
                CutoffCount = CutoffCount + CutoffCount1

                BoundaryCutOff = FLOAT(CutoffCount)/N_ELEMENTS(ObjPtsX)
                IF BoundaryCutOff GT 0.4 THEN BEGIN
                    Reject = 1
                ENDIF
            ENDELSE
        ENDELSE

        ; Move on to the next node
        pThisNode = pNextNode
        pNextNode =    (*pNextNode).Next
        IF Reject THEN BEGIN
            ; Remove the node
            (*pThisNode).Next = PTR_NEW()
            DeleteROIList, pThisNode
            ; Are we deleting the head of the list?
            IF pPriorNode EQ PTR_NEW() THEN pListHead = pNextNode $
                ELSE (*pPriorNode).Next = pNextNode

        ENDIF ELSE BEGIN
            FinalCount = FinalCount +1

            ;Save the Image Cutoff number
            (*pThisNode).CutOff = 100*BoundaryCutOff
            ; For older code
            BoundaryCutOff = FIX(ROUND(31*BoundaryCutOff) < 31)
            (*pThisNode).Dark = ISHFT(BoundaryCutOff, 8) + (*pThisNode).Dark

            ; Add some padding to the region box
            sx = (sx - 10) > 0
            ex = (ex + 10) < ((*pThisNode).dx-1)
            sy = (sy - 10) > 0
            ey = (ey + 10) < ((*pThisNode).dy-1)

            (*pThisNode).SPBox = sy * (*pThisNode).dx + sx
            (*pThisNode).EPBox = ey * (*pThisNode).dx + ex

            pPriorNode = pThisNode
        ENDELSE
    ENDFOR

    ; Fix counter for each element in our new list
    IF FinalCount GT 0 THEN BEGIN
        pNextNode = pListHead
        FOR i=1, FinalCount DO BEGIN
            (*pNextNode).CountNext = FinalCount -i
            pNextNode = (*pNextNode).Next
        ENDFOR
        pListHead = DelRepObjs(pListHead)
    ENDIF

    RETURN, pListHead
END

;===============================================================================
; Img is the gray scale image, possibly filtered
; GradImg is the gradient of Img

FUNCTION FindObjects, Img, GradImg, Threshold, DarkLevel, SEGMENTEDIMG=ThreshImg, $
    BINIMG=CusterImg, HOLESIMG=InvImg, YXAspect=YXAspect, ENLARGE_AREA=Enlarge_Area

IF N_ELEMENTS(DarkLevel) EQ 0 THEN DarkThreshold = -25 ELSE DarkThreshold = DarkLevel
IF N_ELEMENTS(Threshold) EQ 0 THEN Threshold = 40
MinNumPts = 6
DarkThreshold = DarkThreshold < (-15)

; Get image dimensions
ImgSize = size(Img)
dx = ImgSize[1]
dy = ImgSize[2]

;Threshold image
Indx = WHERE(Img LT DarkThreshold, Count)
IF Count LT 2 THEN BEGIN
    RETURN, PTR_NEW()
ENDIF

SmallArea = 25 * 25

; Adjust DarkThreshold for small stuff
IF (Count LT SmallArea) THEN BEGIN
    DarkThreshold = DarkThreshold + 10
    ;Re-Threshold image
    Indx = WHERE(Img LT DarkThreshold, Count)
    IF Count LT 2 THEN RETURN, PTR_NEW()
ENDIF

ThreshImg = BYTARR(dx, dy)
ThreshImg[Indx] = 1

; Erosion/dilation kernel
K = REPLICATE(1, 3, 3)

; Expand objects?
IF KEYWORD_SET(Enlarge_Area) THEN $
    ThreshImg = DILATE(ThreshImg, K)

; =========== Fill in holes ==============
; Label background features
InvImg = ThreshImg XOR 1
InvImg = LABEL_REGION(InvImg)
; Count how many potential holes we have
Holes = Max(InvImg)

FOR h=1, Holes DO BEGIN
    ; Holes are those features which don't touch any image edge
    HIndx = WHERE(InvImg EQ h, HPts)
    IF HPts GT 0 THEN BEGIN
        HPtsX = HIndx MOD dx
        HPtsY = HIndx / dx
        ; Does it touch any edge?
        MinX = Min(HPtsX, MAX=MaxX)
        MinY = Min(HPtsY, MAX=MaxY)

        IF MinX GT 1 AND MaxX LT dx-2 AND MinY GT 1 AND MaxY LT dy-2 THEN BEGIN
            ; Fill in the hole in our thresholded image
            ThreshImg[HIndx] = 1
        ENDIF ELSE BEGIN
            ; Clear-out non-holes
            InvImg[HIndx] = 0
        ENDELSE
    ENDIF
ENDFOR

; Deal with edge effects
LeftPts = WHERE(ThreshImg[0,*] NE 0, LeftCount)
RightPts = WHERE(ThreshImg[dx-1,*] NE 0, RightCount)
BottomPts = WHERE(ThreshImg[*,0] NE 0, BottomCount)
TopPts = WHERE(ThreshImg[*,dy-1] NE 0, TopCount)

; Now Label real features
ThreshImg = LABEL_REGION(ThreshImg, /EIGHT)
; Re-fill edges
IF LeftCount GT 0 THEN ThreshImg[0,LeftPts] = ThreshImg[1,LeftPts]
IF RightCount GT 0 THEN ThreshImg[dx-1,RightPts] = ThreshImg[dx-2,RightPts]
IF BottomCount GT 0 THEN ThreshImg[BottomPts, 0] = ThreshImg[BottomPts, 1]
IF TopCount GT 0 THEN ThreshImg[TopPts, dy-1] = ThreshImg[TopPts, dy-2]

; Boundary image
CusterImg = ThreshImg
ObjPts = WHERE(ERODE(ThreshImg, K), ObjPtsCount)
IF ObjPtsCount GT 0 THEN $
    CusterImg[ObjPts] = 0

; pList is a pointer to the start of the linked list which holds the particle objects
pList = PTR_NEW()
ListLen = 0

; How many features do we have?
FeatureCount = MAX(ThreshImg)

; Process each feature
FOR f=1, FeatureCount DO BEGIN
    ObjPts = WHERE(CusterImg EQ f, ObjPtsCount)

    ; Use gradient image to discriminate non-distinct particles
    GradIndx = WHERE(GradImg[ObjPts] GT Threshold, GradCount)
    IF (GradCount/FLOAT(ObjPtsCount) GT 0.4) AND ObjPtsCount GE MinNumPts THEN BEGIN

        ; This finds the length of the shape (really the longest dimension)
        ObjLen = ShapeLen (ObjPts, dx, lp1, lp2, YXAspect=YXAspect)

        ; covert to 2D coords from 1D
        lp1x = lp1 MOD dx
        lp1y = lp1 / dx
        lp2x = lp2 MOD dx
        lp2y = lp2 / dx

        ; this is the perpendicular angle from longest dimension
        angle = atan(lp2y-lp1y, lp2x-lp1x)

        ; this returns the width of the object
        ObjWidth = ShapeWidth (ObjPts, dx, angle, wp1, wp2, YXAspect=YXAspect)

        ; This is the minimum size of a valid object!
        IF ObjLen GT 1 AND ObjWidth GT 1 THEN BEGIN

            ; create structure of info for this particle

            ; TAF adding tag names to this structure to get over errors
            ;Tag name HOLE_COUNT is undefined for structure <Anonymous>.
            ;Tag name HOLE1_AREA is undefined for structure <Anonymous>.
            ;Hole_Count:0U, $   ; Number of holes found1
            ; Stats including all holes
            ;Hole_Area_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
            ;Hole_Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
            ;Hole_Other_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
            ; Values for the biggest 3 holes
            ;Hole1_Area:0.0,    $
            ;Hole1_Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
            ;Hole1_Perim:0.0,   $
            ;Hole1_Other_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
            ;Hole2_Area:0.0,    $
            ;Hole2_Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
            ;Hole2_Perim:0.0,   $
            ;Hole2_Other_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
            ;Hole3_Area:0.0,    $
            ;Hole3_Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
            ;Hole3_Perim:0.0,   $
            ;Hole3_Other_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$

            pNode = PTR_NEW({Obj:ObjPts, SPBox:ObjPts[0], EPBox:ObjPts[ObjPtsCount-1], $
                Len:ObjLen, Len1:lp1, Len2:lp2, Width:ObjWidth, Wid1:wp1, Wid2:wp2, $
                dx:dx, dy:dy, Dark:-255, Flags:FIX(0), CutOff:FLOAT(0), pImg:PTR_NEW(), Next:PTR_NEW(), $
                CountNext:-1, Hole_Count:0,$
                Hole_Area_Stats:      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
                Hole_Intensity_Stats: [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
                Hole_Other_Stats:     [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
                Hole1_Area:            0.0, $
                Hole1_Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
                Hole1_Perim:           0.0, $
                Hole1_Other_Stats:    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
                Hole2_Area:            0.0, $
                Hole2_Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
                Hole2_Perim:           0.0, $
                Hole2_Other_Stats:    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
                Hole3_Area:            0.0, $
                Hole3_Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
                Hole3_Perim:           0.0, $
                Hole3_Other_Stats:    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]$
                })

            ; update linked list
            IF listlen EQ 0 THEN pList = pNode $
                ELSE (*pPriorNode).Next = pNode
            ListLen = ListLen + 1
            pPriorNode = pNode
        ENDIF
    ENDIF
ENDFOR

; Fix counter for each element
IF ListLen GT 0 THEN BEGIN
    pNextNode = pList
    FOR i=1, ListLen DO BEGIN
        (*pNextNode).CountNext = ListLen - i
        pNextNode = (*pNextNode).Next
    ENDFOR
ENDIF

RETURN, pList

END

;===============================================================================
FUNCTION FilterImg, Img, GradientImg=GradientImg, MedianFilters=MedianFilters, StatFilter=StatFilter

    IF N_ELEMENTS(MedianFilters) NE 1 THEN MedianFilters = 12

    ImgSize = size(Img)
    dx = ImgSize(1)
    dy = ImgSize(2)

    FiltImg = MEDIAN(Img, 3)
    MedianCount = MedianFilters -1

    ; Apply MedianFilters of median filters
    FOR m=1, MedianCount DO $
        FiltImg = MEDIAN(TEMPORARY(FiltImg), 3)

    ; Apply statistical filter?
    IF dx GT 6 AND dy GT 6 AND KEYWORD_SET(StatFilter) THEN BEGIN
        sigma = STDEV(FiltImg,ImgMean)
        sigma = 0.001 > sigma
        FiltImg = LEEFILT(TEMPORARY(FiltImg),3,sigma*0.5)
    ENDIF

    GradientImg = SOBEL(FiltImg)
    ; Clean-up edges of image
    GradientImg(0:1,*) = 0
    GradientImg(dx-2:*,*) = 0
    GradientImg(*,0:1) = 0
    GradientImg(*,dy-2:*) = 0

    RETURN, FiltImg

END


;===============================================================================
; ConvertROIList2Array: Converts a linked-list of ROIs into an array of
;   imageheader30 structures.

FUNCTION ConvertROIList2Array, pObjList

    IF NOT PTR_VALID(pObjList) THEN RETURN, -1
    IF N_TAGS(*pObjList) LT 2 THEN RETURN, -1

    ; How many Objects in list?
    ListLen = (*pObjList).CountNext +1
    IF ListLen LT 1 THEN RETURN, -1

    ; Allocate array of imageheader31 structures
    ROIArray = REPLICATE({imageheader31}, ListLen)

    pThisNode = pObjList
    FOR i=0, ListLen-1 DO BEGIN
        ; Copy over info fields
        ROIArray[i].PixBytes = 0
        ROIArray[i].Flags = (*pThisNode).Flags
        ROIArray[i].Dark = (*pThisNode).Dark
        ROIArray[i].Area = (*pThisNode).Area
        ROIArray[i].Perim = (*pThisNode).Perimeter
        ROIArray[i].Len = (*pThisNode).Len
        ROIArray[i].Wid = (*pThisNode).Width
        ROIArray[i].Centroid = (*pThisNode).Centroid
        ROIArray[i].Roundness1 = (*pThisNode).Roundness1
        ROIArray[i].Roundness2 = (*pThisNode).Roundness2
        ROIArray[i].Roundness3 = (*pThisNode).Roundness3
        ROIArray[i].Sample_period = 1
        ROIArray[i].Focus = ROUND((*pThisNode).Focus[4])
        ROIArray[i].FocusMin = ROUND((*pThisNode).Focus[2])
        ROIArray[i].FocusMax = ROUND((*pThisNode).Focus[3])
        ROIArray[i].FocusStdDev = (*pThisNode).Focus[7]
        ROIArray[i].FocusGoodSamples = ROUND((*pThisNode).Focus[1])
        ROIArray[i].Focus_Stats = (*pThisNode).Focus[1:*]
        ROIArray[i].CutOff = ROUND((*pThisNode).CutOff)
        ROIArray[i].Crystal = (*pThisNode).Crystal
        ;JS Mod
        if((size((*pThisNode).Harmonics,/dim))[0] eq 1)then harmonic_len = 0 else harmonic_len = N_ELEMENTS((*pThisNode).Harmonics)
        ROIArray[i].Harmonic_len = harmonic_len
        ROIArray[i].Harmonics = PTR_NEW((*pThisNode).Harmonics)

        IF N_ELEMENTS((*pThisNode).Radii) GT 1 THEN BEGIN
            ROIArray[i].Radii_len = N_ELEMENTS((*pThisNode).Radii)
            ROIArray[i].Radii = PTR_NEW((*pThisNode).Radii)
        ENDIF ELSE BEGIN
            ROIArray[i].Radii_len = 0
            ROIArray[i].Radii = PTR_NEW()
        ENDELSE
        ROIArray[i].Radii_Stats = (*pThisNode).Radii_Stats
        ROIArray[i].Intensity_Stats = (*pThisNode).Intensity_Stats
        ROIArray[i].Intensity_HoleLess_Stats = (*pThisNode).Intensity_HoleLess_Stats
        ROIArray[i].Gradient_Stats = (*pThisNode).Gradient_Stats

        ROIArray[i].Hole_Count = (*pThisNode).Hole_Count
        ROIArray[i].Hole_Area_Stats = (*pThisNode).Hole_Area_Stats
        ROIArray[i].Hole_Intensity_Stats = (*pThisNode).Hole_Intensity_Stats
        ROIArray[i].Hole1_Area = (*pThisNode).Hole1_Area
        ROIArray[i].Hole1_Intensity_Stats = (*pThisNode).Hole1_Intensity_Stats
        ROIArray[i].Hole2_Area = (*pThisNode).Hole2_Area
        ROIArray[i].Hole2_Intensity_Stats = (*pThisNode).Hole2_Intensity_Stats
        ROIArray[i].Hole3_Area = (*pThisNode).Hole3_Area
        ROIArray[i].Hole3_Intensity_Stats = (*pThisNode).Hole3_Intensity_Stats

        IF N_ELEMENTS((*pThisNode).Curvature) GT 1 THEN BEGIN
            ROIArray[i].Curvature_len = N_ELEMENTS((*pThisNode).Curvature)
            ROIArray[i].Curvature = PTR_NEW((*pThisNode).Curvature)
        ENDIF ELSE BEGIN
            ROIArray[i].Curvature_len = 0
            ROIArray[i].Curvature = PTR_NEW()
        ENDELSE
        ROIArray[i].Curvature_Stats = (*pThisNode).Curvature_Stats
        ROIArray[i].Curvature_Window = (*pThisNode).Curvature_Window

        ; Correct coordinate information
        ; All coordinates in the ROI list are w.r.t. the ROI image
        ; We need to make them relative to the bounding box specified by SPBox,EPBox
        ImgPosX = (*pThisNode).Obj[0]
        ImgPosY = (*pThisNode).Obj[1]

        ; sx, sy, ex, ey w.r.t. ROI image
        ;sx = (*pThisNode).SPBox MOD (*pThisNode).dx
        ;sy = (*pThisNode).SPBox / (*pThisNode).dx
        ;ex = (*pThisNode).EPBox MOD (*pThisNode).dx
        ;ey = (*pThisNode).EPBox / (*pThisNode).dx
        ;ParticleDX = ex - sx + 1
        ;ParticleDY = ey - sy + 1

        ; TEMPORARY FIX BEGIN
        ; Make all coordinates relative to the original ROI box
        sx = 0
        sy = 0
        ex = (*pThisNode).dx -1
        ey = (*pThisNode).dy -1
        ParticleDX = (*pThisNode).dx
        ParticleDY = (*pThisNode).dy
        ; TEMPORARY FIX END

        ; Make everything relative to bounding box
        x = (*pThisNode).Len1 MOD (*pThisNode).dx
        y = (*pThisNode).Len1 / (*pThisNode).dx
        x = (x - sx) > 0
        y = (y - sy) > 0
        ROIArray[i].SLen = x + y*ParticleDX

        x = (*pThisNode).Len2 MOD (*pThisNode).dx
        y = (*pThisNode).Len2 / (*pThisNode).dx
        x = (x - sx) > 0
        y = (y - sy) > 0
        ROIArray[i].ELen = x + y*ParticleDX

        x = (*pThisNode).Wid1 MOD (*pThisNode).dx
        y = (*pThisNode).Wid1 / (*pThisNode).dx
        x = (x - sx) > 0
        y = (y - sy) > 0
        ROIArray[i].SWid = x + y*ParticleDX

        x = (*pThisNode).Wid2 MOD (*pThisNode).dx
        y = (*pThisNode).Wid2 / (*pThisNode).dx
        x = (x - sx) > 0
        y = (y - sy) > 0
        ROIArray[i].EWid = x + y*ParticleDX

        ; Boundary points
        x = ((*pThisNode).Obj)[2:*] MOD (*pThisNode).dx
        y = ((*pThisNode).Obj)[2:*] / (*pThisNode).dx
        x = (x - sx) > 0
        y = (y - sy) > 0
        ROIArray[i].Boundary = PTR_NEW(x + y*ParticleDX)
        ROIArray[i].Bound_len = N_ELEMENTS(x)

        ; Bounding box coordinates need to be relative to the full frame
        ROIArray[i].sx = ImgPosX + sx
        ROIArray[i].ex = ImgPosX + ex
        ROIArray[i].sy = ImgPosY + sy
        ROIArray[i].ey = ImgPosY + ey

        ; Move on to the next node
        pThisNode =    (*pThisNode).Next
    ENDFOR

    RETURN, ROIArray

END

;===============================================================================
; FindParticles:
;===============================================================================
FUNCTION FindParticles, Img, GradientThresh, Thresh, MedianFilters, YXAspect, ImgPos, ENLARGE_AREA=Enlarge_Area, Roundness2=Roundness2,Roundness3=Roundness3

    ;errno = 0
    ;catch,errno
    ;if(errno ne 0)then begin
    ;    PRINT, 'Error message:', !ERR_STRING
    ;    help,/trace
    ;    return,-1
    ;endif


    ; Compute filtered and gradient image
    FilteredImg = FilterImg (Img, GradientImg=GradientImg, MedianFilters=MedianFilters)

    pSubObjList = FindObjects(FilteredImg, GradientImg, GradientThresh, Thresh, $
        SEGMENTEDIMG=SegmentedImg, BINIMG=BinImg, HOLESIMG=HolesImg, YXAspect=YXAspect, $
        ENLARGE_AREA=Enlarge_Area)

    IF (PTR_VALID(pSubObjList)) THEN BEGIN
        ; Screen for bogus particles
        pSubObjList = ScreenObjects(pSubObjList)
        ; Compute particle parameters: area, perim, etc.
        pSubObjList = ComputeParticleParams(pSubObjList, Img, GradientImg, SegmentedImg, BinImg, HolesImg, YXAspect=YXAspect, IMGPOS=ImgPos, Roundness2=Roundness2, Roundness3=Roundness3)

        IF (PTR_VALID(pSubObjList)) THEN BEGIN
            ROIArray = ConvertROIList2Array(pSubObjList)
            ; Delete the source list, as we are done with it
            DeleteROIList, pSubObjList

            RETURN, ROIArray
        ENDIF
    ENDIF

    RETURN, -1
END; FindParticles


;===============================================================================
; ProcessParticles: Returns an array of imageheader30 structures
;===============================================================================
FUNCTION ProcessParticles, FrameData, backImage, GradientThreshold=GradientThresh, Threshold=Thresh, $
    MedianFilters=MedianFilters, YXAspect=YXAspect, display = display, ENLARGE_AREA=Enlarge_Area

; How many ROIs?
ListLen = FrameData.FrameHeader.ROIsCount
IF ListLen LT 1 THEN RETURN, -1
; Sanity checks
IF ListLen NE N_ELEMENTS(*(FrameData.pFrame)) THEN RETURN, -1

; Default processing parameters
IF N_ELEMENTS(GradientThresh) EQ 0 THEN GradientThresh = 40
IF N_ELEMENTS(Thresh) NE 1 THEN Thresh = -25
IF N_ELEMENTS(MedianFilters) NE 1 THEN MedianFilters = 12
IF N_ELEMENTS(YXAspect) NE 1 THEN YXAspect = 1

; NOTE: Need to deal here with ROIs which may have been incorrectly split
; by the real-time CPI code
; For now, we assume that each ROI image we process below is
; independent of all others
FOR i=0, ListLen-1 DO BEGIN
    pROIImg = (*(FrameData.pFrame))[i].pROI
    imHead = (*(FrameData.pFrame))[i].imageheader

    ; remove backimage
    backROI = BackImage[imHead.sx:imHead.ex,imHead.sy:imHead.ey]
    ; save this back into the data array
    *pROIImg = *pROIImg - fix(backROI)

    NewParticles = FindParticles(*pROIImg, GradientThresh, Thresh, $
        MedianFilters, YXAspect, ENLARGE_AREA=Enlarge_Area, $
        [imHead.sx, imHead.sy])

    ; DISPLAY
    IF KEYWORD_SET(display) THEN BEGIN
        img = *pROIImg
        DMin = MIN(img, MAX=DMax)
        Img = BYTSCL(Img, MIN=DMin, MAX=DMax, TOP=!d.table_size - 2)

        dim = size(img,/dim)
        max_dim = max(dim, min = min_dim)
        scale = fix(300*fix(min_dim)/max_dim)
    ENDIF

    ;Add new particles to our array
    IF N_TAGS(NewParticles[0]) GT 10 THEN BEGIN
        IF KEYWORD_SET(display) THEN $
            FOR k=0, N_ELEMENTS(NewParticles)-1 DO BEGIN
            ; Modify boundary points to be refenced w.r.t ROI sx,sy
            dx = NewParticles[k].ex - NewParticles[k].sx + 1
            BoundX = (*NewParticles[k].boundary MOD dx) + (NewParticles[k].sx - imHead.sx)
            BoundY = (*NewParticles[k].boundary / dx) + (NewParticles[k].sy - imHead.sy)
            ; Mark boundary for display
            img[BoundX, BoundY] = !d.table_size - 1
        ENDFOR

        IF N_ELEMENTS(ParticleArray) LT 1 THEN ParticleArray = NewParticles $
            ELSE ParticleArray = [ParticleArray, NewParticles]
    ENDIF

    IF KEYWORD_SET(display) THEN BEGIN
        erase
        if(dim[0] eq max_dim)then begin
            ; force to 300 columns
            tv,congrid(img,300,scale)
        endif else begin
            ; for to 300 rows
            tv,congrid(img,scale,300)
        endelse
        wait,.1
    ENDIF
ENDFOR

IF N_ELEMENTS(ParticleArray) LT 1 THEN RETURN, -1
stop
RETURN, ParticleArray
END; old processparticle.pro

; ;;;;;;;;;;;;;;;;;;  added following new parts for aaron's stuff  **********************


FUNCTION processparticles_ab, cimage, dp, yxaspect,Img,gradientThresh,Thresh,MedianFilters,Sxy,hole_count,total_hole_area, Roundness2=Roundness2, Roundness3=Roundness3
    ; the last 5 input is for calculating carlos focus only.
    ;stop

    ; this is from arraon bamster
    Forward_function particleshapetype_mo, shapeHarmonics_mo,  shapelen_mo,  shapewidth_mo, mode_mo,abparams2
    forward_function get_mean_brightness_inside_boundary

    data = {imageHeader31,BlockMark:0U,    $
        ROIInfoSize:0UL,$  ; this is variable because of the last two fields in the header and the data size
        ROIVer:31U,       $ ; should be 31
        sx:0U,        $ ; upper left corner X of image in full frame co-ords
        sy:0U,        $ ; upper left corner Y of image in full frame co-ords
        ex:0U,        $ ; lower right corner X of image in full frame co-ords
        ey:0U,        $ ; lower right corner Y of image in full frame co-ords
        PixBytes:0U,   $    ; if this is zero, no image data is included!
        Flags:0U,   $   ; contains info about crstal type, focus, reject and manual parameters
        Len:0.0,     $    ; length of major axis in pixels
        SLen:0UL,   $   ; 1D index of length vector start point in Sub-array coords
        ELen:0UL,   $   ; 1D index of length vector stop point in Sub-array coords
        Wid:0.0,     $    ; lenght of axis perpendicular to major axis in pixels
        SWid:0UL,   $   ; 1D index of width vector start point in Sub-array coords
        EWid:0UL,   $   ; 1D index of width vector start point in Sub-array coords
        Dark:0U,     $    ; this field is called usROIDepth in header, contains depth and cutoff info
        Area:0.0,   $   ; computed area   in pixels
        Perim:0.0,     $  ; computed perimeter in pixels
        Centroid:[0.,0.],$ ; Centroid of particle ( 2 elements )
        Roundness1:0.0, $   ; Measure of particle roundness [using method 1, area vs circular area for diameter]
        X_Moment:0.0,  $    ; 1/2 area on each side of vert line
        Y_Moment:0.0,  $    ; 1/2 area on each side of horiz line
        Rubber_Band:0.0,$  ; Measure of boundary roughness
        Fractal_dim:0.0,$  ; Fractal dimension
        Harmonic_len:0U,$  ; Number of elements in Harmonic array
        Bound_len:0U,  $    ; Number of elements in Boundary array
        Sample_period:0.0,$    ; Sample period of harmonic array in pixels
        Focus:0B,   $   ; % focus 0-100. This is the average focus value for multiple samples along the particle boundary
        FocusMin:0B,   $ ; Min % focus of samples taken
        FocusMax:0B,   $ ; Max % focus of samples taken
        FocusStdDev:0.0,$  ; Standard deviation of focus samples taken
        FocusGoodSamples:0B,$; Percent of samples which yielded valid focus values: 0 - 100
        CutOff:0B,     $  ; % particle is cut-off 0 - 100
        Crystal:0B,       $ ; Scalar type index meaning TBD
        Confidence:0B, $   ; Factor assessing confidence in particle use 0-10
        P1:0.0,         $    ; used in manualFeature as input of status of manual setting, sent by view_event
        Roundness2:0.0,         $    ; Roundness 2, stddev of distances to centroid
        Roundness3:0.0,         $    ; Roundness 3, ratio of perimeter angles total to expected
        P4:0.0,         $    ; Place holder

        Radii_len:0U,  $    ; Number of elements in the radii array below
        Curvature_len:0U,$ ; Number of elements in the radius of curvature array below
        Curvature_Window:0U,$; Number of points used in the local radius of curvature measurement
        ; For the stats arrays that follow, the elements are:
        ; N: number of data points used,
        ; Min, Max, Mean, Median, Mode, Standard Deviation, Skewness, Kurtosis
        Focus_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], $
        Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], $
        Intensity_HoleLess_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],    $
        Gradient_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  $
        Radii_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], $
        ; Local radius of curvature stats
        Curvature_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], $
        ; Below are stats computed for holes within the particle
        Hole_Count:0U, $   ; Number of holes found1
        ; Stats including all holes
        Hole_Area_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
        Hole_Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
        Hole_Other_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
        ; Values for the biggest 3 holes
        Hole1_Area:0.0,    $
        Hole1_Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
        Hole1_Perim:0.0,   $
        Hole1_Other_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
        Hole2_Area:0.0,    $
        Hole2_Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
        Hole2_Perim:0.0,   $
        Hole2_Other_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
        Hole3_Area:0.0,    $
        Hole3_Intensity_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$
        Hole3_Perim:0.0,   $
        Hole3_Other_Stats:[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],$

        Harmonics:ptr_new(),$; Circular harmonics (float) of boundary(var. len)
        Boundary:ptr_new(),    $; Location (long) of boundary pixels in ROI array ( 1D indicies) (var. len)
        Radii:ptr_new(), $  ; Radial distance (float) from the centroid to the perimeter of the particle measured in Sample_period pixel steps
        Curvature:ptr_new()} ; Local radius of curvature (float) of the particle boundary, in pixels


    s=size(cimage,/dim)
    ab=abparams2(cimage,dp = dp / !CPI_MICRONS_PER_PIXEL, contour=0.6, res=1.0)  ;pixel coordinates
    IF ab.error THEN return,-1

    ;roundness=ab.area/(!pi/4*(ab.diam^2))
    ; added this by Qixu Mo on Feb 15,2008.
    make_xy_vectors_with_maximum_eq_1_distances,ab.x,ab.y,xn,yn,xnn,ynn

    mean_brightness=get_mean_brightness_inside_boundary(img,ab.x,ab.y);  added May8, 2009.

    if mean_brightness lt 0 then begin
        ;   stop
    endif

    ; This finds the length of the shape (really the longest dimension)
    ObjLen = ShapeLen_mo(xnn, ynn,lp1, lp2, YXAspect=YXAspect);xnn,ynn are input. Lp1 lp2 are outputs.

    ; this is the perpendicular angle from longest dimension
    angle = atan(ynn[lp2]-ynn[lp1], xnn[lp2]-xnn[lp1])

    ; this returns the width of the object
    ObjWidth = ShapeWidth_mo (xnn,ynn,angle, wp1, wp2, YXAspect=YXAspect)


    ; insert following to calculate the carlos focus values.
    ;**************************************************************************************
    FilteredImg = FilterImg (Img, GradientImg=GradientImg, MedianFilters=MedianFilters)
    Focus_carlos = ComputeFocus(xnn, ynn, GradientImg);    index (4) is the focus values of the image. ==the 5th element.
    focus_of_image =focus_carlos(4)
    ;*********************************************************************************

    ;recaluclate roundness, perimeter, area and diameter of the xnn ynn curve.
    ;*************************************************************************
    ;see if it is a sphere. for a sphere   y2=r2-x2
    xc=mean(xnn)
    yc=mean(ynn)
    x=xnn-xc
    y=ynn-yc
    x2=x*x
    y2=y*y
    ris=sqrt(x2+y2)
    mean_r=mean(ris)
    absdev=mean(abs(ris-mean_r))

    Pmo=total(sqrt((x-shift(x,1))^2+(y-shift(y,1))^2));  perimeter
    Amo=total((x*shift(y,1)-y*shift(x,1)))/2.0 ; area;  it can be negative depending on the loop direction.
    roundness1=4.0*abs(Amo)/(!Pi*objLen*objLen)
    Area=abs(Amo)
    perimeter=Pmo;   ; x and y in here are closed loop boundary line.


    ; TAF this code is executed AFTER the main process particles, and is highly redundant.
    ; adding Roundness2 and Roundness3 to function call to add to the data structure here, temporarily




    ;     it has  been tested than uisng x y and using xnn ynn give the same Pmo Amo results.

    ;   window,2,ysize=72*4.5,xsize=72*8
    ;   !P.thick=1
    ;  !P.CHARSIZE=1
    ;  !P.multi=[0,1,1]
    ;  loadct,0
    ;  white=!D.N_COLORS-1
    ;  !P.BACKGROUND=white

    ;    plot,ris, color=0,title='  New Rshape',$
    ;   xstyle=1,ystyle=1
    ;   write_png,'C:/New_rshape_x.png',tvrd(true=1),/order

    ; stop
    ;************************************************************************
    ;Rejection
    Reject=0
    Aspect = float(ObjLen)/(ObjWidth)
    IF Aspect LT 1 THEN Aspect = 1/Aspect
    IF Aspect GT 20 THEN Reject=1

    IF ObjLen lt 2 THEN Reject=1

    cutoff=1-float(ab.ngood)/n_elements(ab.x)
    IF cutoff gt 0.4 THEN Reject=1  ;More than 40% on edge

    Harmonics = -1L
    rShape = -1L
    IF N_ELEMENTS(ab.x) GT 3  THEN BEGIN
        centroid=[xc,yc]
        Harmonics=ShapeHarmonics_mo(xnn, ynn, centroid,rShape=rShape);   the rshape in here is almost the same as ris above.

        ;window,2,ysize=72*4.5,xsize=72*8
        ;plot,rshape, color=0,title='  New Rshape',$
        ;xstyle=1,ystyle=1
        ;write_png,'C:/New_rshape_x.png',tvrd(true=1),/order


        IF N_ELEMENTS(rShape) GT 1 THEN BEGIN
            MinVal = MIN(rShape, MAX=MaxVal)
            Moments = MOMENT(rShape, SDEV=Sigma)
            Radii_Stats = [N_ELEMENTS(rShape),MinVal,MaxVal,Moments[0],MEDIAN(rShape,/EVEN),mode_mo(rShape),Sigma,Moments[2],Moments[3]]
        ENDIF
    ENDIF

    ;Crystal
    Crystal=ParticleShapeType_mo(xnn, ynn,xc,yc,objLen,objWidth, Area, Perimeter,hole_count,total_hole_area,mean_brightness,$
        HARMONICS=Harmonics, Roundness1=Roundness1, Roundness2=Roundness2, Roundness3=Roundness3, rShape=rShape)



    ; above crystal is a value of 1, 15 or so to see if it is sphere or unknown type.


    ; ========= Intensity Stats =================
    ImgVals = cimage
    MinVal = MIN(ImgVals, MAX=MaxVal)
    Moments = MOMENT(ImgVals, SDEV=Sigma)
    Intensity_Stats = [N_ELEMENTS(ImgVals),MinVal,MaxVal,Moments[0],MEDIAN(ImgVals,/EVEN),mode_mo(ImgVals),Sigma,Moments[2],Moments[3]]

    ; ========= Gradient Stats =================
    ;dummy=filterimg(cimage,gradientimg=gradientimg)
    ;GradVals = GradientImg(cimage)
    ;MinVal = MIN(GradVals, MAX=MaxVal)
    ;Moments = MOMENT(GradVals, SDEV=Sigma)
    ;GradStats = [N_ELEMENTS(GradVals),MinVal,MaxVal,Moments[0],MEDIAN(GradVals,/EVEN),mode_mo(GradVals),Sigma,Moments[2],Moments[3]]


    OutFocus = (100 - ab.Focus[0]) > 0
    Crystal = Crystal AND '000F'X
    OutFocus = BYTE(OutFocus) + ISHFT(Crystal,8)

    ;Change the boundary to 1-D coordinates
    roundx=round(ab.x)
    roundy=round(ab.y)
    boundary_1d=s[0]*roundy+roundx;  which is not xnn, ynn. it should be replace with xnn, ynn.

    ;******************************* correct teh length, width, perimeter and area by focus values.
    ; factor_by_focus=(objlen-(100.0-(ab.focus < 100))/4.0)/objlen;   take this away on June 27, 2008.
    ;**********************************************************************************************
    ;Stuff variables
    data.Flags = FIX(OutFocus)
    data.len=objlen;  in pixels.   This is Lob---Length of observed length in pixels.
    data.SLen=lp1
    data.ELen=lp2
    data.Wid=objwidth; This is Wob---width of observed width in pixels.
    data.SWid=wp1
    data.EWid=wp2
    data.Dark=ab.thresh
    data.Area=(ab.area);  area of observed, in pixels.
    data.Perim=(ab.perimeter);*factor_by_focus
    data.Centroid=[ab.xcentroid,ab.ycentroid];*factor_by_focus
    data.Roundness1=roundness1
    data.Roundness2=roundness2
    data.Roundness3=roundness3
    ;stop
    data.Harmonic_len=N_ELEMENTS(Harmonics)
    data.Bound_len=n_elements(ab.x)
    data.Sample_period=1
    data.Focus=focus_of_image;     changed in May 5, 2009, by mo -----ab.focus
    data.FocusGoodSamples=100
    data.CutOff=cutoff*100
    data.Crystal=crystal

    data.Radii_len=n_elements(rshape)
    data.Intensity_Stats=Intensity_Stats
    ;data.Gradient_Stats=gradstats
    data.Radii_Stats=data.radii_stats
    data.Harmonics=ptr_new(harmonics)
    data.Boundary=ptr_new(boundary_1d)  ;([[ab.x],[ab.y]])
    data.Radii=ptr_new(rshape)

    ;if data.len lt -2.0 and data.wid lt -2.0 then begin
    ; print,factor_by_focus
    ;        ; stop
    ; endif

    return,data
END; end of processparticles_ab

;===============================================================================

;===============================================================================
;===============================================================================
;===============================================================================
FUNCTION ParticleShapeType_mo, x, y, centroidx, centroidy, Len, Width, Area, Perimeter,$
    hole_count,total_hole_area,mean_brightness,HARMONICS=Harmonics, Roundness1=Roundness1, Roundness2=Roundness2, Roundness3=Roundness3, rShape=rShape

;************************************************************************************
;print, "in ParticleShapeType_mo"
;stop
IF N_ELEMENTS(rShape) LT 4 THEN begin
    RETURN, 15
endif

; Classify only spherical water drops for now
; TAF attempting to revert to CPI 22 to recover
IF ABS(Roundness1 - 1) LT 0.25 THEN BEGIN

    ; IF len lt 4.0 then begin
    ;    if ABS(Roundness1) gt 0.50 and mean_brightNess lt 95.0 then RETURN, 1   ; Water = 1
    ;ENDIF

    ;if Len lt 30.0/2.3 then begin
    ;     rnd_crit=0.30
    ; endif else begin
    ;     if Len gt 75.0/2.3 then rnd_crit=0.75 else rnd_crit=Len/2.3/100.0;   should it be 1.0 for maximum ?
    ; endelse


    Sigma = STDEV(rShape, MeanRadius)
    MinR = MIN(rShape, MAX=MaxR)

    IF Sigma LE 2 THEN Sigma = 0

    IF ABS(MinR - MeanRadius) LT 2 THEN MinR = MeanRadius
    IF ABS(MaxR - MeanRadius) LT 2 THEN MaxR = MeanRadius

    MaxDeviate = 0.1*MeanRadius

    ;IF(Sigma LT MaxDeviate AND ABS(MaxR - MeanRadius) LT MaxDeviate $
    ;   AND ABS(MinR - MeanRadius) LT MaxDeviate and $; $ mean_brightNess lt 95.0 AND hole_count eq 0
    ;   Roundness1 gt !Roundness1_min and Roundness1 lt !Roundness1_max and $
    ;   Roundness2 gt !Roundness2_min and Roundness2 lt !Roundness2_max and $
    ;   Roundness3 gt !Roundness3_min and Roundness3 lt !Roundness3_max) then begin

    IF (Sigma LT MaxDeviate AND ABS(MaxR - MeanRadius) LT MaxDeviate $
        AND ABS(MinR - MeanRadius) LT MaxDeviate )  then  RETURN,1;   and $; $ mean_brightNess lt 95.0 AND hole_count eq 0
        ;Roundness1 gt .75 and Roundness1 lt 1.25) then begin
        ;TAF removed roundness1 below, added roundness1-3 above
        ;roundness1 > rnd_crit) then begin

        ; TAF insert Roundness2 and Roundness3 criteria here
        ;    RETURN, 1   ; Water = 1

        ;IF Sigma LT MaxDeviate AND ABS(MaxR - MeanRadius) LT MaxDeviate $
        ;    AND ABS(MinR - MeanRadius) LT MaxDeviate THEN RETURN, 1   ; Water = 1

    ENDIF

RETURN, 15 ; Ice = 15
END



;===============================================================================
FUNCTION ShapeHarmonics_mo, BoundaryPtsX, BoundaryPtsY, Centroid, rShape=rShape
    BFCount = N_ELEMENTS(BoundaryPtsX)
    IF BFCount LT 4 THEN RETURN, -1

    ; Compute r, pd w.r.t. centroid (pd: perimeter distance)
    ; First compute r, pd for each boundary pt.
    DeltaX = FLOAT(BoundaryPtsX - Centroid[0])
    DeltaY = FLOAT(BoundaryPtsY - Centroid[1])
    BoundaryR = SQRT(DeltaX^2 + DeltaY^2)

    DeltaX = ABS(BoundaryPtsX - SHIFT(BoundaryPtsX, 1))
    DeltaY = ABS(BoundaryPtsY - SHIFT(BoundaryPtsY, 1))
    BoundaryPD = FLTARR(BFCount)
    BoundaryPD[0] = 0

    FOR i=1, BFCount-1 DO BEGIN
        IF DeltaX[i] LT 2 AND DeltaY[i] LT 2 THEN BEGIN
            IF DeltaX[i] + DeltaY[i] EQ 2 THEN BoundaryPD[i] = BoundaryPD[i-1] + 1.34 $
                ELSE BoundaryPD[i] = BoundaryPD[i-1] + 0.948
        ENDIF ELSE BEGIN
            BoundaryPD[i] = BoundaryPD[i-1] + SQRT(DeltaX[i]^2 + DeltaY[i]^2)
        ENDELSE
    ENDFOR

    ; Using a fixed sampling period of ~1 pixel
    Samples = ROUND(BoundaryPD[BFCount-1] +1)
    r = FLTARR(Samples)
    r[0] = BoundaryR[0]
    dpd = BoundaryPD[BFCount-1]/(Samples-1)
    pd = dpd
    pt2 = 1

    ; Now do Samples steps along perimeter
    i = 1
    REPEAT BEGIN
        WHILE pd GT (BoundaryPD[pt2] + 0.001) AND pt2 LT BFCount-1 DO pt2 = pt2 + 1
        IF pt2 GE BFCount-1 THEN BEGIN
            r[i] = BoundaryR[BFCount-1]
        ENDIF ELSE BEGIN
            ; Interpolate for r
            Slope = (BoundaryR[pt2] - BoundaryR[pt2-1])/(BoundaryPD[pt2] - BoundaryPD[pt2-1])
            r[i] = BoundaryR[pt2-1] + (pd - BoundaryPD[pt2-1])*Slope
        ENDELSE

        i = i + 1
        pd = pd + dpd
    ENDREP UNTIL (pd GT (BoundaryPD[BFCount-1] + 0.001))

    IF i LT Samples THEN r[i:Samples-1] = r[i-1]

    LastWind = -1
    ;IF !D.WINDOW GT -1 THEN LastWind = !D.WINDOW
    ;WINDOW, 1, XPOS=0, YPOS=0, xsize = 300, ysize=300,TITLE='Object Shape'
    ;WINDOW, 2, XPOS=0, YPOS=330, xsize = 300, ysize=300,TITLE='Object Harmonics'

    ; Plot it
    ;WSET, 1
    ;PLOT, r

    rShape = r

    r = (ABS(FFT(r)))[0:Samples/2]
    r = r/MAX(r)
    PlotPts = N_ELEMENTS(r)
    ;WSET, 2
    ;PLOT, r[0:(PlotPts-1) < 40]
    ;IF LastWind GT -1 THEN WSET, LastWind
    if n_elements(r) lt 6 then r=[r,0,0,0,0,0,0]
    RETURN, r
END



;**************************************************************************
;===============================================================================
FUNCTION ShapeLen_mo, ObjPtsX, ObjPtsY, lp1, lp2, YXASPECT=YXAspect

    IF N_ELEMENTS(YXASPECT) EQ 0 THEN YXASPECT = 1

    ObjPtsCount = N_ELEMENTS(ObjPtsX)

    MaxDist = 0.0
    FOR i=0L, ObjPtsCount-2 DO BEGIN
        ; these are the delta x and delta y from the ith point to all the others
        dx = long(ObjPtsX[i] - ObjPtsX[i+1:ObjPtsCount-1])
        dy = long(ObjPtsY[i] - ObjPtsY[i+1:ObjPtsCount-1])

        ; length from ith point to all others
        len = sqrt(FLOAT(dx*YXASPECT)^2 + FLOAT(dy)*dy)
        ;update MaxDist to be the longest so far
        MaxDist = max(len) > MaxDist

        indx = where(len EQ MaxDist, Matches)
        IF (Matches GT 0) THEN BEGIN
            ; here if there are more than 1 longest vectors; lp1 lp2 are the longest length two points.
            IF ObjPtsX[i] LT ObjPtsX[i+1+indX[0]] THEN BEGIN
                lp1 = i   ;ab changed all these to indices
                lp2 = i+1+indX[0]
            ENDIF ELSE BEGIN
                lp2 = i
                lp1 = i+1+indX[0]
            ENDELSE
        ENDIF
    ENDFOR

    return, MaxDist
END

pro make_xy_vectors_with_maximum_eq_1_distances_old,x,y,xn,yn,xnn,ynn
    nnn=n_elements(x)
    xn=x(0)
    yn=y(0)
    for i=1, nnn-1 do begin
        ; There are 3 case;  xi > xi_1    xi <xi_1    xi=xi_1
        ;case 1;   x(i) gt x(i-1)
        ;*****************************************************************
        if x(i) gt x(i-1) then begin; ;There are 2 case;   xb ge xa or xa ge xb
            slope=double(y(i)-y(i-1))/double(x(i)-x(i-1))
            xb=floor(x(i))
            xa=ceil(x(i-1))
            deltx=xb-xa; if deltx gt 0 then there is at least one pixel distance between the two points.

            if deltx ge 1 then begin
                xinsert=indgen(deltx)+xa
                yinsert=slope*(Xinsert-x(i-1)) +y(i-1)
                xn=[temporary(xn),x(i-1),xinsert,x(i)]
                yn=[temporary(yn),y(i-1),yinsert,y(i)]
            endif else begin;  deltx =0 or less than 0.
                ;xn=[temporary(xn),x(i-1),x(i)]
                ;yn=[temporary(yn),y(i-1),y(i)]
                xn=[temporary(xn),x(i)]
                yn=[temporary(yn),y(i)]
            endelse

        endif;  if x(i)  gt x(i-1)

        ;case 2
        ;***************************************************************
        if x(i) lt x(i-1) then begin
            slope=double(y(i)-y(i-1))/double(x(i)-x(i-1))
            xa=floor(x(i-1))
            xb=ceil(x(i))
            deltx=xa-xb

            if deltx ge 1 then begin
                xinsert=[reverse(indgen(deltx)+xb)]
                yinsert=slope*(Xinsert-x(i-1))+y(i-1)
                xn=[temporary(xn),x(i-1),xinsert,x(i)]
                yn=[temporary(yn),y(i-1),yinsert,y(i)]
            endif else begin;  deltx =0
                xn=[temporary(xn),x(i-1),x(i)]
                yn=[temporary(yn),y(i-1),y(i)]
            endelse
        endif;if x(i) lt x(i-1)

        ;case 3.
        ;****************************************************
        if x(i) eq x(i-1) then begin
            xn=[temporary(xn),x(i)]
            yn=[temporary(yn),y(i)]
        endif


    endfor

    ; do it again for yn xn
    xt=yn
    yt=xn
    nnn=n_elements(xt)
    xnn=xt(0)
    ynn=yt(0)
    for i=1, nnn-1 do begin
        if xt(i) gt xt(i-1) then begin; ;There are 2 case;   xtb ge xta or xta ge xtb
            slope=double(yt(i)-yt(i-1))/double(xt(i)-xt(i-1))
            xb=floor(xt(i))
            xa=ceil(xt(i-1))
            deltx=xb-xa; if deltx gt 0 then there is at least one pixel distance between the two points.

            if deltx ge 1 then begin
                xinsert=indgen(deltx)+xa
                yinsert=slope*(Xinsert-xt(i-1)) +yt(i-1)
                xnn=[temporary(xnn),xt(i-1),xinsert,xt(i)]
                ynn=[temporary(ynn),yt(i-1),yinsert,yt(i)]
            endif else begin;  deltx =0
                xnn=[temporary(xnn),xt(i-1),xt(i)]
                ynn=[temporary(ynn),yt(i-1),yt(i)]
            endelse

        endif;  if x(i)  gt x(i-1)

        ;***************************************************************
        if xt(i) lt xt(i-1) then begin
            slope=double(yt(i)-yt(i-1))/double(xt(i)-xt(i-1))
            xa=floor(xt(i-1))
            xb=ceil(xt(i))
            deltx=xa-xb
            if deltx ge 1 then begin
                xinsert=[reverse(indgen(deltx)+xb)]
                yinsert=slope*(Xinsert-xt(i-1))+yt(i-1)
                xnn=[temporary(xnn),xt(i-1),xinsert,xt(i)]
                ynn=[temporary(ynn),yt(i-1),yinsert,yt(i)]
            endif else begin;  deltx =0
                xnn=[temporary(xnn),xt(i-1),xt(i)]
                ynn=[temporary(ynn),yt(i-1),yt(i)]
            endelse
        endif;  if x(i) lt x(i-1)
        ;****************************************************
        if xt(i) eq xt(i-1) then begin
            xnn=[temporary(xnn),xt(i)]
            ynn=[temporary(ynn),yt(i)]
        endif
    endfor
end; make_xy_vectors_with_maximum_eq_1_distances_old,x,y,xn,yn
;===============================================================================

pro make_xy_vectors_with_maximum_eq_1_distances,x,y,xn,yn,xnn,ynn
    nnn=n_elements(x)
    xk=round(x);  round them to be integers. and then +
    yk=round(y)
    xn=xk(0)
    yn=yk(0)

    for i=1, nnn-1 do begin
        xd=xk(i)-xk(i-1);  x-distance.
        if xd gt 1 then begin; insert npoints between
            npoints=abs(xd)-1LL
            slope=double(y(i)-y(i-1))/double(x(i)-x(i-1))
            xinsert=indgen(npoints)+1+xk(i-1)
            yinsert=round(  slope*(Xinsert-x(i-1)) +y(i-1)  )
            xn=[temporary(xn),xinsert,xk(i)]
            yn=[temporary(yn),yinsert,yk(i)]
        endif

        if xd eq 1 or xd eq 0 or xd eq -1 then begin; points are natural, no inserts.
            xn=[temporary(xn),xk(i)]
            yn=[temporary(yn),yk(i)]
        endif

        if xd lt -1 then begin; insert npoints between.
            npoints=abs(xd)-1LL;     suppose sequence is 8, 5, npoints=2.  ===8,7,6,5===> ([0,1]+6)
            slope=double(y(i)-y(i-1))/double(x(i)-x(i-1));        reverse([6,7])=[7,6]
            xinsert=[reverse(indgen(npoints)+1+xk(i))]
            yinsert=slope*(Xinsert-x(i-1))+y(i-1)
            xn=[temporary(xn),xinsert,xk(i)]
            yn=[temporary(yn),yinsert,yk(i)]
        endif
    endfor

    ; do it again for yn xn
    xt=yn
    yt=xn

    nnn=n_elements(xt)
    xk=round(xt);  round them to be integers. and then +
    yk=round(yt)
    xnn=xk(0)
    ynn=yk(0)

    for i=1, nnn-1 do begin
        xd=xk(i)-xk(i-1);  x-distance.
        if xd gt 1 then begin; insert npoints between
            npoints=abs(xd)-1LL
            slope=double(yt(i)-yt(i-1))/double(xt(i)-xt(i-1))
            xinsert=indgen(npoints)+1+xk(i-1)
            yinsert=round(  slope*(Xinsert-xt(i-1)) +yt(i-1)  )
            xnn=[temporary(xnn),xinsert,xk(i)]
            ynn=[temporary(ynn),yinsert,yk(i)]
        endif

        if xd eq 1 or xd eq 0 or xd eq -1 then begin; points are natural, no inserts.
            xnn=[temporary(xnn),xk(i)]
            ynn=[temporary(ynn),yk(i)]
        endif

        if xd lt -1 then begin; insert npoints between.
            npoints=abs(xd)-1LL;     suppose sequence is 8, 5, npoints=2.  ===8,7,6,5===> ([0,1]+6)
            slope=double(yt(i)-yt(i-1))/double(xt(i)-xt(i-1));        reverse([6,7])=[7,6]
            xinsert=[reverse(indgen(npoints)+1+xk(i))]
            yinsert=slope*(Xinsert-xt(i-1))+yt(i-1)
            xnn=[temporary(xnn),xinsert,xk(i)]
            ynn=[temporary(ynn),yinsert,yk(i)]
        endif
    endfor

end; make_xy_vectors_with_maximum_eq_1_distances,x,y,xn,yn
;===============================================================================


FUNCTION ShapeWidth_Mo, ObjPtsX, ObJPtsY,angle, wp1, wp2, YXASPECT=YXAspect
    IF N_ELEMENTS(YXASPECT) EQ 0 THEN YXASPECT = 1
    ;Slope = -tan(angle)
    IF angle LT 0 THEN angle = angle + !PI
    IF angle LT !PI/2 THEN angle = angle + !PI/2 $
        ELSE angle = angle - !PI/2

    ObjPtsCount = N_ELEMENTS(ObjPtsX)
    wp1=0
    wp2=0
    ; search along this direction for the largest distance ( ie. the width) (see shapeLength)
    MaxDist = 0.0
    FOR i=0L, ObjPtsCount-2 DO BEGIN
        dx = long(ObjPtsX[i+1:ObjPtsCount-1] - ObjPtsX[i])
        dy = long(ObjPtsY[i+1:ObjPtsCount-1] - ObjPtsY[i])
        AngleNow = atan(dy,dx)
        WIndx = where(AngleNow LT 0, Matches)
        IF (Matches GT 0) THEN AngleNow[WIndx] = AngleNow[WIndx] + !PI

        WIndx = where(abs(AngleNow - angle) LT 0.1, Matches)

        IF (Matches GT 0) THEN BEGIN
            len = sqrt(FLOAT(YXASPECT*dx[WIndx])^2 + FLOAT(dy[WIndx])^2)
            MaxDist = max(len) > MaxDist
            indx = where(len EQ MaxDist, Matches)
            IF (Matches GT 0) THEN BEGIN
                IF ObjPtsX[i] LT ObjPtsX[i+1+WIndx[indx[0]]] THEN BEGIN
                    wp1 = i           ;ab changed these all to indices
                    wp2 = i+1+WIndx[indx[0]]
                ENDIF ELSE BEGIN
                    wp2 = i
                    wp1 = i+1+WIndx[indx[0]]
                ENDELSE
            ENDIF
        ENDIF
        ;print,i, maxDist,matches,wp1,wp2
        ;stop
    ENDFOR
    return, MaxDist
END; shapwidth_mo

;**************************************
;===============================================================================
FUNCTION mode_mo, x
    ;A faster way of finding the mode
    ;Aaron Bansemer, NCAR, 2/2007
    sx=x[sort(x)]
    u=uniq(sx)
    dummy=max(u-shift(u,1),i)
    return,sx[u[i]]
END;mode_mo,

Function Get_mean_brightness_inside_boundary, img,BoundaryPtsX, boundaryPTsY
    ;***********************************************************
    sxsy=size(img,/dim);   which is dims
    BoundaryPtsX=[BoundaryPtsX, boundaryPtsX(0)];  close the loop.
    BoundaryPtsy=[BoundaryPtsy, boundaryPtsy(0)];  close the loop
    ;                   BoundaryPts_index= BoundartPtsY*DX +BoundaryPTsX
    ROIout2=obj_new('IDLgrRoi',fix(BoundaryPTsX),fix(boundaryPTsY),color=[255,255,255])
    maskResult = ROIout2 -> ComputeMask(DIMENSIONS = sxsy);  maskresult is either 0 or 255.
    obj_destroy,ROIout2;  destroy teh object so that the computer will run faster

    mask = (maskResult GT 0);   which is just a 0 or 1 matrix.
    maskImg = img * mask

    ;window,4,xs=400,ys=400;  This display a red images of the droplet.
    ;TV,maskResult
    ;TV,maskImg;    ; This display a blue images. maskedimage is enclosed by boundary line.
    index=where(maskImg gt 0,contGGG)
    if contGGG eq 0 then return, 0
    mean_brightness=total(maskimg(index))/float(contGGG)
    return, mean_brightness
end; Function Get_mean_brightness_inside_boundary, img,dx,dy,BoundaryPtsX, boundaryPTsY, mean_brightness
