function read_backGnd,FileUnit,ImgDataVer, FPos, frameheader=frameheader

    ; routine to read background image frames closest to the indicated position
    ; Parameters
    ;   fileunit: Input      open and valid logical file unit to read from
    ;   ImgDataVer:   Input        file version number from PIFileheader
    ;   FPos:     Input/Output    filepos to read from

    ;   Keyword
    ;   FrameHeader   Output         If present, return the frameheader in this variable

    ;   Return value
    ;     The form is
    ;      {ImageHeader:ImageHeader, pBackImg:PTR_NEW()}
    ;      where
    ;          ImageHeader is described in read_frame.pro
    ;          pBackImg points to an array containing the actual background data

    point_lun,FileUnit,fpos
    ;stop
    ;create data structure for roi image
    CASE ImgDataVer OF
        25:frameheader = {frameheader25}
        23:frameheader = {frameheader23}
        21:frameheader = {frameheader21}
        else: BEGIN           ;changed at nov12 of UAV balloon version 40.

            IF ImgDataVer GT 30 THEN frameheader = {frameheader25} $
                ELSE r = dialog_message('Frame version not supported',/error)
        END
    endcase

    ; read the frame header
    readu,FileUnit,FrameHeader
    ; stop
    case FrameHeader.Ver of
        21:ImageHeader = {imageheader21}
        23:ImageHeader = {imageheader21} ;??? are there changes in this version???
        25:ImageHeader = {imageheader25}
        else: BEGIN
            ; stop
            IF FrameHeader.Ver GT 30 THEN ImageHeader = {imageheader25} $
                ELSE r = 1;dialog_message('Frame version not supported',/error)
        END
    endcase

    ; read the image header
    readu, FileUnit, ImageHeader

    ; Allocate a buffer for the Background image
    dx = ImageHeader.EX + 1
    dy = ImageHeader.EY + 1
    case abs(ImageHeader.PixBytes) of
        1: BackImg = BYTARR(dx, dy)
        2: BackImg = INTARR(dx, dy)
        4: BackImg = LONARR(dx, dy)
        else:r=1;dialog_message('Unsupported Image Data Type')
    endcase

    ; Read the Background image
    readu, FileUnit, BackImg
    ; stop
    ; we return the image frame header directly and a pointer to the background image
    return,{ImageHeader:ImageHeader,pBackImg:ptr_new(BackImg,/no_copy)}
end
