; File:  a3vcpiview_3_view_widget.pro
; Author:  Ted Fisher <ted@specinc.com> <fishnuke@gmail.com>
; Purpose:  Part of the 3V-CPI View data processing suite. a3vcpiview_3_view_widget implements a GUI
;   to display CPI images and the associated triggering 2D-S images, as well as output particle parameters.
;   It is invoked via a button in the 3V-CPI CPI GUI.





; 2016-03-28 fixed CPI Hex Time display (had word_1b repeated), removed some prints, changed CPI min/max limits and displaying more digits
; 2016-03-15 TAF merged _temp version





pro draw_2D_particle_image, HorV, index, pStateInfo
    ; Draw on 2D particle with the desired index into alldata, using image data from file
    ; Draw the particle to either the H or V display in the 3 Viewer widget

    Common  DataBlock, alldataH, alldataV, ADataH, ADataV, FirstData, LastData, StripData, $
        FirstDataV, LastDataV, StripDataV, OneHzH, OneHzV, aveStripdata, aveStripdataV

    ;Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy  ; get our info structure

    subindex = index
    Hsubindex = subindex
    Vsubindex = subindex

    if(HorV eq 'H') then begin
        aveStripData      = {_DerivedDataRev2}
        ;subindex          = *(ave_info.Hsubindex)
        nnnn              = n_elements(subindex) ; there are total of nnnn particles to be displayed
        sub_foffset       = (*alldataH)(subindex).imageoffset
        L1size_array      = (*alldataH)(subindex).Nslice
        imgword_array     = (*alldataH)(subindex).imageWords
        nanoseconds_array = (*alldataH)(subindex).nanoseconds
        CPI_trigger_array = (*alldataH)(subindex).CPI_trigger
        PartID_array      = (*alldataH)(subindex).PartID

        timestamp = SPEC_Nano2HMS((*alldataH)(subindex).Nanoseconds)
        hex_timestamp = string(format = '(%"0x%012X")', (*alldataH)(subindex).ProbeCounter)

        ;print, "draw_2D_particle_image H", index
    endif

    if(HorV eq 'V') then begin
        aveStripDataV     = {_DerivedDataRev2}
        ;subindex          = *(ave_info.Vsubindex)
        nnnn              = n_elements(subindex) ; there are total of nnnn particles to be displayed
        sub_foffset       = (*alldataV)(subindex).imageoffset
        L1size_array      = (*alldataV)(subindex).Nslice
        imgword_array     = (*alldataV)(subindex).imageWords
        nanoseconds_array = (*alldataV)(subindex).nanoseconds
        CPI_trigger_array = (*alldataV)(subindex).CPI_trigger
        PartID_array      = (*alldataV)(subindex).PartID

        timestamp = SPEC_Nano2HMS((*alldataV)(subindex).Nanoseconds)
        hex_timestamp = string(format = '(%"0x%012X")', (*alldataV)(subindex).ProbeCounter)

        ;print, "draw_2D_particle_image V", index
    endif

    if(L1size_array gt 0) then begin

        i=0
        yoffset = 0

        foffsets_array    =       sub_foffset(i)
        L1size_array      =      L1size_array(i)
        imgword_array     =     imgword_array(i)
        nanoseconds_array = nanoseconds_array(i)
        CPI_trigger_array = CPI_trigger_array(i)
        PartID_array      =      PartID_array(i)
        use_index         =          subindex(i)

        ; create a filename that is one directory above the Data file's horizyymmddhhmmss.bin that was selected above
        ; Here it doesn't matter what the current file is because we will always have a horiz filename at this point to work with
        ;fname = strmid(!currentHPname,0,strlen(!currentHPname)-5) + 'SORTED' +   strmid(!currentHFname,strpos(!currentHFname,"horiz")+5,12) + '.2DS'
        ; substrsF=strsplit(!currentHFname,'\',/extract)
        ; short_filename=substrsF(n_elements(substrsF)-1)
        ; datestamp = strmid(Short_filename,5,strlen(Short_filename)-9)
        ; fname = strmid(!currentHPname,0,strlen(!currentHPname)-5) + 'SORTED' + datestamp+ '.2DS'

        file_chars_to_remove = 4

        my_str1  = strmid(!currentHPname, 0,strlen(!currentHPname)-1);   that is the C:\****\data1  for example
        my_pos   = strpos(my_str1,'\',/reverse_search); get upper directory of data1.
        fname    = strmid(my_str1,0,my_pos+1)+'SORTED'
        horzname = strmid(!currentHFname,strpos(!currentHFname,"horiz")+5,strlen(!currentHFname))
        ;horzname= strmid(!currentHFname,strpos(!currentHFname,"horiz")+8,strlen(!currentHFname))
        datetime = strmid(horzname,0,strlen(horzname)-file_chars_to_remove)
        ;fname = fname+datetime+'.2DS';   get the sorted****.2ds full name.
        fname = fname + datetime + '.2DSCPI' ; get the sorted****.2ds full name.
        OpenR, lun, fname, /Get_Lun

        ; draw the graph in the Strip Widget
        if(HorV eq 'H') then begin
            wid = (*(*pStateInfo).p3ViewState).H_WID
        endif

        if(HorV eq 'V') then begin
            wid = (*(*pStateInfo).p3ViewState).V_WID
        endif

        WSet, wid

        ; temporary workaround for CPI and 2DS color compatibility
        device, decomposed = 0
        loadct, !ct_2DS, /silent

        erase, color='FFFFFF'x  ; set background to white

        WordCount = ImgWord_array(i)
        Slices    =  L1size_array(i)

        if(WordCount eq 0) then begin
            print, "WordCount = 0, cannot draw image -- Have you loaded 2D-S data?"

            return
        endif

        Image = BytArr(WordCount * 2)

        POINT_LUN, lun, fOffsets_array(i)
        ReadU, lun, Image

        out_byte_array = BytArr(128L, Slices)

        HSums=intarr(slices)
        HMax  = 0
        Area = 0UL
        Midpoint = 0
        Edges = 0
        L3 = 0.0
        L4 = 0
        L5 = 0
        PC3x = 0.0
        PC3y = 0.0
        PC4x = 0.0
        PC4y = 0.0
        A3 = 0.0
        Roundness1 = 0.0
        Roundness2 = 0.0
        Roundness3 = 0.0
        L1r = 0.0
        maxLength = 0.0
        maxWidth  = 0.0
        perimeter = 0UL
        perimeterum = 0.
        csp = 0.
        Err=0

        ml_p1x = 0
        ml_p1y = 0
        ml_p2x = 0
        ml_p2y = 0
        mw_p1x = 0
        mw_p1y = 0
        mw_p2x = 0
        mw_p2y = 0

        perimeter_points_x = intarr(128L*Slices)
        perimeter_points_y = intarr(128L*Slices)

        ProcessParticle, Image, out_byte_array, Wordcount,Slices, HSums, HMax, Area, Midpoint, Edges,Err , $
            L3,L4,L5,PC3x,PC3y,PC4x,PC4y,A3, Roundness1, Roundness2, Roundness3, L1r, maxLength, maxWidth, perimeter, crystal_type, perimeterum, csp, $
            ml_p1x=ml_p1x, ml_p1y=ml_p1y, ml_p2x=ml_p2x, ml_p2y=ml_p2y, mw_p1x=mw_p1x, mw_p1y=mw_p1y, mw_p2x=mw_p2x, mw_p2y=mw_p2y, $
            perimeter_points_x=perimeter_points_x, perimeter_points_y=perimeter_points_y, HorV=HorV,largest_satellite_area=largest_satellite_area

        ; Mass from CSP using parameterization for normal shaded pixel area

        csp_mass = (10.0^(-9.0 * !csp_beta)) * !csp_alpha * csp^!csp_beta ; here csp in um, area in mm^2 ; sizes is im um. mass in mg

        ; recolor CPI_trigger particles for display

        CPI_trigger = CPI_trigger_array[i]
        PartID = PartID_array[i]
        Touch_edges= (Edges and 3)

        ; recolor the CPI trigger particles

        ;CPI_trigger_color_index = 140

        aa = where(out_byte_array ne 255, aacount, complement=bb)

        if(aa(0) ne -1) then begin ; TAF what is -1 a flag for again?
            ;if(CPI_trigger) then begin
            ;out_byte_array(aa) = 55 ; deep blue for CPI trigger
            ;    out_byte_array(aa) = CPI_trigger_color_index ; bright green for CPI trigger
            ;stop
            ;endif else begin
            out_byte_array(aa) = 0
            ;endelse
        endif

        ;make a color display of the images


        if(HorV eq 'H') then begin
            if((*alldataH)(use_index(i)).reject ge 1) then begin
                yy = rotate(out_byte_array, !IMAGE_ROTATION);        ; change 1 to 4 will switch leading andtail, y-don't change.
                nm = size(yy, /dimension)
                colorimage = bytarr(3,nm(0),nm(1))
                colorimage(0,*,*) = 255
                colorimage(1,*,*) = yy ; colorimage(2,*,*) = 0
                tv, colorimage, yoffset, 0, true=1
            endif else begin
                white_index = where(out_byte_array eq 255, white_count)

                if(slices gt 1) then begin
                    nm = size(out_byte_array, /dimension)
                endif else begin
                    nm = [!PIXELS, 1]
                endelse

                colorimage = bytarr(3, nm[0], nm[1])
                colorimage[*] = 255

                rplane = out_byte_array
                gplane = out_byte_array
                bplane = out_byte_array

                if(!show_CPI_trigger_3viewer and CPI_trigger and aacount gt 0) then begin
                    rplane[aa] = 0
                    gplane[aa] = 128 ; darker green
                    bplane[aa] = 0
                endif

                if((!show_perimeter_3viewer) and (white_count gt 0) and (Area gt 1) and (slices gt 1)) then begin
                    if(perimeter gt 0) then begin
                        ; for 3V-CPI inverted images, temporarily invert the perimeter

                        rplane[perimeter_points_x[0:perimeter-1], perimeter_points_y[0:perimeter-1]] = 205 ; dim magenta
                        gplane[perimeter_points_x[0:perimeter-1], perimeter_points_y[0:perimeter-1]] =   0
                        bplane[perimeter_points_x[0:perimeter-1], perimeter_points_y[0:perimeter-1]] = 205
                    endif

                    ; draw a line along the length
                    length_line = raster_line(ml_p1x, ml_p1y, ml_p2x, ml_p2y)

                    rplane[length_line[0,*], length_line[1,*]] =  40 ; blueish
                    gplane[length_line[0,*], length_line[1,*]] = 126
                    bplane[length_line[0,*], length_line[1,*]] = 255

                    ; draw a line along the width
                    width_line = raster_line(mw_p1x, mw_p1y, mw_p2x, mw_p2y)

                    rplane[width_line[0,*], width_line[1,*]] =  87 ; greenyellow
                    gplane[width_line[0,*], width_line[1,*]] = 228
                    bplane[width_line[0,*], width_line[1,*]] =  27
                endif

                colorimage(0,*,*) = rplane
                colorimage(1,*,*) = gplane
                colorimage(2,*,*) = bplane

                ; TAF for 1:1 zoom
                ;colorimage2 = bytarr(3, nm[1], nm[0])

                ;colorimage2[0, *, *] = rotate(reform(colorimage[0, *, *]), !IMAGE_ROTATION)
                ;colorimage2[1, *, *] = rotate(reform(colorimage[1, *, *]), !IMAGE_ROTATION)
                ;colorimage2[2, *, *] = rotate(reform(colorimage[2, *, *]), !IMAGE_ROTATION)

                ;tv, colorimage2, yoffset, 0, true = 1
                ; TAF for 1:1 zoom

                ; TAF for 1:!zoom_factor_3viewer zoom
                colorimage2 = bytarr(3, nm[1]*!zoom_factor_3viewer, nm[0]*!zoom_factor_3viewer)

                colorimage2[0, *, *] = congrid(rotate(reform(colorimage[0, *, *]), !IMAGE_ROTATION), nm[1]*!zoom_factor_3viewer, nm[0]*!zoom_factor_3viewer)
                colorimage2[1, *, *] = congrid(rotate(reform(colorimage[1, *, *]), !IMAGE_ROTATION), nm[1]*!zoom_factor_3viewer, nm[0]*!zoom_factor_3viewer)
                colorimage2[2, *, *] = congrid(rotate(reform(colorimage[2, *, *]), !IMAGE_ROTATION), nm[1]*!zoom_factor_3viewer, nm[0]*!zoom_factor_3viewer)

                tv, colorimage2, yoffset, 0, true = 1
                ; TAF for 1:!zoom_factor_3viewer zoom

                ; display the 2D-S timestamps
                widget_control, (*(*pStateInfo).p3ViewState).txt_H_hex,  set_value = hex_timestamp
                widget_control, (*(*pStateInfo).p3ViewState).txt_H_time, set_value = timestamp

                ; display the 2D-S particle ID
                (*(*pStateInfo).p3ViewState).val_H_HID = PartID
                HID = string(format='(%"0x%04x")', PartID)
                widget_control, (*(*pStateInfo).p3ViewState).txt_H_HID,  set_value = HID

                if(HorV eq 'H') then begin
                    microns_per_pixel = !microns_per_pixelH
                endif else begin
                    microns_per_pixel = !microns_per_pixelV
                endelse




                alexei1 = 0
                find_Image_Parameters, out_byte_array, alexei1  ; this is to find alexi1 max area.
                Get_alexei_correct_size,Alexei1,area,L5*microns_per_pixel,L7;  L7 in unit um.
                L7=L7/microns_per_pixel
                if L5 gt 40  or area gt 0.9*alexei1 then L7=L5;     no correction if size is great  or if it is in focus.

                ; for m2 m4 m7 method, touching edes partickles should not be displayed.  M7 ? maxlength greater than 127 also ?  No.
                ;Throw_out= this_sizes eq 127 or touch_edges or This_sizes eq 0



                L1=slices
                temp_area =Area
                This_sizes =maxLength
                alpha = 0.115D
                beta = 1.218D
                mass1=alpha*double(temp_area*microns_per_pixel*microns_per_pixel/1000000.0)^beta; ;here sizes in um, area in mm^2 ; sizes is im um.
                mass2=0.917*(0.1^9.0*!pi/6.0)*(This_sizes*microns_per_pixel)^3.0; size in um water-mass. milligrams.
                bradpaul_massM7 = mass1 < mass2;    % take the smaller one as the value
                get_Icsp_from_adj_n_csp,M7cspmass,1,csp,this_sizes*microns_per_pixel
                Throw_outM7= this_sizes eq 127 or touch_edges or This_sizes eq 0

                ;calculate L6  which is L1 or L5  in focus.
                if area gt 0.9*alexei1 then begin
                    if L1 gt 26.5 then L6=L1 else L6=L5;  ; Hmax is L2.
                endif else L6=0.0

                ;M1areaMass
                temp_area =Area < (!pi*L1*L1/4.0)
                This_sizes=L1
                mass1=alpha*double(temp_area*microns_per_pixel*microns_per_pixel/1000000.0)^beta; ;here sizes in um, area in mm^2 ; sizes is im um.
                mass2=0.917*(0.1^9.0*!pi/6.0)*(This_sizes*microns_per_pixel)^3.0; size in um water-mass. milligrams.
                bradpaul_massM1 = mass1 < mass2
                get_Icsp_from_adj_n_csp,M1cspmass,1,csp,this_sizes*microns_per_pixel



                ;M4areaMass
                temp_area =Area < (!pi*L7*L7/4.0)
                This_sizes=L7
                mass1=alpha*double(temp_area*microns_per_pixel*microns_per_pixel/1000000.0)^beta; ;here sizes in um, area in mm^2 ; sizes is im um.
                mass2=0.917*(0.1^9.0*!pi/6.0)*(This_sizes*microns_per_pixel)^3.0; size in um water-mass. milligrams.
                bradpaul_massM4 = mass1 < mass2
                get_Icsp_from_adj_n_csp,M4cspmass,1,csp,this_sizes*microns_per_pixel
                Throw_outM4= this_sizes eq 127 or touch_edges or This_sizes eq 0

                ;M4areaMassu
                temp_area =Area < (!pi*L5*L5/4.0)
                This_sizes=L5
                mass1=alpha*double(temp_area*microns_per_pixel*microns_per_pixel/1000000.0)^beta; ;here sizes in um, area in mm^2 ; sizes is im um.
                mass2=0.917*(0.1^9.0*!pi/6.0)*(This_sizes*microns_per_pixel)^3.0; size in um water-mass. milligrams.
                bradpaul_massM4u = mass1 < mass2
                get_Icsp_from_adj_n_csp,M4cspmassu,1,csp,this_sizes*microns_per_pixel
                Throw_outM4u= this_sizes eq 127 or touch_edges or This_sizes eq 0


                ;M6areaMassu
                if L6 gt 0.0 then begin
                    temp_area =Area < (!pi*L6*L6/4.0)
                    This_sizes=L6
                    mass1=alpha*double(temp_area*microns_per_pixel*microns_per_pixel/1000000.0)^beta; ;here sizes in um, area in mm^2 ; sizes is im um.
                    mass2=0.917*(0.1^9.0*!pi/6.0)*(This_sizes*microns_per_pixel)^3.0; size in um water-mass. milligrams.
                    bradpaul_massM6 = mass1 < mass2
                    get_Icsp_from_adj_n_csp,M6cspmass,1,csp,this_sizes*microns_per_pixel
                    Throw_outM6= this_sizes eq 127 or touch_edges or This_sizes eq 0
                endif else begin
                    bradpaul_massM6=0.0
                    M6cspmass=0.0
                    Throw_outM6=1
                endelse

                ; display the particle parameters (length, width, area, etc)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HL1,   set_value = strtrim(string(L1        * microns_per_pixel, format='(%"%7.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HL4,   set_value = strtrim(string(L7        * microns_per_pixel, format='(%"%7.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HL4u,  set_value = strtrim(string(L5        * microns_per_pixel, format='(%"%7.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HL6,   set_value = strtrim(string(L6        * microns_per_pixel, format='(%"%7.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HmaxL, set_value = strtrim(string(maxLength * microns_per_pixel, format='(%"%7.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HmaxW, set_value = strtrim(string(maxwidth  * microns_per_pixel, format='(%"%7.0f")'), 2)

                alexei1_ratio   = area/Alexei1
                M1area  =(area < (!pi*L1*L1/4.0))*microns_per_pixel*microns_per_pixel
                M4area  =(area < (!pi*L7*L7/4.0))*microns_per_pixel*microns_per_pixel
                M4areaun=(area < (!pi*L5*L5/4.0))*microns_per_pixel*microns_per_pixel
                M6area  =(area < (!pi*L6*L6/4.0))*microns_per_pixel*microns_per_pixel
                M7area  =(area < (!pi*maxlength*maxlength/4.0))*microns_per_pixel*microns_per_pixel

                widget_control, (*(*pStateInfo).p3ViewState).txt_HM1area,     set_value = strtrim(string(M1area,        format='(%"%9.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HM4area,     set_value = strtrim(string(M4area,        format='(%"%9.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HM4areaun,   set_value = strtrim(string(M4areaun,      format='(%"%9.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HM6area,     set_value = strtrim(string(M6area,        format='(%"%9.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HM7area,     set_value = strtrim(string(M7area,        format='(%"%9.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HfocusRatio, set_value = strtrim(string(alexei1_ratio, format='(%"%5.2f")'), 2)



                widget_control, (*(*pStateInfo).p3ViewState).txt_HM1cspMass,    set_value = strtrim(string(csp_mass,        format='(%"%11.3e")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HM1areaMass,   set_value = strtrim(string(bradpaul_massM1, format='(%"%11.3e")'), 2)

                widget_control, (*(*pStateInfo).p3ViewState).txt_HM4cspMass,    set_value = strtrim(string(M4cspmass,       format='(%"%11.3e")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HM4areaMass,   set_value = strtrim(string(bradpaul_massM4u,format='(%"%11.3e")'), 2)

                widget_control, (*(*pStateInfo).p3ViewState).txt_HM4cspMassu,   set_value = strtrim(string(M4cspmassu,      format='(%"%11.3e")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HM4areaMassu,  set_value = strtrim(string(bradpaul_massM4u,format='(%"%11.3e")'), 2)

                widget_control, (*(*pStateInfo).p3ViewState).txt_HM6cspMass,    set_value = strtrim(string(M6cspmass,       format='(%"%11.3e")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HM6areaMass,   set_value = strtrim(string(bradpaul_massM6, format='(%"%11.3e")'), 2)

                widget_control, (*(*pStateInfo).p3ViewState).txt_HM7cspMass,    set_value = strtrim(string(M7cspmass,       format='(%"%11.3e")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_HM7areaMass,   set_value = strtrim(string(bradpaul_massM7, format='(%"%11.3e")'), 2)


                if L6  eq 0.0 then begin
                    widget_control, (*(*pStateInfo).p3ViewState).txt_HM6areaMass,   set_value = ' '
                    widget_control, (*(*pStateInfo).p3ViewState).txt_HM6area,       set_value = ' '
                    widget_control, (*(*pStateInfo).p3ViewState).txt_HL6,           set_value = ' '
                    widget_control, (*(*pStateInfo).p3ViewState).txt_HM6cspMass,    set_value = ' '
                endif



            endelse
        endif else if(HorV eq 'V') then begin
            if((*alldataV)(use_index(i)).reject ge 1) then begin
                yy = rotate(out_byte_array, !IMAGE_ROTATION)
                nm = size(yy, /dimension)
                colorimage = bytarr(3, nm(0), nm(1))
                colorimage(0,*,*) = 255
                colorimage(1,*,*) = yy ; colorimage(2,*,*) = 0
                tv, colorimage, yoffset, 0, true = 1
            endif else begin
                white_index = where(out_byte_array eq 255, white_count)

                if(slices gt 1) then begin
                    nm = size(out_byte_array, /dimension)
                endif else begin
                    nm = [!PIXELS, 1]
                endelse

                colorimage = bytarr(3, nm[0], nm[1])
                colorimage[*] = 255

                rplane = out_byte_array
                gplane = out_byte_array
                bplane = out_byte_array

                if(!show_CPI_trigger_3viewer and CPI_trigger and aacount gt 0) then begin
                    rplane[aa] = 0
                    gplane[aa] = 128 ; darker green
                    bplane[aa] = 0
                endif

                if((!show_perimeter_3viewer) and (white_count gt 0) and (Area gt 1) and (slices gt 1)) then begin
                    if(perimeter gt 0) then begin
                        ; for 3V-CPI inverted images, temporarily invert the perimeter

                        rplane[perimeter_points_x[0:perimeter-1], perimeter_points_y[0:perimeter-1]] = 205 ; dim magenta
                        gplane[perimeter_points_x[0:perimeter-1], perimeter_points_y[0:perimeter-1]] =   0
                        bplane[perimeter_points_x[0:perimeter-1], perimeter_points_y[0:perimeter-1]] = 205
                    endif

                    ; draw a line along the length
                    length_line = raster_line(ml_p1x, ml_p1y, ml_p2x, ml_p2y)

                    rplane[length_line[0,*], length_line[1,*]] =  40 ; blueish
                    gplane[length_line[0,*], length_line[1,*]] = 126
                    bplane[length_line[0,*], length_line[1,*]] = 255

                    ; draw a line along the width
                    width_line = raster_line(mw_p1x, mw_p1y, mw_p2x, mw_p2y)

                    rplane[width_line[0,*], width_line[1,*]] =  87 ; greenyellow
                    gplane[width_line[0,*], width_line[1,*]] = 228
                    bplane[width_line[0,*], width_line[1,*]] =  27
                endif

                colorimage(0,*,*) = rplane
                colorimage(1,*,*) = gplane
                colorimage(2,*,*) = bplane

                ; TAF for 1:1 zoom
                ;colorimage2 = bytarr(3, nm[1], nm[0])

                ;colorimage2[0, *, *] = rotate(reform(colorimage[0, *, *]), !IMAGE_ROTATION)
                ;colorimage2[1, *, *] = rotate(reform(colorimage[1, *, *]), !IMAGE_ROTATION)
                ;colorimage2[2, *, *] = rotate(reform(colorimage[2, *, *]), !IMAGE_ROTATION)

                ;tv, colorimage2, yoffset, 0, true = 1
                ; TAF for 1:1 zoom

                ; TAF for 1:!zoom_factor_3viewer zoom
                colorimage2 = bytarr(3, nm[1]*!zoom_factor_3viewer, nm[0]*!zoom_factor_3viewer)

                colorimage2[0, *, *] = congrid(rotate(reform(colorimage[0, *, *]), !IMAGE_ROTATION), nm[1]*!zoom_factor_3viewer, nm[0]*!zoom_factor_3viewer)
                colorimage2[1, *, *] = congrid(rotate(reform(colorimage[1, *, *]), !IMAGE_ROTATION), nm[1]*!zoom_factor_3viewer, nm[0]*!zoom_factor_3viewer)
                colorimage2[2, *, *] = congrid(rotate(reform(colorimage[2, *, *]), !IMAGE_ROTATION), nm[1]*!zoom_factor_3viewer, nm[0]*!zoom_factor_3viewer)

                tv, colorimage2, yoffset, 0, true = 1
                ; TAF for 1:!zoom_factor_3viewer zoom

                widget_control, (*(*pStateInfo).p3ViewState).txt_V_hex,  set_value = hex_timestamp
                widget_control, (*(*pStateInfo).p3ViewState).txt_V_time, set_value = timestamp

                (*(*pStateInfo).p3ViewState).val_V_VID = PartID
                VID = string(format='(%"0x%04x")', PartID)
                widget_control, (*(*pStateInfo).p3ViewState).txt_V_VID,  set_value = VID

                if(HorV eq 'H') then begin
                    microns_per_pixel = !microns_per_pixelH
                endif else begin
                    microns_per_pixel = !microns_per_pixelV
                endelse

                ; display the particle parameters (length, width, area, etc)

                alexei1 = 0
                find_Image_Parameters, out_byte_array, alexei1  ; this is to find alexi1 max area.
                Get_alexei_correct_size,Alexei1,area,L5*microns_per_pixel,L7;  L7 in unit um.
                L7=L7/microns_per_pixel

                if L5 gt 40  or area gt 0.9*alexei1 then L7=L5;     no correction if size is great  or if it is in focus.


                L1=slices
                temp_area =Area
                This_sizes =maxLength
                alpha = 0.115D
                beta = 1.218D
                mass1=alpha*double(temp_area*microns_per_pixel*microns_per_pixel/1000000.0)^beta; ;here sizes in um, area in mm^2 ; sizes is im um.
                mass2=0.917*(0.1^9.0*!pi/6.0)*(This_sizes*microns_per_pixel)^3.0; size in um water-mass. milligrams.
                bradpaul_massM7 = mass1 < mass2;    % take the smaller one as the value
                get_Icsp_from_adj_n_csp,M7cspmass,1,csp,this_sizes*microns_per_pixel


                ;calculate L6  which is L1 or L2  in focus.
                if area gt 0.9*alexei1 then begin
                    if L1 gt 26.5 then L6=L1 else L6=L5;  ; Hmax is L2.
                endif else L6=0.0

                ;M1areaMass
                temp_area =Area < (!pi*L1*L1/4.0)
                This_sizes=L1
                mass1=alpha*double(temp_area*microns_per_pixel*microns_per_pixel/1000000.0)^beta; ;here sizes in um, area in mm^2 ; sizes is im um.
                mass2=0.917*(0.1^9.0*!pi/6.0)*(This_sizes*microns_per_pixel)^3.0; size in um water-mass. milligrams.
                bradpaul_massM1 = mass1 < mass2
                get_Icsp_from_adj_n_csp,M1cspmass,1,csp,this_sizes*microns_per_pixel

                ;M4areaMass
                temp_area =Area < (!pi*L7*L7/4.0)
                This_sizes=L7
                mass1=alpha*double(temp_area*microns_per_pixel*microns_per_pixel/1000000.0)^beta; ;here sizes in um, area in mm^2 ; sizes is im um.
                mass2=0.917*(0.1^9.0*!pi/6.0)*(This_sizes*microns_per_pixel)^3.0; size in um water-mass. milligrams.
                bradpaul_massM4 = mass1 < mass2
                get_Icsp_from_adj_n_csp,M4cspmass,1,csp,this_sizes*microns_per_pixel

                ;M4areaMassu
                temp_area =Area < (!pi*L5*L5/4.0)
                This_sizes=L5
                mass1=alpha*double(temp_area*microns_per_pixel*microns_per_pixel/1000000.0)^beta; ;here sizes in um, area in mm^2 ; sizes is im um.
                mass2=0.917*(0.1^9.0*!pi/6.0)*(This_sizes*microns_per_pixel)^3.0; size in um water-mass. milligrams.
                bradpaul_massM4u = mass1 < mass2
                get_Icsp_from_adj_n_csp,M4cspmassu,1,csp,this_sizes*microns_per_pixel


                ;M6areaMassu
                if L6 gt 0.0 then begin
                    temp_area =Area < (!pi*L6*L6/4.0)
                    This_sizes=L6
                    mass1=alpha*double(temp_area*microns_per_pixel*microns_per_pixel/1000000.0)^beta; ;here sizes in um, area in mm^2 ; sizes is im um.
                    mass2=0.917*(0.1^9.0*!pi/6.0)*(This_sizes*microns_per_pixel)^3.0; size in um water-mass. milligrams.
                    bradpaul_massM6 = mass1 < mass2
                endif else bradpaul_massM6=0.0
                get_Icsp_from_adj_n_csp,M6cspmass,1,csp,this_sizes*microns_per_pixel


                ; display the particle parameters (length, width, area, etc)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VL1,   set_value = strtrim(string(L1        * microns_per_pixel, format='(%"%7.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VL4,   set_value = strtrim(string(L7        * microns_per_pixel, format='(%"%7.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VL4u,  set_value = strtrim(string(L5        * microns_per_pixel, format='(%"%7.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VL6,   set_value = strtrim(string(L6        * microns_per_pixel, format='(%"%7.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VmaxL, set_value = strtrim(string(maxLength * microns_per_pixel, format='(%"%7.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VmaxW, set_value = strtrim(string(maxwidth  * microns_per_pixel, format='(%"%7.0f")'), 2)

                alexei1_ratio   = area/Alexei1
                M1area  =(area < (!pi*L1*L1/4.0))*microns_per_pixel*microns_per_pixel
                M4area  =(area < (!pi*L7*L7/4.0))*microns_per_pixel*microns_per_pixel
                M4areaun=(area < (!pi*L5*L5/4.0))*microns_per_pixel*microns_per_pixel
                M6area  =(area < (!pi*L6*L6/4.0))*microns_per_pixel*microns_per_pixel
                M7area  =(area < (!pi*maxlength*maxlength/4.0))*microns_per_pixel*microns_per_pixel

                widget_control, (*(*pStateInfo).p3ViewState).txt_VM1area,     set_value = strtrim(string(M1area,        format='(%"%9.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VM4area,     set_value = strtrim(string(M4area,        format='(%"%9.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VM4areaun,   set_value = strtrim(string(M4areaun,      format='(%"%9.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VM6area,     set_value = strtrim(string(M6area,        format='(%"%9.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VM7area,     set_value = strtrim(string(M7area,        format='(%"%9.0f")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VfocusRatio, set_value = strtrim(string(alexei1_ratio, format='(%"%5.2f")'), 2)



                widget_control, (*(*pStateInfo).p3ViewState).txt_VM1cspMass,    set_value = strtrim(string(csp_mass,        format='(%"%11.3e")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VM1areaMass,   set_value = strtrim(string(bradpaul_massM1, format='(%"%11.3e")'), 2)

                widget_control, (*(*pStateInfo).p3ViewState).txt_VM4cspMass,    set_value = strtrim(string(M4cspmass,       format='(%"%11.3e")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VM4areaMass,   set_value = strtrim(string(bradpaul_massM4u,format='(%"%11.3e")'), 2)

                widget_control, (*(*pStateInfo).p3ViewState).txt_VM4cspMassu,   set_value = strtrim(string(M4cspmassu,      format='(%"%11.3e")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VM4areaMassu,  set_value = strtrim(string(bradpaul_massM4u,format='(%"%11.3e")'), 2)

                widget_control, (*(*pStateInfo).p3ViewState).txt_VM6cspMass,    set_value = strtrim(string(M6cspmass,       format='(%"%11.3e")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VM6areaMass,   set_value = strtrim(string(bradpaul_massM6, format='(%"%11.3e")'), 2)

                widget_control, (*(*pStateInfo).p3ViewState).txt_VM7cspMass,    set_value = strtrim(string(M7cspmass,       format='(%"%11.3e")'), 2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_VM7areaMass,   set_value = strtrim(string(bradpaul_massM7, format='(%"%11.3e")'), 2)


                if L6  eq 0.0 then begin
                    widget_control, (*(*pStateInfo).p3ViewState).txt_VM6areaMass,   set_value = ' '
                    widget_control, (*(*pStateInfo).p3ViewState).txt_VM6area,       set_value = ' '
                    widget_control, (*(*pStateInfo).p3ViewState).txt_VL6,           set_value = ' '
                    widget_control, (*(*pStateInfo).p3ViewState).txt_VM6cspMass,    set_value = ' '
                endif
            endelse
        endif ; HorV test

        Close,   lun
        Free_Lun,lun

    endif ; L1size_array gt 0 test

    ;Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
end ; draw_2D_particle_image



pro draw_2D_blank, HorV, pStateInfo
    if(HorV eq 'H') then begin
        wid = (*(*pStateInfo).p3ViewState).H_WID

        widget_control, (*(*pStateInfo).p3ViewState).txt_H_hex,    set_value = '0xXXXXXXXXXXXX'
        widget_control, (*(*pStateInfo).p3ViewState).txt_H_time,   set_value = 'HH:MM:SS.000.000.000'
        widget_control, (*(*pStateInfo).p3ViewState).txt_H_HID,    set_value = '0xXXXX'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HL1,  set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HL4,  set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HL4u,  set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HL6,  set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HmaxL, set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HmaxW,  set_value = 'X'
        ;widget_control, (*(*pStateInfo).p3ViewState).txt_H_perim,  set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HM1area,    set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HM4area,   set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HM4areaun, set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HM6area, set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HM7area, set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HfocusRatio, set_value='X'

        widget_control, (*(*pStateInfo).p3ViewState).txt_HM1cspmass,  set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HM1areamass, set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HM4areamass, set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HM4areamassu,set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HM6areamass, set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_HM7areamass, set_value='X'




    endif else if(HorV eq 'V') then begin
        wid = (*(*pStateInfo).p3ViewState).V_WID

        widget_control, (*(*pStateInfo).p3ViewState).txt_V_hex,    set_value = '0xXXXXXXXXXXXX'
        widget_control, (*(*pStateInfo).p3ViewState).txt_V_time,   set_value = 'HH:MM:SS.000.000.000'
        widget_control, (*(*pStateInfo).p3ViewState).txt_V_VID,    set_value = '0xXXXX'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VL1,  set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VL4,  set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VL4u,  set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VL6,  set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VmaxL, set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VmaxW,  set_value = 'X'
        ;widget_control, (*(*pStateInfo).p3ViewState).txt_V_perim,  set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VM1area,    set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VM4area,   set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VM4areaun, set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VM6area, set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VM7area, set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VfocusRatio, set_value='X'

        widget_control, (*(*pStateInfo).p3ViewState).txt_VM1cspmass,  set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VM1areamass, set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VM4areamass, set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VM4areamassu,set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VM6areamass, set_value = 'X'
        widget_control, (*(*pStateInfo).p3ViewState).txt_VM7areamass, set_value='X'

    endif

    wset, wid

    device, decomposed = 0
    loadct, !ct_2DS, /silent

    erase, color='FFFFFF'x  ; set background to white
end ; draw_2D_blank



pro cpi_slider_event, event
    Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy  ; get our info structure

    ;print, "cpi_slider_event"

    ;view_info.cpi_index = event.value

    ;print, view_info.cpi_index

    ;pStateInfo = !POINTER_STATE_INFO

    if(ptr_valid((*pStateInfo).pFrameData))then begin  ; this needs to be a defined structure to continue
        widget_control, (*(*pStateInfo).p3ViewState).txt_min_cpi_size,  get_value = min_cpi_size_string
        widget_control, (*(*pStateInfo).p3ViewState).txt_max_cpi_size,  get_value = max_cpi_size_string

        ; some minor error checking and ranging on the input
        const_cpi_low_range =        .0      ; microns
        const_cpi_hi_range  = 5000000.       ; microns
        const_cpi_min_diff  =        .00001  ; microns -- enforce at least this much between hi and low range

        min_cpi_size = double(min_cpi_size_string)
        min_cpi_size = (min_cpi_size ge const_cpi_low_range) ? min_cpi_size : const_cpi_low_range
        min_cpi_size = (min_cpi_size le const_cpi_hi_range)  ? min_cpi_size : const_cpi_hi_range

        max_cpi_size = double(max_cpi_size_string)
        max_cpi_size = (max_cpi_size ge const_cpi_low_range) ? max_cpi_size : const_cpi_low_range
        max_cpi_size = (max_cpi_size le const_cpi_hi_range)  ? max_cpi_size : const_cpi_hi_range

        ; make sure min is less than max, and the min_diff is enforced

        if(min_cpi_size eq max_cpi_size) then begin ; if they are equal, set the difference to min_diff
            if(max_cpi_size eq const_cpi_hi_range) then begin ; at the high end, decrease min
                min_cpi_size = min_cpi_size - const_cpi_min_diff
            endif else begin ; at the low end or in the middle, increase max
                max_cpi_size = max_cpi_size + const_cpi_min_diff
            endelse
        endif else if(min_cpi_size gt max_cpi_size) then begin ; if the min is greater than the max, switch them
            temp_val = min_cpi_size
            min_cpi_size = max_cpi_size
            max_cpi_size = temp_val
        endif

        ; update the text fields with the potentially modified values
        widget_control, (*(*pStateInfo).p3ViewState).txt_min_cpi_size,  set_value = strtrim(string(format='(%"%9.1f")', min_cpi_size), 2)
        widget_control, (*(*pStateInfo).p3ViewState).txt_max_cpi_size,  set_value = strtrim(string(format='(%"%9.1f")', max_cpi_size), 2)

        ; get the list of frames to use ( this is set via *pImgPosIndx which may have backgrounds depending on gui setting)
        ImgList = (*(*pStateInfo).pFileScan)[*(*pStateInfo).pImgPosIndx]

        ; handle the case when this is a background
        background = !false
        LastPos = (*pStateInfo).ImgPos
        (*pStateInfo).ImgPos = event.value

        dir =  (*pStateInfo).ImgPos - LastPos

        ; advance in the direction, find the next image and test it for size, if the size is OK then display, or if the start/end the image is hit

        done  = 0
        index = 0 ; currently only looking at first image in each frame
        loopcount = 0
        loopbailcount = 1000

        while(~done) do begin ; loop through the CPI images in dir direction, searching for images between min_cpi_size and max_cpi_size, or quit after loopbailcount tries or reaching the end of the file
            if(event.value ne 0 and ImgList[event.value-1].imgtype eq 2 or loopcount gt 0)then begin
                ; If at end of file step back one else step over one
                if((*pStateInfo).SkipBkgd)then begin ; TAF why bother skipping the background here?
                    if(dir gt 0)then begin
                        (*pStateInfo).ImgPos = ((*pStateInfo).ImgPos eq (*pStateInfo).TotImgCount) ? (*pStateInfo).ImgPos-1 :(*pStateInfo).ImgPos+1
                    endif else begin
                        (*pStateInfo).ImgPos = ((*pStateInfo).ImgPos eq 1) ? (*pStateInfo).ImgPos+1 :(*pStateInfo).ImgPos-1
                    endelse
                endif else background = !true
            endif

            ; delete current frame
            delete_frame, (*pStateInfo).pFrameData

            ; get the frame indicated via position in scan list

            frameData = read_frame((*pStateInfo).InFileLun,(*pStateInfo).FileHead.Ver,(ImgList[(*pStateInfo).ImgPos-1]).fpos,date=(*pStateInfo).date)
            ; TAF why two identical calls to read_frame here???
            ;frameData = read_frame((*pStateInfo).InFileLun,(*pStateInfo).FileHead.Ver,(ImgList[(*pStateInfo).ImgPos-1]).fpos,date=(*pStateInfo).date)
            ;stop;  here frame data is the same as in process_time_lib2.pro after feamedata=read_frame. good!, March 18/05

            if(size(frameData, /type) eq 8)then begin
                ;if(size(frameData, /type) eq 8)then begin
                ; display frame time
                ; comment following 4 lines out and replace it with 3 new words----september 1, 2009.
                ;frame_time = convertFromSeconds(frameData.frameheader.Time,(*pStateInfo).fileHead.year)
                ;frame_time.msec  = ROUND(CalcStrobeTime(frameData.frameheader.ArrivalT1, frameData.frameheader.ArrivalT2,frameData.frameheader.Ver))
                ;msg = 'frame#:'+strtrim(fix((*pStateInfo).ImgPos),2) + ' Frame time:' + strtrim(fix(frame_time.day),2) + ':' + strtrim(fix(frame_time.hour),2) $
                ;     + ':' + strtrim(fix(frame_time.minute),2) + ':'+  strtrim(fix(frame_time.sec),2) + ':'+  strtrim(fix(frame_time.msec),2) + $
                ;      ' Missed Particles: ' + strtrim(FrameData.frameHeader.Missed-1,2)
                ;ShowMsg, (*pStateInfo).InstLabel,msg
                ;time1= 'CPI: ' + strtrim(fix(frame_time.hour),2) + ':' + strtrim(fix(frame_time.minute),2) + ':'+  strtrim(fix(frame_time.sec),2) + ':'+  strtrim(fix(frame_time.msec),2)

                ; display the CPI hex time

                word1 = string(format='(%"%04X")', frameData.frameheader.Time)
                word2 = string(format='(%"%04X")', frameData.frameheader.arrivalT1)
                word3 = string(format='(%"%04X")', frameData.frameheader.arrivalT2)

                word1_a = strmid(word1,0,2)
                word1_b = strmid(word1,2,2)
                word2_a = strmid(word2,0,2)
                word2_b = strmid(word2,2,2)
                word3_a = strmid(word3,0,2)
                word3_b = strmid(word3,2,2)

                time_hex = '0x' + word3_a + word3_b + word2_a + word2_b + word1_a + word1_b

                time_dec = string(format='(%"%02d:%02d:%02d.%03d")', $
                    fix(frameData.frameheader.hour),   $
                    fix(frameData.frameheader.minute), $
                    fix(frameData.frameheader.sec),    $
                    fix(frameData.frameheader.msec))

                widget_control, (*(*pStateInfo).p3ViewState).txt_CPI_hex,  set_value = time_hex
                widget_control, (*(*pStateInfo).p3ViewState).txt_CPI_time, set_value = time_dec

                ; put the CPI time string into a system variable for use in naming single png output images
                ; using . instead of : for file naming

                CPIfilename = strmid((*pStateInfo).InFileName,    strpos((*pStateInfo).InFileName, '\' , /reverse_search) + 1)
                CPIfilepath = strmid((*pStateInfo).InFileName, 0, strpos((*pStateInfo).InFileName, '\' , /reverse_search) + 1)

                time_png = string(format='(%"%sCPI_3views_%s_%02d.%02d.%02d.%03d.png")', $
                    CPIfilepath,                       $
                    CPIfilename,                       $
                    fix(frameData.frameheader.hour),   $
                    fix(frameData.frameheader.minute), $
                    fix(frameData.frameheader.sec),    $
                    fix(frameData.frameheader.msec))

                defsysv, '!CPI_3views_png_filename', exists = name_exists

                if(name_exists) then begin
                    !CPI_3views_png_filename = time_png
                endif else begin
                    defsysv, '!CPI_3views_png_filename', time_png
                endelse

                ; display the CPI H and V particle IDs
                (*(*pStateInfo).p3ViewState).val_CPI_HID = frameData.frameHeader.Pheight1
                HID = string(format='(%"0x%04X")', frameData.frameHeader.Pheight1)
                widget_control, (*(*pStateInfo).p3ViewState).txt_CPI_HID,  set_value = HID

                (*(*pStateInfo).p3ViewState).val_CPI_VID = frameData.frameHeader.Pheight2
                VID = string(format='(%"0x%04X")', frameData.frameHeader.Pheight2)
                widget_control, (*(*pStateInfo).p3ViewState).txt_CPI_VID,  set_value = VID

                ; initally, no background removed
                (*pStateInfo).backRemoved = !false

                ; save into state structure pFrameData always holds the current frame
                (*pStateInfo).pFrameData = ptr_new(frameData, /no_copy)

                if((*pStateInfo).Bkgs_exist and (*pStateInfo).BkgPos ge 0)then begin
                    ; determine if the current background is still the best
                    position = (*pStateInfo).ImgPos-1
                    ImgscanList = (*(*pStateInfo).pFileScan)[*(*pStateInfo).pImgPosIndx]
                    backScanList = where(ImgscanList.imgtype eq 2)

                    if(ptr_valid((*pStateInfo).pBackPosIndx))then get_best_backGnd_pos,ImgscanList,backScanList,position

                    if(position ne (*pStateInfo).BkgPos)then begin
                        ;ShowMsg, (*pStateInfo).InstLabel,'Current Background at Frame # '+ strtrim(position+1,2)

                        BackFrame = read_backGnd((*pStateInfo).InFileLun,(*(*pStateInfo).pFrameData).frameheader.ver,imgList[position].fpos)  ; contains the structure of the background

                        (*pStateInfo).BackFrame = ptr_new(BackFrame,/no_copy)
                        (*pStateInfo).BkgPos =  position
                    endif; else print,'Current background OK'
                endif

                ; set min/max to gui ( this removes a background if needed )
                set_min_max, pStateInfo

                ; Show main info line
                ;disp_frame_info, pStateInfo, 0

                ;get Obj info if it exists
                if(ptr_valid((*pStateInfo).pObjList) and not background)then begin
                    ;return,{FrameHeader:FrameHeader,pFrame:pFrame,pImHDfPos:ptr_new(imageHeaderFPOS)}
                    ; clean up existing obj if it exists
                    if(ptr_valid((*pStateInfo).pObjData))then begin
                        ptr_free,(*(*(*pStateInfo).pObjData).pFrame).imageheader.Harmonics
                        ptr_free,(*(*(*pStateInfo).pObjData).pFrame).imageheader.Boundary
                        ptr_free,(*(*pStateInfo).pObjData).pFrame
                        ptr_free,(*(*pStateInfo).pObjData).pImHDfPos
                        ptr_free,(*pStateInfo).pObjData
                    endif

                    ;BkgPosList = TEMPORARY(WHERE((*(*pStateInfo).pFileScan).ImgType eq 2, BkgCount))
                    ;where((*pStateInfo).ImgPos-1 gt
                    ;objData = get_obj_header((*pStateInfo).ObjLUN,(*(*pStateInfo).pObjList)[(*pStateInfo).ImgPos-1-BkgCount].fPos)
                    ;objTime = float(objData.frameHeader.time) + calcStrobeTime(objData.frameHeader.arrivalT1,objData.frameHeader.arrivalT2)/1000.0D

                    ; need to get calcStrobe time working for 3V-CPI
                    ;roiTime = float((*(*pStateInfo).pFrameData).frameheader.time) + calcStrobeTime((*(*pStateInfo).pFrameData).frameheader.arrivalT1,(*(*pStateInfo).pFrameData).frameheader.arrivalT2,(*(*pStateInfo).pFrameData).frameheader.Ver)/1000.0D

                    roiTime = float((*(*pStateInfo).pFrameData).frameheader.time); mo change this on august 15, 2012.
                    roitime =(*(*pStateInfo).pFrameData).frameheader.hour*3600.0+(*(*pStateInfo).pFrameData).frameheader.minute*60.0 + $
                        (*(*pStateInfo).pFrameData).frameheader.sec*1.0+(*(*pStateInfo).pFrameData).frameheader.msec/1000.0D
                    ;print,roiTime,objTime,format='(2(f15.5,x))'
                    ;if(objTime ne roiTime)then begin

                    indexOBJ = where(roiTime eq (*(*pStateInfo).pObjList).time, cntOBJ)
                    ;print,roiTime,(*(*pStateInfo).pObjList)[indexOBJ[0]].time,format='(2(f15.5,x))'
                    if(cntOBJ gt 0)then begin
                        ; read a frame of info from the obj file
                        objData = get_obj_header((*pStateInfo).ObjLUN,(*(*pStateInfo).pObjList)[indexOBJ[0]].fPos)
                        ; store obj info
                        (*pStateInfo).pObjData = ptr_new(objData,/no_copy)
                    endif else print, 'Error: obj frame not found'
                    ;endif else (*pStateInfo).pObjData = ptr_new(objData,/no_copy)
                endif else begin
                    ; clean up existing obj if it exists
                    if(ptr_valid((*pStateInfo).pObjData))then begin
                        ptr_free,(*(*(*pStateInfo).pObjData).pFrame).imageheader.Harmonics
                        ptr_free,(*(*(*pStateInfo).pObjData).pFrame).imageheader.Boundary
                        ptr_free,(*(*pStateInfo).pObjData).pFrame
                        ptr_free,(*(*pStateInfo).pObjData).pImHDfPos
                        ptr_free,(*pStateInfo).pObjData
                    endif

                    (*pStateInfo).pObjData = ptr_new()
                endelse

                if ptr_valid((*pStateInfo).pObjData) eq 1 then begin

                    objData = (*(*pStateInfo).pObjData)

                    pixsize = (*pStateInfo).Settings.PixSize ; microns/pixel
                    length  = (*objData.pFrame)[index].imageheader.len   * pixsize ; microns

                    ;print, 'ImgPos = ', (*pStateInfo).ImgPos
                    ;print, 'length = ', length
                    ;print, 'min_cpi_size = ', min_cpi_size
                    ;print, ''

                    ; test for exit conditions, CPI image length in specified range, beginning or end of CPI image buffer reached, or loopbailcount exceeded
                    if(((length ge min_cpi_size) and (length le max_cpi_size)) or $
                        ((*pStateInfo).ImgPos ge (*pStateInfo).TotImgCount-1) or $
                        ((*pStateInfo).ImgPos le 1) or $
                        (loopcount gt loopbailcount)) then begin
                        done = 1
                    endif
                endif  else begin; ptr_valid((*pStateInfo).pObjData) eq 1 then begin
                    if( ((*pStateInfo).ImgPos ge (*pStateInfo).TotImgCount-1) or $
                        ((*pStateInfo).ImgPos le 1) or $
                        (loopcount gt loopbailcount)) then begin
                        done = 1
                    endif
                endelse

                loopcount++
            endif

        endwhile

        ; set the slider to the current value (may have moved more than the user intended if there were smaller images skipped by min_cpi_size
        widget_control, (*(*pStateInfo).p3ViewState).cpi_slider, set_value = (*pStateInfo).ImgPos

        ;WIDGET_CONTROL, (*pStateInfo).ImgSliderID, SET_VALUE=(*pStateInfo).ImgPos

        ; delete current frame
        delete_frame, (*pStateInfo).pFrameData

        ; get the frame indicated via position in scan list

        frameData = read_frame((*pStateInfo).InFileLun,(*pStateInfo).FileHead.Ver,(ImgList[(*pStateInfo).ImgPos-1]).fpos,date=(*pStateInfo).date)
        ; TAF why two identical calls to read_frame here???
        ;frameData = read_frame((*pStateInfo).InFileLun,(*pStateInfo).FileHead.Ver,(ImgList[(*pStateInfo).ImgPos-1]).fpos,date=(*pStateInfo).date)
        ;stop;  here frame data is the same as in process_time_lib2.pro after feamedata=read_frame. good!, March 18/05

        if(size(frameData, /type) eq 8)then begin
            ; display frame time
            ; comment following 4 lines out and replace it with 3 new words----september 1, 2009.
            ;frame_time = convertFromSeconds(frameData.frameheader.Time,(*pStateInfo).fileHead.year)
            ;frame_time.msec  = ROUND(CalcStrobeTime(frameData.frameheader.ArrivalT1, frameData.frameheader.ArrivalT2,frameData.frameheader.Ver))
            ;msg = 'frame#:'+strtrim(fix((*pStateInfo).ImgPos),2) + ' Frame time:' + strtrim(fix(frame_time.day),2) + ':' + strtrim(fix(frame_time.hour),2) $
            ;     + ':' + strtrim(fix(frame_time.minute),2) + ':'+  strtrim(fix(frame_time.sec),2) + ':'+  strtrim(fix(frame_time.msec),2) + $
            ;      ' Missed Particles: ' + strtrim(FrameData.frameHeader.Missed-1,2)
            ;ShowMsg, (*pStateInfo).InstLabel,msg
            ;time1= 'CPI: ' + strtrim(fix(frame_time.hour),2) + ':' + strtrim(fix(frame_time.minute),2) + ':'+  strtrim(fix(frame_time.sec),2) + ':'+  strtrim(fix(frame_time.msec),2)

            ; display the CPI hex time, swap around endianess for display

            word1 = string(format='(%"%04X")', frameData.frameheader.Time)
            word2 = string(format='(%"%04X")', frameData.frameheader.arrivalT1)
            word3 = string(format='(%"%04X")', frameData.frameheader.arrivalT2)

            word1_a = strmid(word1,0,2)
            word1_b = strmid(word1,2,2)
            word2_a = strmid(word2,0,2)
            word2_b = strmid(word2,2,2)
            word3_a = strmid(word3,0,2)
            word3_b = strmid(word3,2,2)

            time_hex = '0x' + word3_a + word3_b + word2_a + word2_b + word1_a + word1_b

            time_dec = string(format='(%"%02d:%02d:%02d.%03d")', $
                fix(frameData.frameheader.hour),   $
                fix(frameData.frameheader.minute), $
                fix(frameData.frameheader.sec),    $
                fix(frameData.frameheader.msec))

            widget_control, (*(*pStateInfo).p3ViewState).txt_CPI_hex,  set_value = time_hex
            widget_control, (*(*pStateInfo).p3ViewState).txt_CPI_time, set_value = time_dec

            ; put the CPI time string into a system variable for use in naming single png output images
            ; using . instead of : for file naming

            CPIfilename = strmid((*pStateInfo).InFileName,    strpos((*pStateInfo).InFileName, '\' , /reverse_search) + 1)
            CPIfilepath = strmid((*pStateInfo).InFileName, 0, strpos((*pStateInfo).InFileName, '\' , /reverse_search) + 1)

            time_png = string(format='(%"%sCPI_3views_%s_%02d.%02d.%02d.%03d.png")', $
                CPIfilepath,                       $
                CPIfilename,                       $
                fix(frameData.frameheader.hour),   $
                fix(frameData.frameheader.minute), $
                fix(frameData.frameheader.sec),    $
                fix(frameData.frameheader.msec))

            defsysv, '!CPI_3views_png_filename', exists = name_exists

            if(name_exists) then begin
                !CPI_3views_png_filename = time_png
            endif else begin
                defsysv, '!CPI_3views_png_filename', time_png
            endelse

            ; display the CPI H and V particle IDs
            (*(*pStateInfo).p3ViewState).val_CPI_HID = frameData.frameHeader.Pheight1
            HID = string(format='(%"0x%04x")', frameData.frameHeader.Pheight1)
            widget_control, (*(*pStateInfo).p3ViewState).txt_CPI_HID,  set_value = HID

            (*(*pStateInfo).p3ViewState).val_CPI_VID = frameData.frameHeader.Pheight2
            VID = string(format='(%"0x%04x")', frameData.frameHeader.Pheight2)
            widget_control, (*(*pStateInfo).p3ViewState).txt_CPI_VID,  set_value = VID

            ; initally, no background removed
            (*pStateInfo).backRemoved = !false

            ; save into state structure pFrameData always holds the current frame
            (*pStateInfo).pFrameData = ptr_new(frameData, /no_copy)

            if((*pStateInfo).Bkgs_exist and (*pStateInfo).BkgPos ge 0)then begin
                ; determine if the current background is still the best
                position = (*pStateInfo).ImgPos-1
                ImgscanList = (*(*pStateInfo).pFileScan)[*(*pStateInfo).pImgPosIndx]
                backScanList = where(ImgscanList.imgtype eq 2)

                if(ptr_valid((*pStateInfo).pBackPosIndx))then get_best_backGnd_pos,ImgscanList,backScanList,position

                if(position ne (*pStateInfo).BkgPos)then begin
                    ;ShowMsg, (*pStateInfo).InstLabel,'Current Background at Frame # '+ strtrim(position+1,2)

                    BackFrame = read_backGnd((*pStateInfo).InFileLun,(*(*pStateInfo).pFrameData).frameheader.ver,imgList[position].fpos)  ; contains the structure of the background

                    (*pStateInfo).BackFrame = ptr_new(BackFrame,/no_copy)
                    (*pStateInfo).BkgPos =  position
                endif; else print,'Current background OK'
            endif

            ; set min/max to gui ( this removes a background if needed )
            set_min_max, pStateInfo

            ; Show main info line
            ;disp_frame_info, pStateInfo, 0

            ;get Obj info if it exists
            if(ptr_valid((*pStateInfo).pObjList) and not background)then begin
                ;return,{FrameHeader:FrameHeader,pFrame:pFrame,pImHDfPos:ptr_new(imageHeaderFPOS)}
                ; clean up existing obj if it exists
                if(ptr_valid((*pStateInfo).pObjData))then begin
                    ptr_free,(*(*(*pStateInfo).pObjData).pFrame).imageheader.Harmonics
                    ptr_free,(*(*(*pStateInfo).pObjData).pFrame).imageheader.Boundary
                    ptr_free,(*(*pStateInfo).pObjData).pFrame
                    ptr_free,(*(*pStateInfo).pObjData).pImHDfPos
                    ptr_free,(*pStateInfo).pObjData
                endif

                ;BkgPosList = TEMPORARY(WHERE((*(*pStateInfo).pFileScan).ImgType eq 2, BkgCount))
                ;where((*pStateInfo).ImgPos-1 gt
                ;objData = get_obj_header((*pStateInfo).ObjLUN,(*(*pStateInfo).pObjList)[(*pStateInfo).ImgPos-1-BkgCount].fPos)
                ;objTime = float(objData.frameHeader.time) + calcStrobeTime(objData.frameHeader.arrivalT1,objData.frameHeader.arrivalT2)/1000.0D

                ; need to get calcStrobe time working for 3V-CPI
                ;roiTime = float((*(*pStateInfo).pFrameData).frameheader.time) + calcStrobeTime((*(*pStateInfo).pFrameData).frameheader.arrivalT1,(*(*pStateInfo).pFrameData).frameheader.arrivalT2,(*(*pStateInfo).pFrameData).frameheader.Ver)/1000.0D

                roiTime = float((*(*pStateInfo).pFrameData).frameheader.time); mo change this on august 15, 2012.
                roitime =(*(*pStateInfo).pFrameData).frameheader.hour*3600.0+(*(*pStateInfo).pFrameData).frameheader.minute*60.0 + $
                    (*(*pStateInfo).pFrameData).frameheader.sec*1.0+(*(*pStateInfo).pFrameData).frameheader.msec/1000.0D
                ;print,roiTime,objTime,format='(2(f15.5,x))'
                ;if(objTime ne roiTime)then begin

                indexOBJ = where(roiTime eq (*(*pStateInfo).pObjList).time, cntOBJ)
                ;print,roiTime,(*(*pStateInfo).pObjList)[indexOBJ[0]].time,format='(2(f15.5,x))'
                if(cntOBJ gt 0)then begin
                    ; read a frame of info from the obj file
                    objData = get_obj_header((*pStateInfo).ObjLUN,(*(*pStateInfo).pObjList)[indexOBJ[0]].fPos)
                    ; store obj info
                    (*pStateInfo).pObjData = ptr_new(objData,/no_copy)
                endif else print, 'Error: obj frame not found'
                ;endif else (*pStateInfo).pObjData = ptr_new(objData,/no_copy)
            endif else begin
                ; clean up existing obj if it exists
                if(ptr_valid((*pStateInfo).pObjData))then begin
                    ptr_free,(*(*(*pStateInfo).pObjData).pFrame).imageheader.Harmonics
                    ptr_free,(*(*(*pStateInfo).pObjData).pFrame).imageheader.Boundary
                    ptr_free,(*(*pStateInfo).pObjData).pFrame
                    ptr_free,(*(*pStateInfo).pObjData).pImHDfPos
                    ptr_free,(*pStateInfo).pObjData
                endif

                (*pStateInfo).pObjData = ptr_new()
            endelse

            pBoundary = PTR_NEW()
            tvlct,255,255,0,255

            ; now display frame
            ;if((*pStateInfo).displayPerim and not background)then begin
            if(!show_perimeter_3viewer and not background)then begin
                ; if an object file exist toggle perimeter
                if(ptr_valid((*pStateInfo).pObjList))then begin
                    ROI_Count = (*(*pStateInfo).pObjData).Frameheader.ROIsCount

                    if((*(*pStateInfo).pFrameData).Frameheader.ROIsCount eq ROI_Count)then begin
                        pBoundary = ptrarr(ROI_count)

                        for index = 0,ROI_count-1 do begin
                            header = (*(*(*pStateInfo).pObjData).pFrame)[index].imageheader
                            if(header.bound_len ne 0)then pBoundary[index] = header.boundary
                        endfor
                    endif
                endif
            endif

            ; temporary workaround for CPI and 2DS color compatibility
            wset, (*(*pStateInfo).p3ViewState).CPI_WID
            device, decomposed = 0
            loadct, !ct_CPI, /silent

            bounds = DispFrame(pStateInfo, pBoundary=pBoundary, dispwindow = (*(*pStateInfo).p3ViewState).CPI_WID)

            ptr_free, (*pStateInfo).pBounds
            (*pStateInfo).pBounds =  ptr_new(bounds)

            if ptr_valid((*pStateInfo).pObjData) eq 0 then begin
                Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
                return
            endif;   ; Mo added this 2014-12-29.


            ; display the CPI particle information (Length, Width, Area, etc)
            objData = (*(*pStateInfo).pObjData)
            index   = 0
            pixsize = (*pStateInfo).Settings.PixSize ; microns/pixel

            cpi_length    = (*objData.pFrame)[index].imageheader.len   * pixsize ; microns
            cpi_width     = (*objData.pFrame)[index].imageheader.wid   * pixsize ; microns
            cpi_perimeter = (*objData.pFrame)[index].imageheader.perim * pixsize ; microns
            cpi_area      = (*objData.pFrame)[index].imageheader.area  * pixsize * pixsize ; microns^2

            length_string = strtrim(string(cpi_length,    format='(%"%7.0f")'), 2)
            width_string = strtrim(string(cpi_width,     format='(%"%7.0f")'), 2)
            perimeter_string = strtrim(string(cpi_perimeter, format='(%"%7.0f")'), 2)
            Shaded_area_string = strtrim(string(cpi_area,      format='(%"%9.0f")'), 2)
            PiR2_area_string   = strtrim(string(!pi*cpi_length*cpi_length/4.0,format='(%"%9.0f")'),2)
            PiW2_area_string   = strtrim(string(!pi*cpi_width*cpi_width/4.0,format='(%"%9.0f")'),2)

            ;csp = ((area * microns_per_pixel * microns_per_pixel) * maxWidth*microns_per_pixel * (2.*maxLength*microns_per_pixel + 2.*maxWidth*microns_per_pixel)) / perimeterum
            if(cpi_perimeter gt 0) then begin
                cpi_csp = (cpi_area * cpi_width * (2.*cpi_length + 2.*cpi_width)) / cpi_perimeter
            endif else cpi_csp = -1. ; end if

            ; Mass from CSP using parameterization for normal shaded pixel area


            cpi_csp_mass = (10.0^(-9.0 * !csp_beta)) * !csp_alpha * cpi_csp^!csp_beta ; here csp in um, area in mm^2 ; sizes is im um. mass in mg

            csp_string = strtrim(string(cpi_csp_mass, format='(%"%11.3e")'), 2)

            temp_area =CPI_Area
            This_sizes =CPI_Length

            alpha = 0.115D
            beta = 1.218D

            mass1=alpha*double(temp_area)^beta; ;here sizes in um, area in mm^2 ; sizes is im um.
            mass2=0.917*(0.1^9.0*!pi/6.0)*This_sizes^3.0; size in um water-mass. milligrams.
            CPI_bradpaul_mass = mass1 < mass2
            CPI_bradPaul_str  = strtrim(string(cpi_bradPaul_mass, format='(%"%11.3e")'), 2)



            ; TAF can we filter the CPI images by size here, or above somewhere?  Need a checkbox and a size input field

            widget_control, (*(*pStateInfo).p3ViewState).txt_cpiL, set_value =    length_string
            widget_control, (*(*pStateInfo).p3ViewState).txt_cpiW, set_value =     width_string
            ;widget_control, (*(*pStateInfo).p3ViewState).txt_cpi_perim,  set_value = perimeter_string
            widget_control, (*(*pStateInfo).p3ViewState).txt_cshad_area, set_value =    shaded_area_string
            ;widget_control, (*(*pStateInfo).p3ViewState).txt_cpi_maxL_area,   set_value =     PiR2_area_string
            ;widget_control, (*(*pStateInfo).p3ViewState).txt_cpi_maxW_area,   set_value =     PiW2_area_string
            widget_control, (*(*pStateInfo).p3ViewState).txt_cpi_csp,    set_value =       csp_string
            widget_control, (*(*pStateInfo).p3ViewState).txt_cpi_iwc,    set_value =  cpi_bradpaul_str

            ; TAF set the CPI focus here, also find the H or V positions (distance from edges)

            ;stop


            ; find and draw the relevant 2D-S images
            if(dir gt 0) then begin
                H_find_next_ID, pStateInfo
                V_find_next_ID, pStateInfo
            endif else begin
                H_find_prev_ID, pStateInfo
                V_find_prev_ID, pStateInfo
            endelse
        endif else begin
            print,'Invalid Frame'
        endelse
    endif

    Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
end ; cpi_slider_event



;pro cpi_match_2DS_part_IDs_event, event
;    Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy  ; get our info structure

;    print, "cpi_match_2DS_part_IDs_event"

;    Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
;end ; cpi_match_2DS_part_IDs_event



pro H_slider_event, event
    Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy  ; get our info structure

    ;print, "H_slider_event"

    HorV = 'H'

    draw_2D_particle_image, HorV, event.value, pStateInfo

    Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
end ; H_slider_event



pro V_slider_event, event
    Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy  ; get our info structure

    ;print, "V_slider_event"

    HorV = 'V'

    draw_2D_particle_image, HorV, event.value, pStateInfo

    Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
end ; V_slider_event



pro H_find_prev_ID_event, event
    Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy  ; get our info structure

    ;print, "H_find_prev_ID_event"

    H_find_prev_ID, pStateInfo

    Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
end ; H_find_prev_ID_event



pro H_find_prev_ID, pStateInfo
    Common  DataBlock, alldataH, alldataV, ADataH, ADataV, FirstData, LastData, StripData, $
        FirstDataV, LastDataV, StripDataV, OneHzH, OneHzV, aveStripdata, aveStripdataV

    ; get H slider value

    widget_control, (*(*pStateInfo).p3ViewState).H_slider, get_value = H_slider_index

    ; get the CPI H Particle ID
    ; TAF currently stored in system variable !val_CPI_HID

    ; find all the H Particle IDs, and select the one previous to this one to draw

    index = where((*alldataH).PartID eq (*(*pStateInfo).p3ViewState).val_CPI_HID, count)

    if(count gt 0) then begin
        subindex = where(index lt H_slider_index, count)

        if(count gt 0) then begin
            ; the last item in the subindex points to the previous matching particle

            new_index = index[subindex[(size(subindex))[1]-1]]

            HorV = 'H'

            draw_2D_particle_image, HorV, new_index, pStateInfo
            widget_control, (*(*pStateInfo).p3ViewState).H_slider, set_value = new_index
        endif else begin
            ; TAF note currently misses the index if you just turned around in the search direction, recovers after scanning, but not perfect

            ;print, "H_find_prev_ID_event, no matching index found"

            draw_2D_blank, 'H', pStateInfo
        endelse
    endif else begin
        ;print, "H_find_prev_ID_event, PartID not found"

        draw_2D_blank, 'H', pStateInfo
    endelse
end ; H_find_prev_ID



pro H_find_next_ID_event, event
    Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy  ; get our info structure

    H_find_next_ID, pStateInfo

    Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
end ; H_find_next_ID_event



pro H_find_next_ID, pStateInfo
    Common  DataBlock, alldataH, alldataV, ADataH, ADataV, FirstData, LastData, StripData, $
        FirstDataV, LastDataV, StripDataV, OneHzH, OneHzV, aveStripdata, aveStripdataV

    ; get H slider value

    widget_control, (*(*pStateInfo).p3ViewState).H_slider, get_value = H_slider_index

    ; get the CPI H Particle ID
    ; TAF currently stored in system variable !val_CPI_HID

    ; find all the H Particle IDs, and select the one previous to this one to draw

    index = where((*alldataH).PartID eq (*(*pStateInfo).p3ViewState).val_CPI_HID, count)

    if(count gt 0) then begin
        subindex = where(index gt H_slider_index, count)

        if(count gt 0) then begin
            ; the first item in the subindex points to the previous matching particle

            new_index = index[subindex[0]]

            HorV = 'H'

            draw_2D_particle_image, HorV, new_index, pStateInfo
            widget_control, (*(*pStateInfo).p3ViewState).H_slider, set_value = new_index
        endif else begin
            ; TAF note currently misses the index if you just turned around in the search direction, recovers after scanning, but not perfect

            print, "H_find_next_ID_event, no matching index found"

            draw_2D_blank, 'H', pStateInfo
        endelse
    endif else begin
        print, "H_find_next_ID_event, PartID not found"

        draw_2D_blank, 'H', pStateInfo
    endelse
end ; H_find_next_ID



pro V_find_prev_ID_event, event
    Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy  ; get our info structure

    ;print, "V_find_prev_ID_event"

    V_find_prev_ID, pStateInfo

    Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
end ; V_find_prev_ID_event



pro V_find_prev_ID, pStateInfo
    Common  DataBlock, alldataH, alldataV, ADataH, ADataV, FirstData, LastData, StripData, $
        FirstDataV, LastDataV, StripDataV, OneHzH, OneHzV, aveStripdata, aveStripdataV

    ; get V slider value

    widget_control, (*(*pStateInfo).p3ViewState).V_slider, get_value = V_slider_index

    ; get the CPI V Particle ID
    ; TAF currently stored in system variable !val_CPI_VID

    ; find all the V Particle IDs, and select the one previous to this one to draw

    index = where((*alldataV).PartID eq (*(*pStateInfo).p3ViewState).val_CPI_VID, count)

    if(count gt 0) then begin
        subindex = where(index lt V_slider_index, count)

        if(count gt 0) then begin
            ; the last item in the subindex points to the previous matching particle

            new_index = index[subindex[(size(subindex))[1]-1]]

            HorV = 'V'

            draw_2D_particle_image, HorV, new_index, pStateInfo
            widget_control, (*(*pStateInfo).p3ViewState).V_slider, set_value = new_index
        endif else begin
            ; TAF note currently misses the index if you just turned around in the search direction, recovers after scanning, but not perfect

            ;print, "V_find_prev_ID_event, no matching index found"

            draw_2D_blank, 'V', pStateInfo
        endelse
    endif else begin
        ;print, "V_find_prev_ID_event, PartID not found"

        draw_2D_blank, 'V', pStateInfo
    endelse
end ; V_find_prev_ID



pro V_find_next_ID_event, event
    Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy  ; get our info structure

    ;print, "V_find_next_ID_event"

    V_find_next_ID, pStateInfo

    Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
end ; V_find_next_ID_event



pro V_find_next_ID, pStateInfo
    Common  DataBlock, alldataH, alldataV, ADataH, ADataV, FirstData, LastData, StripData, $
        FirstDataV, LastDataV, StripDataV, OneHzH, OneHzV, aveStripdata, aveStripdataV

    ; get V slider value

    widget_control, (*(*pStateInfo).p3ViewState).V_slider, get_value = V_slider_index

    ; get the CPI V Particle ID
    ; TAF currently stored in system variable !val_CPI_VID

    ; find all the V Particle IDs, and select the one previous to this one to draw

    index = where((*alldataV).PartID eq (*(*pStateInfo).p3ViewState).val_CPI_VID, count)

    if(count gt 0) then begin
        subindex = where(index gt V_slider_index, count)

        if(count gt 0) then begin
            ; the first item in the subindex points to the previous matching particle

            new_index = index[subindex[0]]

            HorV = 'V'

            draw_2D_particle_image, HorV, new_index, pStateInfo
            widget_control, (*(*pStateInfo).p3ViewState).V_slider, set_value = new_index
        endif else begin
            ; TAF note currently misses the index if you just turned around in the search direction, recovers after scanning, but not perfect

            ;print, "V_find_next_ID_event, no matching index found"

            draw_2D_blank, 'V', pStateInfo
        endelse
    endif else begin
        ;print, "V_find_next_ID_event, PartID not found"

        draw_2D_blank, 'V', pStateInfo
    endelse
end ; V_find_next_ID




pro show_perimeter_3viewer_event, event
    ; event for handling the "Show Perimeters" checkbox in 3 Viewer
    ; toggles the state flag for showing perimeters
    ; TAF will have to work on the redraw feature

    ;Common  DataBlock, alldataH, alldataV, ADataH, ADataV, FirstData, LastData, StripData, $
    ;        FirstDataV, LastDataV, StripDataV, OneHzH, OneHzV, aveStripdata, aveStripdataV

    Widget_Control, event.top, Get_UValue=strip_info, /No_Copy  ; get our info structure

    if !show_perimeter_3viewer eq 0 then !show_perimeter_3viewer=1 else !show_perimeter_3viewer=0

    ;ave_info = strip_info.ave_info
    ;foffset  = !Foffset

    ;newoffset = Fill_averaged_Strip(foffset, ave_info, strip_info, 0, 'H')

    Widget_Control, event.top, Set_UValue=strip_info, /No_Copy  ; release our info structure
end ; show_perimeter_3viewer_event



pro show_CPI_trigger_3viewer_event, event
    ; event for handling the "Show Perimeters" checkbox in 3 Viewer
    ; toggles the state flag for showing perimeters
    ; TAF will have to work on the redraw feature

    ;Common  DataBlock, alldataH, alldataV, ADataH, ADataV, FirstData, LastData, StripData, $
    ;        FirstDataV, LastDataV, StripDataV, OneHzH, OneHzV, aveStripdata, aveStripdataV

    Widget_Control, event.top, Get_UValue=strip_info, /No_Copy  ; get our info structure

    if !show_CPI_trigger_3viewer eq 0 then !show_CPI_trigger_3viewer=1 else !show_CPI_trigger_3viewer=0

    ;ave_info = strip_info.ave_info
    ;foffset  = !Foffset

    ;newoffset = Fill_averaged_Strip(foffset, ave_info, strip_info, 0, 'H')

    Widget_Control, event.top, Set_UValue=strip_info, /No_Copy  ; release our info structure
end ; show_CPI_trigger_3viewer_event




;pro txt_n_before, event
;    Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy  ; get our info structure

;    ;print, "txt_n_before"

;    widget_control, event.top, get_value = in_string

;    (*(*pStateInfo).p3ViewState).n_before = 0 > fix(in_string) < !MAX_2DS_IMAGES_3VIEW

;    ; TAF call update here

;    Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
;end ; txt_n_before



;pro txt_n_after, event
;    Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy  ; get our info structure

;    ;print, "txt_n_after"

;    widget_control, event.top, get_value = in_string

;    (*(*pStateInfo).p3ViewState).n_after = 0 > fix(in_string) < !MAX_2DS_IMAGES_3VIEW

;    ; TAF call update here

;    Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
;end ; txt_n_after



pro output_single_png_event, event
    ; Generates a png image from the CPI image and the H and V 2D-S images
    ; prints out the CPI and 2D-S times and particle IDs

    Common  DataBlock, alldataH, alldataV, ADataH, ADataV, FirstData, LastData, StripData, $
        FirstDataV, LastDataV, StripDataV, OneHzH, OneHzV, aveStripdata, aveStripdataV

    Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy

    ; get H and V slider values

    ;widget_control, (*(*pStateInfo).p3ViewState).H_slider, get_value = H_slider_index
    ;widget_control, (*(*pStateInfo).p3ViewState).V_slider, get_value = V_slider_index

    ; get the CPI H and V Particle ID
    ; TAF currently stored in system variable !val_CPI_VID

    ;print, 'output_single_png_event'




    ; capture the CPI image, as well as the 2D-S H and V images

    wset, (*(*pStateInfo).p3ViewState).CPI_WID

    cpi_image = tvrd(/true)

    wset, (*(*pStateInfo).p3ViewState).H_WID

    h_image = tvrd(/true)

    wset, (*(*pStateInfo).p3ViewState).V_WID

    v_image = tvrd(/true)

    x_cpi = (size(cpi_image))[2]
    y_cpi = (size(cpi_image))[3]

    x_2ds = (size(h_image))[2]

    gray_color = 80

    ; build the 3 view image in a buffer, with vertical gray bars separating the 3 images

    image_3views = bytarr(3, 2 + x_cpi + 2 * x_2ds, y_cpi)

    image_3views[*, 0: x_cpi-1, *] = cpi_image
    image_3views[*, x_cpi, *] = gray_color
    image_3views[*, 1 + x_cpi : x_cpi + x_2ds, *] = h_image
    image_3views[*, 1 + x_cpi + x_2ds, *] = gray_color
    image_3views[*, 2 + x_cpi + x_2ds : 1 + x_cpi + 2 * x_2ds, *] = v_image

    pnger, image = image_3views, !CPI_3views_png_filename, /rotated

    Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
end ; output_single_png_event



pro output_all_pngs_event, event
    ; Generates png images for each CPI and 2D-S images with matching particle IDs
    ; prints out the CPI and 2D-S times and particle IDs

    Common  DataBlock, alldataH, alldataV, ADataH, ADataV, FirstData, LastData, StripData, $
        FirstDataV, LastDataV, StripDataV, OneHzH, OneHzV, aveStripdata, aveStripdataV

    Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy

    ; get H and V slider values

    ;widget_control, (*(*pStateInfo).p3ViewState).H_slider, get_value = H_slider_index
    ;widget_control, (*(*pStateInfo).p3ViewState).V_slider, get_value = V_slider_index

    ; get the CPI H and V Particle ID
    ; TAF currently stored in system variable !val_CPI_VID

    ;print, 'output_all_pngs_event'

    Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
end ; output_all_pngs_event



pro output_3view_csv_event, event
    ; Generates png images for each CPI and 2D-S images with matching particle IDs
    ; prints out the CPI and 2D-S times and particle IDs

    Common  DataBlock, alldataH, alldataV, ADataH, ADataV, FirstData, LastData, StripData, $
        FirstDataV, LastDataV, StripDataV, OneHzH, OneHzV, aveStripdata, aveStripdataV

    Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy

    ; get H and V slider values

    ;widget_control, (*(*pStateInfo).p3ViewState).H_slider, get_value = H_slider_index
    ;widget_control, (*(*pStateInfo).p3ViewState).V_slider, get_value = V_slider_index

    ; get the CPI H and V Particle ID
    ; TAF currently stored in system variable !val_CPI_VID


    print, 'output_3view_csv starting:', systime()
    startseconds = systime(/seconds)

    CPIfilename = strmid((*pStateInfo).InFileName,    strpos((*pStateInfo).InFileName, '\' , /reverse_search) + 1)
    CPIfilepath = strmid((*pStateInfo).InFileName, 0, strpos((*pStateInfo).InFileName, '\' , /reverse_search) + 1)

    file_chars_to_remove = 4
    my_str1  = strmid(!currentHPname, 0,strlen(!currentHPname)-1);   that is the C:\****\data1  for example
    my_pos   = strpos(my_str1,'\',/reverse_search); get upper directory of data1.
    fname    = strmid(my_str1,0,my_pos+1)+'SORTED'
    horzname = strmid(!currentHFname,strpos(!currentHFname,"horiz")+5,strlen(!currentHFname))
    ;horzname= strmid(!currentHFname,strpos(!currentHFname,"horiz")+8,strlen(!currentHFname))
    datetime = strmid(horzname,0,strlen(horzname)-file_chars_to_remove)

    csv_filename = string(format='(%"%sCPI_3views_%s_2D_%s.csv")', $
        CPIfilepath, $
        CPIfilename, $
        datetime)

    openw,  olun, /get, csv_filename
    printf, olun, $
        'CPI Time UTC, 2D-S-H-Time, 2D-S-V-Time, CPI-Hex-Time, H-Hex, V-Hex, H-ID, V-ID, CPImaxL(um), HmaxLen(um), VmaxLen(um), CPImaxW(um), '+ $
        '  HmaxWid(um), VmaxWid(um), CPIPerim(um), HPerim(um), VPerim(um), CPIshadedarea(um^2), Hshadedarea(um^2), Vshadedarea(um^2), CPIcspM(mg), HcspM(mg), VcspM(mg),  ' +$
        '  CPIareaM(mg),  HareaM(mg),   VareaM(mg),  2DSHL1(um), 2DSHL5(um), 2DSHL7(um), 2DSVL1(um), 2DSVL5(um), 2DSVL7(um), HPiL1^2/4, HPiL5^2/4, '+ $
        '  HpiL7^2/4, HPimaxL^2/4, Hsatlit_area,  VPiL1^2/4, VPiL5^2/4, VpiL7^2/4, VPimaxL^2/4, Vsatlit_area,  ' +  $
        '  CPI_PiL^2/4, Hratio(focus),   Vratio(focus) '

    ;    for i=0L, hk_count-1 do begin
    ;        printf, olun, format='(%"%02d:%02d:%02d, %5d, %3d")', $
    ;                         hk_timestamp_array[i].hour, $
    ;                         hk_timestamp_array[i].minute, $
    ;                         hk_timestamp_array[i].second, $
    ;                         cpi_frame_rate[i], $
    ;                         cpi_min_pix[i]

    ;    endfor



    if(ptr_valid((*pStateInfo).pFrameData))then begin  ; this needs to be a defined structure to continue


        widget_control, (*(*pStateInfo).p3ViewState).txt_min_cpi_size,  get_value = min_cpi_size_string
        widget_control, (*(*pStateInfo).p3ViewState).txt_max_cpi_size,  get_value = max_cpi_size_string

        ; some minor error checking and ranging on the input
        const_cpi_low_range =      .5 ; microns
        const_cpi_hi_range  = 50000.  ; microns
        const_cpi_min_diff  =     1.  ; microns -- enforce at least this much between hi and low range

        min_cpi_size = float(min_cpi_size_string)
        min_cpi_size = (min_cpi_size ge const_cpi_low_range) ? min_cpi_size : const_cpi_low_range
        min_cpi_size = (min_cpi_size le const_cpi_hi_range)  ? min_cpi_size : const_cpi_hi_range

        max_cpi_size = float(max_cpi_size_string)
        max_cpi_size = (max_cpi_size ge const_cpi_low_range) ? max_cpi_size : const_cpi_low_range
        max_cpi_size = (max_cpi_size le const_cpi_hi_range)  ? max_cpi_size : const_cpi_hi_range

        ; make sure min is less than max, and the min_diff is enforced

        if(min_cpi_size eq max_cpi_size) then begin ; if they are equal, set the difference to min_diff
            if(max_cpi_size eq const_cpi_hi_range) then begin ; at the high end, decrease min
                min_cpi_size = min_cpi_size - const_cpi_min_diff
            endif else begin ; at the low end or in the middle, increase max
                max_cpi_size = max_cpi_size + const_cpi_min_diff
            endelse
        endif else if(min_cpi_size gt max_cpi_size) then begin ; if the min is greater than the max, switch them
            temp_val = min_cpi_size
            min_cpi_size = max_cpi_size
            max_cpi_size = temp_val
        endif

        ; update the text fields with the potentially modified values
        widget_control, (*(*pStateInfo).p3ViewState).txt_min_cpi_size,  set_value = strtrim(string(format='(%"%5.1f")', min_cpi_size), 2)
        widget_control, (*(*pStateInfo).p3ViewState).txt_max_cpi_size,  set_value = strtrim(string(format='(%"%5.1f")', max_cpi_size), 2)

        ; get the list of frames to use ( this is set via *pImgPosIndx which may have backgrounds depending on gui setting)
        ImgList = (*(*pStateInfo).pFileScan)[*(*pStateInfo).pImgPosIndx]

        ; handle the case when this is a background
        background = !false
        ;LastPos = (*pStateInfo).ImgPos
        ;(*pStateInfo).ImgPos = event.value

        ;dir =  (*pStateInfo).ImgPos - LastPos

        ; advance in the direction, find the next image and test it for size, if the size is OK then display, or if the start/end the image is hit

        done  = 0
        cpi_index = 0 ; currently only looking at first image in each frame
        loopcount = 0

        if((*pStateInfo).Bkgs_exist and (*pStateInfo).BkgPos ge 0)then begin
            ; determine if the current background is still the best
            position = 1
            ImgscanList = (*(*pStateInfo).pFileScan)[*(*pStateInfo).pImgPosIndx]
            backScanList = where(ImgscanList.imgtype eq 2)

            if(ptr_valid((*pStateInfo).pBackPosIndx))then get_best_backGnd_pos,ImgscanList,backScanList,position

            if(position ne (*pStateInfo).BkgPos)then begin
                ;ShowMsg, (*pStateInfo).InstLabel,'Current Background at Frame # '+ strtrim(position+1,2)

                BackFrame = read_backGnd((*pStateInfo).InFileLun,(*(*pStateInfo).pFrameData).frameheader.ver,imgList[position].fpos)  ; contains the structure of the background

                (*pStateInfo).BackFrame = ptr_new(BackFrame,/no_copy)
                (*pStateInfo).BkgPos =  position
            endif; else print,'Current background OK'
        endif

        H_index = 0
        V_index = 0
        n_triplets = 0

        ; temporary workaround for CPI and 2DS color compatibility
        wset, (*(*pStateInfo).p3ViewState).CPI_WID
        device, decomposed = 0
        loadct, !ct_CPI, /silent


        ;*************************  to show the progress ---please wait bar  ******************************
        Widget_Control, event.top, TLB_Get_Offset=offsets
        info = ShowProgress(Message='Please wait,processing data...', XOffset=offsets(0)+50, $
            YOffset=offsets(1)+50, DrawSize=200, ButtonTitle='Cancel Local Operation')


        ; Set counters. These will indicate how often you want to update
        ; the "progress" display and check for a "cancel operation"
        ; event.

        counter = 1 > ((*pStateInfo).TotImgCount+1)/100
        percentDone = 0.0

        ; Set the y coordinates of the red box. These never change

        ybox_coords = [0, 10, 10, 0, 0]

        ; Take over color index 1 to draw a red box in the draw widget

        TVLct, r, g, b, /Get
        oldDrawColor = [r(1), g(1), b(1)]
        TVLct, 255, 0, 0, 1
        ;***************************************************************************************************
        N_of_loop =((*pStateInfo).TotImgCount+1) > 1

        for j=1L, (*pStateInfo).TotImgCount-1 do begin
            ; delete current frame
            delete_frame, (*pStateInfo).pFrameData

            ; get the frame indicated via position in scan list

            (*pStateInfo).ImgPos = j

            frameData = read_frame((*pStateInfo).InFileLun,(*pStateInfo).FileHead.Ver,(ImgList[(*pStateInfo).ImgPos-1]).fpos,date=(*pStateInfo).date)

            if(size(frameData, /type) eq 8)then begin
                ; display the CPI hex time

                word1 = string(format='(%"%04X")', frameData.frameheader.Time)
                word2 = string(format='(%"%04X")', frameData.frameheader.arrivalT1)
                word3 = string(format='(%"%04X")', frameData.frameheader.arrivalT2)

                word1_a = strmid(word1,0,2)
                word1_b = strmid(word1,2,2)
                word2_a = strmid(word2,0,2)
                word2_b = strmid(word2,2,2)
                word3_a = strmid(word3,0,2)
                word3_b = strmid(word3,2,2)

                cpi_time_hex = '0x' + word3_a + word3_b + word2_a + word2_b + word1_a + word1_b

                cpi_time_dec = string(format='(%"%02d:%02d:%02d.%03d")', $
                    fix(frameData.frameheader.hour),   $
                    fix(frameData.frameheader.minute), $
                    fix(frameData.frameheader.sec),    $
                    fix(frameData.frameheader.msec))

                ;widget_control, (*(*pStateInfo).p3ViewState).txt_CPI_hex,  set_value = time_hex
                ;widget_control, (*(*pStateInfo).p3ViewState).txt_CPI_time, set_value = time_dec

                ; display the CPI H and V particle IDs
                (*(*pStateInfo).p3ViewState).val_CPI_HID = frameData.frameHeader.Pheight1
                ;HID = string(format='(%"0x%04X")', frameData.frameHeader.Pheight1)
                ;widget_control, (*(*pStateInfo).p3ViewState).txt_CPI_HID,  set_value = HID

                (*(*pStateInfo).p3ViewState).val_CPI_VID = frameData.frameHeader.Pheight2
                ;VID = string(format='(%"0x%04X")', frameData.frameHeader.Pheight2)
                ;widget_control, (*(*pStateInfo).p3ViewState).txt_CPI_VID,  set_value = VID

                ; initally, no background removed
                (*pStateInfo).backRemoved = !false

                ; save into state structure pFrameData always holds the current frame
                (*pStateInfo).pFrameData = ptr_new(frameData, /no_copy)

                ; set min/max to gui ( this removes a background if needed )
                set_min_max, pStateInfo

                ;get Obj info if it exists
                if(ptr_valid((*pStateInfo).pObjList) and not background)then begin
                    ; clean up existing obj if it exists
                    if(ptr_valid((*pStateInfo).pObjData))then begin
                        ptr_free, (*(*(*pStateInfo).pObjData).pFrame).imageheader.Harmonics
                        ptr_free, (*(*(*pStateInfo).pObjData).pFrame).imageheader.Boundary
                        ptr_free,   (*(*pStateInfo).pObjData).pFrame
                        ptr_free,   (*(*pStateInfo).pObjData).pImHDfPos
                        ptr_free,     (*pStateInfo).pObjData
                    endif

                    roiTime = float((*(*pStateInfo).pFrameData).frameheader.time)
                    roitime =(*(*pStateInfo).pFrameData).frameheader.hour*3600.0+(*(*pStateInfo).pFrameData).frameheader.minute*60.0 + $
                        (*(*pStateInfo).pFrameData).frameheader.sec*1.0+(*(*pStateInfo).pFrameData).frameheader.msec/1000.0D

                    indexOBJ = where(roiTime eq (*(*pStateInfo).pObjList).time, cntOBJ)

                    if(cntOBJ gt 0)then begin
                        ; read a frame of info from the obj file
                        objData = get_obj_header((*pStateInfo).ObjLUN,(*(*pStateInfo).pObjList)[indexOBJ[0]].fPos)
                        ; store obj info
                        (*pStateInfo).pObjData = ptr_new(objData,/no_copy)
                    endif else print, 'Error: obj frame not found'

                endif else begin
                    ; clean up existing obj if it exists
                    if(ptr_valid((*pStateInfo).pObjData))then begin
                        ptr_free, (*(*(*pStateInfo).pObjData).pFrame).imageheader.Harmonics
                        ptr_free, (*(*(*pStateInfo).pObjData).pFrame).imageheader.Boundary
                        ptr_free,   (*(*pStateInfo).pObjData).pFrame
                        ptr_free,   (*(*pStateInfo).pObjData).pImHDfPos
                        ptr_free,     (*pStateInfo).pObjData
                    endif

                    (*pStateInfo).pObjData = ptr_new()
                endelse

                ; TAF NOTE example2 3 view data 2014-10-03
                ; crashed here on an invalid frame?  was able to skip along to keep processing, should automate
                if ptr_valid((*pStateInfo).pObjData) eq 0 then begin;  Mo added this to skip along.
                    goto, nextJJJJ
                endif

                ; upate the wait bar.  ---start
                ;******************************************************************
                IF (j mod counter) eq 0  THEN BEGIN

                    percentDone =float(j)/float(N_of_loop)

                    ; Slow the loop at bit.

                    Wait, 0.2

                    ; Update the progress box in the display

                    ; Make the draw widget the current window.

                    WSet, info.wid

                    ; Calculate the extent of the box to be drawn in the draw widget

                    xextent = Fix(info.drawsize*percentDone)
                    xbox_coords = [0, 0, xextent, xextent, 0]

                    ; Draw the box

                    Polyfill, xbox_coords, ybox_coords, Color=1, /Device

                    ; Did the "Cancel Operation" button get clicked?
                    ; Look for an event. Don't turn off the hourglass cursor!

                    progressEvent = Widget_Event(info.cancel, /NoWait)

                    ; Is this a button event?

                    eventName = Tag_Names(progressEvent, /Structure_Name)
                    IF eventName EQ 'WIDGET_BUTTON' THEN BEGIN

                        ; If it IS a button event, destroy the widget program and
                        ; issue an informational message to the user to alert him or her.

                        Widget_Control, info.top, /Destroy
                        result = Widget_Message('Operation Canceled!!', /Information)

                        ; Escape from the loop. Interrupt code would go here.

                        GoTo, outSideLoop

                    ENDIF

                ENDIF
                ; upate the wait bar.  ---end
                ;******************************************************************

                objData = (*(*pStateInfo).pObjData)

                pixsize = (*pStateInfo).Settings.PixSize ; microns/pixel
                length  = (*objData.pFrame)[cpi_index].imageheader.len   * pixsize ; microns

                ; test for  CPI image length in specified range
                if((length ge min_cpi_size) and (length le max_cpi_size)) then begin
                    pBoundary = PTR_NEW()
                    tvlct,255,255,0,255

                    ; now display frame
                    if(!show_perimeter_3viewer and not background)then begin
                        ; if an object file exist toggle perimeter
                        if(ptr_valid((*pStateInfo).pObjList))then begin
                            ROI_Count = (*(*pStateInfo).pObjData).Frameheader.ROIsCount

                            if((*(*pStateInfo).pFrameData).Frameheader.ROIsCount eq ROI_Count)then begin
                                pBoundary = ptrarr(ROI_count)

                                for cpi_index = 0,ROI_count-1 do begin
                                    header = (*(*(*pStateInfo).pObjData).pFrame)[cpi_index].imageheader
                                    if(header.bound_len ne 0)then pBoundary[cpi_index] = header.boundary
                                endfor
                            endif
                        endif
                    endif

                    bounds = DispFrame(pStateInfo, pBoundary=pBoundary, dispwindow = (*(*pStateInfo).p3ViewState).CPI_WID)

                    ptr_free, (*pStateInfo).pBounds
                    (*pStateInfo).pBounds =  ptr_new(bounds)

                    ; display the CPI particle information (Length, Width, Area, etc)
                    objData = (*(*pStateInfo).pObjData)
                    cpi_index   = 0
                    pixsize = (*pStateInfo).Settings.PixSize ; microns/pixel

                    cpi_length    = (*objData.pFrame)[cpi_index].imageheader.len   * pixsize ; microns
                    cpi_width     = (*objData.pFrame)[cpi_index].imageheader.wid   * pixsize ; microns
                    cpi_perimeter = (*objData.pFrame)[cpi_index].imageheader.perim * pixsize ; microns
                    cpi_area      = (*objData.pFrame)[cpi_index].imageheader.area  * pixsize * pixsize ; microns^2

                    ;   length_string = strtrim(string(length,    format='(%"%7.1f")'), 2)
                    ;    width_string = strtrim(string(width,     format='(%"%7.1f")'), 2)
                    ;perimeter_string = strtrim(string(perimeter, format='(%"%7.1f")'), 2)
                    ;     area_string = strtrim(string(area,      format='(%"%9.1f")'), 2)

                    ;csp = ((area * microns_per_pixel * microns_per_pixel) * maxWidth*microns_per_pixel * (2.*maxLength*microns_per_pixel + 2.*maxWidth*microns_per_pixel)) / perimeterum
                    if(cpi_perimeter gt 0) then begin
                        ;csp = ((area * pixsize * pixsize) * width*pixsize * (2.*length*pixsize + 2.*width*pixsize)) / perimeter

                        ;csp = ((Area*microns_per_pixel*microns_per_pixel) * maxWidth*microns_per_pixel * (2.*maxLength*microns_per_pixel + 2.*maxWidth*microns_per_pixel)) / perimeterum

                        cpi_csp = ((cpi_area) * cpi_width * (2.*cpi_length + 2.*cpi_width)) / cpi_perimeter
                    endif else cpi_csp = -1. ; end if

                    ; Mass from CSP using parameterization for normal shaded pixel area
                    cpi_csp_mass = (10.0^(-9.0 * !csp_beta)) * !csp_alpha * cpi_csp^!csp_beta ; here csp in um, area in mm^2 ; sizes is im um. mass in mg

                    ;*****************************************************
                    cpi_length_string = strtrim(string(cpi_length,    format='(%"%7.0f")'), 2)
                    cpi_width_string = strtrim(string(cpi_width,     format='(%"%7.0f")'), 2)
                    cpi_perimeter_string = strtrim(string(cpi_perimeter, format='(%"%7.0f")'), 2)
                    cpi_Shaded_area_string = strtrim(string(cpi_area,      format='(%"%9.0f")'), 2)
                    cpi_PiR2_area_string   = strtrim(string(!pi*cpi_length*cpi_length/4.0,format='(%"%9.0f")'),2)
                    cpi_PiW2_area_string   = strtrim(string(!pi*cpi_width*cpi_width/4.0,format='(%"%9.0f")'),2)
                    cpi_csp_string = strtrim(string(cpi_csp_mass, format='(%"%11.3e")'), 2)
                    temp_area =CPI_Area
                    This_sizes =CPI_Length
                    alpha = 0.115D
                    beta = 1.218D
                    mass1=alpha*double(temp_area)^beta; ;here sizes in um, area in mm^2 ; sizes is im um.
                    mass2=0.917*(0.1^9.0*!pi/6.0)*This_sizes^3.0; size in um water-mass. milligrams.
                    CPI_bradpaul_mass = mass1 < mass2
                    CPI_bradPaul_str  = strtrim(string(cpi_bradPaul_mass, format='(%"%11.3e")'), 2)


                    ;************************************


                    ;****************************************************

                    ;csp_string = strtrim(string(csp_mass, format='(%"%11.3e")'), 2)

                    ; find the relevant 2D-S images

                    ; get the 2D-S H particle information
                    have_H = 0

                    index = where((*alldataH).PartID eq (*(*pStateInfo).p3ViewState).val_CPI_HID, count)

                    if(count gt 0) then begin
                        subindex = where(index gt H_index, count)

                        if(count gt 0) then begin
                            ; the first item in the subindex points to the previous matching particle

                            subindex = index[subindex[0]]

                            HorV = 'H'

                            ;draw_2D_particle_image, HorV, new_index, pStateInfo
                            ;widget_control, (*(*pStateInfo).p3ViewState).H_slider, set_value = new_index

                            aveStripData      = {_DerivedDataRev2}
                            ;subindex          = *(ave_info.Hsubindex)
                            nnnn              = n_elements(subindex) ; there are total of nnnn particles to be displayed
                            sub_foffset       = (*alldataH)(subindex).imageoffset
                            L1size_array      = (*alldataH)(subindex).Nslice
                            imgword_array     = (*alldataH)(subindex).imageWords
                            nanoseconds_array = (*alldataH)(subindex).nanoseconds
                            CPI_trigger_array = (*alldataH)(subindex).CPI_trigger
                            PartID_array      = (*alldataH)(subindex).PartID

                            H_timestamp = SPEC_Nano2HMS((*alldataH)(subindex).Nanoseconds)
                            H_hex_timestamp = string(format = '(%"0x%012X")', (*alldataH)(subindex).ProbeCounter)

                            if(L1size_array gt 0) then begin

                                i=0
                                yoffset = 0

                                foffsets_array    =       sub_foffset(i)
                                L1size_array      =      L1size_array(i)
                                imgword_array     =     imgword_array(i)
                                nanoseconds_array = nanoseconds_array(i)
                                CPI_trigger_array = CPI_trigger_array(i)
                                PartID_array      =      PartID_array(i)
                                use_index         =          subindex(i)

                                ; create a filename that is one directory above the Data file's horizyymmddhhmmss.bin that was selected above
                                ; Here it doesn't matter what the current file is because we will always have a horiz filename at this point to work with
                                ;fname = strmid(!currentHPname,0,strlen(!currentHPname)-5) + 'SORTED' +   strmid(!currentHFname,strpos(!currentHFname,"horiz")+5,12) + '.2DS'
                                ; substrsF=strsplit(!currentHFname,'\',/extract)
                                ; short_filename=substrsF(n_elements(substrsF)-1)
                                ; datestamp = strmid(Short_filename,5,strlen(Short_filename)-9)
                                ; fname = strmid(!currentHPname,0,strlen(!currentHPname)-5) + 'SORTED' + datestamp+ '.2DS'

                                file_chars_to_remove = 4

                                my_str1  = strmid(!currentHPname, 0,strlen(!currentHPname)-1);   that is the C:\****\data1  for example
                                my_pos   = strpos(my_str1,'\',/reverse_search); get upper directory of data1.
                                fname    = strmid(my_str1,0,my_pos+1)+'SORTED'
                                horzname = strmid(!currentHFname,strpos(!currentHFname,"horiz")+5,strlen(!currentHFname))
                                ;horzname= strmid(!currentHFname,strpos(!currentHFname,"horiz")+8,strlen(!currentHFname))
                                datetime = strmid(horzname,0,strlen(horzname)-file_chars_to_remove)
                                ;fname = fname+datetime+'.2DS';   get the sorted****.2ds full name.
                                fname = fname + datetime + '.2DSCPI' ; get the sorted****.2ds full name.
                                OpenR, lun, fname, /Get_Lun

                                WordCount = ImgWord_array(i)
                                Slices    =  L1size_array(i)

                                if(WordCount eq 0) then begin
                                    print, "WordCount = 0, cannot draw image -- Have you loaded 2D-S data?"

                                    ;return
                                endif else begin
                                    Image = BytArr(WordCount * 2)

                                    POINT_LUN, lun, fOffsets_array(i)
                                    ReadU, lun, Image

                                    out_byte_array = BytArr(128L, Slices)

                                    HSums=intarr(slices)
                                    HMax  = 0
                                    H_Area = 0UL
                                    Midpoint = 0
                                    Edges = 0
                                    L3 = 0.0
                                    L4 = 0
                                    L5 = 0
                                    PC3x = 0.0
                                    PC3y = 0.0
                                    PC4x = 0.0
                                    PC4y = 0.0
                                    A3 = 0.0
                                    Roundness1 = 0.0
                                    Roundness2 = 0.0
                                    Roundness3 = 0.0
                                    L1r = 0.0
                                    H_maxLength = 0.0
                                    H_maxWidth  = 0.0
                                    perimeter = 0UL
                                    H_perimeterum = 0.
                                    H_csp = 0.
                                    Err=0

                                    ml_p1x = 0
                                    ml_p1y = 0
                                    ml_p2x = 0
                                    ml_p2y = 0
                                    mw_p1x = 0
                                    mw_p1y = 0
                                    mw_p2x = 0
                                    mw_p2y = 0

                                    perimeter_points_x = intarr(128L*Slices)
                                    perimeter_points_y = intarr(128L*Slices)

                                    ProcessParticle, Image, out_byte_array, Wordcount,Slices, HSums, HMax, H_Area, Midpoint, Edges,Err , $
                                        L3,L4,L5,PC3x,PC3y,PC4x,PC4y,A3, Roundness1, Roundness2, Roundness3, L1r, H_maxLength, H_maxWidth, perimeter, crystal_type, H_perimeterum, H_csp, $
                                        ml_p1x=ml_p1x, ml_p1y=ml_p1y, ml_p2x=ml_p2x, ml_p2y=ml_p2y, mw_p1x=mw_p1x, mw_p1y=mw_p1y, mw_p2x=mw_p2x, mw_p2y=mw_p2y, $
                                        perimeter_points_x=perimeter_points_x, perimeter_points_y=perimeter_points_y, HorV=HorV,largest_satellite_area=largest_satellite_area


                                    ; Mass from CSP using parameterization for normal shaded pixel area

                                    H_csp_mass = (10.0^(-9.0 * !csp_beta)) * !csp_alpha * H_csp^!csp_beta ; here csp in um, area in mm^2 ; sizes is im um. mass in mg

                                    microns_per_pixel = !microns_per_pixelH
                                    alexei1 = 0
                                    find_Image_Parameters, out_byte_array, alexei1  ; this is to find alexi1 max area.
                                    Get_alexei_correct_size,Alexei1,H_area,L4*microns_per_pixel,L7;  L7 in unit um.


                                    temp_area =H_Area
                                    This_sizes =H_maxLength
                                    alpha = 0.115D
                                    beta = 1.218D

                                    mass1=alpha*double(temp_area*microns_per_pixel*microns_per_pixel/1000000.0)^beta; ;here sizes in um, area in mm^2 ; sizes is im um.
                                    mass2=0.917*(0.1^9.0*!pi/6.0)*(This_sizes*microns_per_pixel)^3.0; size in um water-mass. milligrams.

                                    H_bradpaul_mass = mass1 < mass2;    % take the smaller one as the value
                                    ;******************************************************************************
                                    HL1=slices
                                    HmaxL_str= strtrim(string(H_maxLength   * microns_per_pixel, format='(%"%7.0f")'), 2)
                                    HmaxW_str= strtrim(string(H_maxwidth    * microns_per_pixel, format='(%"%7.0f")'), 2)
                                    HL1_str  = strtrim(string(HL1     * microns_per_pixel, format='(%"%7.0f")'), 2)
                                    HL5_str  = strtrim(string(L5    * microns_per_pixel, format='(%"%7.0f")'), 2)
                                    HL7_str  = strtrim(string(L7, format='(%"%7.0f")'), 2)
                                    H_alexei1_ratio   = H_area/Alexei1
                                    H_alx_area_ration_str= strtrim(string(H_Alexei1_ratio, format='(%"%5.2f")'), 2)
                                    Hsat_area_str  = strtrim(string(largest_satellite_area, format='(%"%9.0f")'), 2)
                                    HPimaxL_area_str= strtrim(string(!pi*(H_maxlength*microns_per_pixel/2.0)^2, format='(%"%9.0f")'), 2)
                                    HPimaxW_area_str= strtrim(string(!pi*(H_maxwidth*microns_per_pixel/2.0)^2, format='(%"%9.0f")'), 2)
                                    HPiL1_area_str  = strtrim(string(!pi*(HL1*microns_per_pixel/2.0)^2, format='(%"%9.0f")'), 2)
                                    HPiL5_area_str  = strtrim(string(!pi*(L5*microns_per_pixel/2.0)^2, format='(%"%9.0f")'), 2)
                                    HPiL7_area_str  = strtrim(string(!pi*(L7/2.0)^2, format='(%"%9.0f")'), 2)
                                    H_iwc_str       = strtrim(string(H_bradpaul_mass, format='(%"%11.3e")'), 2)
                                    ;******************************************************************************

                                    have_H = 1
                                endelse

                                free_lun, lun
                            endif
                        endif
                    endif

                    ; get the 2D-S V particle information
                    have_V = 0

                    index = where((*alldataV).PartID eq (*(*pStateInfo).p3ViewState).val_CPI_VID, count)

                    if(count gt 0) then begin
                        subindex = where(index gt V_index, count)

                        if(count gt 0) then begin
                            ; the first item in the subindex points to the previous matching particle

                            subindex = index[subindex[0]]

                            HorV = 'V'

                            ;draw_2D_particle_image, HorV, new_index, pStateInfo
                            ;widget_control, (*(*pStateInfo).p3ViewState).V_slider, set_value = new_index

                            aveStripDataV     = {_DerivedDataRev2}
                            ;subindex          = *(ave_info.Vsubindex)
                            nnnn              = n_elements(subindex) ; there are total of nnnn particles to be displayed
                            sub_foffset       = (*alldataV)(subindex).imageoffset
                            L1size_array      = (*alldataV)(subindex).Nslice
                            imgword_array     = (*alldataV)(subindex).imageWords
                            nanoseconds_array = (*alldataV)(subindex).nanoseconds
                            CPI_trigger_array = (*alldataV)(subindex).CPI_trigger
                            PartID_array      = (*alldataV)(subindex).PartID

                            V_timestamp = SPEC_Nano2HMS((*alldataV)(subindex).Nanoseconds)
                            V_hex_timestamp = string(format = '(%"0x%012X")', (*alldataV)(subindex).ProbeCounter)

                            if(L1size_array gt 0) then begin
                                i=0
                                yoffset = 0

                                foffsets_array    =       sub_foffset(i)
                                L1size_array      =      L1size_array(i)
                                imgword_array     =     imgword_array(i)
                                nanoseconds_array = nanoseconds_array(i)
                                CPI_trigger_array = CPI_trigger_array(i)
                                PartID_array      =      PartID_array(i)
                                use_index         =          subindex(i)

                                ; create a filename that is one directory above the Data file's horizyymmddhhmmss.bin that was selected above
                                ; Here it doesn't matter what the current file is because we will always have a horiz filename at this point to work with
                                ;fname = strmid(!currentHPname,0,strlen(!currentHPname)-5) + 'SORTED' +   strmid(!currentHFname,strpos(!currentHFname,"horiz")+5,12) + '.2DS'
                                ; substrsF=strsplit(!currentHFname,'\',/extract)
                                ; short_filename=substrsF(n_elements(substrsF)-1)
                                ; datestamp = strmid(Short_filename,5,strlen(Short_filename)-9)
                                ; fname = strmid(!currentHPname,0,strlen(!currentHPname)-5) + 'SORTED' + datestamp+ '.2DS'

                                file_chars_to_remove = 4

                                my_str1  = strmid(!currentHPname, 0,strlen(!currentHPname)-1);   that is the C:\****\data1  for example
                                my_pos   = strpos(my_str1,'\',/reverse_search); get upper directory of data1.
                                fname    = strmid(my_str1,0,my_pos+1)+'SORTED'
                                horzname = strmid(!currentHFname,strpos(!currentHFname,"horiz")+5,strlen(!currentHFname))
                                ;horzname= strmid(!currentHFname,strpos(!currentHFname,"horiz")+8,strlen(!currentHFname))
                                datetime = strmid(horzname,0,strlen(horzname)-file_chars_to_remove)
                                ;fname = fname+datetime+'.2DS';   get the sorted****.2ds full name.
                                fname = fname + datetime + '.2DSCPI' ; get the sorted****.2ds full name.
                                OpenR, lun, fname, /Get_Lun

                                WordCount = ImgWord_array(i)
                                Slices    =  L1size_array(i)

                                if(WordCount eq 0) then begin
                                    print, "WordCount = 0, cannot draw image -- Have you loaded 2D-S data?"

                                    ;return
                                endif else begin
                                    Image = BytArr(WordCount * 2)

                                    POINT_LUN, lun, fOffsets_array(i)
                                    ReadU, lun, Image

                                    out_byte_array = BytArr(128L, Slices)

                                    HSums=intarr(slices)
                                    HMax  = 0
                                    V_Area = 0UL
                                    Midpoint = 0
                                    Edges = 0
                                    L3 = 0.0
                                    L4 = 0
                                    L5 = 0
                                    PC3x = 0.0
                                    PC3y = 0.0
                                    PC4x = 0.0
                                    PC4y = 0.0
                                    A3 = 0.0
                                    Roundness1 = 0.0
                                    Roundness2 = 0.0
                                    Roundness3 = 0.0
                                    L1r = 0.0
                                    V_maxLength = 0.0
                                    V_maxWidth  = 0.0
                                    perimeter = 0UL
                                    V_perimeterum = 0.
                                    V_csp = 0.
                                    Err=0

                                    ml_p1x = 0
                                    ml_p1y = 0
                                    ml_p2x = 0
                                    ml_p2y = 0
                                    mw_p1x = 0
                                    mw_p1y = 0
                                    mw_p2x = 0
                                    mw_p2y = 0

                                    perimeter_points_x = intarr(128L*Slices)
                                    perimeter_points_y = intarr(128L*Slices)

                                    ProcessParticle, Image, out_byte_array, Wordcount,Slices, HSums, HMax, V_Area, Midpoint, Edges,Err , $
                                        L3,L4,L5,PC3x,PC3y,PC4x,PC4y,A3, Roundness1, Roundness2, Roundness3, L1r, V_maxLength, V_maxWidth, perimeter, crystal_type, V_perimeterum, V_csp, $
                                        ml_p1x=ml_p1x, ml_p1y=ml_p1y, ml_p2x=ml_p2x, ml_p2y=ml_p2y, mw_p1x=mw_p1x, mw_p1y=mw_p1y, mw_p2x=mw_p2x, mw_p2y=mw_p2y, $
                                        perimeter_points_x=perimeter_points_x, perimeter_points_y=perimeter_points_y, HorV=HorV,largest_satellite_area=largest_satellite_area


                                    ; Mass from CSP using parameterization for normal shaded pixel area

                                    V_csp_mass = (10.0^(-9.0 * !csp_beta)) * !csp_alpha * V_csp^!csp_beta ; here csp in um, area in mm^2 ; sizes is im um. mass in mg

                                    microns_per_pixel = !microns_per_pixelV
                                    alexei1 = 0
                                    find_Image_Parameters, out_byte_array, alexei1  ; this is to find alexi1 max area.
                                    Get_alexei_correct_size,Alexei1,V_area,L4*microns_per_pixel,L7;  L7 in unit um.

                                    temp_area =V_Area
                                    This_sizes=V_maxLength
                                    alpha = 0.115D
                                    beta = 1.218D

                                    mass1=alpha*double(temp_area*microns_per_pixel*microns_per_pixel/1000000.0)^beta; ;here sizes in um, area in mm^2 ; sizes is im um.
                                    mass2=0.917*(0.1^9.0*!pi/6.0)*(This_sizes*microns_per_pixel)^3.0; size in um water-mass. milligrams.


                                    V_bradpaul_mass = mass1 < mass2;    % take the smaller one as the value
                                    ;******************************************************************************
                                    VL1=slices
                                    VmaxL_str= strtrim(string(V_maxLength   * microns_per_pixel, format='(%"%7.0f")'), 2)
                                    VmaxW_str= strtrim(string(V_maxwidth    * microns_per_pixel, format='(%"%7.0f")'), 2)
                                    VL1_str  = strtrim(string(VL1     * microns_per_pixel, format='(%"%7.0f")'), 2)
                                    VL5_str  = strtrim(string(L5    * microns_per_pixel, format='(%"%7.0f")'), 2)
                                    VL7_str  = strtrim(string(L7, format='(%"%7.0f")'), 2)
                                    V_alexei1_ratio   = V_area/Alexei1
                                    V_alx_area_ration_str= strtrim(string(V_Alexei1_ratio, format='(%"%5.2f")'), 2)
                                    Vsat_area_str  = strtrim(string(largest_satellite_area, format='(%"%9.0f")'), 2)
                                    VPimaxL_area_str= strtrim(string(!pi*(V_maxlength*microns_per_pixel/2.0)^2, format='(%"%9.0f")'), 2)
                                    VPimaxW_area_str= strtrim(string(!pi*(V_maxwidth*microns_per_pixel/2.0)^2, format='(%"%9.0f")'), 2)
                                    VPiL1_area_str  = strtrim(string(!pi*(VL1*microns_per_pixel/2.0)^2, format='(%"%9.0f")'), 2)
                                    VPiL5_area_str  = strtrim(string(!pi*(L5*microns_per_pixel/2.0)^2, format='(%"%9.0f")'), 2)
                                    VPiL7_area_str  = strtrim(string(!pi*(L7/2.0)^2, format='(%"%9.0f")'), 2)
                                    V_iwc_str       = strtrim(string(V_bradpaul_mass, format='(%"%11.3e")'), 2)
                                    ;******************************************************************************
                                    have_V = 1
                                endelse

                                free_lun, lun
                            endif
                        endif
                    endif

                    if(have_H and have_V) then begin
;'CPI Time UTC, 2D-S H Time, 2D-S V Time, CPI Hex Time, H Hex, V Hex, H ID, V ID, CPI L(um), H L(um), V L(um), CPI W(um), H W(um), V W(um), CPI P(um), H P(um), V P(um), CPI A(um^2), H A(um^2), V A(um^2), CPI M(mg), H M(mg), V M(mg)'
                        n_triplets++
                        format_str='(%" %s, %s, %s, %s, %s, %s, 0x%04X, 0x%04X, %7.1f, %7.1f, %7.1f, %7.1f, %7.0f, %7.1f, %7.1f, %7.1f, %7.1f, %9.1f, %9.1f, %9.1f, %11.3e, %11.3e, %11.3e, '+$
                            '  %s, %s, %s, %s, %s,  %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s,  %s ")';  added 22 parameter string to print out.
                        ;printf, olun, format='(%" %s, %s, %s, %s, %s, %s, 0x%04X, 0x%04X, %7.1f, %7.1f, %7.1f, %7.1f, %7.0f, %7.1f, %7.1f, %7.1f, %7.1f, %9.1f, %9.1f, %9.1f, %11.3e, %11.3e, %11.3e, 22(%s,), %s")' ,$

                        printf, olun, format=format_str, $
                            cpi_time_dec, $
                            H_timestamp, $
                            V_timestamp, $
                            cpi_time_hex, $
                            H_hex_timestamp, $
                            V_hex_timestamp, $
                            (*(*pStateInfo).p3ViewState).val_CPI_HID, $
                            (*(*pStateInfo).p3ViewState).val_CPI_VID, $
                            cpi_length, $
                            H_maxLength * !microns_per_pixelH, $
                            V_maxLength * !microns_per_pixelV, $
                            cpi_width, $
                            H_maxWidth * !microns_per_pixelH, $
                            V_maxWidth * !microns_per_pixelV, $
                            cpi_perimeter, $
                            H_perimeterum, $
                            V_perimeterum, $
                            cpi_area, $
                            H_Area * !microns_per_pixelH * !microns_per_pixelH, $
                            V_Area * !microns_per_pixelV * !microns_per_pixelV, $
                            cpi_csp_mass, $
                            H_csp_mass, $
                            V_csp_mass, $;  added following 23 parameters to be print out.
                            CPI_bradPaul_str, $    ;CPI_iwc_str,$
                            H_iwc_str,  $
                            V_iwc_str , $    ;     = strtrim(string(V_bradpaul_mass, format='(%"%11.3e")'), 2)
                            HL1_str,    $
                            HL5_str,$
                            HL7_str,$
                            VL1_str,$;   = strtrim(string(VL1     * microns_per_pixel, format='(%"%7.0f")'), 2)
                            VL5_str,$;   = strtrim(string(L5    * microns_per_pixel, format='(%"%7.0f")'), 2)
                            VL7_str,$;   = strtrim(string(L7, format='(%"%7.0f")'), 2)
                            HPiL1_area_str, $
                            HPiL5_area_str, $
                            HPiL7_area_str, $
                            HPimaxL_area_str,  $  ;          HPimaxW_area_str,  $
                            Hsat_area_str, $
                            VPiL1_area_str, $;     = strtrim(string(!pi*(L1*microns_per_pixel/2.0)^2, format='(%"%9.0f")'), 2)
                            VPiL5_area_str, $;  = strtrim(string(!pi*(L5*microns_per_pixel/2.0)^2, format='(%"%9.0f")'), 2)
                            VPiL7_area_str, $;  = strtrim(string(!pi*(L7/2.0)^2, format='(%"%9.0f")'), 2)
                            VPimaxL_area_str,  $;            VPimaxW_area_str,  $
                            Vsat_area_str,   $  ;      = strtrim(string(largest_satellite_area, format='(%"%9.0f")'), 2)
                            cpi_PiR2_area_string, $;         cpi_PiW2_area_string,$
                            H_alx_area_ration_str, $
                            V_alx_area_ration_str;  = strtrim(string(V_Alexei1_ratio, format='(%"%5.2f")'), 2)

                    endif

                endif

                loopcount++
            endif

            nextJJJJ:  ; do nothing here. just for skip objdata not valid.

        endfor
        outsideloop:
    endif

    free_lun, olun

    print, 'output_3view_csv finished at:', systime()
    endseconds = systime(/seconds)
    print, 'elapsed seconds:', endseconds - startseconds
    print, 'generated ', n_triplets, ' matched triplets'

    ;close the wait bar
    ;***********************************
    ;***********************
    ; Make the draw widget the current window.

    WSet, info.wid
    ; Calculate the extent of the box to be drawn (full box).
    xbox_coords = [0, 0, Fix(info.drawsize), Fix(info.drawsize), 0]
    ; Draw the box
    Polyfill, xbox_coords, ybox_coords, Color=1, /Device
    wait,1
    ; Destroy the "show progress" widget program
    Widget_Control, info.top, /Destroy
    ;*****************************
    ;************************************

    Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
end ; output_3view_csv_event



pro load_new_cpi_file_event, event
    ; Loads a new CPI roi file, opens a file dialog for user to select file

    Common  DataBlock, alldataH, alldataV, ADataH, ADataV, FirstData, LastData, StripData, $
        FirstDataV, LastDataV, StripDataV, OneHzH, OneHzV, aveStripdata, aveStripdataV

    Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy

    ;    print, 'load_new_cpi_file_event'

    LoadFile, pStateInfo
    ; mo modified this on April 25  to added change Time  to the png pages to sync with 2DS images and other images.
    ;if !DT_CPI eq (-99999L) then begin

    if(!disable_time_popups eq 1) then begin
        !Time_synced = 1
    endif else begin
        !Time_synced=(moqixu_input_single_number_pop_up(title='If Time is synced enter 1 else 0: ',label='Time synced:',value='0'))*1L

        if !Time_synced gt 0 then begin
            !DT_CPI=(moqixu_input_single_number_pop_up(title='For relabeling CPI png page image only',label='CPI_Time_offset:',value='0'))*1L
        endif else begin
            !DT_CPI=0
        endelse
    endelse
    parameters = extractgui(pStateInfo)

    initpTimeInfo, pStateInfo,multiple = 0

    if(*pStateInfo).objLUN  eq -1  then begin
        process_frames , parameters,pStateInfo;   : this call makes a new obj file.  May 20, 2011.
    endif

    ; update pStateInfo with current objData
    if(ptr_valid((*pStateInfo).pObjList))then begin
        objData = get_obj_header((*pStateInfo).ObjLUN,(*(*pStateInfo).pObjList)[(*pStateInfo).ImgPos-1].fPos)
        ; store obj info
        (*pStateInfo).pObjData = ptr_new(objData)
    endif

    Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
end ; load_new_cpi_file_event



pro load_new_2ds_files_event, event
    ; Loads new 2D-S .bin files, opens a file dialog for user to select files

    Common  DataBlock, alldataH, alldataV, ADataH, ADataV, FirstData, LastData, StripData, $
        FirstDataV, LastDataV, StripDataV, OneHzH, OneHzV, aveStripdata, aveStripdataV

    Widget_Control, event.top, Get_UValue = pStateInfo, /No_Copy

    rrr=dialog_message('Please go to the 2D-S GUI, open a new 2DS .bin file which overlaps with CPI.',/info)

    Widget_Control, event.top, Set_UValue = pStateInfo, /No_Copy  ; release our info structure
end ; load_new_2ds_files_event



pro a3vcpiview_3_view_widget, pStateInfo
    Common  DataBlock, alldataH, alldataV, ADataH, ADataV, FirstData, LastData, StripData, $
        FirstDataV, LastDataV, StripDataV, OneHzH, OneHzV, aveStripdata, aveStripdataV
    ;Common avestripdata, avestripdataV ; The 2D-S H and V data, pStateInfo has the CPI data

    ; Sets up the Widget GUI window, displays, and buttons for the 3 Viewer


    if(ptr_valid((*pStateInfo).pFileScan) eq 0) then begin
        print, 'Error: No CPI ROI file data available. Open a CPI data file first.'
        ;stop

        return
    endif

    if(n_elements(alldataH) eq 0) then begin
        ;print, 'Error: No 2D-S data available. Open a 2D-S data file first.'
        rrr=dialog_message('Error: No 2D-S data available. Open a 2D-S data file first.',/info)
        ;stop
        return
    endif

    ;defsysv, '!PIXELS', 128

    defsysv, '!zoom_factor_3viewer', 4 ; the integer zoom in pixels to display 2D-S data -- 1 shaded bit will be shown on screen as a square
    ; that is !zoom_factor_3viewer on each side

    defsysv, '!show_perimeter_3viewer',   0 ; boolean flag to draw perimeters or not in the 3 Viewer
    defsysv, '!show_CPI_trigger_3viewer', 0 ; boolean flag to draw 2D-S CPI trigger particles in a different color

    defsysv, '!MAX_2DS_IMAGES_3VIEW', 10 ; the maximum allowed for n_before and n_after

    ;defsysv, '!POINTER_STATE_INFO', pStateInfo

    cpi_max = (size((*(*pStateInfo).pFileScan)[*(*pStateInfo).pImgPosIndx]))[1] - 1L
    H_max = (size((*alldataH)))[1] - 1L
    V_max = (size((*alldataV)))[1] - 1L

    cpi_image_draw_width     = 512
    cpi_image_draw_height    = 512
    ;hv_2DS_image_draw_width  = !PIXELS * 3 * !zoom_factor_3viewer ; for wider view
    hv_2DS_image_draw_width  = !PIXELS * !zoom_factor_3viewer / 2
    hv_2DS_image_draw_height = !PIXELS * !zoom_factor_3viewer

    image_layout_buffer = 10 ; pixels, used to space out the layout, bigger = more spread out


    ;    ; The Window Layout
    ;    layout_base = WIDGET_BASE(/row, title='3V-CPI View 3 Viewer', UVALUE='3V-CPI View 3 Viewer', FRAME=1, MAP=1) ; the lowest level widget base for the window

    ;    cpi_base = widget_base(layout_base, /col, /frame)

    ;    cpi_image_base     = widget_base(cpi_base, /col)
    ;    cpi_control_base   = widget_base(cpi_base, /row)
    ;    cpi_control_base_1 = widget_base(cpi_control_base, /col) ; for Frame Slider, and times
    ;    cpi_control_base_2 = widget_base(cpi_control_base, /col, xs = 250) ; for CPI 2DS ID H
    ;    cpi_control_base_3 = widget_base(cpi_control_base, /col) ; for CPI 2DS ID V

    ;    HV_2DS_base    = widget_base(layout_base, /col)

    ;    H_base         = widget_base(HV_2DS_base, /row, /frame)
    ;    H_image_base   = widget_base(H_base, /col)
    ;    H_control_base = widget_base(H_base, /col)

    ;    V_base         = widget_base(HV_2DS_base, /row, /frame)
    ;    V_image_base   = widget_base(V_base, /col)
    ;    V_control_base = widget_base(V_base, /col)

    ;    ; the Draw areas
    ;     cpi_image_draw = widget_draw(cpi_image_base, xs=cpi_image_draw_width, ys=cpi_image_draw_height)

    ;    H2ds_image_draw = widget_draw(H_image_base, xs=hv_2DS_image_draw_width, ys=hv_2DS_image_draw_height)
    ;    V2ds_image_draw = widget_draw(V_image_base, xs=hv_2DS_image_draw_width, ys=hv_2DS_image_draw_height)


    ;    ; the Controls and Text fields
    ;    cpi_txt1 = widget_text(cpi_control_base_1, value = 'CPI Hex Time')
    ;    cpi_txt2 = widget_text(cpi_control_base_1, value = 'CPI Time')

    ;    cpi_slider = widget_slider(cpi_control_base_1, uvalue = 'CPI Slider', maximum = cpi_max, event_pro = 'cpi_slider_event', scroll = 1)

    ;    ;cpi_match_2DS_part_IDs = widget_button(cpi_control_base_2, value = 'Match 2DS IDs', event_pro = 'cpi_match_2DS_part_IDs_event')

    ;    cpi_txt3 = widget_text(cpi_control_base_3, value = 'CPI 2DS ID H')
    ;    cpi_txt4 = widget_text(cpi_control_base_3, value = 'CPI 2DS ID V')

    ;    H_txt1 = widget_text(H_control_base, value = 'H Hex Time')
    ;    H_txt2 = widget_text(H_control_base, value = 'H Time')
    ;    H_txt3 = widget_text(H_control_base, value = '2DS ID H')

    ;    H_slider = widget_slider(H_control_base, uvalue = 'H Slider', maximum = H_max, event_pro = 'H_slider_event', scroll = 1)

    ;    H_control_prev_next_base = widget_base(H_control_base, /row)
    ;    H_find_prev_ID = widget_button(H_control_prev_next_base, value = "Prev H ID", event_pro = 'H_find_prev_ID_event')
    ;    H_find_next_ID = widget_button(H_control_prev_next_base, value = "Next H ID", event_pro = 'H_find_next_ID_event')

    ;    V_txt1 = widget_text(V_control_base, value = 'V Hex Time')
    ;    V_txt2 = widget_text(V_control_base, value = 'V Time')
    ;    V_txt3 = widget_text(V_control_base, value = '2DS ID V')

    ;    V_slider = widget_slider(V_control_base, uvalue = 'V Slider', maximum = V_max, event_pro = 'V_slider_event', scroll = 1)

    ;    V_control_prev_next_base = widget_base(V_control_base, /row)
    ;    V_find_prev_ID = widget_button(V_control_prev_next_base, value = "Prev V ID", event_pro = 'V_find_prev_ID_event')
    ;    V_find_next_ID = widget_button(V_control_prev_next_base, value = "Next V ID", event_pro = 'V_find_next_ID_event')


    ; The Window Layout
    layout_base = WIDGET_BASE(/col, title='3V-CPI View 3 Viewer', UVALUE='3V-CPI View 3 Viewer', FRAME=1, MAP=1) ; the lowest level widget base for the window

    image_base = widget_base(layout_base, /row, /frame)

    ; from original zoom 1:1
    ;cpi_image_base = widget_base(image_base, /col)
    ;HV_2DS_base    = widget_base(image_base, /col)
    ;H_base         = widget_base(HV_2DS_base, /row)
    ;H_image_base   = widget_base(H_base, /col)
    ;V_base         = widget_base(HV_2DS_base, /row)
    ;V_image_base   = widget_base(V_base, /col)

    cpi_image_base = widget_base(image_base, /col)
    ;H_2DS_base    = widget_base(image_base, /row)
    ;H_base        = widget_base(image_base, /col)
    H_image_base   = widget_base(image_base, /col)
    ;V_2DS_base    = widget_base(image_base, /row)
    ;V_base        = widget_base(image_base, /col)
    V_image_base   = widget_base(image_base, /col)

    label_offset = 10

    ; the Draw areas

    cpi_image_draw = widget_draw( cpi_image_base, xs =   cpi_image_draw_width,  ys =    cpi_image_draw_height + label_offset)
    junk = widget_label(cpi_image_base, value = 'CPI Frame')

    H2ds_image_draw = widget_draw(   H_image_base, xs = hv_2DS_image_draw_width, ys = hv_2DS_image_draw_height + label_offset)
    junk = widget_label(  H_image_base, value = '2D-S H Image')

    V2ds_image_draw = widget_draw(   V_image_base, xs = hv_2DS_image_draw_width, ys = hv_2DS_image_draw_height + label_offset)
    junk = widget_label(  V_image_base, value = '2D-S V Image')

    parameter_base = widget_base(image_base, /col, /frame)
    ;*****************************************************************
    ; added  Length Parameter for comparison
    ;************************************************************
    ;*************************************************************
    parameter_base4 = widget_base(parameter_base,/frame)
    p4_base = widget_base(parameter_base4, /grid, col = 4)

    junk = widget_label(p4_base, value = 'Size(um): ')
    junk = widget_label(p4_base, value = 'M1(L1)')
    junk = widget_label(p4_base, value = 'M4(L7)')
    junk = widget_label(p4_base, value = 'M4uncorr(L5)')
    junk = widget_label(p4_base, value = 'M6(L1,L5)')
    junk = widget_label(p4_base, value = 'M7(maxLen)')
    junk = widget_label(p4_base, value = 'maxWid')

    junk = widget_label(p4_base, value = 'CPI')
    txt_Junk   = widget_text(p4_base, value = '', xsize=10)
    txt_Junk   = widget_text(p4_base, value = '', xsize=10)
    txt_Junk   = widget_text(p4_base, value = '', xsize=10)
    txt_Junk   = widget_text(p4_base, value = '', xsize=10)
    txt_CPIL  = widget_text(p4_base, value = '', xsize=10)
    txt_CPIW  = widget_text(p4_base, value = '', xsize=10)


    junk = widget_label(p4_base, value = '2D-S H')
    txt_HL1   = widget_text(p4_base, value = '', xsize=10)
    txt_HL4   = widget_text(p4_base, value = '', xsize=10)
    txt_HL4u   = widget_text(p4_base, value = '', xsize=10)
    txt_HL6   = widget_text(p4_base, value = '', xsize=10)
    txt_HMaxL = widget_text(p4_base, value = '', xsize=10)
    txt_HmaxW = widget_text(p4_base, value = '', xsize=10)

    junk = widget_label(p4_base, value = '2D-S V')
    txt_VL1   = widget_text(p4_base, value = '', xsize=10)
    txt_VL4   = widget_text(p4_base, value = '', xsize=10)
    txt_VL4u   = widget_text(p4_base, value = '', xsize=10)
    txt_VL6   = widget_text(p4_base, value = '', xsize=10)
    txt_VmaxL = widget_text(p4_base, value = '', xsize=10)
    txt_VmaxW = widget_text(p4_base, value = '', xsize=10)

    ;******************************************************************
    parameter_base3 = widget_base(parameter_base,/frame)
    p3_base = widget_base(parameter_base3, /grid, col = 4)

    junk = widget_label(p3_base, value = 'Area(um^2):')
    junk = widget_label(p3_base, value = 'M1')
    junk = widget_label(p3_base, value = 'M4')
    junk = widget_label(p3_base, value = 'M4uncorr')
    junk = widget_label(p3_base, value = 'M6')
    junk = widget_label(p3_base, value = 'M7')
    junk = widget_label(p3_base, value = 'Ratio(focus)')


    junk = widget_label(p3_base, value = 'CPI')
    txt_CM1area      = widget_text(p3_base, value = '', xsize=10)
    txt_CM4area       = widget_text(p3_base, value = '', xsize=10)
    txt_CM4areaun     = widget_text(p3_base, value = '', xsize=10)
    txt_cpi_maxL_area = widget_text(p3_base, value = '', xsize=10)
    txt_Cshad_area    = widget_text(p3_base, value = '', xsize=10)
    txt_CRatio    = widget_text(p3_base, value = '', xsize=10)


    junk = widget_label(p3_base, value = '2D-S H')
    txt_HM1area      = widget_text(p3_base, value = '', xsize=10)
    txt_HM4area       = widget_text(p3_base, value = '', xsize=10)
    txt_HM4areaun     = widget_text(p3_base, value = '', xsize=10)
    txt_HM6area       = widget_text(p3_base, value = '', xsize=10)
    txt_HM7area       = widget_text(p3_base, value = '', xsize=10)
    txt_HfocusRatio   = widget_text(p3_base, value = '', xsize=10)


    junk = widget_label(p3_base, value = '2D-S V')
    txt_VM1area      = widget_text(p3_base, value = '', xsize=10)
    txt_VM4area       = widget_text(p3_base, value = '', xsize=10)
    txt_VM4areaun     = widget_text(p3_base, value = '', xsize=10)
    txt_VM6area       = widget_text(p3_base, value = '', xsize=10)
    txt_VM7area       = widget_text(p3_base, value = '', xsize=10)
    txt_VfocusRatio  = widget_text(p3_base, value = '', xsize=10)

    ;********************************************************************

    ;********************************************************************


    p1_base = widget_base(parameter_base, /grid, col = 4, /frame)

    junk = widget_label(p1_base, value = 'Mass(mg):')
    junk = widget_label(p1_base, value = 'M1(csp) ')
    junk = widget_label(p1_base, value = 'M1(area)')
    junk = widget_label(p1_base, value = 'M4(csp) ')
    junk = widget_label(p1_base, value = 'M4(area)')
    junk = widget_label(p1_base, value = 'M4uc(csp)')
    junk = widget_label(p1_base, value = 'M4uc(area)')
    junk = widget_label(p1_base, value = 'M6(csp) ')
    junk = widget_label(p1_base, value = 'M6(area)')
    junk = widget_label(p1_base, value = 'M7(csp)')
    junk = widget_label(p1_base, value = 'M7(area)')
    ;junk = widget_label(p1_base, value = ' ')


    junk = widget_label(p1_base, value = 'CPI')
    junk                = widget_text(p1_base, value = '', xsize=10)
    junk                = widget_text(p1_base, value = '', xsize=10)
    junk                = widget_text(p1_base, value = '', xsize=10)
    junk                = widget_text(p1_base, value = '', xsize=10)
    junk                = widget_text(p1_base, value = '', xsize=10)
    junk                = widget_text(p1_base, value = '', xsize=10)
    junk                = widget_text(p1_base, value = '', xsize=10)
    junk                = widget_text(p1_base, value = '', xsize=10)
    txt_cpi_csp         = widget_text(p1_base, value = '', xsize=10)
    txt_cpi_iwc         = widget_text(p1_base, value = '', xsize=10)

    junk = widget_label(p1_base, value = '2D-S H')
    txt_HM1cspMass    = widget_text(p1_base, value = '', xsize=10)
    txt_HM1areaMass    = widget_text(p1_base, value = '', xsize=10)
    txt_HM4cspMass    = widget_text(p1_base, value = '', xsize=10)
    txt_HM4areaMass    = widget_text(p1_base, value = '', xsize=10)
    txt_HM4cspMassu   = widget_text(p1_base, value = '', xsize=10)
    txt_HM4areaMassu    = widget_text(p1_base, value = '', xsize=10)
    txt_HM6cspMass    = widget_text(p1_base, value = '', xsize=10)
    txt_HM6areaMass    = widget_text(p1_base, value = '', xsize=10)
    txt_HM7cspMass    = widget_text(p1_base, value = '', xsize=10)
    txt_HM7areaMass    = widget_text(p1_base, value = '', xsize=10)


    junk = widget_label(p1_base, value = '2D-S V')
    txt_VM1cspMass    = widget_text(p1_base, value = '', xsize=10)
    txt_VM1areaMass    = widget_text(p1_base, value = '', xsize=10)
    txt_VM4cspMass    = widget_text(p1_base, value = '', xsize=10)
    txt_VM4areaMass    = widget_text(p1_base, value = '', xsize=10)
    txt_VM4cspMassu   = widget_text(p1_base, value = '', xsize=10)
    txt_VM4areaMassu    = widget_text(p1_base, value = '', xsize=10)
    txt_VM6cspMass    = widget_text(p1_base, value = '', xsize=10)
    txt_VM6areaMass    = widget_text(p1_base, value = '', xsize=10)
    txt_VM7cspMass     = widget_text(p1_base, value = '', xsize=10)
    txt_VM7areaMass    = widget_text(p1_base, value = '', xsize=10)

    ;**************************************************

    control_base = widget_base(layout_base, col=8, /frame); don't use grids in here.

    ;    c1_base = widget_base(control_base, /col)
    ;    c2_base = widget_base(control_base, /col)
    ;    c3_base = widget_base(control_base, /col)
    ;    c4_base = widget_base(control_base, /col)
    ;    c5_base = widget_base(control_base, /col)
    ;    c6_base = widget_base(control_base, /col)
    ;    c7_base = widget_base(control_base, /col)

    junk = widget_label(control_base, value = '  ',xsize=40,ysize=20)
    junk = widget_label(control_base, value = 'CPI',xsize=40,ysize=20)
    junk = widget_label(control_base, value = '2D-S H',xsize=40,ysize=20)
    junk = widget_label(control_base, value = '2D-S V',xsize=40,ysize=20)

    junk     = widget_label(control_base, value = 'Hex Time',xsize=50)
    cpi_txt1 = widget_text(control_base,  value = 'CPI Hex Time',xsize=30)
    H_txt1   = widget_text(control_base,  value = 'H Hex Time',xsize=30)
    V_txt1   = widget_text(control_base,  value = 'V Hex Time',xsize=30)

    junk     = widget_label(control_base, value = 'Time',xsize=50)
    cpi_txt2 = widget_text(control_base,  value = 'CPI Time',xsize=30)
    H_txt2   = widget_text(control_base,  value = 'H Time',xsize=30)
    V_txt2   = widget_text(control_base,  value = 'V Time',xsize=30)

    junk     = widget_label(control_base, value = ' ',ysize=16)
    junk     = widget_label(control_base, value = 'CPI IDs', ysize=20)
    cpi_txt3 = widget_text(control_base,  value = 'CPI 2DS ID H',xsize=30)
    cpi_txt4 = widget_text(control_base,  value = 'CPI 2DS ID V',xsize=30)

    junk     = widget_label(control_base, value = '    ',ysize=16)
    junk     = widget_label(control_base, value = '2D-S IDs',ysize=20)
    H_txt3   = widget_text(control_base,  value = '2DS ID H',xsize=30)
    V_txt3   = widget_text(control_base,  value = '2DS ID V',xsize=30)


    cpi_slider = widget_slider(control_base, uvalue = 'CPI Slider', minimum = 1, maximum = cpi_max, event_pro = 'cpi_slider_event', scroll = 1)
    H_slider = widget_slider(control_base, uvalue = 'H Slider', maximum = H_max, event_pro = 'H_slider_event', scroll = 1)
    V_slider = widget_slider(control_base, uvalue = 'V Slider', maximum = V_max, event_pro = 'V_slider_event', scroll = 1)
    junk       = widget_label(control_base, value = ' ')

    n_before = 5
    n_after  = 5
    txt_n_bf_width = 5

    junk       = widget_label(control_base, value = ' ',ysize=16)
    junk       = widget_label(control_base, value = 'Search IDs',ysize=20)

    H_control_prev_next_base = widget_base(control_base, /row)
    H_find_prev_ID = widget_button(H_control_prev_next_base, value = 'Prev H ID', event_pro = 'H_find_prev_ID_event')
    H_find_next_ID = widget_button(H_control_prev_next_base, value = 'Next H ID', event_pro = 'H_find_next_ID_event')

    V_control_prev_next_base = widget_base(control_base, /row)
    V_find_prev_ID = widget_button(V_control_prev_next_base, value = 'Prev V ID', event_pro = 'V_find_prev_ID_event')
    V_find_next_ID = widget_button(V_control_prev_next_base, value = 'Next V ID', event_pro = 'V_find_next_ID_event')

    ;following is 8th column
    input_out_control_base1  =widget_base(control_base)
    show_png_output_buttons = widget_base(input_out_control_base1, /row)
    ;output_single_png2      = widget_button(show_png_output_buttons, value = 'Output Single PNG', event_pro = 'output_2single_png_event')
    ;output_all_pngs2        = widget_button(show_png_output_buttons, value = 'Output All PNGs',   event_pro = 'output_2all_pngs_event')
    ; checkbox for toggling display of perimeters in 3 Viewer, for all images: CPI, 2D-S H and V
    base_show_perimeter1 = widget_base(show_png_output_buttons,/nonexclusive)
    ck_show_perimeter = widget_button(base_show_perimeter1, value = 'Show Perimeters', Event_pro = 'show_perimeter_3viewer_event')
    widget_control, ck_show_perimeter, set_button = !show_perimeter_3viewer

    ; checkbox for toggling display of CPI Trigger 2D-S images as a different color in 3 Viewer
    base_show_CPI_trigger = widget_base(show_png_output_buttons, /nonexclusive)
    ck_show_CPI_trigger = widget_button(base_show_CPI_trigger, value = 'Show CPI Trigger 2D-S as Green', Event_pro = 'show_CPI_trigger_3viewer_event')
    widget_control, ck_show_CPI_trigger, set_button = !show_CPI_trigger_3viewer

    ; Filter CPI images by minimum size
    input_out_control_base2  =widget_base(control_base)
    base_cpi_size    = widget_base(input_out_control_base2, /row)
    junk             = widget_label(base_cpi_size, value = 'CPI Size Range (microns) Min:')
    txt_min_cpi_size = widget_text(base_cpi_size,  value = '100', xsize=5, /editable)
    junk1             = widget_label(base_cpi_size, value = '   Max:')
    txt_max_cpi_size = widget_text(base_cpi_size,  value = '500', xsize=5, /editable)



    input_out_control_base3  =widget_base(control_base)
    show_png_output_buttons = widget_base(input_out_control_base3, /row)
    output_single_png      = widget_button(show_png_output_buttons, value = 'Output Single PNG', event_pro = 'output_single_png_event')
    ;output_all_pngs        = widget_button(show_png_output_buttons, value = 'Output All PNGs',   event_pro = 'output_all_pngs_event')
    output_3view_csv       = widget_button(show_png_output_buttons, value = 'Output CSV File',   event_pro = 'output_3view_csv_event')


    input_out_control_base4  =widget_base(control_base)
    load_file_buttons = widget_base(input_out_control_base4, /row)
    load_new_cpi_file  = widget_button(load_file_buttons, value = 'Load New CPI File',   event_pro = 'load_new_cpi_file_event')
    load_new_2ds_files = widget_button(load_file_buttons, value = 'Load New 2D-S Files', event_pro = 'load_new_2ds_files_event')


    ;HV_control_n_bf_base = widget_base(control_base, /row)
    ;txt_n_before = widget_text(HV_control_n_bf_base, value = strtrim(string(n_before),2), xs=txt_n_bf_width, event_pro = 'txt_n_before_event')
    ;txt_n_after  = widget_text(HV_control_n_bf_base, value = strtrim(string(n_after ),2), xs=txt_n_bf_width, event_pro = 'txt_n_after_event')



    widget_control, layout_base, Set_UValue=pStateInfo, /No_Copy
    widget_control, layout_base, /REALIZE

    ; get the realized window IDs for the draw widgets
    widget_control,  cpi_image_draw, get_value = cpi_WID
    widget_control,  cpi_image_draw, set_draw_view = [0,0] ; TAF not sure if needed, test

    widget_control, H2ds_image_draw, get_value = H_WID
    widget_control, H2ds_image_draw, set_draw_view = [0,0]

    widget_control, V2ds_image_draw, get_value = V_WID
    widget_control, V2ds_image_draw, set_draw_view = [0,0]



    threeViewState = {  p3ViewState,                        $
        cpi_slider:cpi_slider,              $
        H_slider:H_slider,                  $
        V_slider:V_slider,                  $
        n_before:n_before,                  $
        n_after:n_after,                    $
        txt_CPI_hex:CPI_txt1,               $
        txt_CPI_time:CPI_txt2,              $
        txt_CPI_HID:CPI_txt3,               $
        txt_CPI_VID:CPI_txt4,               $
        txt_H_hex:H_txt1,                   $
        txt_H_time:H_txt2,                  $
        txt_H_HID:H_txt3,                   $
        txt_V_hex:V_txt1,                   $
        txt_V_time:V_txt2,                  $
        txt_V_VID:V_txt3,                   $
        val_CPI_HID:0,                      $
        val_CPI_VID:0,                      $
        val_H_HID:0,                        $
        val_V_VID:0,                        $
        CPI_WID:cpi_WID,                    $
        H_WID:H_WID,                        $
        V_WID:V_WID,                        $
        txt_min_cpi_size:txt_min_cpi_size,  $
        txt_max_cpi_size:txt_max_cpi_size,  $
        txt_cpiL:txt_cpiL,      $
        txt_cpiW:txt_cpiW,        $

        txt_HL1:txt_HL1,                  $
        txt_HL4:txt_HL4,                  $
        txt_HL4u:txt_HL4u,            $
        txt_HL6:txt_HL6,              $
        txt_HmaxL:txt_HmaxL,          $
        txt_HmaxW:txt_HmaxW,            $


        txt_VL1:txt_VL1,                  $
        txt_VL4:txt_VL4,                  $
        txt_VL4u:txt_VL4u,                 $
        txt_VL6:txt_VL6,                  $
        txt_VmaxL:txt_VmaxL,          $
        txt_VmaxW:txt_VmaxW,            $

        txt_Cshad_area:txt_Cshad_area,  $
        txt_cpi_maxL_area:txt_cpi_maxL_area,    $

        txt_HM1area:txt_HM1area,              $
        txt_HM4area:txt_HM4area,              $
        txt_HM4areaun:txt_HM4areaun,          $
        txt_HM6area:txt_HM6area,                  $
        txt_HM7area:txt_HM7area,                  $
        txt_HfocusRatio:txt_HfocusRatio,           $

        txt_VM1area:txt_VM1area,              $
        txt_VM4area:txt_VM4area,              $
        txt_VM4areaun:txt_VM4areaun,          $
        txt_VM6area:txt_VM6area,                  $
        txt_VM7area:txt_VM7area,                  $
        txt_VfocusRatio:txt_VfocusRatio,           $

        txt_cpi_csp:txt_cpi_csp,            $
        txt_cpi_iwc:txt_cpi_iwc,            $

        txt_HM1cspMass:txt_HM1cspMass,                  $
        txt_HM1areaMass:txt_HM1areaMass,                  $
        txt_HM4cspMass: txt_HM4cspMass,                 $
        txt_HM4areaMass: txt_HM4areaMass,                 $
        txt_HM4cspMassu: txt_HM4cspMassu,$
        txt_HM4areaMassu: txt_HM4areaMassu,$
        txt_HM6cspMass:txt_HM6cspMass,$
        txt_HM6areaMass:txt_HM6areaMass,$
        txt_HM7cspmass:txt_HM7cspmass,$
        txt_HM7areamass:txt_HM7areamass,$


        txt_VM1cspMass:txt_VM1cspMass,                  $
        txt_VM1areaMass:txt_VM1areaMass,                  $
        txt_VM4cspMass: txt_VM4cspMass,                 $
        txt_VM4areaMass: txt_VM4areaMass,                 $
        txt_VM4cspMassu: txt_VM4cspMassu,$
        txt_VM4areaMassu: txt_VM4areaMassu,$
        txt_VM6cspMass:txt_VM6cspMass,$
        txt_VM6areaMass:txt_VM6areaMass,$
        txt_VM7cspmass:txt_VM7cspmass,$
        txt_VM7areamass:txt_VM7areamass $

        }

    ; TAF assign the p3ViewState pointer in pStateInfo
    ; but first have to re-grab pStateInfo, since it was set earlier with the call to widget_control with /realize
    ; which was needed to get valid Window IDs
    widget_control, layout_base, get_UValue=pStateInfo, /no_copy
    (*pStateInfo).p3ViewState = ptr_new(threeViewState, /no_copy)
    widget_control, layout_base, Set_UValue=pStateInfo, /no_copy

    XMANAGER, 'Widget_3View', layout_base, /JUST_REG, /no_block

end ; a3vcpiview_3_view_widget


