pro DrawTimeSeries, ave_info

    Common  DataBlock,alldataH,alldataV,ADataH,ADataV,FirstData,LastData,StripData,FirstDataV,LastDataV,StripDataV,OneHzH,OneHzV


    forward_function SPEC_Nano2HMS

    Widget_Control, ave_info.ZoomWindowSeriesDrawID, Get_Value=wid

    bluecolor = GetColor("Navy",  !D.Table_Size-2)
    redcolor = GetColor("Red",   !D.Table_Size-2)
    whitecolor = GetColor("White", !D.Table_Size-2)
    ;-------------------------------------------------------------
    ;This routine use ave_info --user selected start-time and endtime to get
    ;particles within the time period to make high resolution time series plots,
    ;always 300 bins. concentration, extinction, IWC vs time.

    local_ADataH = REPLICATE({_AverageData},300)  ; prepare the 300 bin averaging array
    local_ADataV = REPLICATE({_AverageData},300)
    ; stop
    ;call a subroutine to get all 300 bins time series data :local_AdataH and local_AdataV
    ;get_300bin_HV_time_series,ave_info,'H','L1',local_adataH
    ;get_300bin_HV_time_series,ave_info,'V','L1',local_adataV
    ;stop
    ; Look at the first and last particles in this
    ;  file for the range of timestamps.
    ;---------------------------------------------

    plot_string =  !ZOOM_START_STRING +  $
        " to " + !ZOOM_END_STRING
    plot_string2 =' Images: '+ave_info.Hpartcont+'  '+ave_info.Vpartcont; ave_info.Hpartcont is string.

    chart1 =    [0.10, 0.700, 0.95, 0.945]
    chart2 =    [0.10, 0.700 - 0.31, 0.95, 0.945-0.31]
    chart3 =    [0.10, 0.700 - 0.62, 0.95, 0.945-0.62]

    axisColor = GetColor("Navy",  !D.Table_Size-2)
    dataColor = GetColor("Navy",  !D.Table_Size-2)
    backColor = GetColor("White", !D.Table_Size-2)

    thisCharsize = Str_Size('A Sample String', 0.25)

    strf = 'aaa';SPEC_Nano2HMS(local_FirstData.Nanoseconds)
    strl = 'bbb'; SPEC_Nano2HMS(local_LastData.Nanoseconds)

    fullrange = 0ULL
    binwidth = 0ULL
    ;fullrange = local_LastData.Nanoseconds - local_FirstData.Nanoseconds
    fullrange = ave_info.endNanoseconds - ave_info.startNanoseconds
    binwidth = fullrange/8
    curbin = ave_info.startNanoseconds

    tnames = StrArr(9)
    FOR i=0,8 DO BEGIN
        tnames(i) =  strtrim(strmid(SPEC_Nano2HMS(curbin),0,12),2)
        curbin += binwidth
    ENDFOR

    ;-------------------------------
    ;Set up X axis tick labels.
    ;-------------------------------
    !X.TICKNAME = tnames
    !X.TICKS = 8

    !P.Multi = [0,3,1,0,1]  ; Make sure we're plotting at the top position
    Xdata = local_ADataH.Concentration

    H_dead = 'H= ' + strtrim(string(format = '(%"%15.3f Secs")',ave_info.Hdeadtime),2)
    V_dead = 'V= ' + strtrim(string(format = '(%"%15.3f Secs")',ave_info.Vdeadtime),2)
    main_title = 'Particle Concentration (H=BLUE, V=RED) Deadtime '+ H_dead + V_dead
    y_title = 'Concentration(#/L)'
    ;----------------------------------------------------
    ; Make sure the window we're plotting to is selected
    ;----------------------------------------------------
    Widget_Control, ave_info.ZoomWindowSeriesDrawID, Get_Value=wid
    WSet, wid

    thisCharsize = 1.8

    Plot, Xdata, $
        Background=backcolor, $           ; The background color of the display.
        Charsize=thisCharsize, $          ; The character size, as determined by Str_Size.
        Color=axiscolor, $                ; The color of the axes.
        XStyle=1, $                       ; Exact axis scaling. No autoscaled axes.
        ; XTickformat='(I6)', $
        Title = main_title, $
        YTitle = y_title , $
        Position=chart1 ; , $
        ;XTICKFORMAT='MYXTICKS'


        Xdata = local_ADataV.Concentration
    Oplot,Xdata, $
        color = redcolor

    Xdata = local_ADataH.Extinction
    main_title = 'Particle Extinction (H=BLUE, V=RED)'
    y_title = 'Extinction(1/km)'


    Plot, Xdata, $
        Background=backcolor, $           ; The background color of the display.
        Charsize=thisCharsize, $          ; The character size, as determined by Str_Size.
        Color=axiscolor, $                ; The color of the axes.
        XStyle=1, $                       ; Exact axis scaling. No autoscaled axes.
        ; XTickformat='(I6)', $
        Title = main_title, $
        YTitle = y_title , $
        Position=chart2 , $
        /NOERASE ;, $
        ; XTICKFORMAT='MYXTICKS'


        Xdata = local_ADataV.Extinction
    Oplot,Xdata, $
        color = redcolor

    Xdata = local_ADataH.IWC
    main_title = 'Particle IWC (H=BLUE, V=RED)'
    y_title = 'ice Water Content(g/m^3)'


    Plot, Xdata, $
        Background=backcolor, $           ; The background color of the display.
        Charsize=thisCharsize, $          ; The character size, as determined by Str_Size.
        Color=axiscolor, $                ; The color of the axes.
        XStyle=1, $                       ; Exact axis scaling. No autoscaled axes.
        ; XTickformat='(I6)', $
        Title = main_title, $
        YTitle = y_title , $
        Position=chart3 , $
        /NOERASE ;, $
        ;    XTICKFORMAT='MYXTICKS'

        Xdata = local_ADataV.IWC
    Oplot,Xdata, $
        color = redcolor


    !X.TICKNAME = ''
    !X.TICKS = 0

    XYOUTS,25,590, 'In.File: ' + strmid(!CurrentHFname,strlen(!CurrentHPname)) + '  Time from ' + plot_string + ' ' + plot_string2 ,color=bluecolor, size=1, /device
    device, CURSOR_STANDARD= !ARROW_CURSOR  ; show hourglass while we process the data
end

;=============================================================================
FUNCTION ZOOM_MYXTICKS, axis, index, value
    Common  DataBlock,alldataH,alldataV,ADataH,ADataV,FirstData,LastData,StripData,FirstDataV,LastDataV,StripDataV,OneHzH,OneHzV


    ; str = SPEC_Nano2HMS(FirstData.Nanoseconds)
    fullrange = 0ULL
    binwidth = 0ULL
    fullrange = LastData.Nanoseconds - FirstData.Nanoseconds
    binwidth = fullrange/7
    curbin = FirstData.Nanoseconds  + (index * binwidth)


    ; WIDGET_CONTROL ,strip_info.stripLabel ,TLB_GET_SIZE= a
    ; widget_info(drawID,/GEOMETRY)

    xlabel = strmid(SPEC_Nano2HMS(curbin),0,12)
    ;xoffset = (index * ((0.95-0.15)/7.0)) + 0.153
    xoffset = (index * ((0.95-0.15)/7.0)) + 0.172


    yoffset = 0.681
    charsize = !CHAR_SIZE  * 0.85 ; use the sizing we calculated from the last window resize
    bluecolor = GetColor("Navy", !D.Table_Size-2)
    XYOutS, xoffset, yoffset, xlabel, color = bluecolor, $
        Alignment=1.0, Orientation=0.0,size = charsize, /NORMAL

    ;print, "MYXTICKS",axis,index,value
    ;return , string(index)
    return, ""

end
;*************************************************************

