pro generate_1hz_file,HorV,Seconds,min_time,info,version_string,header_balance,oneHz_filename
    ; input is  seconds,min_time, info, ****, all are input.

    Common  DataBlock,alldataH,alldataV,ADataH,ADataV,FirstData,LastData,StripData,FirstDataV,LastDataV,StripDataV,OneHzH,OneHzV

    forward_function get_IWC_from_adj_n_sizes

    StructSize = N_TAGS( {_DerivedDataRev2} , /DATA_LENGTH )

    ; Now generate the PSD bins for the 1 second data
    ; Bin edges are:

    if strupcase(HorV) eq 'H' then begin
        microns_per_pixel =!microns_per_pixelH
        alldata = alldataH
        oneHz =oneHzH
    endif else begin
        microns_per_pixel =!microns_per_pixelV
        alldata = alldataV
        oneHz=oneHzV
    endelse

    xpts_pixels = [1, 2,3,4,5,6,7,8,9,10,12,15,20,25,32,40,50,60,70,80,90,100,120,140,1001]
    xpts=xpts_pixels*microns_per_pixel

    NanosecondsPerSecond = 1000000000.0D  ; use double for some calculations
    NPS = 1000000000ULL

    max_sec_string = strtrim(string(Seconds),2)
    min_index = 0
    max_index = Seconds

    dww = !DWW_main; 63.0mm in 2D-S, 50.8 mm in 3V-CPI adjust in a3VCPIView.pro
    pxs = microns_per_pixel/1000.0;    0.01D   ; 10 micrometers or 0.01mm optically realized size of the array elements
    strbsz = microns_per_pixel/1000.0;  0.01D  ; 10 micrometers or 0.01mm pixel size in the TAS direction (strobe size)
    DeltaT = 0.0D
    SamplesPerBin = 0UL


    ;variables for IWC
    alpha = 0.115D
    beta = 1.218D
    constant = pxs * strbsz * 0.000001D  ; for IWC calculation

    start_bin = min_time
    end_bin = start_bin + NPS


    ;-----------------------------------------------------------------------
    ; Generate Horizontal 1Hz file
    FOR i=0L,Seconds-1 DO BEGIN
        ;-----------------------------------------------------
        ; Prepare pointers and time stamps for later use
        ; (like in the strip charts of bitmaps)
        ;-----------------------------------------------------
        OneHz[i].NSOffset = start_bin  ; always store the nanoseconds of the start of each 1 second bin
        aa = where( ((*alldata)(*).Nanoseconds GE start_bin) and ((*alldata)(*).Nanoseconds LT end_bin),thepoint)
        Widget_Control, info.statusID, DYNAMIC_RESIZE=1, $
            Set_Value= string( format = '("Generating 1Hz Data File" , I6)',i+1) + " of " + max_sec_string + " bins completed"

        if (thepoint eq 0) then begin; No particles in this bin. so we look forward to get a first particle for it.
            zz = where( ((*alldata)(*).Nanoseconds GE end_bin) and ((*alldata)(*).Area gt 0),zz_count)
            if (zz_count eq 0) then begin
                OneHz[i].FileOffset = 0 ;flag that there are no more particles beyond this 1 sec bin
            endif else begin
                OneHz[i].FileOffset = zz[0] ; save the index of the first particle from some future 1 sec bin
            endelse
        endif else begin; If we've got 1 or more points, then perform the averaging

            ;calculate the sample volume first.
            Get_Total_DTs_btw_T1T2,HorV,start_bin,end_bin,total_DTs,FullTicks,SampleVolume
            if SampleVolume lt 0.0 then SampleVolume=2.0*microns_per_pixel

            ; get all particles sizes, area, adjs for this time period.
            index=where((*alldata).area gt 0.0 and (*alldata).Nanoseconds GE start_bin and $
                (*alldata).Nanoseconds LT end_bin, contG)

            if (contG eq 0) then begin  ; no good particles in this 1Hz bin
                zz = where((*alldata).Nanoseconds GE end_bin and (*alldata).Area gt 0,zz_count)
                if (zz_count eq 0) then begin
                    OneHz[i].FileOffset = 0 ;flag that there are no more particles beyond this 1 sec bin
                endif else begin
                    OneHz[i].FileOffset = zz[0] ; save the index of the first particle from some future 1 sec bin
                endelse
            endif else begin  ; else, there are none fake particles in this 1Hz bin
                OneHz[i].FileOffset = index[0] ;save the index of the first particle from some future 1 sec bin
                adjs=(*alldata)(index).adj1*(*alldata)(index).adj0
                sizes=microns_per_pixel*(*alldata)(index).L1n;  in um unit.
                areas=(microns_per_pixel*microns_per_pixel*0.1^6)*(*alldata)(index).area; in unit mm2 now.

                OneHz[i].Concentration = total(adjs)/SampleVolume
                OneHz[i].Extinction = 2.0*total(adjs *Areas)/SampleVolume
                get_IWC_from_adj_n_sizes,total_mass,alpha,beta,adjs,sizes,areas
                OneHz[i].IWC = total_mass/Samplevolume

                ; get the psd of the current time period.
                L1n=sizes/microns_per_pixel
                FOR bin=0,23 DO BEGIN
                    somarr=where(L1n lt xpts_pixels[bin+1] and L1n ge xpts_pixels[bin], contG) ; Handle all but last bins
                    binwidth=(xpts[bin+1]-xpts[bin])
                    LiterTum=binwidth*SampleVolume
                    if contG gt 0 then begin;  if there is no particles, area and mass should be zero
                        OneHz[i].bins[bin] = total(adjs(somarr)) /LiterTum  ;    #/L/um
                    endif
                ENDFOR
            endelse
        endelse;  of else begin; If we've got 1 or more points, then perform the averaging
        ;done H-channel-----1Hz

        start_bin = end_bin
        end_bin += NPS

    ENDFOR
    ;------------------------------------------------------------------------------------
    ; Replace the particle index with the actual byte offset of that particle's structure
    ;  because the FillStrip routine expects it in this format.
    ;-------------------------------------------------------------------------------------
    if strupcase(HorV) eq 'H' then begin
        oneHzH=oneHz
        OneHzH.FileOffset = OneHz.FileOffset * StructSize
        OpenW, onehz_lun, oneHz_filename, /Get_Lun

        WRITEU, onehz_lun, version_string
        WRITEU, onehz_lun, header_balance
        writeu, onehz_lun,long(Seconds)
        writeu, onehz_lun,OneHzH
        close, onehz_lun
        Free_Lun,onehz_lun
    endif else begin;  ; note that if there is only one channel, for HVPS, it need to do this also. Care has to be taken.
        oneHzV=oneHz
        OneHzV.FileOffset = OneHz.FileOffset * StructSize
        OpenU, onehz_lun, oneHz_filename, /Get_Lun,/Append
        writeu, onehz_lun,OneHzV
        close, onehz_lun
        Free_Lun,onehz_lun
    endelse

end ; end of generate_1Hz_file,HorV.*****

;=======================================================================
;-----------------------------------------------------------
;=======================================================================
;-----------------------------------------------------------
; Here is the code that we use to figure out how many of the
; one second averaged samples should be grouped into a single
; bin in order to best try to show 1000 bins for the
; time series display.
; If there are more than 1000 bins (more than will nicely fit across the screen)
;  then we begin to group the 1Hz bins together. Start with 5 1Hz bins, then 10,
;  then 15, etc. When we've figured out a number (divisible by 5) that will get
;  us less then 1000 total bins, use that number. This routine calculates
;  that number.
;  Start by passing the total number of 1Hz bins in num.
;  We return the width of each bin in BinWidth (will be 5,10,15,20, etc.)
;  And we also return the total number of bins that should be created.
;  For example, if you have 4000 seconds of data, then you pass the variable num
;   num = 4000
;   BinWidth = 5
;   routine will return 800 for the number of bins to create
;-----------------------------------------------------------
FUNCTION CalcNumBins, num, BinWidth

    if (num gt 1000.0) then begin
        NumBins = 1001.0
        BinWidth = 0.0
        while NumBins gt 1000.0 do begin
            BinWidth += 5.0
            NumBins = num/BinWidth
        endwhile
    endif else begin
        NumBins = num
        BinWidth = 1.0
    endelse
    NumBins = ceil(NumBins) ; round to remove decimal part
    return, NumBins
END
;=======================================================================

