; File:  id.pro
; Author:  Ted Fisher <ted@specinc.com> <fishnuke@gmail.com>
; Purpose:  decompresses 2D-S image data and displays it in an IDL window


;   The c version to extract clear and shaded bits

;   if(t == 0x7fff) // 0x7fff is defined as 128 clear bits
;   {
;       clear_bits  = PIXELS
;       shaded_bits = 0
;   }
;   else
;   {
;       if(t == 0x4000) // 0x4000 is defined as 128 shaded bits
;       {
;           shaded_bits = PIXELS
;           clear_bits = 0
;       }
;       else
;       {
;           clear_bits  = t & 0x7f
;           shaded_bits = ((t >> 7) & 0x7f)
;       } // end if
;   } // end if


; From the 2D-S manual:

; Bit Number           Value          Description

;   15                   0              Indicates that this is an image word
;   14                   1              Indicates the start of a slice (FIRST)
;                        0              Indicates the continuation of a slice
;   7-13                                Number of shaded pixels (bit 13 is MSB)
;   0- 6                                Number of clear  pixesl (bit  6 is MSB)
;          0x7FFF = 128 clear bits
;          0x4000 = 128 set   bits (shaded)







pro id, in_data


    PIXELS = 128

    mask = byte([128, 64, 32, 16, 8, 4, 2, 1])
    mask_index = 0

    clear_flag = '7FFF'X
    shaded_flag = '4000'X


    in = in_data ; safeguard the image from changes here

    ; da  = out_array       ; binary data array
    ; dba = out_byte_array  ; byte data array, for displaying image


    slices = (size(in))[1]/2


    ; the image will be PIXELS x slices

    image = bytarr(PIXELS, slices)
    index = 0 ; index into input compressed image

    byte_array = bytarr(PIXELS * slices)
    byte_index = -PIXELS ; index into the 1D decompressed image, representing each pixel as a byte in the array
    last_byte_index = -PIXELS ; TAF NOTE: this start value is to accomodate Chris Wieland's really bizarre loop logic where he imcrements at the start of every loop -- need to change this for sure!

    print, ''

    still_filling_bitmap = 1

    while(1) do begin

        t = ishft(in[index+1], 8) + in[index]

        if(t eq clear_flag) then begin
            ; 128 clear bits this slice
            clear_bits = PIXELS
            shaded_bits = 0
        endif else if(t eq shaded_flag) then begin
            ; 128 shaded bits this slice
            clear_bits = 0
            shaded_bits = PIXELS
        endif else begin
            ; mixed clear and shaded, decode
            clear_bits =  t      and '7F'XB
            shaded_bits = (t/128) and '7F'XB ; or use ishft(t, -3)
        endelse



        if(t and '4000'X) then begin ; start of a slice
            if(still_filling_bitmap)
            byte_index = last_byte_index + PIXELS


            c =  clear_bits
            s = shaded_bits

            print, 'c', c, '    s', s

            ;stop


            index += 2

            if(index ge 2*slices) then begin
                break
            endif
        endwhile



        print, ''










        ;ted, dba


    end ; id
