; File: two_uints2float.pro
; Author:  Ted Fisher <ted@specinc.com> <fishnuke@gmail.com>
; Purpose:  implements the two_uints2float function, which needed to be in a separate file
;  for IDL namespace



function two_uints2float, in_array_hi, in_array_lo
    ; convert two unsigned integers 16 bit binary data to a single 32 bit floating point IEEE 754 single precision

    ;    UPDATE:  The real way to do IDL type casting is:

    ;  Thanks to David Fanning for the enlightenment on IDL type casting -- the trick is the Offset in byte and float functions

    ;    ; convert the raw bytes to TAS using IDL type casting

    ;    hi_bytes = byte(hk_array[*, 74], 0, 2, hk_count)
    ;    lo_bytes = byte(hk_array[*, 75], 0, 2, hk_count)
    ;    tas       = float([lo_bytes, hi_bytes], 0, hk_count)

    ; or more generally:

    ;    count = (size(in_array_hi))[1]

    ;    hi_bytes = byte(in_array_hi, 0, 2, count)
    ;    lo_bytes = byte(in_array_lo, 0, 2, count)

    ;    float_values = float([lo_bytes, hi_bytes], 0, count)


    ; used to assemble raw binary data to useable floating point numbers in IDL

    ; tasbin = dec2bin(data(HKDataIDX+74)*2^16 + data(HKDataIDX+75));%binary representation
    ; S=bin2dec(tasbin(1)); %floating point sign bit
    ; E=bin2dec(tasbin(2:9));%floating point EEEEEEEE (exponential) field
    ; Fraction = 1 + bin2dec(tasbin(10:32))/8388608

    ; tas = [tas,(-1)^S*2^(E-127)*Fraction]

    ; tas = two_uints2float(hk_array[*, 74], hk_array[*, 75])

    ; NOTES:
    ; the matlab is an incomplete conversion and not the full IEEE 754 single precision floating point specification
    ; the IDL as of 2014-05-01 is incomplete as well, though seems OK for most TAS values -- either need to finish the implementation,
    ; or clamp it against bad TAS values, thus making it a TAS converter not a full binary to float converter...
    ; might be able to do a reads on the binary expanded as a hex string and read into a double, but might be slow?

    ;    raw_floats = ishft(ulong(in_array_hi), 16) + ulong(in_array_lo)

    ;    s = double(ishft(ulong('80000000'X and raw_floats), -31)) ; highest bit is the sign bit

    ;    e = long(ishft(ulong('7F800000'X) and raw_floats, -23)) ; 8 bits of exponent

    ;    f = dblarr(n_elements(e))

    ;    index = where(e eq 0, count)
    ;    if(count gt 0) then begin
    ;        f[index] = double(ulong('007FFFFF'X) and raw_floats[index])/(2.^23.) ; 23 bits of fraction, leading 24th bit is 0
    ;    endif

    ;    index = where(e ne 0, count)
    ;    if(count gt 0) then begin
    ;        f[index] = double(ulong('00800000'X) or (ulong('007FFFFF'X) and raw_floats[index]))/(2.^23.) ; 23 bits of fraction, leading 24th bit is 1
    ;    endif

    ;    ; -1^sign * (1 + sum[(from i to 23) of bit23-i * 2^-i) * 2^(e-127)
    ;    ; or more simply:

    ;    float_values = float((-1.)^s * f * 2^(e - 127))

    count = size(in_array_hi, /n_elements)

    hi_bytes = byte(in_array_hi, 0, 2, count)
    lo_bytes = byte(in_array_lo, 0, 2, count)

    float_values = float([lo_bytes, hi_bytes], 0, count)


    return, float_values
end ; two_uints2float
